// ===========================================================================
//	CComparisonWindow.cp
// ===========================================================================


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <QuickDraw.h>
#include <Menus.h>
#include <ImageCompression.h>
#include <LDocument.h>
#include <LString.h>

#include "RAVE.h"
#include "CTestInfo.h"
#include "CComparisonWindow.h"
#include "S3Utilities.h"
#include "S3Alert.h"
#include "S3Messages.h"


CComparisonWindow* CComparisonWindow::sWindow = nil;

	
// ===========================================================================
//		 CreateComparisonWindow
// ===========================================================================
//	Return a newly created Window object initialized from a PPob resource

	CComparisonWindow*
CComparisonWindow::CreateComparisonWindow(
	ResIDT		inWindowID,
	LCommander	*inSuperCommander)
{
	if (sWindow != nil) {
			// we've already created a comparison window, so just return that one,
			// because we only want one in existence at a time.
		return sWindow;
	}
	
	SetDefaultCommander(inSuperCommander);

	CComparisonWindow	*theWindow =
			(CComparisonWindow*) UReanimator::ReadObjects('PPob', inWindowID);

	Assert_(theWindow);
	
	theWindow->FinishCreate();

	if (theWindow->HasAttribute(windAttr_ShowNew)) {
		theWindow->Show();
	}
	
	return theWindow;
}


// ===========================================================================
//		 CreateComparisonWindowStream
// ===========================================================================
//	Return a newly created Window object initialized with data from a Stream

	CComparisonWindow*
CComparisonWindow::CreateComparisonWindowStream(
	LStream	*inStream)
{
	if (sWindow != nil) {
			// we've already created a comparison window, so just return that one,
			// because we only want one in existence at a time.
		return sWindow;
	}
	
	return (new CComparisonWindow(inStream));
}


// ===========================================================================
	void
CComparisonWindow::CompareWindows(
	CTinselTestWindow* inWindow1, 
	CTinselTestWindow* inWindow2)
{
	if (sWindow == nil || inWindow1 == nil || inWindow2 == nil) {
			// we can't create the comparison window now because we need to have
			// a pointer to the app to use as its super commander
		return;
	}
	
		// this may take a while, so show the watch cursor
	::SetCursor(*(::GetCursor(watchCursor)));
	
		// first get the window's LGWorld pointers.  this is important to avoid a
		// serious crashing bug.  before, if the user selected Compare Windows before
		// running a test, this routine would call GetGWorld to get the LGWorld
		// (which would be nil), then called the GetMacGWorld method of that pointer,
		// without checking for nil.  this would return a garbage GWorld.  comparing 
		// these bad GWorlds would spank the computer hard.
	LGWorld* lGWorld1 = inWindow1->GetGWorld();
	LGWorld* lGWorld2 = inWindow2->GetGWorld();

	if (lGWorld1 == nil || lGWorld2 == nil) {
			// if the LGWorlds haven't been allocated yet, we can't get their
			// Mac GWorlds, so we can't compare the windows 
		return;
	}
	
		// now get the Mac GWorlds out of the LGWorlds
	GWorldPtr gWorld1 = lGWorld1->GetMacGWorld();
	GWorldPtr gWorld2 = lGWorld2->GetMacGWorld();

	if (gWorld1 == nil || gWorld2 == nil) {
			// we can't compare the windows if their LGWorlds didn't have enough
			// memory to allocate Mac GWorlds
		return;
	}
	
		// compare the two windows
	sWindow->CompareGWorlds(gWorld1, gWorld2);

		// bring the window forward
	sWindow->Select();
	
		// we're done with the comparison, so turn the cursor back to an arrow
	::InitCursor();
	
		// default to the average comparison mode
	sWindow->SetCompareMode(eCompare_Average);
	
		// update the window so the user can see the average difference
		// between the two windows
	sWindow->DrawSelf();
}


// ===========================================================================
//		 CComparisonWindow
// ===========================================================================
//	Default Constructor

CComparisonWindow::CComparisonWindow()
	: LWindow()
{
	InitComparisonWindow();
}


// ===========================================================================
//		 CComparisonWindow(SWindowInfo&)
// ===========================================================================
//	Construct Window from the data in a struct

CComparisonWindow::CComparisonWindow(
	const SWindowInfo	&inWindowInfo)
	: LWindow(inWindowInfo)
{
	InitComparisonWindow();
}


// ===========================================================================
//		 CComparisonWindow(ResIDT, Uint32, LCommander*)
// ===========================================================================
//	Construct a Window from a WIND Resource with the specified attributes
//	and SuperCommander

CComparisonWindow::CComparisonWindow(
	ResIDT		inWINDid,
	Uint32		inAttributes,
	LCommander	*inSuper)
		: LWindow(inWINDid, inAttributes, inSuper)
{
	InitComparisonWindow();
}


// ===========================================================================
//		 CComparisonWindow(LStream*)
// ===========================================================================
//	Construct Window from the data in a stream

CComparisonWindow::CComparisonWindow(
	LStream	*inStream)
	: LWindow(inStream)
{
	InitComparisonWindow();
}


// ===========================================================================
CComparisonWindow::~CComparisonWindow()
{
	delete mComparisonGWorld;
	mComparisonGWorld = nil;
	delete mChannelGWorld;
	mChannelGWorld = nil;
	
	sWindow = nil;
}


// ===========================================================================
	void
CComparisonWindow::InitComparisonWindow()
{
	mGWorld = nil;
	mComparisonGWorld = nil;
	mChannelGWorld = nil;
	mLastSwitchTime = 0;
	mIsAlternating = false;
	
	SetCompareMode(eCompare_Average);
	
		// save a pointer to this window
	sWindow = this;
}


// ===========================================================================
//		 ObeyCommand
// ===========================================================================
//	Respond to commands
	Boolean
CComparisonWindow::ObeyCommand(
	CommandT	inCommand,
	void		*ioParam)
{
	Boolean	cmdHandled = true;
	
	switch (inCommand) {
//		case cmd_Save:
//			AskSaveAs();
//			break;

		case cmd_CompareRed:
			SetCompareMode(eCompare_Red);
			break;
		
		case cmd_CompareGreen:
			SetCompareMode(eCompare_Green);
			break;
		
		case cmd_CompareBlue:
			SetCompareMode(eCompare_Blue);
			break;
		
		case cmd_CompareAlpha:
			SetCompareMode(eCompare_Alpha);
			break;
		
		case cmd_CompareAverage:
			SetCompareMode(eCompare_Average);
			break;
		
		case cmd_CompareLeft:
			SetCompareMode(eCompare_Left);
			break;
		
		case cmd_CompareRight:
			SetCompareMode(eCompare_Right);
			break;
		
		case cmd_CompareAlternate:
			SetCompareMode(eCompare_Alternate);
			break;
		
		default:
			cmdHandled = LWindow::ObeyCommand(inCommand, ioParam);
			break;
	}
	
	return cmdHandled;
}


// ---------------------------------------------------------------------------
//		 FindCommandStatus
// ---------------------------------------------------------------------------
//	Pass back status of a (menu) command
	void
CComparisonWindow::FindCommandStatus(
	CommandT	inCommand,
	Boolean		&outEnabled,
	Boolean		&outUsesMark,
	Char16		&outMark,
	Str255		outName)
{
	outUsesMark = false;
	
	switch (inCommand) {
		case cmd_Print:
		case cmd_PrintOne:
		case cmd_PageSetup:
				// we don't handle printing
			outEnabled = false;
			break;
			
		case cmd_CompareRed:
  		outEnabled = true;
  		outUsesMark = true;
  		outMark = (GetCompareMode() == eCompare_Red) ? checkMark : noMark;
  		break;
  		
		case cmd_CompareGreen:
  		outEnabled = true;
  		outUsesMark = true;
  		outMark = (GetCompareMode() == eCompare_Green) ? checkMark : noMark;
  		break;
  		
		case cmd_CompareBlue:
  		outEnabled = true;
  		outUsesMark = true;
  		outMark = (GetCompareMode() == eCompare_Blue) ? checkMark : noMark;
  		break;
  		
		case cmd_CompareAlpha:
  		outEnabled = true;
  		outUsesMark = true;
  		outMark = (GetCompareMode() == eCompare_Alpha) ? checkMark : noMark;
  		break;
  		
		case cmd_CompareAverage:
  		outEnabled = true;
  		outUsesMark = true;
  		outMark = (GetCompareMode() == eCompare_Average) ? checkMark : noMark;
  		break;
  		
		case cmd_CompareLeft:
  		outEnabled = true;
  		outUsesMark = true;
  		outMark = (GetCompareMode() == eCompare_Left && !mIsAlternating) ? checkMark : noMark;
  		break;
  		
  	case cmd_CompareRight:
  		outEnabled = true;
  		outUsesMark = true;
  		outMark = (GetCompareMode() == eCompare_Right && !mIsAlternating) ? checkMark : noMark;
  		break;
  		
  	case cmd_CompareAlternate:
  		outEnabled = true;
  		outUsesMark = true;
  		outMark = (mIsAlternating) ? checkMark : noMark;
  		break;
  		
//	  case cmd_Save:
//				// the test window is not a document, but we can save its contents
//				// to a file, so enable the Save command.  Save As... would be more
//				// appropriate, since every time Save is selected, the user has to
//				// give a new filename, but it's more convenient to have a command-key.
//				// so enable Save, as long as our GWorld isn't empty.
//			if (mGWorld != nil) {
//				outEnabled = true;
//			} else {
//				outEnabled = false;
//			}
//			break;
//			
//	  case cmd_SaveAs:
//				// for the test windows, Save and Save As... are equivalent, so 
//				// disable Save As... and just leave Save.
//			outEnabled = false;
//			break;
			
		default:
			LWindow::FindCommandStatus(inCommand, outEnabled, outUsesMark,
								outMark, outName);
			break;
	}
}


// ===========================================================================
	void
CComparisonWindow::SpendTime(const EventRecord& /* inMacEvent */)
{
	if (!mIsAlternating) {
			// the user has turned off the alternating option by switching to
			// one of the other comparison views, rather than by hitting 
			// command-8 again, so stop this idle nonsense.
		StopIdling();
		return;
	}

	unsigned long currentTime;
	
	::GetDateTime(&currentTime);
	
	if (currentTime - mLastSwitchTime > 0) {
			// at least a second has passed since the last time we switched views,
			// so switch to either the left or right test output views.  we can't
			// just call SetCompareMode because that will turn alternating off.
			// so we have to do everything that routine does ourselves.
		if (mCompareMode == eCompare_Left) {
			mCompareMode = eCompare_Right;

				// get the right window's title
			CTinselTestWindow* testWindow = CTinselTestWindow::GetRightWindow();
			Assert_(testWindow);
			LStr255 windowTitle;
			testWindow->GetDescriptor(windowTitle);
			
				// append that to the title of this window
			windowTitle = "\pComparison : " + windowTitle;
			SetDescriptor(windowTitle);

				// redraw the window, this time showing the right window
			DrawSelf();
		} else {
			mCompareMode = eCompare_Left;

				// get the left window's title
			CTinselTestWindow* testWindow = CTinselTestWindow::GetLeftWindow();
			Assert_(testWindow);
			LStr255 windowTitle;
			testWindow->GetDescriptor(windowTitle);
			
				// append that to the title of this window
			windowTitle = "\pComparison : " + windowTitle;
			SetDescriptor(windowTitle);

				// redraw the window, this time showing the left window
			DrawSelf();
		}
	}
	
	::GetDateTime(&mLastSwitchTime);
}


// ===========================================================================
	void
CComparisonWindow::PutOnDuty()
{
	if (mIsAlternating) {
			// we were alternating the last time we were on duty, so start
			// alternating again
		StartIdling();
	}
}


// ===========================================================================
	void
CComparisonWindow::TakeOffDuty()
{
		// since we're no longer going to be the topmost window, we stop idling,
		// because we don't want to draw on top of other windows.
	StopIdling();
}


// ===========================================================================
	void
CComparisonWindow::CompareGWorlds (
	const GWorldPtr	inGWorld1,
	const GWorldPtr	inGWorld2)
{
	if (inGWorld1 == nil || inGWorld2 == nil) {
			// for some reason, the test windows have not been created
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoWindowsToCompare, nil);
		delete mComparisonGWorld;
		mComparisonGWorld = nil;
		return;
	}
	
	PixMapHandle pixMapH1 = ::GetGWorldPixMap(inGWorld1);
	PixMapHandle pixMapH2 = ::GetGWorldPixMap(inGWorld2);

	if ((**pixMapH1).pixelSize != (**pixMapH2).pixelSize) {
			// the two GWorlds have to have the same depth
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NeedSameBitDepth, nil);
		delete mComparisonGWorld;
		mComparisonGWorld = nil;
		return;
	}

	Int16 pixelDepth = (**pixMapH1).pixelSize;
	
	if (pixelDepth != 32 && pixelDepth != 16) {
			// we can compare only GWorlds that are 32-bit or 16-bit
		S3StopAlert(STRx_AlertStrings_ID, AlertString_No8bitComparison, nil);
		delete mComparisonGWorld;
		mComparisonGWorld = nil;
		return;
	}
	
	if ((**pixMapH1).bounds.right != (**pixMapH2).bounds.right ||
			(**pixMapH1).bounds.bottom != (**pixMapH2).bounds.bottom) {
			// the two GWorlds must be the same size for us to be able to
			// compare them
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NeedSameSize, nil);
		delete mComparisonGWorld;
		mComparisonGWorld = nil;
		return;
	}
	
	Rect bounds = (**pixMapH1).bounds;

		// make sure we have a GWorld into which we can put the comparison info
	if (mComparisonGWorld == nil) {
			// create a GWorld the same size and depth as the left test window
		mComparisonGWorld = CreateGWorld(bounds, pixelDepth);
	} else {
			// see if the existing comparison GWorld is the same size as the
			// the test windows 
		PixMapHandle comparisonPixMapH = ::GetGWorldPixMap(mComparisonGWorld->GetMacGWorld());
		
		if (!BlocksAreEqual(&(**comparisonPixMapH).bounds, &bounds, sizeof(Rect)) ||
				(**comparisonPixMapH).pixelSize != pixelDepth) {
				// the windows must have changed size or bit-depth since the last 
				// time we extracted any channel info from them, so we have to 
				// create a new GWorld of the appropriate size and depth.  
			delete mComparisonGWorld;
			mComparisonGWorld = CreateGWorld(bounds, pixelDepth);
		}
	}
		
	if (mComparisonGWorld == nil) {
			// we don't have enough memory for the GWorld, so just return,
			// since we can't do the comparison
		S3StopAlert(STRx_AlertStrings_ID, AlertString_CouldntCompareWindows, nil);
		return;
	}

	PixMapHandle pixMapH3 = ::GetGWorldPixMap(mComparisonGWorld->GetMacGWorld());
	
		// for now, we'll show the comparison GWorld when we get redrawn.  this
		// will probably get changed using SetCompareMode before the window
		// is actually redrawn the first time.
	mGWorld = mComparisonGWorld;

	short width = inGWorld1->portRect.right;
	short height = inGWorld1->portRect.bottom;

		// make our window as big as the test windows
	ResizeTo(width, height);

		// find the absolute difference between the argb values in inGWorld1 and 
		// inGWorld2 and put that value into mComparisonGWorld
	MapPixels(&CComparisonWindow::DifferencePixel, inGWorld1, inGWorld2, mComparisonGWorld->GetMacGWorld());
	
	return;
}	


// ===========================================================================
	void
CComparisonWindow::MapPixels(
	PixelFunctionPtr inFunction,
	GWorldPtr	inGWorld1,
	GWorldPtr	inGWorld2,
	GWorldPtr	inGWorld3)
{
	Assert_(inGWorld1 != nil);
	
		// if either of the other two GWorlds are nil, we just set them 
		// equal to the first one so that we can use all three pointers
		// without worrying about whether any of them are nil.  it's up 
		// to the caller to make sure the function passed in has the 
		// pointers it expects.
	if (inGWorld2 == nil) {
		inGWorld2 = inGWorld1;
	}
	
	if (inGWorld3 == nil) {
		inGWorld3 = inGWorld1;
	}
	
	PixMapHandle pixMapH1 = ::GetGWorldPixMap(inGWorld1);
	PixMapHandle pixMapH2 = ::GetGWorldPixMap(inGWorld2);
	PixMapHandle pixMapH3 = ::GetGWorldPixMap(inGWorld3);
	
	Assert_((**pixMapH1).pixelSize == (**pixMapH2).pixelSize);
	Assert_((**pixMapH2).pixelSize == (**pixMapH3).pixelSize);

	Assert_((**pixMapH1).bounds.right == (**pixMapH2).bounds.right);
	Assert_((**pixMapH1).bounds.bottom == (**pixMapH2).bounds.bottom);
	Assert_((**pixMapH2).bounds.right == (**pixMapH3).bounds.right);
	Assert_((**pixMapH2).bounds.bottom == (**pixMapH3).bounds.bottom);

	Uint8* lineStart1 = (Uint8*) ::GetPixBaseAddr(pixMapH1);
	Uint8* lineStart2 = (Uint8*) ::GetPixBaseAddr(pixMapH2);
	Uint8* lineStart3 = (Uint8*) ::GetPixBaseAddr(pixMapH3);
	
	short width = (**pixMapH1).bounds.right;
	short height = (**pixMapH1).bounds.bottom;
	
	short rowBytes1 = PixMapRowBytes(pixMapH1);
	short rowBytes2 = PixMapRowBytes(pixMapH2);
	short rowBytes3 = PixMapRowBytes(pixMapH3);

		// the argb values for each GWorld will be separated out into
		// 3 element arrays, one for each channel
	Uint8 a[3] = { 0, 0, 0 };
	Uint8 r[3] = { 0, 0, 0 };
	Uint8 g[3] = { 0, 0, 0 };
	Uint8 b[3] = { 0, 0, 0 };
	
	if ((**pixMapH1).pixelSize == 32) {
		const short bytesPerPixel = 4;
		
		register Uint8* pixel1 = (Uint8*) lineStart1;
		register Uint8* pixel2 = (Uint8*) lineStart2;
		register Uint8* pixel3 = (Uint8*) lineStart3;
	
		for (short y = 0; y < height; y++) {
			for (short x = 0; x < width; x++) {
					// extract the argb channels for each pixel
				a[0] = pixel1[0];
				r[0] = pixel1[1];
				g[0] = pixel1[2];
				b[0] = pixel1[3];

				a[1] = pixel2[0];
				r[1] = pixel2[1];
				g[1] = pixel2[2];
				b[1] = pixel2[3];

				a[2] = pixel3[0];
				r[2] = pixel3[1];
				g[2] = pixel3[2];
				b[2] = pixel3[3];

					// apply inFunction to this pixel
				(this->*inFunction)(a, r, g, b);
				
					// put the presumably changed argb values back into each pixel
				pixel1[0] = a[0];
				pixel1[1] = r[0];
				pixel1[2] = g[0];
				pixel1[3] = b[0];

				pixel2[0] = a[1];
				pixel2[1] = r[1];
				pixel2[2] = g[1];
				pixel2[3] = b[1];

				pixel3[0] = a[2];
				pixel3[1] = r[2];
				pixel3[2] = g[2];
				pixel3[3] = b[2];

					// move to the next pixel
				pixel1 += bytesPerPixel;
				pixel2 += bytesPerPixel;
				pixel3 += bytesPerPixel;
			} // endfor
			
				// move to the next row of pixels
			lineStart1 += rowBytes1;
			lineStart2 += rowBytes2;
			lineStart3 += rowBytes3;
			
				// point to the pixels at the start of the next row
			pixel1 = lineStart1;
			pixel2 = lineStart2;
			pixel3 = lineStart3;
		} // endfor
	} else if ((**pixMapH1).pixelSize == 16) {
		register Uint16* pixel1 = (Uint16*) lineStart1;
		register Uint16* pixel2 = (Uint16*) lineStart2;
		register Uint16* pixel3 = (Uint16*) lineStart3;

		for (short y = 0; y < height; y++) {
			for (short x = 0; x < width; x++) {
					// extract the argb channels for each pixel.  we convert a 1 in
					// the alpha channel to 0x1F (the max RGB value in 16-bit) in
					// case inFunction averages the pixels.  that way, if all the
					// channels are white, they will average to white, rather than
					// to (1 + 0x1F + 0x1F + 0x1F) / 4.
				a[0] = (*pixel1 & kARGB16mask_Alpha) ? 0x1F : 0;
				r[0] = (*pixel1 & kARGB16mask_Red) >> kARGB16shift_Red;
				g[0] = (*pixel1 & kARGB16mask_Green) >> kARGB16shift_Green;
				b[0] = (*pixel1 & kARGB16mask_Blue) >> kARGB16shift_Blue;
				
				a[1] = (*pixel2 & kARGB16mask_Alpha) ? 0x1F : 0;
				r[1] = (*pixel2 & kARGB16mask_Red) >> kARGB16shift_Red;
				g[1] = (*pixel2 & kARGB16mask_Green) >> kARGB16shift_Green;
				b[1] = (*pixel2 & kARGB16mask_Blue) >> kARGB16shift_Blue;
				
				a[2] = (*pixel3 & kARGB16mask_Alpha) ? 0x1F : 0;
				r[2] = (*pixel3 & kARGB16mask_Red) >> kARGB16shift_Red;
				g[2] = (*pixel3 & kARGB16mask_Green) >> kARGB16shift_Green;
				b[2] = (*pixel3 & kARGB16mask_Blue) >> kARGB16shift_Blue;
				
					// apply inFunction to this pixel
				(this->*inFunction)(a, r, g, b);
				
					// put the presumably changed argb values back into each pixel
				*pixel1 = 0;
				*pixel1 |= (a[0] << kARGB16shift_Alpha) & kARGB16mask_Alpha;
				*pixel1 |= (r[0] << kARGB16shift_Red) & kARGB16mask_Red;
				*pixel1 |= (g[0] << kARGB16shift_Green) & kARGB16mask_Green;
				*pixel1 |= (b[0] << kARGB16shift_Blue) & kARGB16mask_Blue;
				
				*pixel2 = 0;
				*pixel2 |= (a[1] << kARGB16shift_Alpha) & kARGB16mask_Alpha;
				*pixel2 |= (r[1] << kARGB16shift_Red) & kARGB16mask_Red;
				*pixel2 |= (g[1] << kARGB16shift_Green) & kARGB16mask_Green;
				*pixel2 |= (b[1] << kARGB16shift_Blue) & kARGB16mask_Blue;
				
				*pixel3 = 0;
				*pixel3 |= (a[2] << kARGB16shift_Alpha) & kARGB16mask_Alpha;
				*pixel3 |= (r[2] << kARGB16shift_Red) & kARGB16mask_Red;
				*pixel3 |= (g[2] << kARGB16shift_Green) & kARGB16mask_Green;
				*pixel3 |= (b[2] << kARGB16shift_Blue) & kARGB16mask_Blue;
				
					// move to the next pixel
				pixel1++;
				pixel2++;
				pixel3++;
			} // endfor
			
				// move to the next row of pixels
			lineStart1 += rowBytes1;
			lineStart2 += rowBytes2;
			lineStart3 += rowBytes3;
			
				// point to the pixels at the start of the next row
			pixel1 = (Uint16*) lineStart1;
			pixel2 = (Uint16*) lineStart2;
			pixel3 = (Uint16*) lineStart3;
		} // endfor
	}
}


// ===========================================================================
	void
CComparisonWindow::ProcessChannel(PixelFunctionPtr inFunction)
{
	if (mComparisonGWorld == nil) {
			// no windows have been compared yet, so we have nothing to process
		return;
	}
	
	PixMapHandle comparisonPixMapH = ::GetGWorldPixMap(mComparisonGWorld->GetMacGWorld());

		// make sure we have a GWorld into which we can put the channel info
	if (mChannelGWorld == nil) {
		mChannelGWorld = CreateGWorld(
			(**comparisonPixMapH).bounds, (**comparisonPixMapH).pixelSize);
	} else {
			// see if the existing channel GWorld is the same size as the
			// comparison GWorld
		PixMapHandle channelPixMapH = ::GetGWorldPixMap(mChannelGWorld->GetMacGWorld());
		
		if (!BlocksAreEqual(&(**channelPixMapH).bounds, &(**comparisonPixMapH).bounds, sizeof(Rect)) ||
				(**channelPixMapH).pixelSize != (**comparisonPixMapH).pixelSize) {
				// the windows must have changed size or bit-depth since the last 
				// time we extracted any channel info from them, so we have to 
				// create a new GWorld of the appropriate size and depth.  
			delete mChannelGWorld;
			mChannelGWorld = CreateGWorld(
				(**comparisonPixMapH).bounds, (**comparisonPixMapH).pixelSize);
		}
	}
		
	if (mChannelGWorld == nil) {
			// the GWorld allocation we did above was unsuccessful
		S3StopAlert(STRx_AlertStrings_ID, AlertString_CouldntCompareWindows, nil);
		return;
	}

		// apply inFunction to each pixel in mComparisonGWorld and mChannelGWorld
	MapPixels(inFunction, mComparisonGWorld->GetMacGWorld(), 
						mChannelGWorld->GetMacGWorld());
}


// ===========================================================================
	void
CComparisonWindow::ResizeTo(short inWidth, short inHeight)
{
		// make the window be inWidth by inHeight pixels
	Rect newBounds = mUserBounds;
	newBounds.right = newBounds.left + inWidth;
	newBounds.bottom = newBounds.top + inHeight;
		
		// resize the window on screen
	DoSetBounds(newBounds);
}


// ===========================================================================
	void
CComparisonWindow::ResizeAndCenter(const Rect& inRect)
{
		// find the device on which this rect's upper-right corner lies
	GDHandle theDevice = S3GetDeviceFromRect(&inRect, nil);
	Assert_(theDevice);
	
	Rect deviceRect = (**theDevice).gdRect;
	
	short deviceWidth = (deviceRect.right - deviceRect.left);
	short deviceHeight = (deviceRect.bottom - deviceRect.top);
	
		// we want the window to be the same size as inRect
	short windowWidth = inRect.right - inRect.left;
	short windowHeight = inRect.bottom - inRect.top;
	
		// center the window on the screen
	Rect newBounds;
	newBounds.left = deviceRect.left + (deviceWidth - windowWidth) / 2;
	newBounds.right = newBounds.left + windowWidth;
	newBounds.top = deviceRect.top + (deviceHeight - windowHeight) / 2;
	newBounds.bottom = newBounds.top + windowHeight;
		
		// move the window to its new coords
	DoSetBounds(newBounds);
}


// ===========================================================================
	void
CComparisonWindow::DrawSelf()
{
	CTinselTestWindow* testWindow;
	short copyMode = srcCopy;
	
	switch (mCompareMode) {
		case eCompare_Red:
		case eCompare_Green:
		case eCompare_Blue:
  	case eCompare_Alpha:
  	case eCompare_Average:
			mGWorld = mChannelGWorld;
			break;
			
			// if the mode is Left or Right, we set mGWorld to point to the
			// appropriate test window now, rather than in SetCompareMode,
			// so that we always get the most up-to-date GWorld.  otherwise,
			// the GWorld might be deleted between the call to SetCompareMode
			// and now.
		case eCompare_Left:
			testWindow = CTinselTestWindow::GetLeftWindow();
			if (testWindow != nil) {
				mGWorld = testWindow->GetGWorld();
			}
			copyMode = srcCopy + ditherCopy;
			break;
			
		case eCompare_Right:
			testWindow = CTinselTestWindow::GetRightWindow();
			if (testWindow != nil) {
				mGWorld = testWindow->GetGWorld();
			}
			copyMode = srcCopy + ditherCopy;
			break;
	}
	
	FocusDraw();
	
		// let the window do any setup it needs to before we draw into it
	LWindow::DrawSelf();

	if (mGWorld != nil) {
		Rect bounds;
		
		CalcLocalFrameRect(bounds);
		
			// make sure the colors are set correctly for the CopyBits
		::ForeColor(blackColor);
		::BackColor(whiteColor);
		
		mGWorld->CopyImage(GetMacPort(), bounds, copyMode);
	}
}


// ===========================================================================
	LGWorld*
CComparisonWindow::CreateGWorld(Rect& inRect, Int16 inPixelDepth)
{
	LGWorld* theGWorld = nil;
	
	try {
			// try to create a new offscreen GWorld in which to copy the window's 
			// current contents.  
		theGWorld = new LGWorld(inRect, inPixelDepth);
	}

		// we catch all errors here because that's the only way it seems to work
		// with the exceptions thrown by LGWorld
	catch(...) {
		if (theGWorld == nil) {
				// there isn't enough memory to allocate the GWorld in our heap, so
				// try putting it in temporary memory
			theGWorld = new LGWorld(inRect, inPixelDepth, useTempMem);
		}
	}		
	
	return theGWorld;
}


// ===========================================================================
	void
CComparisonWindow::SetCompareMode(ECompareMode inCompareMode)
{
	mCompareMode = inCompareMode;

	switch (inCompareMode) {
		case eCompare_Red:
			SetDescriptor("\pComparison : Red");
			ProcessChannel(&CComparisonWindow::ExtractRed);
			mIsAlternating = false;
			break;
		
		case eCompare_Green:
			SetDescriptor("\pComparison : Green");
			ProcessChannel(&CComparisonWindow::ExtractGreen);
			mIsAlternating = false;
			break;
		
		case eCompare_Blue:
			SetDescriptor("\pComparison : Blue");
			ProcessChannel(&CComparisonWindow::ExtractBlue);
			mIsAlternating = false;
			break;
		
		case eCompare_Alpha:
			SetDescriptor("\pComparison : Alpha");
			ProcessChannel(&CComparisonWindow::ExtractAlpha);
			mIsAlternating = false;
			break;
		
		case eCompare_Average:
			SetDescriptor("\pComparison : Average");
			ProcessChannel(&CComparisonWindow::AveragePixel);
			mIsAlternating = false;
			break;
		
		case eCompare_Left: {
				// get the left window's title
			CTinselTestWindow* testWindow = CTinselTestWindow::GetLeftWindow();
			Assert_(testWindow);
			LStr255 windowTitle;
			testWindow->GetDescriptor(windowTitle);
			
				// append that to the title of this window
			windowTitle = "\pComparison : " + windowTitle;
			SetDescriptor(windowTitle);
			mIsAlternating = false;
			break;
		}
		
		case eCompare_Right: {
				// get the right window's title
			CTinselTestWindow* testWindow = CTinselTestWindow::GetRightWindow();
			Assert_(testWindow);
			LStr255 windowTitle;
			testWindow->GetDescriptor(windowTitle);
			
				// append that to the title of this window
			windowTitle = "\pComparison : " + windowTitle;
			SetDescriptor(windowTitle);
			mIsAlternating = false;
			break;
		}
			
	  case eCompare_Alternate:
			if (mIsAlternating) {
				mIsAlternating = false;
				StopIdling();
			} else {
				mIsAlternating = true;
				StartIdling();
					
					// we call this manually right now so that the mode, mLastSwitchTime
					// and stuff get set up correctly
				EventRecord dummyEvent;
				SpendTime(dummyEvent);
			}
			break;
			
		default:
			SignalPStr_("\pFound undefined ECompareMode in CComparisonWindow::SetCompareMode.");
			mCompareMode = eCompare_Average;
			break;
	} // endswitch

	DrawSelf();
}




