// ===========================================================================
//	CSettingsDoc.cp
// ===========================================================================


#include <limits.h>
#include <TextUtils.h>
#include <QuickDraw.h>
#include <LWindow.h>
#include <LTextEdit.h>
#include <LFile.h>
#include <LPrintout.h>
#include <LPlaceHolder.h>
#include <UMemoryMgr.h>
#include <UWindows.h>
#include <String_Utils.h>
#include <UDebugging.h>
#include <UDesktop.h>

#include "CSettingsDoc.h"
#include "CSneak3viewApp.h"
#include "CComparisonWindow.h"
#include "CSmartTextEdit.h"
#include "CTestPopupMenu.h"
#include "CTestInfo.h"
#include "S3TestC++Interface.h"
#include "S3Alert.h"
#include "S3Messages.h"
#include "StringConversion.h"
#include "S3Utilities.h"
#include "S3TestUtil.h"
#include "UWindowState.h"
#include "LTabViewAlternator.h"
#include "CNewKeyScrollAttachment.h"


const OSType	Sneak3view_SettingsResourceType		= 'S3VS';
const ResIDT	S3VS_Version_ID				= 500;
const ResIDT	S3VS_Settings_ID			= 1000;
const ResIDT	S3VS_CurrentTab_ID		= 2000;
const unsigned char	S3VS_Version_Name[] = "\pVersion";
const unsigned char	S3VS_Settings_Name[] = "\pSettings";
const unsigned char	S3VS_CurrentTab_Name[] = "\pCurrent Tab";

const ResIDT	WIND_DrawContextSettings		= 1000;
const ResIDT	STRx_UntitledSettings		= 400;
const ResIDT	STRx_Sneak3viewPrefs		= 500;

const OSType	Sneak3view_Creator		= 'SN3V';

const PaneIDT	paneID_Scroller		= 1220;
const PaneIDT	paneID_TextEdit		= 1230;
const PaneIDT	paneID_ViewAlternator		= 1001;


// ===========================================================================
	CSettingsDoc*
CSettingsDoc::GetFrontMost()
{
		// init the smallestIndex to a big number so the first time we compare it 
		// to a window index, the window index gets saved
	Int16 smallestIndex = SHRT_MAX;
	
	Int16 currentIndex = 0;
	CSmartSingleDoc* frontmostDoc = nil;
	
		// iterate through all the documents currently open, looking for the 
		// second one from the front
	LListIterator iterator(sDocumentList, iterate_FromStart);
	CSmartSingleDoc* theDoc = nil;
	
	while (iterator.Next(&theDoc)) {
		if (theDoc->GetClassID() == 'Sdoc') {
				// this doc is a CSettingsDoc, so get its window's index
			currentIndex = theDoc->GetWindowIndex();

			if (currentIndex > 0 && currentIndex < smallestIndex) {
					// the index for this doc's window was found (it's greater than 0)
					// and it's smaller than the current smallest, so this is the 
					// frontmost doc that we've found so far
				smallestIndex = currentIndex;
				frontmostDoc = theDoc;
			}
		}
	} // endwhile 

	return (CSettingsDoc*) frontmostDoc;
}


// ===========================================================================
	Boolean
CSettingsDoc::GetPrefsFSSpec(FSSpec& outFSSpec)
{
	Str255 prefsFileName;
	::GetIndString(prefsFileName, STRx_Sneak3viewPrefs, 1);

		// create an FSSpec pointing to where our prefs file should be
	OSErr err = 
		::FindFolder(kOnSystemDisk, 'pref', kCreateFolder, 
			&(outFSSpec.vRefNum), &(outFSSpec.parID));
	err = 
		::FSMakeFSSpec(outFSSpec.vRefNum, outFSSpec.parID, prefsFileName, &outFSSpec);

		// return true if the FSSpec we just made points to an existing
		// file, false otherwise.  this lets the caller know if they can
		// go ahead and open the file or have to create a new one.
	return (err == noErr);
}


// ---------------------------------------------------------------------------
//		 CSettingsDoc(LCommander*, FSSpec*)
// ---------------------------------------------------------------------------
//	Construct a TextDoc associated with the specified file
//
//	If inFileSpec is nil, then create an empty, untitled document
CSettingsDoc::CSettingsDoc(CSneak3viewApp* inSuper, FSSpec* inFileSpec)
	: CSmartSingleDoc(inSuper)
{
		// Create window for our document
	mWindow = CSettingsWindow::CreateSettingsWindow(WIND_DrawContextSettings, this);
	Assert_(mWindow != nil);

		// store a pointer to our test popup menu
	mTestMenu = (CTestPopupMenu*)	mWindow->FindPaneByID(paneID_Test);
	Assert_(mTestMenu != nil);

		// store a pointer to our TextEdit pane where tests can print stuff
	mTextView = (CSmartTextEdit*)	mWindow->FindPaneByID(paneID_TextEdit);
	Assert_(mTextView != nil);

		// create and add an attachment which will allow scrolling the text
		// view with the page up/down keys.  when it's deleted, the text view 
		// will delete the attachment.
	CNewKeyScrollAttachment* keyScrollerP = new CNewKeyScrollAttachment(mTextView);
	mTextView->AddAttachment(keyScrollerP);
	
		// link us and our CTestInfo object to the settings window 
	mCTestInfo.LinkSettingsDoc(this, mWindow);
	
		// link the TinselTest windows to our CTestInfo object, and vice versa
	mCTestInfo.LinkTestWindows(inSuper->GetLeftTinselTestWindow(),
														 inSuper->GetRightTinselTestWindow());
	
	if (inFileSpec == nil) {
			// Set name of untitled window
		NameNewDoc("\p");
		
			// since this is a brand new window, the CTestInfo ReadFromHandle
			// method won't get called, so the following Update method also won't
			// get called.  so we call them explicitly here, to make sure the 
			// TinselTest windows reflect the default settings in this new, topmost 
			// document window
		mCTestInfo.UpdateTestWindows();
		
		LTabViewAlternator* alternator = 
			(LTabViewAlternator*)	mWindow->FindPaneByID(paneID_ViewAlternator);
		
			// have the alternator show the first view
		alternator->ShowView(1);
	} else {
			// Display contents of file in window
		OpenFile(*inFileSpec);
	}
	
		// the Initially Visible flag is not set for the settings window, so 
		// we have to explicitly show it.  we do it here after OpenFile so that
		// it's position can be restored by OpenFile before it's visible
	mWindow->Show();
}


// ===========================================================================
CSettingsDoc::CSettingsDoc(const CSettingsDoc& inSettingsDoc)
	: CSmartSingleDoc(inSettingsDoc.mSuperCommander),
		mCTestInfo(inSettingsDoc.mCTestInfo)
{
		// Create window for our document
	mWindow = CSettingsWindow::CreateSettingsWindow(WIND_DrawContextSettings, this);
	Assert_(mWindow != nil);

		// store a pointer to our test popup menu
	mTestMenu = (CTestPopupMenu*)	mWindow->FindPaneByID(paneID_Test);
	Assert_(mTestMenu != nil);

		// store a pointer to our TextEdit pane where tests can print stuff
	mTextView = (CSmartTextEdit*)	mWindow->FindPaneByID(paneID_TextEdit);
	Assert_(mTextView != nil);

		// create and add an attachment which will allow scrolling the text
		// view with the page up/down keys.  when it's deleted, the text view 
		// will delete the attachment.
	CNewKeyScrollAttachment* keyScrollerP = new CNewKeyScrollAttachment(mTextView);
	mTextView->AddAttachment(keyScrollerP);
	
		// link our CTestInfo object to the settings window 
	mCTestInfo.LinkSettingsDoc(this, mWindow);
	
		// link the TinselTest windows to our CTestInfo object, and vice versa
	CSneak3viewApp* superCommander = (CSneak3viewApp*) inSettingsDoc.mSuperCommander;
	mCTestInfo.LinkTestWindows(
		superCommander->GetLeftTinselTestWindow(),
		superCommander->GetRightTinselTestWindow());

		// name the new window by appending "Copy" to the original name
	LStr255 windowName;
	inSettingsDoc.mWindow->GetDescriptor(windowName);
	windowName += "\p Copy";
	NameNewDoc(windowName);
//	mWindow->SetDescriptor(windowName);

		// get the original window's size
	SDimension16 originalSize;
	inSettingsDoc.mWindow->GetFrameSize(originalSize);

		// get our window's current global bounds
	Rect windowBounds;
	mWindow->CalcPortFrameRect(windowBounds);
	mWindow->PortToGlobalPoint(topLeft(windowBounds));
	mWindow->PortToGlobalPoint(botRight(windowBounds));

		// make it the same size as the original 
	windowBounds.right = windowBounds.left + originalSize.width;
	windowBounds.bottom = windowBounds.top + originalSize.height;
	mWindow->DoSetBounds(windowBounds);
	
		// since this is a brand new window, the CTestInfo ReadFromHandle
		// method won't get called, so the following Update method also won't
		// get called.  so we call them explicitly here, to make sure the 
		// TinselTest windows reflect the default settings in this new, topmost 
		// document window
	mCTestInfo.UpdateTestWindows();
	
		// find the current view of the original's tab view alternator
	LTabViewAlternator* alternator = 
		(LTabViewAlternator*)	inSettingsDoc.mWindow->FindPaneByID(paneID_ViewAlternator);
	
	Int32 originalViewIndex = alternator->GetCurrentViewIndex();
	
		// now show the same view in our tab view alternator
	alternator = 
		(LTabViewAlternator*)	mWindow->FindPaneByID(paneID_ViewAlternator);
	
	alternator->ShowView(originalViewIndex);

		// the Initially Visible flag is not set for the settings window, so 
		// we have to explicitly show it
	mWindow->Show();
}


// ===========================================================================
CSettingsDoc::~CSettingsDoc()
{
	Str255 windowTitle, defaultWindowTitle;
	GetDescriptor(windowTitle);
	::GetIndString(defaultWindowTitle, STRx_UntitledSettings, 1);
	
	if (EqualString(windowTitle, defaultWindowTitle, false, false)) {
		::GetIndString(windowTitle, STRx_Sneak3viewPrefs, 1);
		
		FSSpec prefsSpec;
		OSErr err = 
			::FindFolder(kOnSystemDisk, 'pref', kCreateFolder, 
				&(prefsSpec.vRefNum), &(prefsSpec.parID));
		err = 
			::FSMakeFSSpec(prefsSpec.vRefNum, prefsSpec.parID, windowTitle, &prefsSpec);
		
		delete mFile;
		mFile = new LFile(prefsSpec);
		
		if (err == fnfErr) {
				// Make new file on disk
			mFile->CreateNewFile(Sneak3view_Creator, fileType, 0);
		}
		
		SaveTestInfo();
	}
	
		// save the window position before it closes
	UWindowState::SaveWindowState(mFile, mWindow);
}


// ---------------------------------------------------------------------------
//		 NameNewDoc
// ---------------------------------------------------------------------------
//	Name a new, untitled document window
//
//	Untitled windows start with "untitled", then "untitled 1",
//	"untitled 2", etc. Old numbers are reused, so there won't be
//	gaps in the numbering.
//
//	This routine uses a STR# resource to store the "untitled" string,
//	which can be localized to different languages. The first string
//	is "untitled" and the second is "untitled " (trailing space),
//	which is used when appending a number to the name.
	void
CSettingsDoc::NameNewDoc(Str255 inName)
{
	LStr255	baseName;
	
	if (inName[0] == 0) {
			// Start with the default name ("Untitled Settings")
		::GetIndString(baseName, STRx_UntitledSettings, 1);
	} else {
		baseName = inName;
	}
	
	LStr255 name = baseName;
	
	long	num = 1;
	while (UWindows::FindNamedWindow(name) != nil) {
			// An existing window has the current name
			// Increment counter and try again
		name = baseName + "\p " + LStr255(num++);
	}		
	
	mWindow->SetDescriptor(name);		// Finally, set window title
}


// ---------------------------------------------------------------------------
//		 OpenFile
// ---------------------------------------------------------------------------
//	Open a new document for the specified File
	void
CSettingsDoc::OpenFile(FSSpec	&inFileSpec)
{
		// save the current resource file so that we can restore it later
	Int16 savedResourceRefNum = ::CurResFile();
	
		// Create a new File object, read the entire File contents,
		// put the contents into the text view, and set the Window
		// title to the name of the File.
	Try_ {
			// create a new LFile to operate on the CSettingsDoc we want to open
		mFile = new LFile(inFileSpec);

			// open the document's resource fork for reading
		mFile->OpenResourceFork(fsRdPerm);
		
			// get the Settings resource version from the resource fork we just opened
			// (and don't look for it anywhere else)
		StResource versionH(Sneak3view_SettingsResourceType, S3VS_Version_ID,
												false, true);

			// the default version number for the settings file
		Int16 settingsVersion = 0;
		
		if (Handle(versionH) != nil) {
				// we found a version resource, so update settingsVersion
			settingsVersion = **(Int16**) Handle(versionH);
		}

			// get the Settings resource from the resource fork we just opened
			// (and don't look for it anywhere else)
		StResource settingsH(Sneak3view_SettingsResourceType, S3VS_Settings_ID,
												 false, true);

			// if we found a settings resource, pass it and the version number 
			// to the CTestInfo so it can extract the values and set its own 
			// members appropriately.  if the handle is nil, we don't do 
			// anything, since the CTestInfo will have been initialized in 
			// its constructor.  ReadFromHandle may return false if it can't 
			// handle the settings version, so throw an error if it does.
		if (Handle(settingsH) != nil) {
			if (mCTestInfo.ReadFromHandle(Handle(settingsH), settingsVersion) == false) {
				Throw_(1);
			}
		}

			// get the settings resource which contains the ID of the currently
			// selected view in the tab view alternator
		StResource currentTabIndexH(Sneak3view_SettingsResourceType, S3VS_CurrentTab_ID,
														 false, true);
														 
			// default to showing the first view
		Int32 currentTabIndex = 1;
		
		if (Handle(currentTabIndexH) != nil) {
				// the resource contains just a single long, which is the index
				// of the view to be shown
			currentTabIndex = **(Int32**) Handle(currentTabIndexH);
		} 
			
		LTabViewAlternator* alternator = 
			(LTabViewAlternator*)	mWindow->FindPaneByID(paneID_ViewAlternator);
		
			// have the alternator show the same view as was shown the last
			// time this file was saved (if the resource was found) or else the 
			// first one.  it will ignore bogus tab indexes.
		alternator->ShowView(currentTabIndex);
		
			// we're done reading from the doc's resource fork
		mFile->CloseResourceFork();
	
		// restore the previous resource file
	::UseResFile(savedResourceRefNum);
		
			// read the text from the data fork
		mFile->OpenDataFork(fsRdWrPerm);
		Handle textH = mFile->ReadDataFork();

		if (textH != nil) {
			if (::GetHandleSize(textH) > 0) {
					// there was text in the data fork, so put it into 
					// the LTextEdit view
				mTextView->SetTextHandle(textH);
			}

			::DisposeHandle(textH);
		}
		
			// we're done with the data fork
		mFile->CloseDataFork();
		
		Str255 prefsFileName;
		::GetIndString(prefsFileName, STRx_Sneak3viewPrefs, 1);
		
			// restore the window's position
		UWindowState::RestoreWindowState(mFile, mWindow);
		
		if (EqualString(prefsFileName, inFileSpec.name, false, false)) {
				// this is the Sneak3view Prefs which has been opened.  we want
				// the settings saved in this file to appear in an Untitled Settings
				// window, so change the name of the window to that.
			::GetIndString(prefsFileName, STRx_UntitledSettings, 1);
			mWindow->SetDescriptor(prefsFileName);
			
				// since this window came from the prefs file, we want it to act
				// like it's an untitled document, with no attached file yet.  so
				// we delete the file and set mIsSpecified to false, to indicate that
				// it has not been saved yet.
			delete mFile;
			mFile = nil;
			mIsSpecified = false;
		} else {
				// set the windows's title to the name of the file.  we're specified
				// now because the data is associated with a file.
			mWindow->SetDescriptor(inFileSpec.name);
			mIsSpecified = true;
		}
	}
	
	Catch_(inErr) {
			// restore the previous resource file
		::UseResFile(savedResourceRefNum);
	
			// there was an error trying to open this document,
			// so let the user know
		S3StopAlert(ALRT_StopAlert, AlertString_CouldntReadSettings, nil);
	
		delete this;
		
		Throw_(inErr);
	} EndCatch_
}


// ---------------------------------------------------------------------------
//		 IsModified
// ---------------------------------------------------------------------------
//	Return whether the Document has changed since the last save
	Boolean
CSettingsDoc::IsModified()
{
	Str255 windowTitle, defaultWindowTitle;
	GetDescriptor(windowTitle);
	::GetIndString(defaultWindowTitle, STRx_UntitledSettings, 1);
	
		// the document has changed if the CTestInfo object or the text view 
		// is dirty, as long as the window's title isn't the default title.  the
		// window with a default title will get saved as the app's settings
		// and will be loaded back in the next time the app runs.
	mIsModified = (mCTestInfo.IsDirty() || mTextView->IsDirty()) &&
		!EqualString(windowTitle, defaultWindowTitle, false, false);

	return mIsModified;
}


// ---------------------------------------------------------------------------
//		 DoAESave
// ---------------------------------------------------------------------------
//	Save Document in the specified file with the specified file type
//
//	If file type is fileType_Default, use the normal file type for
//	this document
	void
CSettingsDoc::DoAESave(FSSpec &inFileSpec, OSType inFileType)
{
		// Kill existing file
	delete mFile;						
	
		// Make new file object
	mFile = new LFile(inFileSpec);		
	
		// Find proper file type
	OSType	fileType = CSettingsDoc::fileType;			
	
	if (inFileType != fileType_Default) {
		fileType = inFileType;
	}

	try {
			// Make new file on disk
		mFile->CreateNewFile(Sneak3view_Creator, fileType, 0);
		
			// Write out data
		DoSave();							
		
			// Change window name
		mWindow->SetDescriptor(inFileSpec.name);
		
			// Document now has a specified file
		mIsSpecified = true;				
	}
	
	catch(...) {	
			// there was an error trying to save this document to the file,
			// so let the user know
		S3StopAlert(ALRT_StopAlert, AlertString_CouldntSaveSettings, nil);
	}
}


// ---------------------------------------------------------------------------
//		 DoSave
// ---------------------------------------------------------------------------
//	Save the entire Document to its associated File (which must already exist)
	void
CSettingsDoc::DoSave()
{
		// save the test info stuff into the resource fork
	SaveTestInfo();
	
	Handle textH = mTextView->GetTextHandle();
	
	if (textH == nil) {
			// there's nothing to save
		return;
	}
	
		// open the data fork of the file so we can save the text in the 
		// output pane of the Settings doc
	mFile->OpenDataFork(fsRdWrPerm);

		// lock the handle and write it to the file.  the handle might 0-length,
		// but that's okay, because the user might have deleted all the text,
		// so then we also want the datafork to be 0-length.
	StHandleLocker((Handle) textH);
	
	mFile->WriteDataFork(*textH, ::GetHandleSize(textH));

	mFile->CloseDataFork();
	
		// now that we've saved the file, the text view is not dirty
	mTextView->SetDirty(false);
}


// ---------------------------------------------------------------------------
//		 DoRevert
// ---------------------------------------------------------------------------
//	Revert the Document to the last saved version on disk
	void
CSettingsDoc::DoRevert()
{
		// open the document's resource fork for reading
	mFile->OpenResourceFork(fsRdPerm);
	
		// get the Settings resource version from the resource fork we just opened
		// (and don't look for it anywhere else)
	StResource versionH(Sneak3view_SettingsResourceType, S3VS_Version_ID,
											false, true);

		// the default version number for the settings file
	Int16 settingsVersion = 0;
	
	if (Handle(versionH) != nil) {
			// we found a version resource, so update settingsVersion
		settingsVersion = (Int16) **Handle(versionH);
	}

		// get the Settings resource from the resource fork we just opened
		// (and don't look for it anywhere else)
	StResource settingsH(Sneak3view_SettingsResourceType, S3VS_Settings_ID,
											 false, true);

		// if we found a settings resource, pass it to the CTestInfo so it can
		// extract the values and set its own members appropriately.  if the
		// handle is nil, we don't do anything, since the CTestInfo will have
		// been initialized in its constructor.
	if (Handle(settingsH) != nil) {
		mCTestInfo.ReadFromHandle(Handle(settingsH), settingsVersion);
	}
	
		// we're done reading from the doc's resource fork
	mFile->CloseResourceFork();

		// read the text from the data fork
	mFile->OpenDataFork(fsRdWrPerm);
	Handle textH = mFile->ReadDataFork();

	if (textH != nil) {
		if (::GetHandleSize(textH) > 0) {
				// there was text in the data fork, so put it into 
				// the LTextEdit view
			mTextView->SetTextHandle(textH);
		}

		::DisposeHandle(textH);
	}
	
		// we're done with the data fork
	mFile->CloseDataFork();
	
		// now that we've reverted the file, the text view is not dirty
	mTextView->SetDirty(false);

		// update the text view to show the text we just loaded in
	mTextView->Refresh();
}


// ---------------------------------------------------------------------------
//		 DoPrint
// ---------------------------------------------------------------------------
//	Print the contents of the Document
	void
CSettingsDoc::DoPrint()
{
//	LPrintout		*thePrintout = LPrintout::CreatePrintout(prto_TextDoc);
//	thePrintout->SetPrintRecord(mPrintRecordH);
//	LPlaceHolder	*textPlace = (LPlaceHolder*)
//									thePrintout->FindPaneByID('TBox');
//	textPlace->InstallOccupant(mTextView, atNone);
//	
//	thePrintout->DoPrintJob();
//	delete thePrintout;
}


// ===========================================================================
	void
CSettingsDoc::SaveTestInfo()
{
	Boolean foundError = false;
	
		// save the current resource file so that we can restore it later
	Int16 savedResourceRefNum = ::CurResFile();
	
		// open the document's resource fork for writing
	mFile->OpenResourceFork(fsRdWrPerm);

		// get the Settings resource from the resource fork we just opened
		// (and don't look for it anywhere else)
	StResource settingsH(Sneak3view_SettingsResourceType, S3VS_Settings_ID,
											 false, true);

	Int16 settingsVersion = 0;
	
	if (Handle(settingsH) != nil) {
			// we found an already existing resource in this file, so let 
			// the CTestInfo write its data to this handle and set its version
		if (mCTestInfo.WriteToHandle(Handle(settingsH), &settingsVersion) == false) {
			foundError = true;
		}
		
			// let the resource manager know we've changed this resource
		::ChangedResource(Handle(settingsH));
		
		if (::ResError() != noErr) {
			foundError = true;
		}
	} else {		
			// create a handle for the CTestInfo to write into.  we make it 
			// 0 length since CTestInfo will resize it.
		Handle(settingsH) = ::NewHandle(0);
		
		if (Handle(settingsH) != nil) {
				// let the CTestInfo write its data to this handle and set its version
			if (mCTestInfo.WriteToHandle(Handle(settingsH), &settingsVersion) == false) {
				foundError = true;
			}
			
				// add to the resource fork the settings the CTestInfo just 
				// put into the handle  
			::AddResource(Handle(settingsH), Sneak3view_SettingsResourceType,
										S3VS_Settings_ID, S3VS_Settings_Name);
			
			if (::ResError() != noErr) {
				foundError = true;
			}
		} else {
				// we couldn't allocate the new handle
			foundError = true;
		}
	} 

		// get the Settings resource version from the resource fork we just opened
		// (and don't look for it anywhere else)
	StResource versionH(Sneak3view_SettingsResourceType, S3VS_Version_ID,
											false, true);

	if (Handle(versionH) != nil) {
			// the resource already exists, so just change the version number
		**(Int16**) Handle(versionH) = settingsVersion;
		
			// let the resource manager know we've changed this resource
		::ChangedResource(Handle(versionH));
		
		if (::ResError() != noErr) {
			foundError = true;
		}
	} else {
		Handle(versionH) = ::NewHandle(sizeof(Int16));
		
		if (Handle(versionH) != nil) {
				// save the version number in this new handle
			**(Int16**) Handle(versionH) = settingsVersion;
			
				// add this resource to the resource fork
			::AddResource(Handle(versionH), Sneak3view_SettingsResourceType,
										S3VS_Version_ID, S3VS_Version_Name);
			
			if (::ResError() != noErr) {
				foundError = true;
			}
		} else {
				// we couldn't allocate the new handle
			foundError = true;
		}
	}
	
		// get the settings resource which contains the ID of the currently
		// selected view in the tab view alternator
	StResource currentTabIndexH(Sneak3view_SettingsResourceType, S3VS_CurrentTab_ID,
													 false, true);
													 
	LTabViewAlternator* alternator = 
		(LTabViewAlternator*)	mWindow->FindPaneByID(paneID_ViewAlternator);
	
		// get the currently displayed view from the alternator
	Int32 currentTabIndex = alternator->GetCurrentViewIndex();

	if (Handle(currentTabIndexH) != nil) {
			// the resource already exists, so just change the index
		**(Int32**) Handle(currentTabIndexH) = currentTabIndex;
		
			// let the resource manager know we've changed this resource
		::ChangedResource(Handle(currentTabIndexH));
		
		if (::ResError() != noErr) {
			foundError = true;
		}
	} else {
		Handle(currentTabIndexH) = ::NewHandle(sizeof(Int32));
		
		if (Handle(currentTabIndexH) != nil) {
				// save the index in this new handle
			**(Int32**) Handle(currentTabIndexH) = currentTabIndex;
			
				// add this resource to the resource fork
			::AddResource(Handle(currentTabIndexH), Sneak3view_SettingsResourceType,
										S3VS_CurrentTab_ID, S3VS_Settings_Name);
			
			if (::ResError() != noErr) {
				foundError = true;
			}
		} else {
				// we couldn't allocate the new handle
			foundError = true;
		}
	}
	
	if (Handle(settingsH) != nil || Handle(currentTabIndexH) != nil) {
			// update our resource fork
		::UpdateResFile(mFile->GetResourceForkRefNum());
		
		if (::ResError() != noErr) {
			foundError = true;
		}
	}
		
		// we're done writing to the doc's resource fork
	mFile->CloseResourceFork();
	
		// restore the previous resource file
	::UseResFile(savedResourceRefNum);
	
	if (foundError) {
			// let the calling function know that something went wrong
		throw(-1);
	}
}


// ---------------------------------------------------------------------------
//		 PutOnDuty
// ---------------------------------------------------------------------------
//	A Commander is going on duty
//
//	Subclasses should override this function if they wish to behave
//	differently when on duty than when off duty
	void
CSettingsDoc::PutOnDuty()
{
		// now that we're the front-most document, make sure the TinselTest 
		// windows are sized according to the values stored in our CTestInfo object
	mCTestInfo.UpdateTestWindows();
}


// ---------------------------------------------------------------------------
//		 ObeyCommand
// ---------------------------------------------------------------------------
//	Respond to commands

	Boolean
CSettingsDoc::ObeyCommand(
	CommandT	inCommand,
	void		*ioParam)
{
	Boolean	cmdHandled = true;
	
	switch (inCommand) {
		case cmd_DuplicateSettingsDoc: {
				// duplicate this doc using our copy constructor
			CSettingsDoc* newSettingsDoc = new CSettingsDoc(*this);
			break;
		}

		case cmd_RunTest:
			RunTest();
			break;

		default:
			cmdHandled = inherited::ObeyCommand(inCommand, ioParam);
			break;
	}
	
	return cmdHandled;
}


// ---------------------------------------------------------------------------
//		 FindCommandStatus
// ---------------------------------------------------------------------------
//	Return whether a Command is enabled and/or marked (in a Menu)
	void
CSettingsDoc::FindCommandStatus(
	CommandT	inCommand,
	Boolean		&outEnabled,
	Boolean		&outUsesMark,
	Char16		&outMark,
	Str255		outName)
{
	switch (inCommand) {
		case cmd_DuplicateSettingsDoc:
		case cmd_RunTest:
			outEnabled = true;
			break;  
			
			// we don't handle printing
		case cmd_Print:
		case cmd_PrintOne:
		case cmd_PageSetup:
			outEnabled = false;
			break;
			
		default:
			inherited::FindCommandStatus(inCommand, outEnabled,
									outUsesMark, outMark, outName);
			break;
	}
}


// ===========================================================================
	void
CSettingsDoc::RunTest()
{
	S3tTestTool theTest;
	
	if (mTestMenu->GetData(&theTest) == false || theTest.testMethod == nil) {
			// we got a bogus test, so just return
		return;
	}
	
		// we got the test and it has a valid method pointer so
		// tell the app to show and clear the test windows.  (the 
		// cmd_SelectTestWindows will eventually work its way up to the app.)
	LCommander::GetTarget()->ObeyCommand(cmd_SelectTestWindows, nil);
	
		// have our CTestInfo prepare the S3tTestInfo for the test
	S3tTestInfoP testInfoP = mCTestInfo.GetTestInfo();
	
	if (testInfoP == nil) {
			// something went wrong with the test info, so we can't run the test.
			// we don't need to give any kind of alert here, because GetTestInfo
			// will already have done that.
		return;
	}
	
		// we've got a good TestInfo, so show the watch cursor
	::SetCursor(*(::GetCursor(watchCursor)));
	
		// print out the date, time and name of the test
	unsigned long seconds = 0;
	::GetDateTime(&seconds);
	
	Str255 dateTimeString;
	
	::TimeString(seconds, true, dateTimeString, nil);
	S3PtoCString(dateTimeString);
	
	S3Test_Print("\r%s\r%s\r%s on ", 
				 kS3Divider_Equal, 
				 theTest.testName, 
				 dateTimeString);

	::DateString(seconds, longDate, dateTimeString, nil);
	S3PtoCString(dateTimeString);
	
	S3Test_Print("%s\r\r", dateTimeString);
		
		// run the test
	S3tTestStatus result = (*theTest.testMethod)(testInfoP);

		// let the CTestInfo do what it needs to after the test has run
	mCTestInfo.CleanUpAfterTest();

	if (result == kS3Status_Failure) {
		S3Test_Print("*** Failure code returned ***\r\r");
	}
	
//		// print out the time each engine took to run the test
//	char engineName[kMaxEngineNameLength];
//	S3GetEngineName(S3TestInfo_GetLEngine(testInfoP), engineName, 
//					sizeof(engineName));
//					
//	S3Test_Print("\r%s (seconds): %.8f\r", engineName,
//				 (S3TestInfo_GetLMicroseconds(testInfoP) / 1000000.0));
//
//	S3GetEngineName(S3TestInfo_GetREngine(testInfoP), engineName, 
//					sizeof(engineName));
//	S3Test_Print("%s (seconds): %.8f\r\r%s\r", engineName,
//				 (S3TestInfo_GetRMicroseconds(testInfoP) / 1000000.0), 
//				 kS3Divider_Equal);

	S3Test_Print("\r%s\r\r", kS3Divider_Equal);

		// tell the app to save whatever output 
		// the test just created
	CSneak3viewApp::Instance()->SaveTestOutput();
	
		// we're done with the test, so turn the cursor back to an arrow
	::InitCursor();
	
	if (result == kS3Status_Failure) {
			// the test returned a failture code, so alert the user.  we do 
			// this last so that the test output can be saved and the cursor
			// changed to an arrow before we open the alert box
		S3StopAlert(ALRT_StopAlert, AlertString_TestFailed, nil);
	} else {
			// we always call CompareWindows so that if a comparison window
			// is open, it will be updated now that this test has finished. 
			// if no window is open, CompareWindows will do nothing.
		CComparisonWindow::CompareWindows(
			CTinselTestWindow::GetLeftWindow(),
			CTinselTestWindow::GetRightWindow());
	}
}


// ===========================================================================
//		 S3Print
// ===========================================================================
	void 
CSettingsDoc::S3Print(
	char		*inMessage,
	long		inLength)
{
		// append the message text to the end of the TextEdit view in the
		// frontmost CSettingsDoc
	GetFrontMost()->mTextView->AppendText(inMessage, inLength);
}


// ===========================================================================
//		 RegisterTest
// ===========================================================================
	void 
CSettingsDoc::RegisterTest(
	const char			*inName,
	S3tTestPrereqs	inPrereqs,
	S3tTestMethod		inMethod)
{
	S3tTestTool test = { inName, inPrereqs, inMethod };
	
		// the test's name in the menu defaults to a divider.  if inMethod isn't
		// nil, then the test's actual name is put in the menu.  otherwise, 
		// just a divider shows up.
	Str255 menuItem = "\p-";
	
	if (inMethod != nil) {
			// a valid method pointer was passed in, so convert the test's name
			// into a P-string.  
		S3CtoPString(inName, menuItem);
	}
	
		// add the test data to the test popup
	CTestPopupMenu::AppendSharedItem(menuItem, &test);
}


extern "C" {

	
	void 
CSettingsDoc_S3Print(
	char		*inMessage,
	long		inLength)
{
	CSettingsDoc::S3Print(inMessage, inLength);
}


	void 
CSettingsDoc_RegisterTest(
	const char			*inName,
	S3tTestPrereqs	inPrereqs,
	S3tTestMethod		inMethod)
{
	CSettingsDoc::RegisterTest(inName, inPrereqs, inMethod);
}


}