// ===========================================================================
//	CStringPopupMenu.cp
// ===========================================================================
//
// This derivative of a standard popup menu will broadcast the item text
// of the selected item, rather than the item's number.  The current item
// can also be set by giving the menu a string, which it converts to an 
// item number.


#ifdef PowerPlant_PCH
#include PowerPlant_PCH
#endif


#include <PP_Messages.h>

#include "CStringPopupMenu.h"
#include "StringConversion.h"


// ===========================================================================
//  CStringPopupMenu											   CStringPopupMenu 
// ===========================================================================
	CStringPopupMenu*
CStringPopupMenu::CreateStringPopupMenuStream(LStream	*inStream)
{
	return (new CStringPopupMenu(inStream));
}


// ===========================================================================
//		 CStringPopupMenu
// ===========================================================================
//	Construct from input parameters

CStringPopupMenu::CStringPopupMenu(
	const SPaneInfo	&inPaneInfo,
	MessageT		inValueMessage,
	Int16			inTitleOptions,
	ResIDT			inMENUid,
	Int16			inTitleWidth,
	Int16			inPopupVariation,
	ResIDT			inTextTraitsID,
	Str255			inTitle,
	OSType			inResTypeMENU,
	Int16			inInitialMenuItem)
		: CSmartStdPopupMenu(inPaneInfo, inValueMessage, inTitleOptions,
				inMENUid, inTitleWidth, inPopupVariation, inTextTraitsID,
				inTitle, inResTypeMENU, inInitialMenuItem)
{
	InitStringPopupMenu();
}


// ===========================================================================
//		 CStringPopupMenu
// ===========================================================================
//	Construct from input parameters, using a string to set the initial item

CStringPopupMenu::CStringPopupMenu(
	const SPaneInfo	&inPaneInfo,
	MessageT		inValueMessage,
	Int16			inTitleOptions,
	ResIDT			inMENUid,
	Int16			inTitleWidth,
	Int16			inPopupVariation,
	ResIDT			inTextTraitsID,
	Str255			inTitle,
	OSType			inResTypeMENU,
	Str255			inInitialMenuItemString)
		: CSmartStdPopupMenu(inPaneInfo, inValueMessage, inTitleOptions,
				inMENUid, inTitleWidth, inPopupVariation, inTextTraitsID,
				inTitle, inResTypeMENU, ItemFromString(inInitialMenuItemString))
{
	InitStringPopupMenu();
}


// ===========================================================================
//		 CStringPopupMenu
// ===========================================================================
//	Construct from input parameters and an existing ControlHandle

CStringPopupMenu::CStringPopupMenu(
	const SPaneInfo	&inPaneInfo,
	MessageT		inValueMessage,
	Int32			inMaxValue,
	ResIDT			inTextTraitsID,
	ControlHandle	inMacControlH)
		: CSmartStdPopupMenu(inPaneInfo, inValueMessage, inMaxValue,
						inTextTraitsID, inMacControlH)
{
	InitStringPopupMenu();
}


// ===========================================================================
//		 CStringPopupMenu(LStream*)
// ===========================================================================
//	Construct from data in a Stream

CStringPopupMenu::CStringPopupMenu(
	LStream	*inStream)
		: CSmartStdPopupMenu(inStream)
{	
	InitStringPopupMenu();
}


// ===========================================================================
//		 InitStringPopupMenu
// ===========================================================================
//	Private initializer
//
// CSmartStdPopupMenu::InitStrPopupMenu() should handle setting the initial,
// min, and max values

	void
CStringPopupMenu::InitStringPopupMenu()
{
	mBroadcastString = true;
}


// ===========================================================================
	void
CStringPopupMenu::BroadcastValueMessage()
{
	if (mValueMessage != cmd_Nothing) {
		if (mBroadcastString) {
			Str255 theString;
			GetString(theString);
			
			BroadcastMessage(mValueMessage, (void*) theString);
		} else {
			Int32	value = GetValue();

			BroadcastMessage(mValueMessage, (void*) &value);
		}
	}
}


// ===========================================================================
	void
CStringPopupMenu::GetString(Str255 outString) const
{	
	StringFromItem((Int16) GetValue(), outString);
}


// ===========================================================================
	void
CStringPopupMenu::GetString(char* outString) const
{	
	StringFromItem((Int16) GetValue(), outString);
}


// ===========================================================================
	void
CStringPopupMenu::SetValueByString(Str255 inString)
{	
	Int16 index = ItemFromString(inString);
	
	if (index == 0) {
			// this string wasn't found in the menu, so default to the first item
		index = 1;
	}
	
	SetValue(index);
}


// ===========================================================================
	void
CStringPopupMenu::SetValueByString(char* inString)
{	
	Int16 index = ItemFromString(inString);
	
	if (index == 0) {
			// this string wasn't found in the menu, so default to the first item
		index = 1;
	}
	
	SetValue(index);
}


// ===========================================================================
//		 StringFromItem
// ===========================================================================
//	Return the value for a particular Menu item

	void
CStringPopupMenu::StringFromItem(Int16 inIndex, Str255 outString) const
{
	if (inIndex > CountMItems(mMenuH) || inIndex < 1) {
			// the index is out of range, so make the string empty
		outString[0] = 0;
		
		return;
	}
	
	GetMenuItemText(mMenuH, inIndex, outString);
}


// ===========================================================================
//		 StringFromItem
// ===========================================================================
//	Return the value for a particular Menu item

	void
CStringPopupMenu::StringFromItem(Int16 inIndex, char* outString) const
{
	if (inIndex > CountMItems(mMenuH) || inIndex < 1) {
			// the index is out of range, so make the string empty
		outString = nil;
		
		return;
	}
	
	GetMenuItemText(mMenuH, inIndex, (StringPtr) outString);
	
		// convert the Pascal string to a C string in place
	S3PtoCString((unsigned char*) outString);
//	S3PtoCString((unsigned char*) outString, StrLength((StringPtr) outString));
}


// ===========================================================================
//		 ItemFromString
// ===========================================================================
//	Return the Menu item index number for a particular value
//		Return 0 if the value number is not used for this Menu

	Int16
CStringPopupMenu::ItemFromString(Str255 inString) const
{
		// Search for inString in the items of this Menu
	Str255 itemString;
	short numItems = CountMItems(mMenuH);
	
	for (short theIndex = 1; theIndex <= numItems; theIndex++) {
		GetMenuItemText(mMenuH, theIndex, itemString);
		
		if (EqualString(inString, itemString, true, false)) {
			return theIndex;
		}
	}
	
		// the string wasn't found in this menu
	return 0;
}


// ===========================================================================
//		 ItemFromString
// ===========================================================================
//	Return the Menu item index number for a particular value
//		Return 0 if the value number is not used for this Menu

	Int16
CStringPopupMenu::ItemFromString(char* inString) const
{
	Str255 PString;
	S3CtoPString(inString, PString);
	
			// Search for inString in the items of this Menu
	Str255 itemString;
	short numItems = CountMItems(mMenuH);
	
	for (short theIndex = 1; theIndex <= numItems; theIndex++) {
		GetMenuItemText(mMenuH, theIndex, itemString);
		
		if (EqualString(PString, itemString, true, false)) {
			return theIndex;
		}
	}
	
		// the string wasn't found in this menu
	return 0;
}




