// ===========================================================================
// CTQAColorTable.cp
// ===========================================================================


#include <PictUtil.h>
#include <UDebugging.h>

#include "CTQAColorTable.h"
#include "S3ColorTable.h"


// ===========================================================================
// Constants
// ===========================================================================

const short 				kCL16bit_ElementSize = 				2;
const short 				kCL32bit_ElementSize = 				4;
const short 				kCL8_ElementCount = 				256;
const short 				kCL4_ElementCount = 				16;

const UInt16 				kARGB16shift_alpha =				15;
const UInt16 				kARGB16shift_red =					10;
const UInt16 				kARGB16shift_green =				5;
const UInt16 				kARGB16shift_blue =					0;

const UInt16 				kARGB16mask_alpha =					1 << kARGB16shift_alpha;
const UInt16 				kARGB16mask_red =					0x1F << kARGB16shift_red;
const UInt16 				kARGB16mask_green =					0x1F << kARGB16shift_green;
const UInt16 				kARGB16mask_blue =					0x1F << kARGB16shift_blue;

const UInt32 				kARGB32shift_alpha =				24;
const UInt32 				kARGB32shift_red =					16;
const UInt32 				kARGB32shift_green =				8;
const UInt32 				kARGB32shift_blue =					0;

const UInt32 				kARGB32mask_alpha =					0xFF << kARGB32shift_alpha;
const UInt32 				kARGB32mask_red =					0xFF << kARGB32shift_red;
const UInt32 				kARGB32mask_green =					0xFF << kARGB32shift_green;
const UInt32 				kARGB32mask_blue =					0xFF << kARGB32shift_blue;


// ===========================================================================
CTQAColorTable::CTQAColorTable(
	CTabHandle				inColorTableH,
	TQAColorTableType	 	inColorTableType)
		: mType(inColorTableType), 
		  mDataP(nil)
{
	Assert_(inColorTableH);
	Assert_(mType >= kQAColorTable_CL8_RGB32 && mType <= kQAColorTable_CL4_RGB32);
	
	short					bytesPerElement = 0;
	short					elementCount = 0;
	
	switch (mType) {
//		case kQACL8_RGB16:
//			bytesPerElement = kCL16bit_ElementSize;
//			elementCount = kCL8_ElementCount;
//			break;
		
		case kQAColorTable_CL8_RGB32:
			bytesPerElement = kCL32bit_ElementSize;
			elementCount = kCL8_ElementCount;
			break;
		
//		case kQACL4_RGB16:
//			bytesPerElement = kCL16bit_ElementSize;
//			elementCount = kCL4_ElementCount;
//			break;
		
		case kQAColorTable_CL4_RGB32:
			bytesPerElement = kCL32bit_ElementSize;
			elementCount = kCL4_ElementCount;
			break;
		
		default:
		  SignalCStr_("Illegal TQAColorTableType detected.");
		  break;
	} // endswitch
	
	mDataP = new Byte[elementCount * bytesPerElement];
	
	if (mDataP == nil) {
// should raise an exception here
		return;
	}

	if (bytesPerElement == kCL32bit_ElementSize) {
			// this is a 32-bit lookup table
		UInt32*						elementP = (UInt32*) mDataP;
		UInt32						channelValue;
		
		for (short i = 0; i < elementCount; i++) {
				// turn on all the bits in the alpha channel.  we do this
				// for every entry in the TQAColorTable, regardless of how
				// many entries are in inColorTableH.  this also sets
				// the color to black.
			elementP[i] = kARGB32mask_alpha;
			
			if (i <= (**inColorTableH).ctSize) {
					// we haven't run out of ColorTable entries yet, so reduce
					// each 16-bit channel in the entry to 8 bits and incorporate
					// it into the TQAColorTable entry.
				channelValue = (**inColorTableH).ctTable[i].rgb.red >> 8;
				elementP[i] |= ((channelValue << kARGB32shift_red) & kARGB32mask_red);

				channelValue = (**inColorTableH).ctTable[i].rgb.green >> 8;
				elementP[i] |= ((channelValue << kARGB32shift_green) & kARGB32mask_green);

				channelValue = (**inColorTableH).ctTable[i].rgb.blue >> 8;
				elementP[i] |= ((channelValue << kARGB32shift_blue) & kARGB32mask_blue);
			}
		} // endfor
	} else {
			// this is a 16-bit lookup table
		UInt16*						elementP = (UInt16*) mDataP;
		UInt16						channelValue;
		
		for (short i = 0; i < elementCount; i++) {
				// turn on the alpha channel bit.  we do this for every entry 
				// in the TQAColorTable, regardless of how many entries are 
				// in inColorTableH.  this also sets the color to black.
			elementP[i] = kARGB16mask_alpha;
			
			if (i <= (**inColorTableH).ctSize) {
					// we haven't run out of ColorTable entries yet, so reduce
					// each 16-bit channel in the entry to 5 bits and incorporate
					// it into the TQAColorTable entry.
				channelValue = (**inColorTableH).ctTable[i].rgb.red >> 11;
				elementP[i] |= ((channelValue << kARGB16shift_red) & kARGB16mask_red);

				channelValue = (**inColorTableH).ctTable[i].rgb.green >> 11;
				elementP[i] |= ((channelValue << kARGB16shift_green) & kARGB16mask_green);

				channelValue = (**inColorTableH).ctTable[i].rgb.blue >> 11;
				elementP[i] |= ((channelValue << kARGB16shift_blue) & kARGB16mask_blue);
			}
		} // endfor
	}
}


// ===========================================================================
CTQAColorTable::CTQAColorTable(
	CTabHandle				inColorTableH,
	TQAColorTableType	 	inColorTableType,
	float					inTableBrightness)
		: mType(inColorTableType), 
		  mDataP(nil)
{
	Assert_(inColorTableH);
	Assert_(mType >= kQAColorTable_CL8_RGB32 && mType <= kQAColorTable_CL4_RGB32);
	
	short					bytesPerElement = 0;
	short					elementCount = 0;
	
	switch (mType) {
		case kQAColorTable_CL8_RGB32:
			bytesPerElement = kCL32bit_ElementSize;
			elementCount = kCL8_ElementCount;
			break;
		
		case kQAColorTable_CL4_RGB32:
			bytesPerElement = kCL32bit_ElementSize;
			elementCount = kCL4_ElementCount;
			break;
		
		default:
		  SignalCStr_("Illegal TQAColorTableType detected.");
		  break;
	} // endswitch
	
	mDataP = new Byte[elementCount * bytesPerElement];
	
	if (mDataP == nil) {
// should raise an exception here
		return;
	}

		// this is a 32-bit lookup table
	UInt32*						elementP = (UInt32*) mDataP;
	UInt32						channelValue;
	
	for (short i = 0; i < elementCount; i++) {
			// turn on all the bits in the alpha channel.  we do this
			// for every entry in the TQAColorTable, regardless of how
			// many entries are in inColorTableH.  this also sets
			// the color to black.
		elementP[i] = kARGB32mask_alpha;
		
		if (i <= (**inColorTableH).ctSize) {
				// we haven't run out of ColorTable entries yet, so reduce
				// each 16-bit channel in the entry to 8 bits and incorporate
				// it into the TQAColorTable entry.
			channelValue = (**inColorTableH).ctTable[i].rgb.red >> 8;
			channelValue = UInt32(channelValue * inTableBrightness);
			channelValue = (channelValue > 0xFF) ? 0xFF : channelValue;
			elementP[i] |= ((channelValue << kARGB32shift_red) & kARGB32mask_red);

			channelValue = (**inColorTableH).ctTable[i].rgb.green >> 8;
			channelValue = UInt32(channelValue * inTableBrightness);
			channelValue = (channelValue > 0xFF) ? 0xFF : channelValue;
			elementP[i] |= ((channelValue << kARGB32shift_green) & kARGB32mask_green);

			channelValue = (**inColorTableH).ctTable[i].rgb.blue >> 8;
			channelValue = UInt32(channelValue * inTableBrightness);
			channelValue = (channelValue > 0xFF) ? 0xFF : channelValue;
			elementP[i] |= ((channelValue << kARGB32shift_blue) & kARGB32mask_blue);
		}
	} // endfor
}


// ===========================================================================
CTQAColorTable::~CTQAColorTable()
{
	delete mDataP;
}


	// external C interface to this class
extern "C" {

	
// ===========================================================================
	S3tColorTableP
S3ColorTable_New(
	CTabHandle				inColorTableH,
	TQAColorTableType		inColorTableType)
{
	return S3tColorTableP(new CTQAColorTable(inColorTableH, inColorTableType));
}


// ===========================================================================
	S3tColorTableP
S3ColorTable_NewChangeBrightness(
	CTabHandle				inColorTableH,
	TQAColorTableType		inColorTableType,
	float					inTableBrightness)
{
	return S3tColorTableP(new CTQAColorTable(inColorTableH, inColorTableType, inTableBrightness));
}


// ===========================================================================
	void
S3ColorTable_Delete(
	S3tColorTableP			inS3ColorTable)
{
	Assert_(inS3ColorTable);
	
	delete (CTQAColorTable*) inS3ColorTable;
}


// ===========================================================================
	void*
S3ColorTable_GetData(
	S3tColorTableP			inS3ColorTable)
{
	Assert_(inS3ColorTable);
	
	return ((CTQAColorTable*) inS3ColorTable)->GetData();
}


// ===========================================================================
	long
S3ColorTable_GetType(
	S3tColorTableP			inS3ColorTable)
{
	Assert_(inS3ColorTable);
	
	return ((CTQAColorTable*) inS3ColorTable)->GetType();
}


}
