// ===========================================================================
// CTQAImage.cp
// ===========================================================================


#include <PictUtil.h>
#include <UGWorld.h>
#include <UDebugging.h>

#include "CTQAImage.h"
#include "S3Image.h"
#include "S3Utilities.h"
#include "UExistingHandleBlock.h"


// ===========================================================================
CTQAImage::CTQAImage(
	PicHandle 					inPICTHandle,
	TQAImagePixelType 	inPixelType,
	Boolean							inMakeBlackTransparent,
	Boolean							inMakeMipMap)
{
	Assert_(inPICTHandle);
	
	mPixelType = inPixelType;
	mImages[0].pixmap = nil;
	mMipMaps = nil;
	mIsValidTexture = false;
	mIsValidBitmap = false;
	mIsValidMipMap = false;
	
		// determine the depth the caller wants the texture to have
	Int16 depth = 1;
	short pictInfoVerb = 0;
	short colorsRequested = 0;
	
	switch (inPixelType) {
		case kQAPixel_RGB32:
		case kQAPixel_ARGB32:
			depth = 32;
			break;
		
		case kQAPixel_RGB16:
		case kQAPixel_ARGB16:
			depth = 16;
			break;
		
		case kQAPixel_CL8:
			depth = 8;
			pictInfoVerb = returnColorTable;
			colorsRequested = 256;
			break;
			
		case kQAPixel_CL4:
			depth = 4;
			pictInfoVerb = returnColorTable;
			colorsRequested = 16;
			break;
			
		case kQAPixel_Alpha1:
			depth = 1;
			break;
			
	  default:
		  SignalCStr_("Illegal TQAImagePixelType detected.");
		  break;
	} // endswitch

		// calc the PICT's height and width.  we store the values in the first
		// image in the array because that one will be the same size as the original
		// PICT.  we use GetPictInfo to get the sourceRect of the PICT because that
		// contains the full size of the image and ignores the Hres and Vres
		// values.  so we can use the PICT even if it's not 72 dpi.  we also get
		// the pict's color table if this is a color table texture, using either
		// 16 or 256 colors and the median method for picking the colors.
	PictInfo info;
	::GetPictInfo(inPICTHandle, &info, pictInfoVerb, colorsRequested, medianMethod, 0);	
	Rect pictRect = info.sourceRect;
	StExistingHandleBlockT<CTabHandle> colorTableH(info.theColorTable);
	
	mImages[0].width = pictRect.right - pictRect.left;
	mImages[0].height = pictRect.bottom - pictRect.top;
	
	try {
			// try to create a new offscreen GWorld in which to draw the PICT
		mGWorld = new LGWorld(pictRect, depth, 0, colorTableH);
	}
	
	catch(...) {
		try {
				// there isn't enough memory to allocate the GWorld in our heap, so
				// try putting it in temporary memory
			mGWorld = new LGWorld(pictRect, depth, useTempMem, colorTableH);
		}
		
		catch(...) {
				// we still don't have enough memory for the GWorld, so just return,
				// since we can't do anything if we can't allocate a GWorld 
			return;
		} 
	}
	
		// get ready to draw into the GWorld
	if (mGWorld->BeginDrawing() == false) {
			// the GWorld's memory can't be locked down, so just return
		return;
	}
	
		// draw the PICT into the GWorld
	::DrawPicture(inPICTHandle, &pictRect);
	
	if (depth == 1) {
			// in a 1-bit GWorld, 1 == black and 0 == white, which is the 
			// opposite of what we want, so invert the bitmap
		::InvertRect(&pictRect);
	}

		// we're done drawing into the GWorld
	mGWorld->EndDrawing();
	
	GWorldPtr theGWorldPtr = mGWorld->GetMacGWorld();
	
		// lock the GWorld's pixmap
	::LockPixels(::GetGWorldPixMap(theGWorldPtr));

		// to get the pixmap's actual rowbytes, we have to mask off the top two bits
	mImages[0].rowBytes = (**(::GetGWorldPixMap(theGWorldPtr))).rowBytes & 0x3fff;

		// save the GWorld's base address
	mImages[0].pixmap = GetPixBaseAddr(::GetGWorldPixMap(theGWorldPtr));

	if (inMakeBlackTransparent) {
			// make every black pixel be transparent and every non-black one be opaque
		S3MakeBlackTransparent(mImages[0].width, mImages[0].height, mImages[0].rowBytes, 
													 mImages[0].pixmap, mPixelType);
	}
	
		// if we've gotten here, then we at least have a valid bitmap image,
		// because bitmap dimensions don't have to be powers of 2
	mIsValidBitmap = true;
	
	if (S3IsPowerOf2(mImages[0].width, nil) && S3IsPowerOf2(mImages[0].height, nil)) {
			// the dimensions of a texture must be a power of 2, so we have
			// a valid texture
		mIsValidTexture = true;
		
		if (inMakeMipMap) {
				// the caller wants mipmaps made for this texture
			MakeMipMap();
		}
	}
}


// ===========================================================================
CTQAImage::CTQAImage()
{
	mPixelType = kQAPixel_RGB32;
	mGWorld = nil;
	mMipMaps = nil;
	mImages[0].pixmap = nil;
	mIsValidTexture = false;
	mIsValidBitmap = false;
	mIsValidMipMap = false;
}


// ===========================================================================
CTQAImage::~CTQAImage()
{
	if (mGWorld != nil) {
			// unlock the GWorld's pixels before deleting it
		::UnlockPixels(::GetGWorldPixMap(mGWorld->GetMacGWorld()));
		
		delete mGWorld;
	}

	delete[] mMipMaps;
}


// ===========================================================================
	void
CTQAImage::MakeMipMap()
{
	long			nImages;
	long			width, height;
	long			nPixels;
	long			i;
	long			x, y;
	long			oneHalf;
	
		// we don't have a valid mipmap until we get to the end of this method
	mIsValidMipMap = false;
	
	if (mPixelType == kQAPixel_Alpha1) {
			// we can only mipmap 16- or 32-bit images
		return;
	}
	
if (mPixelType == kQAPixel_CL8 || mPixelType == kQAPixel_CL4) {
		// we can only mipmap 16- or 32-bit images
	return;
}

		// calculate the root-2 of of the width and height of the PICT. they
		// have to be a power of 2 for us to mipmap the image, so return if 
		// they're not.  
	unsigned short widthNBits, heightNBits;
	
	if (S3IsPowerOf2((unsigned) GetWidth(), &widthNBits) == false) {
		return;
	}

	if (S3IsPowerOf2((unsigned) GetHeight(), &heightNBits) == false) {
		return;
	}

		// Allocate working space for the mip map images (starting with the
		// 2^(widthNBits-1) x 2^(heightNBits-1) map), and for the array of
		// TQAImages that describe the mip map and the parent map.
	short pixelSize;
	switch (mPixelType) {
		case kQAPixel_ARGB32:
		case kQAPixel_RGB32:
			pixelSize = 4;
			break;
		
		case kQAPixel_ARGB16:
		case kQAPixel_RGB16:
			pixelSize = 2;
			break;

	  default:
			SignalPStr_("\pFound unrecognized pixel type in CTQAImage::MakeMipMap");
	} // endswitch
	
	nImages = ((widthNBits > heightNBits) ? widthNBits : heightNBits) + 1;
	
		// make sure the width and height are always at least 1
	width = 1 << (((widthNBits - 1) >= 0) ? (widthNBits - 1) : 0);
	height = 1 << (((heightNBits - 1) >= 0) ? (heightNBits - 1) : 0);
	
		// count the number of pixels contained in all the mipmaps
	for (i = nImages - 1, nPixels = 0; i > 0; i--) {
		nPixels += width * height;
		
			// halve the width and height, making sure they're always at least 1
		width >>= 1;
		height >>= 1;
		
		if (width == 0) {
			width = 1;
		}
		
		if (height == 0) {
			height = 1;
		}
	}

		// create an array of bytes of the proper size
	mMipMaps = new Byte[nPixels * pixelSize];
	Byte* mipMapP = (Byte*) mMipMaps;
	
	if (mMipMaps == nil) {
			// we don't have enough memory to create the mipmaps
		return;
	}
	
		// make sure the width and height are always at least 1
	width = 1 << (((widthNBits - 1) >= 0) ? (widthNBits - 1) : 0);
	height = 1 << (((heightNBits - 1) >= 0) ? (heightNBits - 1) : 0);
	
		// fill in the TQAImage info.  the first image is full size, and we 
		// already filled in its info in the constructor.
	for (i = 1; i < nImages; i++) {
		mImages[i].width = width;
		mImages[i].height = height;
		mImages[i].pixmap = mipMapP;
		mImages[i].rowBytes = width * pixelSize;
		
			// go to the beginning of the next pixmap
		mipMapP += width * height * pixelSize;
		
			// halve the width and height, making sure it's always at least 1
		width >>= 1;
		height >>= 1;
		
		if (width == 0) {
			width = 1;
		}
		
		if (height == 0) {
			height = 1;
		}
	}

		// oneHalf is used to round off the box filter of four pixels.
	oneHalf = 4 / 2;

		// Create the mipmapped images.
	for (i = 1; i < nImages; i++) {
		TQAImage		*prevImage, *image;
		long				sourceRowBytes, targetRowBytes;
		Byte* 			sourceLine0, *sourceLine1;
		Byte* 			targetLine;
		
		prevImage = &mImages[i - 1];
		image = &mImages[i];
		
		sourceLine0 = (unsigned char*) prevImage->pixmap;
		
		if (prevImage->height == 1) {
				// Source image has height of 1 (widthNBits must be greater than heightNBits).
				// Make sourceLine1 duplicate sourceLine0, and set sourceRowBytes to 0 so 
				// we just keep re-reading the single source line.
			sourceRowBytes = 0;
			sourceLine1 = sourceLine0;
		} else {
				// Normal case. sourceLine1 points one scanline above sourceLine0.
			sourceRowBytes = prevImage->rowBytes;
			sourceLine1 = sourceLine0 + sourceRowBytes;
		}
		
		targetRowBytes = image->rowBytes;
		targetLine = (Byte*) image->pixmap;

		for (y = image->height; y > 0; y--) {
			Byte* sourcePixel0, *sourcePixel1;
			Byte* targetPixel;
			
			sourcePixel0 = sourceLine0;
			sourcePixel1 = sourceLine1;
			targetPixel = targetLine;
			
			if (prevImage->width == 1) {
					// Source image has width 1 (heightNBits must be greater than widthNBits).
					// Use a special inner loop that doesn't advance past the first pixel of
					// the source lines.
				for (x = image->width; x > 0; x--) {
					long	a, r, g, b;
					
						// Read and average two pixels from source into target.
					a = AlphaChannel(sourcePixel0) + (oneHalf >> 1);
					r = RedChannel(sourcePixel0) + (oneHalf >> 1);
					g = GreenChannel(sourcePixel0) + (oneHalf >> 1);
					b = BlueChannel(sourcePixel0) + (oneHalf >> 1);
					
					a += AlphaChannel(sourcePixel1);
					r += RedChannel(sourcePixel1);
					g += GreenChannel(sourcePixel1);
					b += BlueChannel(sourcePixel1);

					SetARGB(targetPixel, a >> 1, r >> 1, g >> 1, b >> 1);

					targetPixel += pixelSize;
				}
			} else {
					// Source image has at least width 2. This is the normal case.
				for (x = image->width; x > 0; x--) {
					long	a, r, g, b;
					
						// Read and average four pixels from source into target.
					a = AlphaChannel(sourcePixel0) + oneHalf;
					r = RedChannel(sourcePixel0) + oneHalf;
					g = GreenChannel(sourcePixel0) + oneHalf;
					b = BlueChannel(sourcePixel0) + oneHalf;
					
					sourcePixel0 += pixelSize;
					
					a += AlphaChannel(sourcePixel1);
					r += RedChannel(sourcePixel1);
					g += GreenChannel(sourcePixel1);
					b += BlueChannel(sourcePixel1);
					
					sourcePixel1 += pixelSize;
					
					a += AlphaChannel(sourcePixel0);
					r += RedChannel(sourcePixel0);
					g += GreenChannel(sourcePixel0);
					b += BlueChannel(sourcePixel0);
					
					sourcePixel0 += pixelSize;
					
					a += AlphaChannel(sourcePixel1);
					r += RedChannel(sourcePixel1);
					g += GreenChannel(sourcePixel1);
					b += BlueChannel(sourcePixel1);
					
					sourcePixel1 += pixelSize;
					
					SetARGB(targetPixel, a >> 2, r >> 2, g >> 2, b >> 2);
					
					targetPixel += pixelSize;
				}
			}
			
			sourceLine0 += (sourceRowBytes << 1);
			sourceLine1 += (sourceRowBytes << 1);
			targetLine += targetRowBytes;
		}
	}
	
		// if we've gotten here, then we have a valid mipmap
	mIsValidMipMap = true;
}


	// external C interface to this class
extern "C" {

	
// ===========================================================================
	S3Image
S3Image_New(
	PicHandle 					inPICTHandle,
	TQAImagePixelType 	inPixelType,
	Boolean							inMakeBlackTransparent,
	Boolean							inMakeMipMap)
{
	CTQAImage* image = new CTQAImage(inPICTHandle, inPixelType, 
																	 inMakeBlackTransparent, inMakeMipMap);

	if (image == nil) {
			// there's been a serious problem if nil was returned
		return nil;
	}
	
	if (!image->IsValidTexture() && !image->IsValidBitmap()) {
			// the image we created is bad for some reason (probably not enough 
			// memory), so just delete it and return nil
		delete image;
			
		return nil;
	}
	
		// the CTQAImage we just created is valid, so cast it to the opaque
		// type S3Image and return it
	return (S3Image) image;
}


// ===========================================================================
	void
S3Image_Delete(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	delete image;
}


// ===========================================================================
	TQAImage*
S3Image_GetImages(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	return image->GetImages();
}


// ===========================================================================
	long
S3Image_GetWidth(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	return image->GetWidth();
}


// ===========================================================================
	long
S3Image_GetHeight(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	return image->GetHeight();
}


// ===========================================================================
	long
S3Image_GetRowBytes(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	return image->GetRowBytes();
}


// ===========================================================================
	CTabHandle
S3Image_GetColorTableHandle(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	return image->GetColorTableHandle();
}


// ===========================================================================
	Boolean
S3Image_IsValidTexture(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	return image->IsValidTexture();
}


// ===========================================================================
	Boolean
S3Image_IsValidBitmap(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	return image->IsValidBitmap();
}
	

// ===========================================================================
	Boolean
S3Image_IsValidMipMap(
	S3Image 						inImage)
{
	Assert_(inImage);
	
	CTQAImage* image = (CTQAImage*) inImage;
	
	return image->IsValidMipMap();
}
	

}