// ===========================================================================
// CTestInfo.cp
// ===========================================================================


#include <strstream.h>
#include <string.h>
#include <math.h>
#include <QuickDraw.h>
#include <PictUtil.h>
#include <FixMath.h>
#include <LCommander.h>
#include <LFileStream.h>
#include <UGWorld.h>
#include <UDebugging.h>
#include <LString.h>
#include <String_Utils.h>

#include "CTestInfo.h"
#include "CSettingsDoc.h"
#include "CPvalPopupMenu.h"
#include "CEnginePopupMenu.h"
#include "CSmartStdPopupMenu.h"
#include "CBroadcastEditField.h"
#include "CPICTPopupMenu.h"
#include "CTestPopupMenu.h"
#include "CTQAImage.h"
#include "CTQAColorTable.h"
#include "UGWorldState.h"
#include "S3Utilities.h"
#include "S3TestSystem.h"
#include "StringConversion.h"
#include "S3TestUtil.h"
#include "S3PICTList.h"
#include "S3Alert.h"
#include "S3Messages.h"


const Int16 kCurrentSettingsVersion = 1205;


// ===========================================================================
CTestInfo::CTestInfo()
{
	Initialize();
}


// ===========================================================================
CTestInfo::CTestInfo(const CTestInfo& inCTestInfo)
{
		// copy all the fields from inCTestInfo to our testInfo object
	mTestInfo = inCTestInfo.mTestInfo;
	
		// since this is the copy constructor, we know the settings haven't
		// been saved to a file yet, so this object is dirty
	mIsDirty = true;

		// we set the window pointer to nil because we don't want this CTestInfo
		// to point to the original's window.  our CSettingsDoc is repsonsible
		// for linking our settings window and itself to us.
	mSettingsDoc = nil;
	mSettingsWindow = nil;
	
	mPreparedTestInfo = nil;
}


// ===========================================================================
CTestInfo::CTestInfo(const Handle inTestInfoH, Int16 inSettingsVersion)
{
	Initialize();
	
	ReadFromHandle(inTestInfoH, inSettingsVersion);
}


// ===========================================================================
CTestInfo::~CTestInfo()
{
	DisposeTestInfo(&mTestInfo);
}


// ===========================================================================
	void
CTestInfo::Initialize()
{
		// since we're initializing the object, we know none of the settings 
		// have been changed yet
	mIsDirty = false;
	
	mSettingsDoc = nil;
	mSettingsWindow = nil;
	mPreparedTestInfo = nil;

	InitializeTestInfo(&mTestInfo);
}	


// ===========================================================================
	void
CTestInfo::InitializeTestInfo(
	S3tTestInfo* inTestInfo)
{
	if (inTestInfo == nil) {
		return;
	}
	
	inTestInfo->leftMicroseconds = 0;
	inTestInfo->rightMicroseconds = 0;
	
	inTestInfo->testName[0] = 0;
	inTestInfo->argString[0] = 0;

	inTestInfo->leftContext = nil;
	inTestInfo->rightContext = nil;

	inTestInfo->leftEngine = QADeviceGetFirstEngine(nil);
	inTestInfo->rightEngine = QADeviceGetNextEngine(nil, inTestInfo->leftEngine);
	
	if (inTestInfo->rightEngine == nil) {	
			// nil means there's only one engine, so make the right the 
			// same as the left
		inTestInfo->rightEngine = inTestInfo->leftEngine;
	}

		// default to an on-screen device on the deepest monitor
	inTestInfo->leftDevice.deviceType = kQADeviceGDevice;
	inTestInfo->leftDevice.device.gDevice = S3GetDeepestScreen();
	inTestInfo->rightDevice.deviceType = kQADeviceGDevice;
	inTestInfo->rightDevice.device.gDevice = S3GetDeepestScreen();
	inTestInfo->deviceMenu = kS3Device_Window;
	inTestInfo->leftDeviceGWorld = nil;
	inTestInfo->rightDeviceGWorld = nil;

	CTinselTestWindow::GetDefaultGlobalRects(
		nil,
		&inTestInfo->leftGlobalRect, 
		&inTestInfo->rightGlobalRect);

	inTestInfo->clipRegionMenu = kS3ClipRegion_None;

	inTestInfo->leftClip.clipType = kQAClipRgn;
	inTestInfo->leftClip.clip.clipRgn = nil;
	inTestInfo->rightClip.clipType = kQAClipRgn;
	inTestInfo->rightClip.clip.clipRgn = nil;

	inTestInfo->contextFlags = kQAContext_DeepZ;
	inTestInfo->zFunction = kQAZFunction_LT;
	inTestInfo->perspectiveZ = kQAPerspectiveZ_Off;
	inTestInfo->blending = kQABlend_PreMultiply;
	inTestInfo->antialiasing = kQAAntiAlias_Fast;

	inTestInfo->alpha = 0.0;
	inTestInfo->red = 0.0;
	inTestInfo->green = 0.0;
	inTestInfo->blue = 0.0;

	inTestInfo->lineWidth = 1.0;

	memset(&inTestInfo->textureFSSpec, unsigned char(0), sizeof(FSSpec));
	inTestInfo->textureImage = nil;
	inTestInfo->textureWidth = 0;
	inTestInfo->textureHeight = 0;
	inTestInfo->textureRowBytes = 0;
	
	inTestInfo->leftTexture = nil;
	inTestInfo->rightTexture = nil;
	inTestInfo->textureFormat = kQAPixel_RGB32;
	inTestInfo->textureFilter = kQATextureFilter_Fast;
	inTestInfo->textureOp = kQATextureOp_None;
	inTestInfo->textureFlags = kQATexture_None;
	inTestInfo->leftTexture = nil;
	inTestInfo->rightTexture = nil;
	inTestInfo->detachTexture = false;
	inTestInfo->textureMakeBlackTransparent = false;
	inTestInfo->leftTextureColorTable = nil;
	inTestInfo->rightTextureColorTable = nil;
	inTestInfo->textureColorTableType = kQAColorTable_CL8_RGB32;
	inTestInfo->textureUseTransparentIndex = false;
		
	memset(&inTestInfo->bitmapFSSpec, unsigned char(0), sizeof(FSSpec));
	inTestInfo->bitmapImage = nil;
	inTestInfo->bitmapWidth = 0;
	inTestInfo->bitmapHeight = 0;
	inTestInfo->bitmapRowBytes = 0;
	inTestInfo->bitmapFormat = kQAPixel_RGB32;
	inTestInfo->bitmapFlags = kQATexture_None;
	inTestInfo->leftBitmap = nil;
	inTestInfo->rightBitmap = nil;
	inTestInfo->detachBitmap = false;
	inTestInfo->bitmapMakeBlackTransparent = false;
	inTestInfo->leftBitmapColorTable = nil;
	inTestInfo->rightBitmapColorTable = nil;
	inTestInfo->bitmapColorTableType = kQAColorTable_CL8_RGB32;
	inTestInfo->bitmapUseTransparentIndex = false;

	inTestInfo->testName[0] = 0;
	inTestInfo->resetRandomSeed = false;

	inTestInfo->pictList = nil;
}


// ===========================================================================
	void
CTestInfo::DisposeTestInfo(
	S3tTestInfo* inTestInfo)
{
	if (inTestInfo == nil) {
		return;
	}
	
	if (inTestInfo->leftContext != nil)  {
		QADrawContextDelete(inTestInfo->leftContext);
		inTestInfo->leftContext = nil;
	}
		
	if (inTestInfo->rightContext != nil) {
		QADrawContextDelete(inTestInfo->rightContext);
		inTestInfo->rightContext = nil;
	}
	
	delete inTestInfo->leftDeviceGWorld;
	inTestInfo->leftDeviceGWorld = nil;
	
	delete inTestInfo->rightDeviceGWorld;
	inTestInfo->rightDeviceGWorld = nil;
	
	if (inTestInfo->leftClip.clip.clipRgn != nil) {
		::DisposeRgn((RgnHandle) inTestInfo->leftClip.clip.clipRgn);
		inTestInfo->leftClip.clip.clipRgn = nil;
	}
	
	if (inTestInfo->rightClip.clip.clipRgn != nil) {
		::DisposeRgn((RgnHandle) inTestInfo->rightClip.clip.clipRgn);
		inTestInfo->rightClip.clip.clipRgn = nil;
	}
	
	if (inTestInfo->leftTexture != nil) {
		QATextureDelete(inTestInfo->leftEngine, inTestInfo->leftTexture);
		inTestInfo->leftTexture = nil;
	}
	
	if (inTestInfo->rightTexture != nil) {
		QATextureDelete(inTestInfo->rightEngine, inTestInfo->rightTexture);
		inTestInfo->rightTexture = nil;
	}
	
	if (inTestInfo->textureImage != nil) {
		delete inTestInfo->textureImage;
		inTestInfo->textureImage = nil;
	}

	if (inTestInfo->leftTextureColorTable != nil) {
		QAColorTableDelete(inTestInfo->leftEngine, inTestInfo->leftTextureColorTable);
		inTestInfo->leftTextureColorTable = nil;
	}
	
	if (inTestInfo->rightTextureColorTable != nil) {
		QAColorTableDelete(inTestInfo->rightEngine, inTestInfo->rightTextureColorTable);
		inTestInfo->rightTextureColorTable = nil;
	}
	
	if (inTestInfo->leftBitmap != nil) {
		QABitmapDelete(inTestInfo->leftEngine, inTestInfo->leftBitmap);
		inTestInfo->leftBitmap = nil;
	}
	
	if (inTestInfo->rightBitmap != nil) {
		QABitmapDelete(inTestInfo->rightEngine, inTestInfo->rightBitmap);
		inTestInfo->rightBitmap = nil;
	}

	if (inTestInfo->bitmapImage != nil) {
		delete inTestInfo->bitmapImage;
		inTestInfo->bitmapImage = nil;
	}

	if (inTestInfo->leftBitmapColorTable != nil) {
		QAColorTableDelete(inTestInfo->leftEngine, inTestInfo->leftBitmapColorTable);
		inTestInfo->leftBitmapColorTable = nil;
	}
	
	if (inTestInfo->rightBitmapColorTable != nil) {
		QAColorTableDelete(inTestInfo->rightEngine, inTestInfo->rightBitmapColorTable);
		inTestInfo->rightBitmapColorTable = nil;
	}
}


// ===========================================================================
	Boolean
CTestInfo::ReadFromHandle(const Handle inTestInfoH, Int16 inSettingsVersion)
{
		// if the user has reverted this document, we might already have 
		// created drawcontexts, textures, etc., so dispose of them before
		// we overwrite our mTestInfo with the incoming handle
	DisposeTestInfo(&mTestInfo);
	
	mIsDirty = false;

		// make sure we can handle this version of the settings
	if (VerifySettingsVersion(inTestInfoH, inSettingsVersion) == false)
		return false;
	
		// copy the test info settings from the handle to our mTestInfo
	::BlockMoveData(*inTestInfoH, &mTestInfo, sizeof(mTestInfo));
	
		// get a pointer to the alias records which follow the mTestInfo data
	Int8* dataP = (Int8*) *inTestInfoH;
	dataP += sizeof(mTestInfo);

		// get the texture alias
	AliasRecord aliasRec = *(AliasRecord*) dataP;

	if (aliasRec.aliasSize > 0) {
			// an alias for the texture was saved, so copy it into a handle
		StHandleBlock textureAliasH(aliasRec.aliasSize);
		::BlockMoveData(dataP, *Handle(textureAliasH), aliasRec.aliasSize);
		
			// find out where the app was launched from
		FSSpec appFileSpec;
		S3GetAppFileSpec(&appFileSpec);
		
			// we have to initialize to 0 the memory of the FSSpec in mTestInfo
			// because LLDynamicArray::FetchIndexOf uses BlocksAreEqual to  
			// find the index of a piece of data.  CPICTPopupMenu::InitializeMenu
			// also initializses all the FSSpec's it finds to 0, so if the FSSpec
			// saved in the file is the same as one of those in the menu, we 
			// want them to have exactly the same bits in memory.
		memset(&mTestInfo.textureFSSpec, unsigned char(0), sizeof(FSSpec));
		
			// resolve the alias and save the PICT's FSSpec into mTestInfo
		Boolean wasChanged;
		OSErr err = ::ResolveAlias(&appFileSpec, AliasHandle(Handle(textureAliasH)), 
															 &mTestInfo.textureFSSpec, &wasChanged);
		
		if (err != noErr) {
				// there was an error because the alias couldn't be resolved, so
				// clear the textureFSSpec in mTestInfo
			memset(&mTestInfo.textureFSSpec, unsigned char(0), sizeof(FSSpec));
		}
	}
	
		// skip past the texture AliasRecord to the bitmap one
	dataP += aliasRec.aliasSize;

		// get the bitmap alias
	aliasRec = *(AliasRecord*) dataP;

	if (aliasRec.aliasSize > 0) {
			// an alias for the bitmap was saved, so copy it into a handle
		StHandleBlock bitmapAliasH(aliasRec.aliasSize);
		::BlockMoveData(dataP, *Handle(bitmapAliasH), aliasRec.aliasSize);
		
			// find out where the app was launched from
		FSSpec appFileSpec;
		S3GetAppFileSpec(&appFileSpec);
		
			// we have to initialize to 0 the memory of the FSSpec in mTestInfo
			// because LLDynamicArray::FetchIndexOf uses BlocksAreEqual to  
			// find the index of a piece of data.  CPICTPopupMenu::InitializeMenu
			// also initializses all the FSSpec's it finds to 0, so if the FSSpec
			// saved in the file is the same as one of those in the menu, we 
			// want them to have exactly the same bits in memory.
		memset(&mTestInfo.bitmapFSSpec, unsigned char(0), sizeof(FSSpec));
		
			// resolve the alias and save the PICT's FSSpec into mTestInfo
		Boolean wasChanged;
		OSErr err = ::ResolveAlias(&appFileSpec, AliasHandle(Handle(bitmapAliasH)), 
															 &mTestInfo.bitmapFSSpec, &wasChanged);
		
		if (err != noErr) {
				// there was an error because the alias couldn't be resolved, so
				// clear the textureFSSpec in mTestInfo
			memset(&mTestInfo.bitmapFSSpec, unsigned char(0), sizeof(FSSpec));
		}
	}
	
		// set the engine by the name saved in the handle
	mTestInfo.leftEngine = S3GetEngineByName(mTestInfo.leftEngineName);
	
	if (mTestInfo.leftEngine == nil) {
			// the engine name saved in the handle no longer seems to exist,
			// so just set it to the first one we find
		mTestInfo.leftEngine = QADeviceGetFirstEngine(nil);
	}
	
		// set the engine by the name saved in the handle
	mTestInfo.rightEngine = S3GetEngineByName(mTestInfo.rightEngineName);
	
	if (mTestInfo.rightEngine == nil) {
			// the engine name saved in the handle no longer seems to exist,
			// so just set it to the first one we find
		mTestInfo.rightEngine = QADeviceGetFirstEngine(nil);
	}

		// if these saved aliases are to files which are not in the same folder
		// hierarchy as Sneak3view, then they will be added to the end of
		// the list of PICT files it finds when it starts up
	CPICTPopupMenu::AppendPossiblyNewItem(mTestInfo.textureFSSpec.name, &mTestInfo.textureFSSpec);
	CPICTPopupMenu::AppendPossiblyNewItem(mTestInfo.bitmapFSSpec.name, &mTestInfo.bitmapFSSpec);
	
		// we set these pointers to nil regardless of what they were saved
		// as because they're obviously no longer valid
	mTestInfo.leftContext = nil;
	mTestInfo.rightContext = nil;
	mTestInfo.leftDeviceGWorld = nil;
	mTestInfo.rightDeviceGWorld = nil;
	mTestInfo.leftClip.clip.clipRgn = nil;
	mTestInfo.rightClip.clip.clipRgn = nil;
	mTestInfo.leftTexture = nil;
	mTestInfo.rightTexture = nil;
	mTestInfo.leftBitmap = nil;
	mTestInfo.rightBitmap = nil;
	mTestInfo.textureImage = nil;
	mTestInfo.bitmapImage = nil;
	mTestInfo.leftTextureColorTable = nil;
	mTestInfo.rightTextureColorTable = nil;
	mTestInfo.leftBitmapColorTable = nil;
	mTestInfo.rightBitmapColorTable = nil;
	mTestInfo.pictList = nil;
	
	UpdateSettingsWindow();
	UpdateLeftTestWindow();
	UpdateRightTestWindow();
	
	return true;
}


// ===========================================================================
	Boolean
CTestInfo::WriteToHandle(Handle outTestInfoH, Int16* outSettingsVersion)
{
	*outSettingsVersion = kCurrentSettingsVersion;
	
		// we set the handle to which writing to 0, because it might contain
		// a settings resource from an existing file and we don't want 
		// PtrAndHand to concatenate our current data onto the end of this 
		// resource.  we just want it to replace the resource.
	::SetHandleSize(outTestInfoH, 0);
	
		// save the engines' names so later we can find the engines again
		// when we read the file back in.  the engine name isn't kept
		// synchronized with the engine pointer in mTestInfo because the
		// engine pointer can be set from several places and we never have
		// to use the actual engine name except to save the engine selection.
//	S3GetEngineName(mTestInfo.leftEngine, mTestInfo.leftEngineName, 
//									sizeof(mTestInfo.leftEngineName));
//	S3GetEngineName(mTestInfo.rightEngine, mTestInfo.rightEngineName, 
//									sizeof(mTestInfo.rightEngineName));

		// copy mTestInfo to outTestInfoH
	if (::PtrAndHand(&mTestInfo, outTestInfoH, sizeof(mTestInfo)) != noErr)
		return false;

		// create a pointer to the test info in the handle
	S3tTestInfo* testInfoP = (S3tTestInfo*) *outTestInfoH;

	if (mTestInfo.textureFSSpec.name[0] > 0) {
			// we need to create an alias record for the currently selected 
			// texture so that we can find it next time this document is loaded
		FSSpec appFileSpec;
		S3GetAppFileSpec(&appFileSpec);

		if (appFileSpec.name[0] > 0) {
				// we got a good file spec, so create an alias from the position
				// of the app to the selected PICT
			AliasHandle textureH = nil;
			::NewAlias(&appFileSpec, &mTestInfo.textureFSSpec, &textureH);
			
			if (textureH != nil) {
					// we got a good alias handle, so append it to outTestInfoH
				StHandleLocker(Handle(textureH));
				OSErr err = ::HandAndHand((Handle) textureH, outTestInfoH);
				
				::DisposeHandle((Handle) textureH);
				
				if (err != noErr)
					return false;
			} else {
					// an alias couldn't be created for some reason (maybe the file
					// was moved, so the FSSpec no longer points to anything), so 
					// rather than returning an error, just set the texture FSSpec
					// in the test info handle to be empty.  we don't have to bother
					// emptying the one in mTestInfo because that's already been copied
					// to the handle.
				memset(&(testInfoP->textureFSSpec), unsigned char(0), sizeof(FSSpec));
			}
		}
	} 
	
	if (testInfoP->textureFSSpec.name[0] == 0) {
			// no texture file has been selected, or there was a problem in 
			// creating the alias and the FSSpec was cleared, so we create an 
			// empty alias record and append it to outTestInfoH
		AliasRecord aliasRec = { 0, 0 };
		if (::PtrAndHand(&aliasRec, outTestInfoH, sizeof(AliasRecord)) != noErr)
			return false;
	}		
	
	if (mTestInfo.bitmapFSSpec.name[0] > 0) {
			// we need to create an alias record for the currently selected 
			// bitmap so that we can find it next time this document is loaded
		FSSpec appFileSpec;
		S3GetAppFileSpec(&appFileSpec);
		
		if (appFileSpec.name[0] > 0) {
				// we got a good file spec, so create an alias from the position
				// of the app to the selected PICT
			AliasHandle bitmapH = nil;
			::NewAlias(&appFileSpec, &mTestInfo.bitmapFSSpec, &bitmapH);
			
			if (bitmapH != nil) {
					// we got a good alias handle, so append it to outTestInfoH
				StHandleLocker(Handle(bitmapH));
				OSErr err = ::HandAndHand((Handle) bitmapH, outTestInfoH);
				
				::DisposeHandle((Handle) bitmapH);
				
				if (err != noErr)
					return false;
			} else {
					// an alias couldn't be created for some reason (maybe the file
					// was moved, so the FSSpec no longer points to anything), so 
					// rather than returning an error, just set the bitmap FSSpec
					// in the test info handle to be empty.  we don't have to bother
					// emptying the one in mTestInfo because that's already been copied
					// to the handle.
				memset(&(testInfoP->textureFSSpec), unsigned char(0), sizeof(FSSpec));
			}
		}
	} 
	
	if (testInfoP->bitmapFSSpec.name[0] == 0) {
			// no bitmap file has been selected, or there was a problem in 
			// creating the alias and the FSSpec was cleared, so we create an 
			// empty alias record and append it to outTestInfoH
		AliasRecord aliasRec = { 0, 0 };
		if (::PtrAndHand(&aliasRec, outTestInfoH, sizeof(AliasRecord)) != noErr)
			return false;
	}
	
	mIsDirty = false;
	
	return true;
}


// ===========================================================================
	Boolean
CTestInfo::VerifySettingsVersion(const Handle inTestInfoH, Int16 inSettingsVersion)
{
	Boolean settingsVersionIsCompatible = false;

		// each case statement updates the settings handle to a version which the 
		// next case expects.  for instance, case 0 updates it to version 1200, which 
		// case 1200 in turn updates to 1205.  each case falls through to the next 
		// one until the settings handle is completely up to date.  if the settings
		// handle is newer than this version of Sneak3view expects, then none of the
		// cases will satisfy the switch and inSettingsVersion will be false.
	switch (inSettingsVersion) {
	  case 0: {
				// this version of the settings resource doesn't have an argString,
				// so increase the size of the handle to hold a 256-byte string
			::SetHandleSize(inTestInfoH, ::GetHandleSize(inTestInfoH) + sizeof(Str255));
			
			if (::MemError() != noErr) {
					// we can't use this settings handle if we can't make it big
					// enough to hold the new argString field
				return false;
			}
			
				// get a pointer to the first data field after testName, which
				// is two longs and a Str255 past the beginning of the structure,
				// in version 0 of S3tTestInfo
			S3tTestInfoP testInfoP = (S3tTestInfoP) *inTestInfoH;
			Ptr dataAfterTestNameP = 
				Ptr(testInfoP) + sizeof(mTestInfo.leftMicroseconds) + 
				sizeof(mTestInfo.rightMicroseconds) + sizeof(mTestInfo.testName);
			
				// get the size of the data following testName, so we'll know how
				// much to move.  we have to subtract sizeof(mTestInfo.argString) 
				// because sizeof(mTestInfo) includes the argString field, which
				// we know isn't included in this handle.
			long dataAfterTestNameSize = 
				sizeof(mTestInfo) - sizeof(mTestInfo.leftMicroseconds) -
				sizeof(mTestInfo.rightMicroseconds) - sizeof(mTestInfo.testName) -
				sizeof(mTestInfo.argString);

				// move everything after testName 256 bytes over, to make room for
				// the argString field
			::BlockMoveData(dataAfterTestNameP, dataAfterTestNameP + sizeof(Str255),
											dataAfterTestNameSize);
											
				// clear the new argString field
			memset(dataAfterTestNameP, unsigned char(0), sizeof(mTestInfo.argString));
		}
			
		// *** fall through to next case ***
			
		case 1200: {
				// this version of the settings doesn't have the the texture and bitmap 
				// color table data at the end, so we have to increase the size of the 
				// handle to accommodate these.
			::SetHandleSize(inTestInfoH, ::GetHandleSize(inTestInfoH) + 
					sizeof(TQAColorTable*) + sizeof(TQAColorTable*) + 
					sizeof(TQAColorTableType) + sizeof(Boolean) + 
					sizeof(TQAColorTable*) + sizeof(TQAColorTable*) + 
					sizeof(TQAColorTableType) + sizeof(Boolean));
			
			if (::MemError() != noErr) {
					// we can't use this settings handle if we can't make it big
					// enough to hold the new fields
				return false;
			}
			
				// get a pointer to the S3tTestInfo in the handle and set the newly-added
				// fields to default values, since right now there's garbage in them
			S3tTestInfoP testInfoP = (S3tTestInfoP) *inTestInfoH;
			
			testInfoP->leftTextureColorTable = nil;
			testInfoP->rightTextureColorTable = nil;
			testInfoP->textureUseTransparentIndex = false;
			testInfoP->leftBitmapColorTable = nil;
			testInfoP->rightBitmapColorTable = nil;
			testInfoP->bitmapUseTransparentIndex = false;
		}
			
		// *** fall through to next case ***
			
	  case kCurrentSettingsVersion:
			settingsVersionIsCompatible = true;
			break;
	}

	return settingsVersionIsCompatible;
}


// ===========================================================================
	Boolean
CTestInfo::PrepareForTest()
{
		// in case we've already created draw contexts for the current TestInfo,
		// dispose of any handles it contains
//	DisposeTestInfo(&mTestInfo);

	mTestInfo.leftMicroseconds = 0;
	mTestInfo.rightMicroseconds = 0;

	if (mTestInfo.leftEngine == nil || mTestInfo.rightEngine == nil)  {
			// we can't create a draw context without knowing which engines to use
		S3StopAlert(STRx_AlertStrings_ID, AlertString_EngineNotFound, nil);

		return false;
	}
	
	if (mPreparedTestInfo != nil) {
		DisposeTestInfo(mPreparedTestInfo);
		delete mPreparedTestInfo;
		mPreparedTestInfo = nil;
	}
	
		// make a copy of our TestInfo.  this copy is given to the test to use, so
		// that it can change all the fields without affecting the settings window.
	mPreparedTestInfo = new S3tTestInfo(mTestInfo);

	if (mPreparedTestInfo == nil) {
			// we don't have enough memory to make the copy
		return false;
	}
	
		// convert the Pascal argString to a C string, so that the tests can 
		// sscanf to extract the arguments from it.  also convert the name of the test.
	S3PtoCString((StringPtr) &mPreparedTestInfo->argString);
	S3PtoCString((StringPtr) &mPreparedTestInfo->testName);
	
		// make sure we meet the prerequisites for the current test.  we use
		// the test name from mTestInfo rather than from mPreparedTestInfo
		// because we just converted the Pascal format name into a C string.
	Assert_(mTestInfo.testName[0] > 0);
	S3tTestTool theTest;
	CTestPopupMenu::GetSharedDataFromString(mTestInfo.testName, &theTest);
	
	if (theTest.testPrereqs & kS3Prereq_Texture) {
			// this test requires a texture, so make sure the user has selected one
		if (mPreparedTestInfo->textureFSSpec.name[0] > 0) {
				// a texture is selected in the menu, so load the specified PICT 
				// file and create a texture from it for each engine
			if (LoadTextureMap(mPreparedTestInfo) == false)
					// something went wrong and we couldn't create the textures
				return false;
		} else {
				// the test needs a texture, but the user hasn't selected one,
				// so alert them
			S3StopAlert(STRx_AlertStrings_ID, AlertString_TestRequiresTexture, nil);
			
			return false;
		}
	}

	if (theTest.testPrereqs & kS3Prereq_Bitmap) {
			// this test requires a texture, so make sure the user has selected one
		if (mPreparedTestInfo->bitmapFSSpec.name[0] > 0) {
				// a bitmap is selected in the menu, so load the specified PICT 
				// file and create a bitmap from it for each engine
			if (LoadBitmap(mPreparedTestInfo) == false)
					// something went wrong and we couldn't create the bitmaps
				return false;
		} else {
				// the test needs a bitmap, but the user hasn't selected one,
				// so alert them
			S3StopAlert(STRx_AlertStrings_ID, AlertString_TestRequiresBitmap, nil);
			
			return false;
		}
	}

		// set up the devices for the draw contexts
	switch (mPreparedTestInfo->deviceMenu) {
		case kS3Device_Window:
			if (SetupGDevice(mPreparedTestInfo) == false) {
					// the GDevice couldn't be set up correctly, so we can't
					// run the test
				return false;
			}
			break;
		
		case kS3Device_Memory16:
		case kS3Device_Memory32:
			if (SetupMemoryDevice(mPreparedTestInfo) == false) {
					// the memory device couldn't be set up correctly, so we can't
					// run the test
				return false;
			}
			break;
	} // endswitch

		// now make the left clipping region, using local device coords
	mPreparedTestInfo->leftClip.clip.clipRgn = 
		MakeClipRegion(&(mPreparedTestInfo->leftRect), mPreparedTestInfo->clipRegionMenu);
																				 
		// create the draw context itself, using a clipping region if necessary
	TQAError qaError = QADrawContextNew(
		&mPreparedTestInfo->leftDevice, 
		&mPreparedTestInfo->leftRect, 
		(mPreparedTestInfo->leftClip.clip.clipRgn) ? &mPreparedTestInfo->leftClip : nil,
		mPreparedTestInfo->leftEngine,
		mPreparedTestInfo->contextFlags,
		&mPreparedTestInfo->leftContext);
																					 
	if (qaError != kQANoErr || mPreparedTestInfo->leftContext == nil) {
			// we couldn't create the left draw context for some reason
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftContext, nil);
		
		return false;
	}

		// now make the right clipping region, using local device coords
	mPreparedTestInfo->rightClip.clip.clipRgn = 
		MakeClipRegion(&(mPreparedTestInfo->rightRect), mPreparedTestInfo->clipRegionMenu);
																				 
		// create the draw context itself, using a clipping region if necessary
	qaError = QADrawContextNew(
		&mPreparedTestInfo->rightDevice, 
		&mPreparedTestInfo->rightRect, 
		(mPreparedTestInfo->rightClip.clip.clipRgn) ? &mPreparedTestInfo->rightClip : nil, 
		mPreparedTestInfo->rightEngine,
		mPreparedTestInfo->contextFlags,
		&mPreparedTestInfo->rightContext);

	if (qaError != kQANoErr || mPreparedTestInfo->rightContext == nil) {
			// we couldn't create the right draw context for some reason
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightContext, nil);
		
		return false;
	}
	
		// now that the draw contexts have been created, set the state variables
		// for each one
	S3Test_SetInt(mPreparedTestInfo, kQATag_ZFunction, mPreparedTestInfo->zFunction);

	S3Test_SetFloat(mPreparedTestInfo, kQATag_ColorBG_a, mPreparedTestInfo->alpha);

	S3Test_SetFloat(mPreparedTestInfo, kQATag_ColorBG_r, mPreparedTestInfo->red);

	S3Test_SetFloat(mPreparedTestInfo, kQATag_ColorBG_g, mPreparedTestInfo->green);

	S3Test_SetFloat(mPreparedTestInfo, kQATag_ColorBG_b, mPreparedTestInfo->blue);

	S3Test_SetFloat(mPreparedTestInfo, kQATag_Width, mPreparedTestInfo->lineWidth);
	
	S3Test_SetInt(mPreparedTestInfo, kQATag_Antialias, mPreparedTestInfo->antialiasing);

	S3Test_SetInt(mPreparedTestInfo, kQATag_Blend, mPreparedTestInfo->blending);

	S3Test_SetInt(mPreparedTestInfo, kQATag_PerspectiveZ, mPreparedTestInfo->perspectiveZ);

		// if textures have been created, set each draw context's current
		// texture separately
	if (mPreparedTestInfo->leftTexture != nil) {
		QASetPtr(mPreparedTestInfo->leftContext, kQATag_Texture, mPreparedTestInfo->leftTexture);
	}
	
	if (mPreparedTestInfo->rightTexture != nil) {
		QASetPtr(mPreparedTestInfo->rightContext, kQATag_Texture, mPreparedTestInfo->rightTexture);
	}
	
	S3Test_SetInt(mPreparedTestInfo, kQATag_TextureFilter, mPreparedTestInfo->textureFilter);

	S3Test_SetInt(mPreparedTestInfo, kQATag_TextureOp, mPreparedTestInfo->textureOp);
	
	if (mPreparedTestInfo->resetRandomSeed) {
			// reset the seed used for the S3TestUtil_RandomFloat function
		S3TestUtil_ResetRandomSeed();
	}
	
	return true;
}


// ===========================================================================
	void
CTestInfo::CleanUpAfterTest()
{
	if (mPreparedTestInfo == nil) {
		return;
	}
	
	if (mPreparedTestInfo->leftDevice.deviceType == kQADeviceMemory) {
			// we're using a memory device, so we have to copy the pixels from
			// the off-screen memory to the screen
		if (mPreparedTestInfo->leftDeviceGWorld != nil) {
				// save the current GWorld and restore it when we leave the block
			StGWorldState();
			
				// prepare to draw into the left window
			CTinselTestWindow::GetLeftWindow()->FocusDraw();
			GrafPtr windowP = CTinselTestWindow::GetLeftWindow()->GetMacPort();
			
				// make sure the colors are set correctly for the CopyBits
			::ForeColor(blackColor);
			::BackColor(whiteColor);
			
				// copy the offscreen GWorld to the window
			mPreparedTestInfo->leftDeviceGWorld->CopyImage(windowP, windowP->portRect);

				// unlock the pixels now that we're done with them
			::UnlockPixels(::GetGWorldPixMap(mPreparedTestInfo->leftDeviceGWorld->GetMacGWorld()));
		}
	}
	
	if (mPreparedTestInfo->rightDevice.deviceType == kQADeviceMemory) {
			// we're using a memory device, so we have to copy the pixels from
			// the off-screen memory to the screen
		if (mPreparedTestInfo->rightDeviceGWorld != nil) {
				// save the current GWorld and restore it when we leave the block
			StGWorldState();
			
				// prepare to draw into the right window
			CTinselTestWindow::GetRightWindow()->FocusDraw();
			GrafPtr windowP = CTinselTestWindow::GetRightWindow()->GetMacPort();
			
				// make sure the colors are set correctly for the CopyBits
			::ForeColor(blackColor);
			::BackColor(whiteColor);
			
				// copy the offscreen GWorld to the window
			mPreparedTestInfo->rightDeviceGWorld->CopyImage(windowP, windowP->portRect);

				// unlock the pixels now that we're done with them
			::UnlockPixels(::GetGWorldPixMap(mPreparedTestInfo->rightDeviceGWorld->GetMacGWorld()));
		}
	}
	
//		// convert the C argString back to a Pascal string
//	S3CtoPString(mTestInfo.argString);
	
		// now that we're done with the test and have copied any off-screen 
		// data to the screen, we can dispose of whatever memory we were using
	DisposeTestInfo(mPreparedTestInfo);
	
	delete mPreparedTestInfo;
	mPreparedTestInfo = nil;
}


// ===========================================================================
	S3tTestInfoP
CTestInfo::GetTestInfo()
{
	if (PrepareForTest() == false) {
			// for some reason, we couldn't allocate everything we need to to 
			// create a valid S3tTestInfo, so return nil.  but first clean up
			// whatever we were able to actually allocate.
		CleanUpAfterTest();
		
		return nil;
	}
	
	return mPreparedTestInfo;
}


// ===========================================================================
	void
CTestInfo::ListenToMessage(MessageT inMessage, void* ioParam)
{
		// since we got a broadcast, we know one of the controls was changed.
		// LControl doesn't broadcast the value unless it is actually different
		// from the previous one.
	mIsDirty = true;
	
	switch (inMessage) {
		case msg_Width:
		case msg_Height:
		case msg_LeftWindow_Left:
		case msg_LeftWindow_Top:
		case msg_RightWindow_Left:
		case msg_RightWindow_Top:
			SetWindows(inMessage, ioParam);
			break;

		case msg_TestWindow_Drag + TestWindow_Left:
		case msg_TestWindow_Drag + TestWindow_Right:
			SetWindowDrag(inMessage, ioParam);
			break;

		case msg_TestWindow_Grow:
			SetWindowSize(ioParam);
			break;

	  case msg_ResetWindows:
			CTinselTestWindow::GetDefaultGlobalRects(
				nil,
				&mTestInfo.leftGlobalRect, 
				&mTestInfo.rightGlobalRect);
			
				// let the test windows know their rects have changed
			UpdateTestWindows();
			UpdateSettingsWindow();
			break;

		case msg_LeftEngine:
		case msg_RightEngine:
			SetEngine(inMessage, ioParam);
			break;

		case msg_Device:
			mTestInfo.deviceMenu = (S3eDevice) *(unsigned int*) ioParam;
			break;

		case msg_ClipRegion:
			mTestInfo.clipRegionMenu = (S3eClipRegion) *(unsigned int*) ioParam;
			break;

		case msg_Buffering:
			SetBuffering(ioParam);
			break;

		case msg_ZBuffering:
			SetZBuffering(ioParam);
			break;

		case msg_ZFunction:
			mTestInfo.zFunction = *(unsigned long*) ioParam;
			break;

		case msg_PerspectiveZ:
			mTestInfo.perspectiveZ = *(unsigned long*) ioParam;
			break;

		case msg_Blending:
			mTestInfo.blending = *(unsigned long*) ioParam;
			break;

		case msg_Antialiasing:
			mTestInfo.antialiasing = *(unsigned long*) ioParam;
			break;

		case msg_ArgString: {
					// CBroadcastEditField panes broadcast a pointer to themselves 
					// in the ioParam, so we cast it to something we can use and then
					// get the field's descriptor into the argString field
				Assert_(ioParam);
				CBroadcastEditField* theEditField = (CBroadcastEditField*) ioParam;
				theEditField->GetDescriptor((StringPtr) mTestInfo.argString);
			}
			break;

		case msg_Background_Alpha:
		case msg_Background_Red:
		case msg_Background_Green:
		case msg_Background_Blue:
			SetBackground(inMessage, ioParam);
			break;

		case msg_LineWidth: {
					// CBroadcastEditField panes broadcast a pointer to themselves 
					// in the ioParam, so we cast it to something we can use and then
					// get the field's value as a float
				Assert_(ioParam);
				CBroadcastEditField* theEditField = (CBroadcastEditField*) ioParam;
				mTestInfo.lineWidth = theEditField->GetValueAsFloat();
			}
			break;

		case msg_Texture: {
					// CPICTPopupMenu panes broadcast a pointer to themselves 
					// in the ioParam, so we cast it to something we can use
					// and then get the FSSpec of the currently selected texture
				Assert_(ioParam);
				CPICTPopupMenu* pictMenuP = (CPICTPopupMenu*) ioParam;
				pictMenuP->GetData(&mTestInfo.textureFSSpec);
			}
			break;

		case msg_TextureFormat:
			mTestInfo.textureFormat = *(TQAImagePixelType*) ioParam;
			
				// textureFormat and textureColorTableType have to be kept in sync.  if
				// the new textureFormat is CL8 or CL4, check textureColorTableType to
				// get the bit depth of the color table and record the appropriate
				// CL8 or CL4 format.
			if (mTestInfo.textureFormat == kQAPixel_CL8) {
				mTestInfo.textureColorTableType = kQAColorTable_CL8_RGB32;
			} else if (mTestInfo.textureFormat == kQAPixel_CL4) {
				mTestInfo.textureColorTableType = kQAColorTable_CL4_RGB32;
			}
//			if (mTestInfo.textureFormat == kQAPixel_CL8) {
//				mTestInfo.textureColorTableType = 
//						(mTestInfo.textureColorTableType == kQAColorTable_CL8_RGB32 ||
//						 mTestInfo.textureColorTableType == kQAColorTable_CL4_RGB32) ? kQAColorTable_CL8_RGB32 : kQACL8_RGB16;
//			} else if (mTestInfo.textureFormat == kQAPixel_CL4) {
//				mTestInfo.textureColorTableType = 
//						(mTestInfo.textureColorTableType == kQAColorTable_CL8_RGB32 ||
//						 mTestInfo.textureColorTableType == kQAColorTable_CL4_RGB32) ? kQAColorTable_CL4_RGB32 : kQACL4_RGB16;
//			}
			break;

		case msg_TextureFilter:
			mTestInfo.textureFilter = *(unsigned long*) ioParam;
			break;

		case msg_TextureOp_Modulate:
		case msg_TextureOp_Highlight:
		case msg_TextureOp_Decal:
		case msg_TextureOp_Shrink:
			SetTextureOp(inMessage, ioParam);
			break;

		case msg_CreateMipmap:
			if (*(int*) ioParam == Button_On) 
				mTestInfo.textureFlags |= kQATexture_Mipmap;
			else
				mTestInfo.textureFlags &= ~kQATexture_Mipmap;
			break;

		case msg_LockTexture:
			if (*(int*) ioParam == Button_On) 
				mTestInfo.textureFlags |= kQATexture_Lock;
			else
				mTestInfo.textureFlags &= ~kQATexture_Lock;
			break;

		case msg_DetachTexture:
			if (*(int*) ioParam == Button_On) 
				mTestInfo.detachTexture = true;
			else
				mTestInfo.detachTexture = false;
			break;

		case msg_TextureMakeBlackTransparent:
			if (*(int*) ioParam == Button_On) 
				mTestInfo.textureMakeBlackTransparent = true;
			else
				mTestInfo.textureMakeBlackTransparent = false;
			break;

//		case msg_TextureTableFormat: {
//			S3eColorTableFormat tableFormat = (S3eColorTableFormat) *(unsigned int*) ioParam;
//			if (mTestInfo.textureFormat == kQAPixel_CL8) {
//				mTestInfo.textureColorTableType = 
//						(tableFormat == kS3ColorTableFormat_16) ? kQACL8_RGB16 : kQAColorTable_CL8_RGB32;
//			} else if (mTestInfo.textureFormat == kQAPixel_CL4) {
//				mTestInfo.textureColorTableType = 
//						(tableFormat == kS3ColorTableFormat_16) ? kQACL4_RGB16 : kQAColorTable_CL4_RGB32;
//			}
//			break;
//		}

		case msg_TextureUseTransparentIndex: 
			if (*(int*) ioParam == Button_On) 
				mTestInfo.textureUseTransparentIndex = true;
			else
				mTestInfo.textureUseTransparentIndex = false;
			break;

		case msg_Bitmap: {
					// CPICTPopupMenu panes broadcast a pointer to themselves 
					// in the ioParam, so we cast it to something we can use
					// and then get the FSSpec of the currently selected bitmap
				Assert_(ioParam);
				CPICTPopupMenu* pictMenuP = (CPICTPopupMenu*) ioParam;
				pictMenuP->GetData(&mTestInfo.bitmapFSSpec);
			}
			break;
		
		case msg_BitmapFormat:
			mTestInfo.bitmapFormat = *(TQAImagePixelType*) ioParam;
			
				// bitmapFormat and bitmapColorTableType have to be kept in sync.  if
				// the new bitmapFormat is CL8 or CL4, check bitmapColorTableType to
				// get the bit depth of the color table and record the appropriate
				// CL8 or CL4 format.
			if (mTestInfo.bitmapFormat == kQAPixel_CL8) {
				mTestInfo.bitmapColorTableType = kQAColorTable_CL8_RGB32;
			} else if (mTestInfo.bitmapFormat == kQAPixel_CL4) {
				mTestInfo.bitmapColorTableType = kQAColorTable_CL4_RGB32;
			}
//			if (mTestInfo.bitmapFormat == kQAPixel_CL8) {
//				mTestInfo.bitmapColorTableType = 
//						(mTestInfo.bitmapColorTableType == kQAColorTable_CL8_RGB32 ||
//						 mTestInfo.bitmapColorTableType == kQAColorTable_CL4_RGB32) ? kQAColorTable_CL8_RGB32 : kQACL8_RGB16;
//			} else if (mTestInfo.bitmapFormat == kQAPixel_CL4) {
//				mTestInfo.bitmapColorTableType = 
//						(mTestInfo.bitmapColorTableType == kQAColorTable_CL8_RGB32 ||
//						 mTestInfo.bitmapColorTableType == kQAColorTable_CL4_RGB32) ? kQAColorTable_CL4_RGB32 : kQACL4_RGB16;
//			}
			break;

		case msg_LockBitmap:
			if (*(int*) ioParam == Button_On) 
				mTestInfo.bitmapFlags |= kQABitmap_Lock;
			else
				mTestInfo.bitmapFlags &= ~kQABitmap_Lock;
			break;

		case msg_DetachBitmap:
			if (*(int*) ioParam == Button_On) 
				mTestInfo.detachBitmap = true;
			else
				mTestInfo.detachBitmap = false;
			break;

		case msg_BitmapMakeBlackTransparent:
			if (*(int*) ioParam == Button_On) 
				mTestInfo.bitmapMakeBlackTransparent = true;
			else
				mTestInfo.bitmapMakeBlackTransparent = false;
			break;

//		case msg_BitmapTableFormat: {
//			S3eColorTableFormat tableFormat = (S3eColorTableFormat) *(unsigned int*) ioParam;
//			if (mTestInfo.bitmapFormat == kQAPixel_CL8) {
//				mTestInfo.bitmapColorTableType = 
//						(tableFormat == kS3ColorTableFormat_16) ? kQACL8_RGB16 : kQAColorTable_CL8_RGB32;
//			} else if (mTestInfo.bitmapFormat == kQAPixel_CL4) {
//				mTestInfo.bitmapColorTableType = 
//						(tableFormat == kS3ColorTableFormat_16) ? kQACL4_RGB16 : kQAColorTable_CL4_RGB32;
//			}
//			break;
//		}

		case msg_BitmapUseTransparentIndex: 
			if (*(int*) ioParam == Button_On) 
				mTestInfo.bitmapUseTransparentIndex = true;
			else
				mTestInfo.bitmapUseTransparentIndex = false;
			break;

		case msg_ResetRandomSeed:
			if (*(int*) ioParam == Button_On) 
				mTestInfo.resetRandomSeed = true;
			else
				mTestInfo.resetRandomSeed = false;
			break;
		
		case msg_Test: 
			CopyPStr((ConstStr255Param) ioParam, mTestInfo.testName, sizeof(mTestInfo.testName));
			mTestInfo.argString[0] = 0;
	
				// set the argString broadcast edit field to an empty string
			CBroadcastEditField* argStringPane = 
				(CBroadcastEditField*) mSettingsWindow->FindPaneByID(paneID_ArgString);
			argStringPane->SetDescriptor("\p");
			break;
		
		case msg_BroadcasterDied: 
				// we don't really care if a broadcaster has died, since it should
				// only occur when the settings window is deleting all of its 
				// controls, at which time we're about to be destroyed ourselves.
			break;
		
		default:
			SignalPStr_("\pFound unrecognized message in CTestInfo::ListenToMessage");
			break;
	}
}


// ===========================================================================
	void 
CTestInfo::AddEditFieldBroadcaster(PaneIDT inPaneID)
{
		// we have to link the CBroadcastEditField views to us this way because
		// UReanimator::LinkListenerToControls assumes it's linking a listener 
		// to an LControl, which CBroadcastEditField is not
	CBroadcastEditField* theEditField = 
		(CBroadcastEditField*) mSettingsWindow->FindPaneByID(inPaneID);
	theEditField->AddListener(this);
}


// ===========================================================================
	void 
CTestInfo::SetFieldToAcceptFloats(PaneIDT inPaneID)
{
	CBroadcastEditField* theEditField = 
		(CBroadcastEditField*) mSettingsWindow->FindPaneByID(inPaneID);
	theEditField->SetKeyFilter(CBroadcastEditField::UnsignedFloatField);
}


// ===========================================================================
	void 
CTestInfo::SetFieldToAcceptSignedInts(PaneIDT inPaneID)
{
	CBroadcastEditField* theEditField = 
		(CBroadcastEditField*) mSettingsWindow->FindPaneByID(inPaneID);
	theEditField->SetKeyFilter(CBroadcastEditField::SignedIntField);
}


// ===========================================================================
	void
CTestInfo::LinkEnablerViewToControl(
	PaneIDT 									inEnablerViewID, 
	PaneIDT 									inControlID, 
	EnablerFunctionPtr 				inEnablerFunction)
{
	CEnablerView* theEnablerView = 
		(CEnablerView*) mSettingsWindow->FindPaneByID(inEnablerViewID);
		
	theEnablerView->SetEnablerFunction(inEnablerFunction);
		
	LControl* theControl = 
		(LControl*) mSettingsWindow->FindPaneByID(inControlID);
		
	theControl->AddListener(theEnablerView);
}


// ===========================================================================
	void 
CTestInfo::LinkSettingsDoc(CSettingsDoc* inDoc, LWindow* inWindow)
{
	mSettingsDoc = inDoc;
	mSettingsWindow = inWindow;
	
	Assert_(mSettingsDoc);
	Assert_(mSettingsWindow);

		// link the controls in the settings window to us
	UReanimator::LinkListenerToControls(this, mSettingsWindow, RidL_DrawContextSettings);

		// link the enabler view to the format popup so that the view can disable
		// itself when the popup is set to the appropriate values
	LinkEnablerViewToControl(paneID_TextureEnablerView, paneID_TextureFormat, 
			&TextureEnablerFunction);
	LinkEnablerViewToControl(paneID_BitmapEnablerView, paneID_BitmapFormat, 
			&BitmapEnablerFunction);
	
		// link the CBroadcastEditField views in the settings window to us
	AddEditFieldBroadcaster(paneID_Width);
	AddEditFieldBroadcaster(paneID_Height);
	AddEditFieldBroadcaster(paneID_LeftWindow_Left);
	AddEditFieldBroadcaster(paneID_LeftWindow_Top);
	AddEditFieldBroadcaster(paneID_RightWindow_Left);
	AddEditFieldBroadcaster(paneID_RightWindow_Top);
	AddEditFieldBroadcaster(paneID_Background_Alpha);
	AddEditFieldBroadcaster(paneID_Background_Red);
	AddEditFieldBroadcaster(paneID_Background_Green);
	AddEditFieldBroadcaster(paneID_Background_Blue);
	AddEditFieldBroadcaster(paneID_LineWidth);
	AddEditFieldBroadcaster(paneID_ArgString);

		// set the left and top fields to accept positive and negative ints.
		// this is necessary, since these values are in device coordinates,
		// which can be negative
	SetFieldToAcceptSignedInts(paneID_LeftWindow_Left);
	SetFieldToAcceptSignedInts(paneID_LeftWindow_Top);
	SetFieldToAcceptSignedInts(paneID_RightWindow_Left);
	SetFieldToAcceptSignedInts(paneID_RightWindow_Top);
	
		// set the ARGB and line width fields to accept positive floats.
	SetFieldToAcceptFloats(paneID_Background_Alpha);
	SetFieldToAcceptFloats(paneID_Background_Red);
	SetFieldToAcceptFloats(paneID_Background_Green);
	SetFieldToAcceptFloats(paneID_Background_Blue);
	SetFieldToAcceptFloats(paneID_LineWidth);
	
	if (mTestInfo.testName[0] < 1) {
			// by the time the settings window is linked with us, the CTestPopupMenu
			// will have been initialized, so that we can now get the name of the
			// first test in the menu.  it's important to do this here, before 
			// UpdateSettingsWindow gets called, so that when it does, we have a 
			// valid test name for it to set the popup's value with.  we only do
			// this, of course, if we don't already have a valid name in testName.
			// otherwise, we would overwrite the setting we read in from a saved file.
		CTestPopupMenu::GetSharedStringFromItem(1, mTestInfo.testName);
	}
		
		// since this window has just been attached to us, make sure its 
		// controls reflect the values in our mTestInfo instance
	UpdateSettingsWindow();
}


// ===========================================================================
	void 
CTestInfo::LinkTestWindows(CTinselTestWindow* inLeftWindow, 
													 CTinselTestWindow* inRightWindow)
{
		// we broadcast to the test windows when their position and size edit 
		// fields change, and they broadcast to us when they're dragged, so
		// we add each other as listeners
	if (inLeftWindow != nil) {
		AddListener(inLeftWindow);
		inLeftWindow->AddListener(this);
	}
	
	if (inRightWindow != nil) {
		AddListener(inRightWindow);
		inRightWindow->AddListener(this);
	}
}		


// ===========================================================================
	void 
CTestInfo::SetControlValue(PaneIDT inPaneID, Int32 inValue)
{
	LControl* thePane = (LControl*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (thePane == nil)
		return;
		
	thePane->SetValue(inValue);
}


// ===========================================================================
	void 
CTestInfo::SetPvalPopupValue(PaneIDT inPaneID, Int32 inPval)
{
	CPvalPopupMenu* thePane = (CPvalPopupMenu*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (thePane == nil)
		return;
		
	thePane->SetValueByPval(inPval);
}


// ===========================================================================
	void 
CTestInfo::SetEditFieldValue(PaneIDT inPaneID, long inLongValue)
{
	CBroadcastEditField* thePane = (CBroadcastEditField*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (thePane == nil)
		return;
		
	thePane->SetValue(inLongValue);
}


// ===========================================================================
	void 
CTestInfo::SetEditFieldValue(PaneIDT inPaneID, float inFloatValue)
{
	CBroadcastEditField* thePane = (CBroadcastEditField*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (thePane == nil)
		return;
		
	thePane->SetValueAsFloat(inFloatValue);
}


// ===========================================================================
	void 
CTestInfo::SetPICTPopupValue(PaneIDT inPaneID, FSSpec* inFSSpec)
{
	CPICTPopupMenu* thePane = (CPICTPopupMenu*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (thePane == nil)
		return;
		
	thePane->SetValueByData(inFSSpec);
}


// ===========================================================================
	void 
CTestInfo::SetStringPopupValue(PaneIDT inPaneID, char* inString)
{
	CSmartStdPopupMenu* thePane = (CSmartStdPopupMenu*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (thePane == nil)
		return;
		
	thePane->SetValueByString(inString);
}


// ===========================================================================
	void 
CTestInfo::SetStringPopupValue(PaneIDT inPaneID, Str255 inString)
{
	CSmartStdPopupMenu* thePane = (CSmartStdPopupMenu*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (thePane == nil)
		return;
		
	thePane->SetValueByString(inString);
}


// ===========================================================================
	void 
CTestInfo::SetTableFormatPopupValue(PaneIDT inPaneID, TQAColorTableType inTableType)
{
	CPvalPopupMenu* thePane = (CPvalPopupMenu*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (thePane == nil)
		return;

	switch (inTableType) {
//		case kQACL8_RGB16:
//		case kQACL4_RGB16:
//			thePane->SetValueByPval(kS3ColorTableFormat_16);
//			break;
		
		case kQAColorTable_CL8_RGB32:
		case kQAColorTable_CL4_RGB32:
			thePane->SetValueByPval(kS3ColorTableFormat_32);
			break;
	} // endswitch
}


// ===========================================================================
	void
CTestInfo::SetEnablerViewState(
	PaneIDT 									inPaneID, 
	MessageT 									inMessage, 
	void*											inParam)
{
	CEnablerView* theView = (CEnablerView*) mSettingsWindow->FindPaneByID(inPaneID);
	
	if (theView == nil)
		return;

	theView->ListenToMessage(inMessage, inParam);
}


// ===========================================================================
	void 
CTestInfo::UpdateSettingsWindow()
{
	if (mSettingsWindow == nil) 
		return;

		// since we listen to these controls, if we don't stop listening before
		// changing each of their values, we'll get a ListenToMessage call every
		// time they change
	StopListening();

		// get each pane and set its value according to the one stored in mTestInfo
	SetEditFieldValue(paneID_Width, 
										(long) (mTestInfo.leftGlobalRect.right - 
														mTestInfo.leftGlobalRect.left));
	SetEditFieldValue(paneID_Height, 
										(long) (mTestInfo.leftGlobalRect.bottom -
														mTestInfo.leftGlobalRect.top));
	SetEditFieldValue(paneID_LeftWindow_Left, (long) mTestInfo.leftGlobalRect.left);
	SetEditFieldValue(paneID_LeftWindow_Top, (long) mTestInfo.leftGlobalRect.top);
	SetEditFieldValue(paneID_RightWindow_Left, (long) mTestInfo.rightGlobalRect.left);
	SetEditFieldValue(paneID_RightWindow_Top, (long) mTestInfo.rightGlobalRect.top);

	S3GetEngineName(mTestInfo.leftEngine, mTestInfo.leftEngineName, 
									sizeof(mTestInfo.leftEngineName));
	SetStringPopupValue(paneID_LeftEngine, mTestInfo.leftEngineName);

	S3GetEngineName(mTestInfo.rightEngine, mTestInfo.rightEngineName, 
									sizeof(mTestInfo.rightEngineName));
	SetStringPopupValue(paneID_RightEngine, mTestInfo.rightEngineName);
	
	SetPvalPopupValue(paneID_Device, mTestInfo.deviceMenu);
	SetPvalPopupValue(paneID_ClipRegion, mTestInfo.clipRegionMenu);

	SetPvalPopupValue(paneID_Buffering, GetBuffering());
	SetPvalPopupValue(paneID_ZBuffering, GetZBuffering());

	SetPvalPopupValue(paneID_ZFunction, mTestInfo.zFunction);
	SetPvalPopupValue(paneID_PerspectiveZ, mTestInfo.perspectiveZ);
	SetPvalPopupValue(paneID_Blending, mTestInfo.blending);
	SetPvalPopupValue(paneID_Antialiasing, mTestInfo.antialiasing);

	SetEditFieldValue(paneID_Background_Alpha, mTestInfo.alpha);
	SetEditFieldValue(paneID_Background_Red, mTestInfo.red);
	SetEditFieldValue(paneID_Background_Green, mTestInfo.green);
	SetEditFieldValue(paneID_Background_Blue, mTestInfo.blue);
	SetEditFieldValue(paneID_LineWidth, mTestInfo.lineWidth);

	SetPICTPopupValue(paneID_Texture, &mTestInfo.textureFSSpec);
	SetPvalPopupValue(paneID_TextureFormat, mTestInfo.textureFormat);
	SetPvalPopupValue(paneID_TextureFilter, mTestInfo.textureFilter);

	SetControlValue(paneID_TextureOp_Modulate, mTestInfo.textureOp & kQATextureOp_Modulate);
	SetControlValue(paneID_TextureOp_Highlight, mTestInfo.textureOp & kQATextureOp_Highlight);
	SetControlValue(paneID_TextureOp_Decal, mTestInfo.textureOp & kQATextureOp_Decal);
	SetControlValue(paneID_TextureOp_Shrink, mTestInfo.textureOp & kQATextureOp_Shrink);
	SetControlValue(paneID_CreateMipmap, mTestInfo.textureFlags & kQATexture_Mipmap);
	SetControlValue(paneID_LockTexture, mTestInfo.textureFlags & kQATexture_Lock);
	SetControlValue(paneID_DetachTexture, mTestInfo.detachTexture);
	SetControlValue(paneID_TextureMakeBlackTransparent, mTestInfo.textureMakeBlackTransparent);
	SetControlValue(paneID_TextureUseTransparentIndex, mTestInfo.textureUseTransparentIndex);
	SetEnablerViewState(paneID_TextureEnablerView, msg_TextureFormat, &mTestInfo.textureFormat);
	
			// currently there is only one format for the color lookup tables: 32-bit, 
			// so this popup is no longer used
//	SetTableFormatPopupValue(paneID_TextureTableFormat, mTestInfo.textureColorTableType);

	SetPICTPopupValue(paneID_Bitmap, &mTestInfo.bitmapFSSpec);
	SetPvalPopupValue(paneID_BitmapFormat, mTestInfo.bitmapFormat);
	SetControlValue(paneID_LockBitmap, mTestInfo.bitmapFlags & kQABitmap_Lock);
	SetControlValue(paneID_DetachBitmap, mTestInfo.detachBitmap);
	SetControlValue(paneID_BitmapMakeBlackTransparent, mTestInfo.bitmapMakeBlackTransparent);
	SetControlValue(paneID_BitmapUseTransparentIndex, mTestInfo.bitmapUseTransparentIndex);
	SetEnablerViewState(paneID_BitmapEnablerView, msg_BitmapFormat, &mTestInfo.bitmapFormat);

		// currently there is only one format for the color lookup tables: 32-bit, 
		// so this popup is no longer used
//	SetTableFormatPopupValue(paneID_BitmapTableFormat, mTestInfo.bitmapColorTableType);

	SetControlValue(paneID_ResetRandomSeed, mTestInfo.resetRandomSeed);
	SetStringPopupValue(paneID_Test, mTestInfo.testName);

		// store a pointer to the texture popup menu so tests can access the 
		// file specs stored there
	mTestInfo.pictList = (S3PICTList) mSettingsWindow->FindPaneByID(paneID_Texture);
	
		// set the argString broadcast edit field
	CBroadcastEditField* argStringPane = 
		(CBroadcastEditField*) mSettingsWindow->FindPaneByID(paneID_ArgString);
	
	if (argStringPane != nil)
		argStringPane->SetDescriptor((unsigned char*) mTestInfo.argString);

		// we're done changing the controls' values, start listening to them again
	StartListening();
}


// ===========================================================================
	void 
CTestInfo::UpdateLeftTestWindow()
{
	BroadcastMessage(msg_TestWindow_Move + TestWindow_Left, &mTestInfo.leftGlobalRect);
	
		// convert the C string to a Pascal string before sending it to the window
	Str255 engineNamePString;
	S3CtoPString(mTestInfo.leftEngineName, engineNamePString);
	BroadcastMessage(msg_TestWindow_SetTitle + TestWindow_Left, engineNamePString);
}


// ===========================================================================
	void 
CTestInfo::UpdateRightTestWindow()
{
	BroadcastMessage(msg_TestWindow_Move + TestWindow_Right, &mTestInfo.rightGlobalRect);

		// convert the C string to a Pascal string before sending it to the window
	Str255 engineNamePString;
	S3CtoPString(mTestInfo.rightEngineName, engineNamePString);
	BroadcastMessage(msg_TestWindow_SetTitle + TestWindow_Right, engineNamePString);
}


// ===========================================================================
	void 
CTestInfo::UpdateTestWindows()
{
	UpdateLeftTestWindow();
	UpdateRightTestWindow();
}


// ===========================================================================
	void 
CTestInfo::SetWindows(MessageT inMessage, void* ioParam)
{
		// CBroadcastEditField panes broadcast a pointer to themselves 
		// in the ioParam, so we cast it to something we can use and then
		// get the field's value
	CBroadcastEditField* theEditField = (CBroadcastEditField*) ioParam;
	long newValue = theEditField->GetValue();

	long delta = 0;

	switch (inMessage) {
		case msg_Width:
			mTestInfo.leftGlobalRect.right = mTestInfo.leftGlobalRect.left + newValue;
			mTestInfo.rightGlobalRect.right = mTestInfo.rightGlobalRect.left + newValue;
			
				// let the TinselTest windows know their rects have changed
			UpdateLeftTestWindow();
			UpdateRightTestWindow();
			break;
			
		case msg_Height:
				// we subtract one from the height to get the bottom edge 
				// because bottom - top + 1 = height
			mTestInfo.leftGlobalRect.bottom = mTestInfo.leftGlobalRect.top + newValue;
			mTestInfo.rightGlobalRect.bottom = mTestInfo.rightGlobalRect.top + newValue;
			
				// let the TinselTest windows know their rects have changed
			UpdateLeftTestWindow();
			UpdateRightTestWindow();
			break;
			
		case msg_LeftWindow_Left:
			delta = newValue - mTestInfo.leftGlobalRect.left;
			mTestInfo.leftGlobalRect.left += delta;
			mTestInfo.leftGlobalRect.right += delta;
			
				// let the left TinselTest window know its rect has changed
			UpdateLeftTestWindow();
			break;
			
		case msg_LeftWindow_Top:
			delta = newValue - mTestInfo.leftGlobalRect.top;
			mTestInfo.leftGlobalRect.top += delta;
			mTestInfo.leftGlobalRect.bottom += delta;
			
				// let the left TinselTest window know its rect has changed
			UpdateLeftTestWindow();
			break;
			
		case msg_RightWindow_Left:
			delta = newValue - mTestInfo.rightGlobalRect.left;
			mTestInfo.rightGlobalRect.left += delta;
			mTestInfo.rightGlobalRect.right += delta;
			
				// let the right TinselTest window know its rect has changed
			UpdateRightTestWindow();
			break;
			
		case msg_RightWindow_Top:
			delta = newValue - mTestInfo.rightGlobalRect.top;
			mTestInfo.rightGlobalRect.top += delta;
			mTestInfo.rightGlobalRect.bottom += delta;
			
				// let the right TinselTest window know its rect has changed
			UpdateRightTestWindow();
			break;
	}
}


// ===========================================================================
	void 
CTestInfo::SetWindowDrag(MessageT inMessage, void* ioParam)
{
	if (CSettingsDoc::GetFrontMost() != mSettingsDoc) {
			// we only pay attention to test window drag messages when we're part
			// of the frontmost settings doc, because every open settings doc
			// receives a message from the test windows, but only the top one
			// should respond to it.
		return;
	}
	
		// the test windows broadcast a pointer to their new bounds when they've 
		// been dragged, so cast ioParam to a Rect
	Rect newRect = *(Rect*) ioParam;
	
		// since we listen to these controls, if we don't stop listening before
		// changing each of their values, we'll get a ListenToMessage call every
		// time they change
	StopListening();

		// update the edit fields to relfect the window's new position
	switch (inMessage) {
		case msg_TestWindow_Drag + TestWindow_Left:
			mTestInfo.leftGlobalRect = newRect;

			SetEditFieldValue(paneID_LeftWindow_Left, (long) mTestInfo.leftGlobalRect.left);
			SetEditFieldValue(paneID_LeftWindow_Top, (long) mTestInfo.leftGlobalRect.top);
			break;
		
		case msg_TestWindow_Drag + TestWindow_Right:
			mTestInfo.rightGlobalRect = newRect;
		
			SetEditFieldValue(paneID_RightWindow_Left, (long) mTestInfo.rightGlobalRect.left);
			SetEditFieldValue(paneID_RightWindow_Top, (long) mTestInfo.rightGlobalRect.top);
			break;
	} // endswitch

		// we're done changing the controls' values, start listening to them again
	StartListening();
}


// ===========================================================================
	void 
CTestInfo::SetWindowSize(void* ioParam)
{
	if (CSettingsDoc::GetFrontMost() != mSettingsDoc) {
			// we only pay attention to test window resize messages when we're part
			// of the frontmost settings doc, because every open settings doc
			// receives a message from the test windows, but only the top one
			// should respond to it.
		return;
	}
	
		// the test window which was resized broadcasts a pointer to a Rect
		// containing the offsets from the old right and bottom positions to the
		// new ones, so cast ioParam to a Rect
	Rect offsetRect = *(Rect*) ioParam;

		// offset the left and right global rects by the amount in offsetRect
	mTestInfo.leftGlobalRect.right += offsetRect.right;
	mTestInfo.leftGlobalRect.bottom += offsetRect.bottom;
	mTestInfo.rightGlobalRect.right += offsetRect.right;
	mTestInfo.rightGlobalRect.bottom += offsetRect.bottom;
	
		// since we listen to these controls, if we don't stop listening before
		// changing each of their values, we'll get a ListenToMessage call every
		// time they change
	StopListening();

		// update the edit fields with the new width and height
	SetEditFieldValue(paneID_Width, 
										(long) (mTestInfo.leftGlobalRect.right - 
														mTestInfo.leftGlobalRect.left));
	SetEditFieldValue(paneID_Height, 
										(long) (mTestInfo.leftGlobalRect.bottom -
														mTestInfo.leftGlobalRect.top));

		// let the test windows know their rects have changed.  one of them 
		// already knows, because it broadcast the message, but it's easiest
		// just to update them both.
	UpdateTestWindows();

		// we're done changing the controls' values, start listening to them again
	StartListening();
}


// ===========================================================================
	void 
CTestInfo::SetEngine(MessageT inMessage, void* ioParam)
{
		// CEnginePopupMenu broadcasts the current item's string in the ioParam
	StringPtr engineName = (StringPtr) ioParam;
	
		// convert the Pascal string in the ioParam to a C string, in place
	S3PtoCString(engineName);
	
	TQAEngine* engine = S3GetEngineByName((char*) engineName);
	
	switch (inMessage) {
	  case msg_LeftEngine:
				// before we change the engine, we have to delete any bitmaps or
				// textures that are attached to this engine.  otherwise, it will 
				// eventually not be able to create any new ones.
			if (mTestInfo.leftTexture != nil) {
					// delete the attached texture
				QATextureDelete(mTestInfo.leftEngine, mTestInfo.leftTexture);
				mTestInfo.leftTexture = nil;
			}

			if (mTestInfo.leftBitmap != nil) {
					// delete the attached bitmap
				QABitmapDelete(mTestInfo.leftEngine, mTestInfo.leftBitmap);
				mTestInfo.leftBitmap = nil;
			}

				// record the newly selected engine pointer and its name
			mTestInfo.leftEngine = engine;
			S3GetEngineName(mTestInfo.leftEngine, mTestInfo.leftEngineName, 
											sizeof(mTestInfo.leftEngineName));
											
				// let the left window know its title has changed
			UpdateLeftTestWindow();
			break;
			
	  case msg_RightEngine:
				// before we change the engine, we have to delete any bitmaps or
				// textures that are attached to this engine.  otherwise, it will 
				// eventually not be able to create any new ones.
			if (mTestInfo.rightTexture != nil) {
					// delete the attached texture
				QATextureDelete(mTestInfo.rightEngine, mTestInfo.rightTexture);
				mTestInfo.rightTexture = nil;
			}

			if (mTestInfo.rightBitmap != nil) {
					// delete the attached bitmap
				QABitmapDelete(mTestInfo.rightEngine, mTestInfo.rightBitmap);
				mTestInfo.rightBitmap = nil;
			}

				// record the newly selected engine pointer and its name
			mTestInfo.rightEngine = engine;
			S3GetEngineName(mTestInfo.rightEngine, mTestInfo.rightEngineName, 
											sizeof(mTestInfo.rightEngineName));
											
				// let the right window know its title has changed
			UpdateRightTestWindow();
			break;
	} // endswitch
}


// ===========================================================================
	void 
CTestInfo::SetBuffering(void* inParam)
{
	switch ((EBufferingMenu) *(unsigned int*) inParam) {
		case eBufferingMenu_Single:
				// turn the flag off
			mTestInfo.contextFlags &= ~kQAContext_DoubleBuffer;
			break;

		case eBufferingMenu_Double:
				// turn the flag on
			mTestInfo.contextFlags |= kQAContext_DoubleBuffer;
			break;
	}
}


// ===========================================================================
	EBufferingMenu 
CTestInfo::GetBuffering()
{
	if (mTestInfo.contextFlags & kQAContext_DoubleBuffer) {
		return eBufferingMenu_Double;
	} else {
		return eBufferingMenu_Single;
	}
}


// ===========================================================================
	void 
CTestInfo::SetZBuffering(void* inParam)
{
	switch ((EZBufferingMenu) *(unsigned int*) inParam) {
		case eZBufferingMenu_None:
				// turn NoZBuffer on, DeepZ off
			mTestInfo.contextFlags |= kQAContext_NoZBuffer;
			mTestInfo.contextFlags &= ~kQAContext_DeepZ;
			break;

		case eZBufferingMenu_Shallow:
				// turn NoZBuffer off, DeepZ off
			mTestInfo.contextFlags &= ~kQAContext_NoZBuffer;
			mTestInfo.contextFlags &= ~kQAContext_DeepZ;
			break;
			
		case eZBufferingMenu_Deep:
				// turn NoZBuffer off, DeepZ on
			mTestInfo.contextFlags &= ~kQAContext_NoZBuffer;
			mTestInfo.contextFlags |= kQAContext_DeepZ;
			break;
	}
}


// ===========================================================================
	EZBufferingMenu 
CTestInfo::GetZBuffering()
{
		// we add the two flag values so that we can switch on their combined value
	switch (mTestInfo.contextFlags & (kQAContext_NoZBuffer + kQAContext_DeepZ)) {
		case kQAContext_NoZBuffer:
			return eZBufferingMenu_None;
			break;

		case 0:
				// both NoZBuffer off and DeepZ off
			return eZBufferingMenu_Shallow;
			break;
			
		case kQAContext_DeepZ:
			return eZBufferingMenu_Deep;
			break;
			
		default:
				// (kQAContext_NoZBuffer + kQAContext_DeepZ)
			SignalPStr_("\pCan't have both NoZBuffer and DeepZ flags on.");
			return eZBufferingMenu_None;
			break;
	}
}


// ===========================================================================
	void 
CTestInfo::SetBackground(MessageT inMessage, void* inParam)
{
		// CBroadcastEditField panes broadcast a pointer to themselves 
		// in the ioParam, so we cast it to something we can use and then
		// get the field's value as a float
	CBroadcastEditField* theEditField = (CBroadcastEditField*) inParam;
	float newValue = theEditField->GetValueAsFloat();

	switch (inMessage) {
		case msg_Background_Alpha:
			mTestInfo.alpha = newValue;
			break;
			
		case msg_Background_Red:
			mTestInfo.red = newValue;
			break;
			
		case msg_Background_Green:
			mTestInfo.green = newValue;
			break;
			
		case msg_Background_Blue:
			mTestInfo.blue = newValue;
			break;
	}
}


// ===========================================================================
	void 
CTestInfo::SetTextureOp(MessageT inMessage, void* inParam)
{
	switch (inMessage) {
		case msg_TextureOp_Modulate:
			if (*(int*) inParam == Button_On) 
				mTestInfo.textureOp |= kQATextureOp_Modulate;
			else
				mTestInfo.textureOp &= ~kQATextureOp_Modulate;
			break;
			
		case msg_TextureOp_Highlight:
			if (*(int*) inParam == Button_On) 
				mTestInfo.textureOp |= kQATextureOp_Highlight;
			else
				mTestInfo.textureOp &= ~kQATextureOp_Highlight;
			break;
			
		case msg_TextureOp_Decal:
			if (*(int*) inParam == Button_On) 
				mTestInfo.textureOp |= kQATextureOp_Decal;
			else
				mTestInfo.textureOp &= ~kQATextureOp_Decal;
			break;
			
		case msg_TextureOp_Shrink:
			if (*(int*) inParam == Button_On) 
				mTestInfo.textureOp |= kQATextureOp_Shrink;
			else
				mTestInfo.textureOp &= ~kQATextureOp_Shrink;
			break;
	}
}


// ===========================================================================
	RgnHandle 
CTestInfo::MakeClipRegion(const TQARect* inRect, S3eClipRegion inClipRegion)
{
	if (inClipRegion == kS3ClipRegion_None) {
			// an empty clipping region is currently selected in the menu
		return nil;
	}
	
		// save the current port on the stack.  it will get reset on function exit	
//	GrafPtr savedPort;
//	GetPort(&savedPort);
//	StColorPortState(savedPort);
	
	Rect tempRect;
	
		// create a new region and open it
	RgnHandle regionH = ::NewRgn();
	::OpenRgn();
	
	switch (inClipRegion) {
		case kS3ClipRegion_RoundedRectangle:
			{
				long insetH = (inRect->right - inRect->left + 1) / 8;
				long insetV = (inRect->bottom - inRect->top + 1) / 8;
				
				::SetRect(
					&tempRect, 
					inRect->left + insetH,
					inRect->top + insetV,
					inRect->right - insetH,
					inRect->bottom - insetV
				);
				
				::FrameRoundRect (&tempRect, insetH, insetV);
			}
			break;
			
		case kS3ClipRegion_Circle:
			{
				long insetH = (inRect->right - inRect->left + 1) / 8;
				long insetV = (inRect->bottom - inRect->top + 1) / 8;
				
				::SetRect(
					&tempRect, 
					inRect->left + insetH,
					inRect->top + insetV,
					inRect->right - insetH,
					inRect->bottom - insetV
				);
				
				::FrameOval (&tempRect);
			}
			break;
			
		case kS3ClipRegion_HorizontalLines:
			{
				long height = inRect->bottom - inRect->top;
				height = height / 20;
				
				for(int i = 0; i < 10; i++) {
					::SetRect(
						&tempRect,
						inRect->left,
						inRect->top + height * (i*2),
						inRect->right,
						inRect->top + height * (i*2 + 1));
					::FrameRect(&tempRect);
				}
			}
			break;
			
		case kS3ClipRegion_VerticalLines:
			{
				long width = inRect->right - inRect->left;
				width = width / 20;
				
				for(int i = 0; i < 10; i++) {
					::SetRect(
						&tempRect,
						inRect->left + width * (i*2),
						inRect->top,
						inRect->left + width * (i*2 + 1),
						inRect->bottom);
					::FrameRect(&tempRect);
				}
			}
			break;

		case kS3ClipRegion_Checkerboard:
			{
				long height = inRect->bottom - inRect->top;
				height = height / 20;
				
				for(int i = 0; i < 10; i++) {
					::SetRect(
						&tempRect,
						inRect->left,
						inRect->top + height * (i*2),
						inRect->right,
						inRect->top + height * (i*2 + 1));
					::FrameRect(&tempRect);
				}
			}
			{
				long width = inRect->right - inRect->left;
				width = width / 20;
				
				for(int i = 0; i < 10; i++) {
					::SetRect(
						&tempRect,
						inRect->left + width * (i*2),
						inRect->top,
						inRect->left + width * (i*2 + 1),
						inRect->bottom);
					::FrameRect(&tempRect);
				}
			}
			break;
	} // endswitch
	
	::CloseRgn(regionH);
	
	return regionH;
}


// ===========================================================================
// This function assumes the left and right textures, if any, have already
// been properly disposed of (e.g., by DisposeTestInfo()).
//
	Boolean
CTestInfo::LoadTextureMap(S3tTestInfo* inTestInfo)
{
	inTestInfo->leftTexture = nil;
	inTestInfo->rightTexture = nil;

		// load the PICT file into a handle
	PicHandle pictH = S3GetPICTFromFile(&inTestInfo->textureFSSpec);
	
	if (pictH == nil) {
			// alert the user that we couldn't load that PICT file
		S3StopAlert(STRx_AlertStrings_ID, AlertString_PICTFileError, nil);
		
		return false;
	}
	
		// create a TQAImage from this PICT, possibly also making all the black
		// pixels transparent and creating a mipmap
	inTestInfo->textureImage = new CTQAImage(pictH, inTestInfo->textureFormat, 
																				 inTestInfo->textureMakeBlackTransparent,
																				 (inTestInfo->textureFlags & kQATexture_Mipmap));

		// once we've created the CTQAImage, we can dispose of the PICT's handle 
	::DisposeHandle((Handle) pictH);
	
	if (inTestInfo->textureImage == nil) {
			// alert the user that we couldn't create an image from that PICT
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoTextureImage, nil);
		
		return false;
	}
	
	if (!inTestInfo->textureImage->IsValidTexture()) {
			// alert the user that the PICT must have dimensions that are a power of 2
		S3StopAlert(STRx_AlertStrings_ID, AlertString_MustBePowerOf2, nil);
		
		return false;
	}
	
	if ((inTestInfo->textureFlags & kQATexture_Mipmap) && 
			 !inTestInfo->textureImage->IsValidMipMap()) {
			// alert the user that a mipmap for the PICT could not be created 
			// when it should have been
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoTextureMipMap, nil);
		
		return false;
	}
	
		// save the size and rowbytes info for this texture
	inTestInfo->textureWidth = inTestInfo->textureImage->GetWidth();
	inTestInfo->textureHeight = inTestInfo->textureImage->GetHeight();
	inTestInfo->textureRowBytes = inTestInfo->textureImage->GetRowBytes();

		// create the texture for the left engine
	TQAError qaError = QATextureNew(
		inTestInfo->leftEngine,
		inTestInfo->textureFlags,
		inTestInfo->textureFormat,
		inTestInfo->textureImage->GetImages(),
		&inTestInfo->leftTexture);

	if (qaError != kQANoErr) {
			// alert the user that the left engine wouldn't create the texture
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftTexture, nil);
		
		return false;
	}

		// create the texture for the right engine
	qaError = QATextureNew(
		inTestInfo->rightEngine,
		inTestInfo->textureFlags,
		inTestInfo->textureFormat,
		inTestInfo->textureImage->GetImages(),
		&inTestInfo->rightTexture);

	if (qaError != kQANoErr) {
			// alert the user that the right engine wouldn't create the texture
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightTexture, nil);
		
		return false;
	}
	
	if (inTestInfo->textureFormat == kQAPixel_CL4 || 
			inTestInfo->textureFormat == kQAPixel_CL8) {
			// this texture is in color table lookup format, so create the color
			// table for it.  first get the ColorTable out of the GWorld holding\
			// the texture.
// later take color table from resource in a file			
		CTabHandle colorTableH = inTestInfo->textureImage->GetColorTableHandle();
		
		if (colorTableH == nil) {
				// alert the user that we couldn't get the color table fromt the pict
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoColorTableFromPICT, nil);
			
			return false;
		}
		
			// convert the CTabHandle into TQAColorTable format
		CTQAColorTable* colorTableData = 
				new CTQAColorTable(colorTableH, inTestInfo->textureColorTableType);

			// create a new color table in the left engine
		qaError = QAColorTableNew(
			inTestInfo->leftEngine,
			inTestInfo->textureColorTableType,
			colorTableData->GetData(),
			inTestInfo->textureUseTransparentIndex,
			&inTestInfo->leftTextureColorTable);
				
		if (qaError != kQANoErr) {
				// alert the user that the left engine couldn't create a new color table
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftColorTable, nil);
			
			return false;
		}

			// bind this color table to the left texture
		qaError = QATextureBindColorTable(
			inTestInfo->leftEngine,
			inTestInfo->leftTexture,
			inTestInfo->leftTextureColorTable);
		
		if (qaError != kQANoErr) {
				// alert the user that the left engine couldn't bind the color table
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftBindTableTexture, nil);
			
			return false;
		}
		
			// create a new color table in the right engine
		qaError = QAColorTableNew(
			inTestInfo->rightEngine,
			inTestInfo->textureColorTableType,
			colorTableData->GetData(),
			inTestInfo->textureUseTransparentIndex,
			&inTestInfo->rightTextureColorTable);
				
		if (qaError != kQANoErr) {
				// alert the user that the right engine couldn't create a new color table
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightColorTable, nil);
			
			return false;
		}

			// bind this color table to the right texture
		qaError = QATextureBindColorTable(
			inTestInfo->rightEngine,
			inTestInfo->rightTexture,
			inTestInfo->rightTextureColorTable);
		
		if (qaError != kQANoErr) {
				// alert the user that the right engine couldn't bind the color table
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightBindTableTexture, nil);
			
			return false;
		}
	}
			
	if (inTestInfo->detachTexture) {
			// IMPORTANT: we can't detach the textures until after we've created
			// the color tables, if any.  this is because after detaching the
			// textures, we delete the image we used to create them, and we need
			// the image to create the color table.
		qaError = QATextureDetach(
			inTestInfo->leftEngine,
			inTestInfo->leftTexture);
			
		if (qaError != kQANoErr) {
				// alert the user that the left engine wouldn't detach the texture
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftTextureDetach, nil);
			
			return false;
		}

		qaError = QATextureDetach(
			inTestInfo->rightEngine,
			inTestInfo->rightTexture);
			
		if (qaError != kQANoErr) {
				// alert the user that the right engine wouldn't detach the texture
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightTextureDetach, nil);
			
			return false;
		}
	
			// since we've detached the textures, the engines are now responsible for
			// their storage, so we can delete the CTQAImage which is holding them
		delete inTestInfo->textureImage;
	}
	
	return true;
}


// ===========================================================================
// This function assumes the left and right bitmaps, if any, have already
// been properly disposed of (e.g., by DisposeTestInfo()).
//
	Boolean
CTestInfo::LoadBitmap(S3tTestInfo* inTestInfo)
{
	inTestInfo->leftBitmap = nil;
	inTestInfo->rightBitmap = nil;

		// load the PICT file into a handle
	PicHandle pictH = S3GetPICTFromFile(&inTestInfo->bitmapFSSpec);

	if (pictH == nil) {
			// alert the user that we couldn't load that PICT file
		S3StopAlert(STRx_AlertStrings_ID, AlertString_PICTFileError, nil);
		
		return false;
	}
	
		// create a TQAImage from this PICT, possibly also making all the
		// black pixels transparent
	inTestInfo->bitmapImage = new CTQAImage(pictH, inTestInfo->bitmapFormat, 
																				inTestInfo->bitmapMakeBlackTransparent);

		// once we've created the CTQAImage, we can dispose of the PICT's handle 
	::DisposeHandle((Handle) pictH);
	
	if (inTestInfo->bitmapImage == nil) {
			// alert the user that we couldn't create an image from that PICT
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoBitmapImage, nil);
		
		return false;
	}
	
		// save the size and rowbytes info for this bitmap
	inTestInfo->bitmapWidth = inTestInfo->bitmapImage->GetWidth();
	inTestInfo->bitmapHeight = inTestInfo->bitmapImage->GetHeight();
	inTestInfo->bitmapRowBytes = inTestInfo->bitmapImage->GetRowBytes();
	
		// now create new bitmaps for the left and right engines
	TQAError qaError = QABitmapNew(
		inTestInfo->leftEngine,
		inTestInfo->bitmapFlags,
		inTestInfo->bitmapFormat,
		inTestInfo->bitmapImage->GetImages(),
		&inTestInfo->leftBitmap);
		
	if (qaError != kQANoErr) {
			// alert the user that the left engine wouldn't create the bitmap
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftBitmap, nil);
		
		return false;
	}

	qaError = QABitmapNew(
		inTestInfo->rightEngine,
		inTestInfo->bitmapFlags,
		inTestInfo->bitmapFormat,
		inTestInfo->bitmapImage->GetImages(),
		&inTestInfo->rightBitmap);
	
	if (qaError != kQANoErr) {
			// alert the user that the right engine wouldn't create the bitmap
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightBitmap, nil);
		
		return false;
	}
	
	if (inTestInfo->detachBitmap) {
			// detach the bitmaps
		qaError = QABitmapDetach(
			inTestInfo->leftEngine,
			inTestInfo->leftBitmap);
			
		if (qaError != kQANoErr) {
				// alert the user that the left engine wouldn't detach the bitmap
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftBitmapDetach, nil);
			
			return false;
		}

		qaError = QABitmapDetach(
			inTestInfo->rightEngine,
			inTestInfo->rightBitmap);
			
		if (qaError != kQANoErr) {
				// alert the user that the right engine wouldn't detach the bitmap
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightBitmapDetach, nil);
			
			return false;
		}
			
			// since we've detached the bitmaps, the engines are now responsible 
			// their storage, so we can delete the CTQAImage which is holding them
		delete inTestInfo->bitmapImage;
	}
			
	if (inTestInfo->bitmapFormat == kQAPixel_CL4 || 
			inTestInfo->bitmapFormat == kQAPixel_CL8) {
			// this bitmap is in color table lookup format, so create the color
			// table for it.  first get the ColorTable out of the GWorld holding\
			// the bitmap.
// later take color table from resource in a file			
		CTabHandle colorTableH = inTestInfo->bitmapImage->GetColorTableHandle();
		
		if (colorTableH == nil) {
				// alert the user that we couldn't get the color table fromt the pict
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoColorTableFromPICT, nil);
			
			return false;
		}
		
			// convert the CTabHandle into TQAColorTable format
		CTQAColorTable* colorTableData = 
				new CTQAColorTable(colorTableH, inTestInfo->bitmapColorTableType);

			// create a new color table in the left engine
		qaError = QAColorTableNew(
			inTestInfo->leftEngine,
			inTestInfo->bitmapColorTableType,
			colorTableData->GetData(),
			inTestInfo->bitmapUseTransparentIndex,
			&inTestInfo->leftBitmapColorTable);
				
		if (qaError != kQANoErr) {
				// alert the user that the left engine couldn't create a new color table
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftColorTable, nil);
			
			return false;
		}

			// bind this color table to the left bitmap
		qaError = QABitmapBindColorTable(
			inTestInfo->leftEngine,
			inTestInfo->leftBitmap,
			inTestInfo->leftBitmapColorTable);
		
		if (qaError != kQANoErr) {
				// alert the user that the left engine couldn't bind the color table
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoLeftBindTableBitmap, nil);
			
			return false;
		}
		
			// create a new color table in the right engine
		qaError = QAColorTableNew(
			inTestInfo->rightEngine,
			inTestInfo->bitmapColorTableType,
			colorTableData->GetData(),
			inTestInfo->bitmapUseTransparentIndex,
			&inTestInfo->rightBitmapColorTable);
				
		if (qaError != kQANoErr) {
				// alert the user that the right engine couldn't create a new color table
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightColorTable, nil);
			
			return false;
		}

			// bind this color table to the right bitmap
		qaError = QABitmapBindColorTable(
			inTestInfo->rightEngine,
			inTestInfo->rightBitmap,
			inTestInfo->rightBitmapColorTable);
		
		if (qaError != kQANoErr) {
				// alert the user that the right engine couldn't bind the color table
			S3StopAlert(STRx_AlertStrings_ID, AlertString_NoRightBindTableBitmap, nil);
			
			return false;
		}
	}
	
	return true;
}


// ===========================================================================
	Boolean
CTestInfo::SetupGDevice(S3tTestInfo* inTestInfo)
{
		// since the selected device is a screen, make sure that it is the same
		// one on which the test window's upper left corner is located.  the user
		// may have moved the window onto a screen which is not the deepest one,
		// which is where the windows are put by default.
	inTestInfo->leftDevice.deviceType = kQADeviceGDevice;
	Boolean isOnOneDevice = false;
	inTestInfo->leftDevice.device.gDevice = 
		S3GetDeviceFromRect(&inTestInfo->leftGlobalRect, &isOnOneDevice);
	
	if (inTestInfo->leftDevice.device.gDevice == nil || !isOnOneDevice) {
			// the window is hanging off the edge of the screen, so alert the
			// user that we can't run the test
		S3StopAlert(STRx_AlertStrings_ID, AlertString_LeftWindowOffScreen, nil);
		return false;
	}
	
	if (!S3EngineSupportsDevice(inTestInfo->leftEngine, &inTestInfo->leftDevice)) {
			// the chosen engine doesn't support the chosen device, so alert the
			// user that we can't run the test
		S3StopAlert(STRx_AlertStrings_ID, AlertString_LeftNoDeviceSupport, nil);
		return false;
	}

	inTestInfo->rightDevice.deviceType = kQADeviceGDevice;
	inTestInfo->rightDevice.device.gDevice = 
		S3GetDeviceFromRect(&inTestInfo->rightGlobalRect, &isOnOneDevice);

	if (inTestInfo->rightDevice.device.gDevice == nil || !isOnOneDevice) {
			// the window is hanging off the edge of the screen, so alert the
			// user that we can't run the test
		S3StopAlert(STRx_AlertStrings_ID, AlertString_RightWindowOffScreen, nil);
		return false;
	}
	
	if (!S3EngineSupportsDevice(inTestInfo->rightEngine, &inTestInfo->rightDevice)) {
			// the chosen engine doesn't support the chosen device, so alert the
			// user that we can't run the test
		S3StopAlert(STRx_AlertStrings_ID, AlertString_RightNoDeviceSupport, nil);
		return false;
	}
	
		// convert the leftGlobalRect into a TQARect local to the device coords
	Rect tempLeftRect = inTestInfo->leftGlobalRect;
	S3GlobalToDeviceLocal(inTestInfo->leftDevice.device.gDevice,
												&tempLeftRect);
	inTestInfo->leftRect.left = tempLeftRect.left;
	inTestInfo->leftRect.top = tempLeftRect.top;
	inTestInfo->leftRect.right = tempLeftRect.right;
	inTestInfo->leftRect.bottom = tempLeftRect.bottom;

		// convert the rightGlobalRect into a TQARect local to the device coords
	Rect tempRightRect = inTestInfo->rightGlobalRect;
	S3GlobalToDeviceLocal(inTestInfo->rightDevice.device.gDevice,
												&tempRightRect);
	inTestInfo->rightRect.left = tempRightRect.left;
	inTestInfo->rightRect.top = tempRightRect.top;
	inTestInfo->rightRect.right = tempRightRect.right;
	inTestInfo->rightRect.bottom = tempRightRect.bottom;

	return true;
}


// ===========================================================================
	Boolean
CTestInfo::SetupMemoryDevice(S3tTestInfo* inTestInfo)
{
	inTestInfo->leftDevice.deviceType = kQADeviceMemory;
	inTestInfo->rightDevice.deviceType = kQADeviceMemory;
	
		// since a memory device is being used, the device rect's
		// origin is at 0,0
	inTestInfo->leftRect.left = 0;
	inTestInfo->leftRect.top = 0;
	inTestInfo->leftRect.right = inTestInfo->leftGlobalRect.right - inTestInfo->leftGlobalRect.left;
	inTestInfo->leftRect.bottom = inTestInfo->leftGlobalRect.bottom - inTestInfo->leftGlobalRect.top;

	inTestInfo->rightRect = inTestInfo->leftRect;
	
	Uint16 bytesPerPixel;
	Uint16 cmpSize;
	Uint16 depth;
	
		// setup the pixel type of the device, based on what kind the user chose
	if (inTestInfo->deviceMenu == kS3Device_Memory16) {
		inTestInfo->leftDevice.device.memoryDevice.pixelType = kQAPixel_ARGB16;
		bytesPerPixel = 2;
		cmpSize = 5;
		depth = 16;
	} else {
		inTestInfo->leftDevice.device.memoryDevice.pixelType = kQAPixel_ARGB32;
		bytesPerPixel = 4;
		cmpSize = 8;
		depth = 32;
	}

	inTestInfo->leftDevice.device.memoryDevice.width = inTestInfo->leftRect.right;
	inTestInfo->leftDevice.device.memoryDevice.height = inTestInfo->leftRect.bottom;

		// copy all these settings to the right device
	inTestInfo->rightDevice.device.memoryDevice = inTestInfo->leftDevice.device.memoryDevice;

	Rect bounds = { 
		inTestInfo->leftRect.top, 
		inTestInfo->leftRect.left, 
		inTestInfo->leftRect.bottom, 
		inTestInfo->leftRect.right
	};
	
	inTestInfo->leftDeviceGWorld = NewDeviceGWorld(bounds, depth);
	if (inTestInfo->leftDeviceGWorld == nil) {
			// there wasn't enough memory to create a new pixmap, so we can't
			// create the memory device
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoMemoryForDevice, nil);
		return false;
	}
	
	inTestInfo->rightDeviceGWorld = NewDeviceGWorld(bounds, depth);
	if (inTestInfo->rightDeviceGWorld == nil) {
			// there wasn't enough memory to create a new pixmap, so we can't
			// create the memory device
		S3StopAlert(STRx_AlertStrings_ID, AlertString_NoMemoryForDevice, nil);
		return false;
	}
	
	PixMapHandle pixMapH = ::GetGWorldPixMap(inTestInfo->leftDeviceGWorld->GetMacGWorld());
	
		// to get the pixmap's actual rowbytes, we have to mask off the top two bits
	inTestInfo->leftDevice.device.memoryDevice.rowBytes = 
		(**pixMapH).rowBytes & 0x3fff;

		// lock the pixel data of the GWorld so that it won't move when the
		// engine draws into it
	::LockPixels(pixMapH);
	
		// point the memory device to the pixels of the GWorld
	inTestInfo->leftDevice.device.memoryDevice.baseAddr = ::GetPixBaseAddr(pixMapH);
	
	pixMapH = ::GetGWorldPixMap(inTestInfo->rightDeviceGWorld->GetMacGWorld());
	
		// to get the pixmap's actual rowbytes, we have to mask off the top two bits
	inTestInfo->rightDevice.device.memoryDevice.rowBytes = 
		(**pixMapH).rowBytes & 0x3fff;

		// lock the pixel data of the GWorld so that it won't move when the
		// engine draws into it
	::LockPixels(pixMapH);
	
		// point the memory device to the pixels of the GWorld
	inTestInfo->rightDevice.device.memoryDevice.baseAddr = ::GetPixBaseAddr(pixMapH);
	
		// if we've gotten here, then we have a valid memory device 
	return true;
}


// ===========================================================================
	LGWorld*
CTestInfo::NewDeviceGWorld(Rect& inRect, Uint16 inDepth)
{
	LGWorld* theGWorld = nil;
	
	try {
			// try to create a new offscreen GWorld for the memory device
		theGWorld = new LGWorld(inRect, inDepth);
	}
	
	catch(...) {
		try {
				// there isn't enough memory to allocate the GWorld in our heap, so
				// try putting it in temporary memory
			theGWorld = new LGWorld(inRect, inDepth, useTempMem);
		}
		
		catch(...) {
				// we still don't have enough memory for the GWorld
		} 
	}

	return theGWorld;
}


// ===========================================================================
	EEnablerResult
CTestInfo::TextureEnablerFunction(MessageT inMessage, void* ioParam)
{
	EEnablerResult result = eEnablerResult_IgnoreMessage;
	
	if (inMessage == msg_TextureFormat) {
		TQAImagePixelType textureFormat = *(TQAImagePixelType*) ioParam;
		
		if (textureFormat == kQAPixel_CL4 || textureFormat == kQAPixel_CL8) {
			result = eEnablerResult_Enable;
		} else {
			result = eEnablerResult_Disable;
		}
	}
	
	return result;
}


// ===========================================================================
	EEnablerResult
CTestInfo::BitmapEnablerFunction(MessageT inMessage, void* ioParam)
{
	EEnablerResult result = eEnablerResult_IgnoreMessage;
	
	if (inMessage == msg_BitmapFormat) {
		TQAImagePixelType bitmapFormat = *(TQAImagePixelType*) ioParam;
		
		if (bitmapFormat == kQAPixel_CL4 || bitmapFormat == kQAPixel_CL8) {
			result = eEnablerResult_Enable;
		} else {
			result = eEnablerResult_Disable;
		}
	}
	
	return result;
}
