// ===========================================================================
//	CTestPopupMenu.cp
// ===========================================================================
//
// This derivative of a standard popup menu will broadcast the item text
// of the selected item, rather than the item's number.  The current item
// can also be set by giving the menu a string, which it converts to an 
// item number.


#ifdef PowerPlant_PCH
#include PowerPlant_PCH
#endif


#include <PP_Messages.h>

#include "CTestPopupMenu.h"
#include "S3Utilities.h"
#include "S3TestSystem.h"
#include "StringConversion.h"


MenuHandle CTestPopupMenu::sMenuH = nil;
ResIDT CTestPopupMenu::sMenuID = 0;
LList CTestPopupMenu::sPopupMenuList;
LList CTestPopupMenu::sDataList(sizeof(S3tTestTool));
Boolean CTestPopupMenu::sDoUpdate = true;


// ===========================================================================
	void
CTestPopupMenu::InitializeMenu()
{
	Assert_(sMenuH);
	
		// we call this here so that the very first popup menu to be crated
		// and which calls InitializeMenu, will have its min and max values
		// set correctly. since the tests are added to the menu before the
		// popup control is created, CTestPopupMenu doesn't know how long
		// the menu is until UpdateValues gets called.
//	UpdatePopupMenus();

	BeginUpdate();
	
	S3Test_Init();
	
	EndUpdate();
}


// ===========================================================================
	void
CTestPopupMenu::ReinitializeMenu()
{
	// first get rid of old data values
	// then call InitializeMenu()
//	InitializeMenu();
}


// ===========================================================================
	void
CTestPopupMenu::UpdatePopupMenus()
{
	Assert_(sMenuH);
	
	if (!sDoUpdate) {
			// we're in the middle of a BeginUpdate/EndUpdate pair, so we don't
			// do the update until EndUpdate is called
		return;
	}
	
	LListIterator iterator(sPopupMenuList, iterate_FromStart);
	CTestPopupMenu* thePopup = nil;

		// update the control values of each open popup
	while (iterator.Next(&thePopup)) {
		Assert_(thePopup);
		
		thePopup->UpdateValues();
	}
}


// ===========================================================================
	void
CTestPopupMenu::BeginUpdate()
{
	Assert_(sMenuH);
	
		// we set this flag to false so that UpdatePopupMenus() doesn't update
		// every popup after each call to AppendSharedItem, InsertSharedItem, etc.
		// when EndUpdate is called, this flag will be reset.
	sDoUpdate = false;
}


// ===========================================================================
	void
CTestPopupMenu::EndUpdate()
{
	Assert_(sMenuH);
	
		// turn updating back on
	sDoUpdate = true;
	
		// make sure all the popups get updated correctly, now that the caller
		// is done changing the shared menu.
	UpdatePopupMenus();
}


// ===========================================================================
// only one item can be appended at a time
//
	void
CTestPopupMenu::AppendSharedItem(const Str255 inItemString, void* inData, Boolean inIgnoreMetaCharacters)
{
	Assert_(sMenuH);
	
	if (inIgnoreMetaCharacters) {
			// first add the item with a bogus string, then set the item's text
			// to the actual string.  we do this so that we can use metacharacters
			// in the item's name.
		::AppendMenu(sMenuH, "\pTemporary Item String");
		::SetMenuItemText(sMenuH, ::CountMItems(sMenuH), inItemString);
	} else {
			// the caller wants the metacharacters in inItemString to work, so
			// append the item the normal way
		::AppendMenu(sMenuH, inItemString);
	}
	
		// append the data for this item 
	sDataList.InsertItemsAt(1, arrayIndex_Last, inData);
	
	UpdatePopupMenus();
}


// ===========================================================================
// only one item can be inserted at a time
//
	void
CTestPopupMenu::InsertSharedItem(const Str255 inItemString, Int16 inAfterItem, void* inData, Boolean inIgnoreMetaCharacters)
{
	Assert_(sMenuH);
	
	if (inIgnoreMetaCharacters) {
			// first add the item with a bogus string, then set the item's text
			// to the actual string.  we do this so that we can use metacharacters
			// in the item's name.
		::InsertMenuItem(sMenuH, "\pTemporary Item String", inAfterItem);
		::SetMenuItemText(sMenuH, inAfterItem + 1, inItemString);
	} else {
			// the caller wants the metacharacters in inItemString to work, so
			// insert the item the normal way
		::InsertMenuItem(sMenuH, inItemString, inAfterItem);
	}
	
		// insert the data for this item.  inAfterItem is the index of the item
		// after which the new item is inserted, so add 1 to it get the LList index.
	sDataList.InsertItemsAt(1, inAfterItem + 1, inData);
	
	UpdatePopupMenus();
}


// ===========================================================================
	void
CTestPopupMenu::DeleteSharedItem(Int16 inItemNumber)
{
	Assert_(sMenuH);
	
	::DeleteMenuItem(sMenuH, inItemNumber);
	
		// remove the item's corresponding data 
	sDataList.RemoveItemsAt(1, inItemNumber);
	
	UpdatePopupMenus();
}


// ===========================================================================
	Int16
CTestPopupMenu::GetSharedItemFromString(const Str255 inString)
{
	Assert_(sMenuH);
	
		// Search for inString in the items of this Menu
	Str255 itemString;
	short numItems = ::CountMItems(sMenuH);
	
	for (short theIndex = 1; theIndex <= numItems; theIndex++) {
		::GetMenuItemText(sMenuH, theIndex, itemString);
		
		if (::EqualString(inString, itemString, true, false)) {
			return theIndex;
		}
	}
	
		// the string wasn't found in this menu
	return 0;
}


// ===========================================================================
	Int16
CTestPopupMenu::GetSharedItemFromString(const char* inString)
{
	Assert_(sMenuH);
	
	Str255 PString;
	S3CtoPString(inString, PString);
	
			// Search for inString in the items of this Menu
	Str255 itemString;
	short numItems = ::CountMItems(sMenuH);
	
	for (short theIndex = 1; theIndex <= numItems; theIndex++) {
		::GetMenuItemText(sMenuH, theIndex, itemString);
		
		if (::EqualString(PString, itemString, true, false)) {
			return theIndex;
		}
	}
	
		// the string wasn't found in this menu
	return 0;
}


// ===========================================================================
	Boolean
CTestPopupMenu::GetSharedDataFromString(const Str255 inString, void* outData)
{
	Assert_(sMenuH);
	
	Int16 index = GetSharedItemFromString(inString);
	
	return GetSharedDataFromItem(index, outData);
}


// ===========================================================================
	void
CTestPopupMenu::GetSharedStringFromItem(Int16 inIndex, Str255 outString)
{	
	Assert_(sMenuH);
	
	if (inIndex > ::CountMItems(sMenuH) || inIndex < 1) {
			// the index is out of range, so make the string empty
		outString[0] = 0;
		
		return;
	}
	
	::GetMenuItemText(sMenuH, inIndex, outString);
}


// ===========================================================================
	void
CTestPopupMenu::GetSharedStringFromItem(Int16 inIndex, char* outString) 
{	
	Assert_(sMenuH);
	
	if (inIndex > ::CountMItems(sMenuH) || inIndex < 1) {
			// the index is out of range, so make the string empty
		outString[0] = 0;
		
		return;
	}
	
	::GetMenuItemText(sMenuH, inIndex, (unsigned char*) outString);

		// convert the Pascal string to a C string in place
	S3PtoCString((unsigned char*) outString);
}


// ===========================================================================
	Boolean
CTestPopupMenu::GetSharedDataFromString(const char* inString, void* outData)
{
	Assert_(sMenuH);
	
	Int16 index = GetSharedItemFromString(inString);
	
	return GetSharedDataFromItem(index, outData);
}


// ===========================================================================
	Boolean
CTestPopupMenu::GetSharedDataFromItem(Int16 inIndex, void* outData)
{
	Assert_(sMenuH);
	
		// if the index is out of bounds, outData is unchanged and FetchItemAt
		// returns false
	return sDataList.FetchItemAt(inIndex, outData);
}


// ===========================================================================
	Int16
CTestPopupMenu::GetSharedItemFromData(const void* inData)
{
	Assert_(sMenuH);
	
		// if the index is out of bounds, data will remain nil
	Int32 index = sDataList.FetchIndexOf(inData);
	
	if (index < 1) {
			// the data wasn't found in the list
		return 0;
	}
	
	return (Int16) index;
}


// ===========================================================================
//  CTestPopupMenu											   CTestPopupMenu 
// ===========================================================================
	CTestPopupMenu*
CTestPopupMenu::CreateTestPopupMenuStream(LStream	*inStream)
{
	return (new CTestPopupMenu(inStream));
}


// ===========================================================================
//		 CTestPopupMenu(LStream*)
// ===========================================================================
//	Construct from data in a Stream

CTestPopupMenu::CTestPopupMenu(
	LStream	*inStream)
		:	CSmartStdPopupMenu(inStream)
//		: CSharedPopupMenu(inStream)
{	
	InitTestPopupMenu();
}


// ===========================================================================
//		 ~CTestPopupMenu
// ===========================================================================
CTestPopupMenu::~CTestPopupMenu()
{	
		// remove us from the list of popup menus
	sPopupMenuList.Remove(&this);
	
		// get a handle to the popup's control record
	ControlHandle controlH = GetMacControl();
	
		// get a handle to the popup's private data record in the control record
	PopupPrivateDataHandle	dataH =
		(PopupPrivateDataHandle) (**controlH).contrlData;

		// get a handle to our menu record and make a copy of it
	MenuHandle menuH = (**dataH).mHandle;
	MenuHandle menuCopyH = menuH;
	::HandToHand((Handle*) &menuCopyH);
	
		// we dispose of the control here, after we've saved the menu record, 
		// so that we can then reinsert the menu without ~LStdControl()
		// disposing of it again
	if (mMacControlH != nil) {
		FocusDraw();
		::HidePen();
		::DisposeControl(mMacControlH);
		::ShowPen();
		mMacControlH = nil;
	}
	
		// reinsert the menu record into the menu list as a popup menu, but
		// first clear its resource bit so that the Resource Manager doesn't
		// try to load it back in from the resource fork.
	::HClrRBit((Handle) menuCopyH);
	::InsertMenu(menuCopyH, -1);

// should probably go through and set each control's menu handle to point to the new menu	
	
		// point our static menu handle to the menu we just inserted so that
		// all the instances will share this menu
	sMenuH = ::GetMenuHandle(sMenuID);
}


// ===========================================================================
//		 InitStringPopupMenu
// ===========================================================================
//	Private initializer
//
// LStdPopupMenu::InitStdPopupMenu() should handle setting the initial,
// min, and max values

	void
CTestPopupMenu::InitTestPopupMenu()
{	
		// add us to the list of existing popup menus.  we have to do this 
		// before initialzing the menu for the first time, because anything
		// that gets added to the menu won't show up in the very first popup
		// unless there's already a pointer to it in the sPopupMenuList.
	sPopupMenuList.InsertItemsAt(1, arrayIndex_Last, &this);
	
	if (sMenuH == nil) {
			// this is the first popup instance created, so we have to set the 
			// sMenuH class variable to the menu handle created for us by PowerPlant
			// so that all subsequent instances share the same menu
		sMenuH = inherited::GetMacMenuH();
		sMenuID = (**sMenuH).menuID;
		
		::HNoPurge((Handle) sMenuH);
		
		Int32 itemCount = GetMaxValue();
		Int32 dataCount = sDataList.GetCount();
		
			// these two branches can be true only if this is the
			// first popup menu to be created.  after that, the insertion and
			// deletion methods should keep the data list and the menu in sync.
		if (itemCount > dataCount) {
				// the menu has more items than we have data values for, so append
				// an empty S3tTestTool to the end of sDataList, one for each item 
				// that has no data
			S3tTestTool emptyTest = { "", kS3Prereq_None, nil };
			sDataList.InsertItemsAt(itemCount - dataCount,
														  arrayIndex_Last, &emptyTest);
		} else if (itemCount < dataCount) {
				// the menu has fewer items than we have data values for, so remove
				// the extra data values from the end of sDataList
			sDataList.RemoveItemsAt(dataCount - itemCount, itemCount + 1);
		}
		
			// we also have to initialize the menu data to hold the texture file specs
		InitializeMenu();
	}
	
		// get a handle to the popup's private data record in the control record
	PopupPrivateDataHandle	dataH =
		(PopupPrivateDataHandle) (**GetMacControl()).contrlData;

		// make the menu handle of this popup point to the static menu shared
		// by all instances of this class
	(**dataH).mHandle = sMenuH;
	(**dataH).mID = sMenuID;

	 // save the string of the item currently selected in the popup
	GetStringFromItem(GetValue(), mValueString);
	
		// broadcast the current item's string 
	SetBroadcastString(true);
}


// ===========================================================================
	void
CTestPopupMenu::SetValueByData(const void* inData)
{	
	Int16 index = GetSharedItemFromData(inData);
	
	if (index == 0) {
			// this string wasn't found in the menu, so default to the first item
		index = 1;
	}
	
	SetValue(index);
}


// ===========================================================================
	Boolean
CTestPopupMenu::GetData(void* outData) const
{
		// get the data for the currently selected item and return it in outData
	return GetSharedDataFromItem(GetValue(), outData);
}


// ===========================================================================
	void
CTestPopupMenu::SetValue(Int32 inValue)
{
	inherited::SetValue(inValue);
	
		// save the string of the item we just set
	GetStringFromItem(GetValue(), mValueString);
}

// ===========================================================================
	void
CTestPopupMenu::AppendItem(const Str255 inItemString)
{
	AppendSharedItem(inItemString);
}


// ===========================================================================
	void
CTestPopupMenu::InsertItem(const Str255 inItemString, Int16 inAfterItem)
{
	InsertSharedItem(inItemString, inAfterItem);
}


// ===========================================================================
	void
CTestPopupMenu::DeleteItem(Int16 inItemNumber)
{
	DeleteSharedItem(inItemNumber);
}


// ===========================================================================
	void
CTestPopupMenu::UpdateValues()
{
		// determine the maximum value by counting the number of items in 
		// the menu resource
	SetMaxValue(::CountMItems(GetMacMenuH()));
	
	if (GetMaxValue() > 0) {
			// there's at least one item in the menu, so make sure the min
			// value is 1, since we might have just added the first item in the
			// menu, and the min value might still be 0
		SetMinValue(1);
		
			// if items in the menu were changed, the text of the item with 
			// the same number as the value of the popup might be different 
			// than what it was before, because the order of the items has
			// shifted, or the previously selected item was deleted.  although
			// the popup's value is still the same, to the user it won't look 
			// like the same item is selected since a different string appears
			// next to the title.  so reset the value according to the string
			// of the item that was previously selected.  if the item has been
			// deleted (and thus its string can't be found in the menu), 
			// SetValueByString will default to item 1.
		SetValueByString(mValueString);
	} else {
			// there are no items in the menu, so set the other values to
			// reflect that
		SetMinValue(0);
		SetValue(0);
	}
}


// ===========================================================================
	void
CTestPopupMenu::BroadcastValueMessage()
{
	if (mValueMessage != cmd_Nothing) {
		if (mBroadcastString) {
			Str255 theString;
			GetString(theString);
			
			BroadcastMessage(mValueMessage, (void*) theString);
		} else {
// should have a method to switch between several broadcasting types:
// string, pointer, value, data
//			Int32	value = GetValue();
//
//			BroadcastMessage(mValueMessage, (void*) &value);

				// broadcast a pointer to this instance
			BroadcastMessage(mValueMessage, (void*) this);
		}
	}
}

