// ===========================================================================
// LTab.cp
// ===========================================================================
//
// changing the tab's descriptor after it's been constructed will not
// force the tabs to adjust themselves to the longer/shorter name
//	
// assumes the tab's size is not changed after FinishCreateSelf.  would
// have to call CreateRegion again to update the region
//
// LTab should really be an abstract subclass of LControl, then CTextTab
// would be a subclass of LTab
//
// must be a subview of a LTabViewAlternator
//
// assumes LTabViewAlternator draws a frame around the alternating views


#include <QuickDraw.h>
#include <QuickDrawText.h>
#include <TextEdit.h>
#include <Events.h>

#include <UTextTraits.h>
#include <UDrawingState.h>

#include "LTab.h"
#include "LTabViewAlternator.h"


// ===========================================================================
	LTab*
LTab::CreateTabStream(
	LStream* inStream)
{
	return new LTab(inStream);
}


// ===========================================================================
LTab::LTab()
{
	// unused constructor
}


// ===========================================================================
LTab::LTab(
	LStream* inStream)
	: LTextButton(inStream)
{
		// get the ID of the view that this tab controls, which was entered
		// by the user in Constructor
	inStream->ReadData(&mTabViewID, sizeof(PaneIDT));

	inStream->ReadData(&mSelectedFillColor, sizeof(RGBColor));
	inStream->ReadData(&mUnselectedFillColor, sizeof(RGBColor));
	
	inStream->ReadData(&mLabelOffsetLeft, sizeof(Int16));
	inStream->ReadData(&mLabelOffsetTop, sizeof(Int16));
	inStream->ReadData(&mLabelOffsetRight, sizeof(Int16));
	inStream->ReadData(&mLabelOffsetBottom, sizeof(Int16));
	
	mTabIndex = 1;
	mDrawLayer = 1;
	mTabRegionFull = nil;
	mTabRegionLayered = nil;
	mTabPoly = nil;
}


// ===========================================================================
LTab::~LTab()
{
	if (mTabRegionFull != nil) {
		::DisposeRgn(mTabRegionFull);
		mTabRegionFull = nil;
	}

	if (mTabRegionLayered != nil) {
		::DisposeRgn(mTabRegionLayered);
		mTabRegionLayered = nil;
	}

	if (mTabPoly != nil) {
		::KillPoly(mTabPoly);
		mTabPoly = nil;
	}
}


// ===========================================================================
	void
LTab::FinishCreateSelf()
{
		// adjust our size accordning to our label and the size of the text.
		// we do this now rather than in the constructor so that it can be 
		// overridden easily.  it also has to come before we add ourselves
		// to the view alternator so that it will know how big we are.
	AdjustSize();
	
		// assume our super view is the LTabViewAlternator
	LTabViewAlternator* alternator = (LTabViewAlternator*) mSuperView;

		// add us to this alternator
	alternator->AddTab(this);
	
		// create the shape of this tab.  we call this after AddTab because
		// that method changes the tab's position in the LTabViewAlternator.
	CreateRegion();
}


// ===========================================================================
	void
LTab::DrawSelf()
{
	StTextState origTextState;
	StColorPenState origCPenState;
	StClipRgnState origClipRegion;

	::SetClip(mTabRegionLayered);
	
	DrawTab(false);
	
	DrawLabel(false);
	
	return;
}


// ===========================================================================
// draws the background and outline of the tab.  if the tabs overlap, the
// clip region for this tab must already be set before this is called.
// otherwise, it is likely to draw on top of another tab.
//
// because a tab can have background colors, we don't want to just invert
// it when it's clicked, which would most likely produce ugly colors.  
// instead, you can pass a flag to this function to draw the inverted
// version of the tab.
//
// inInverse: pass true to draw the tab as it looks when it is inverted
//
// RETURNS: void
//
	void
LTab::DrawTab(Boolean inInverse)
{
	if (inInverse) {
			// draw the inverted version of this tab, which is simply filled 
			// with black
		::ForeColor(blackColor);
	} else if (mValue == Button_On) {
		::RGBForeColor(&mSelectedFillColor);
	} else {
		::RGBForeColor(&mUnselectedFillColor);
	}
	
		// fill the tab
	::PaintPoly(mTabPoly);

		// outline the tab in black
	::ForeColor(blackColor);
	::FramePoly(mTabPoly);

	if (mValue == Button_On && !inInverse) {
			// since we're the frontmost tab, draw a line covering the
			// black frame around the alternating view, to make it look like
			// the tab is connected to the view.  the line will be drawn
			// with whatever the tab's fill color is.  we don't do this if the
			// tab is being drawn inverted because it looks weird if we do.
		Rect theFrame;
		CalcLocalFrameRect(theFrame);
	
		::RGBForeColor(&mSelectedFillColor);

			// we have to indent the line slightly so that we don't overwrite the
			// left and right edges of the tab outline
		::MoveTo(theFrame.left + 1, theFrame.bottom - 1);
		::LineTo(theFrame.right - 2, theFrame.bottom - 1);
	}
}


// ===========================================================================
// draws the label of the tab.  if the tabs overlap, the
// clip region for this tab must already be set before this is called.
// otherwise, it is likely to draw on top of another tab.
//
// because a tab can have background colors, we don't want to just invert
// it when it's clicked, which would most likely produce ugly colors.  
// instead, you can pass a flag to this function to draw the inverted
// version of the tab label.
//
// inInverse: pass true to draw the tab as it looks when it is inverted
//
// RETURNS: void
//
	void
LTab::DrawLabel(Boolean inInverse)
{
	Rect theFrame;
	CalcLocalFrameRect(theFrame);

		// offset the tab's title from its frame 
	theFrame.left += mLabelOffsetLeft;
	theFrame.right -= mLabelOffsetRight;
	theFrame.top += mLabelOffsetTop;
	theFrame.bottom -= mLabelOffsetBottom;
	
	Int16 oldStyle;
	if ((mTextTraitsH) && (mValue != Button_Off)) {
		oldStyle = (**mTextTraitsH).style;
		(**mTextTraitsH).style ^= mSelectedStyle;
	}
	
	Int16 theJustification = 
		UTextTraits::SetPortTextTraits(mTextTraitsH ? *mTextTraitsH : nil);
	
	if ((mTextTraitsH) && (mValue != Button_Off)) {
		(**mTextTraitsH).style = oldStyle;
	}

	if (theJustification == teFlushDefault) {
		theJustification = ::GetSysDirection();
	}
	
	if (inInverse) {
		::ForeColor(whiteColor);
	} else {
		::ForeColor(blackColor);
	}

		// change the mode to Or to make sure we don't overwrite the background
	::TextMode(srcOr);
	
	UTextDrawing::DrawWithJustification((Ptr) &mText[1], mText[0], 
										 theFrame, theJustification);	
}


// ===========================================================================
	void
LTab::CreateRegion()
{
	Rect theFrame;
	CalcLocalFrameRect(theFrame);
	
		// the tab we're going to draw looks roughly like this
		//	______________
		//	|             \
		//	|              \
		//	----------------
		//
		// create a polygon in which to store the outline of the tab
	mTabPoly = ::OpenPoly();
	Assert_(mTabPoly);
	
		// move to the bottom left corner and draw a line to the bottom right.
		// we have to subtract 1 from the bottom and right coords because the
		// poly outline hangs to the bottom and to the right of the coords.
	::MoveTo(theFrame.left, theFrame.bottom - 1);
	::LineTo(theFrame.right - 1, theFrame.bottom - 1);
	
		// draw a 45-degree line from the bottom right to the top of the frame
	::LineTo(theFrame.right - 1 - (theFrame.bottom - theFrame.top - 1), theFrame.top);

		// draw a line from the end ofthe diagonal line to the top-left corner,
		// then to the bottom left, where we started
	::LineTo(theFrame.left, theFrame.top);
	::LineTo(theFrame.left, theFrame.bottom - 1);
	
	::ClosePoly();
	
		// create a region which is equivalent to the filled polygon
	mTabRegionFull = ::NewRgn();
	Assert_(mTabRegionFull);

	::OpenRgn();
	
		// for some reason, simply framing the polygon we created above for
		// the region doesn't include the bottom edge of the tab.  so we repeat
		// the above drawing commands, except going all the way to the right
		// and bottom coordinates.
		// move to the bottom left corner and draw a line to the bottom right.
	::MoveTo(theFrame.left, theFrame.bottom);
	::LineTo(theFrame.right, theFrame.bottom);
	
		// draw a 45-degree line from the bottom right to the top of the frame
	::LineTo(theFrame.right - (theFrame.bottom - theFrame.top), theFrame.top);

		// draw a line from the end ofthe diagonal line to the top-left corner,
		// then to the bottom left, where we started
	::LineTo(theFrame.left, theFrame.top);
	::LineTo(theFrame.left, theFrame.bottom);

	::CloseRgn(mTabRegionFull);

		// initialize mTabRegionLayered to a copy of the full region
	mTabRegionLayered = ::NewRgn();
	ThrowIfNil_(mTabRegionLayered);
	
	::CopyRgn(mTabRegionFull, mTabRegionLayered);
}


// ===========================================================================
	void
LTab::HotSpotAction(
	Int16	/* inHotSpot */,
	Boolean	inCurrInside,
	Boolean	inPrevInside)
{
		// set the clipping region so we don't overwrite neighboring tabs
	StClipRgnState origClipRegion;

	::SetClip(mTabRegionLayered);
	
	if (inCurrInside) {
		if (!inPrevInside) {
			DrawTab(true);
			DrawLabel(true);
		}
	} else {
		if (inPrevInside) {
			DrawTab(false);
			DrawLabel(false);
		}
	}
}


// ===========================================================================
	void
LTab::HotSpotResult(
	Int16	/* inHotSpot */)
{
	if (mValue == Button_Off) {
		SetValue(Button_On);		// Turn the button on
									// Tell everyone that we clicked
		BroadcastMessage(msg_ControlClicked, (void*) this);
	} else {
			// we've been clicked while we're already the frontmost tab (mValue ==
			// Button_On), so we have to redraw the non-inverted tab and label,  
			// which were just inverted when we got the mouse down event.
		DrawTab(false);
		DrawLabel(false);
	}
}


// ===========================================================================
	void
LTab::SetValue(
	Int32 inValue)
{
	Int32 oldValue = mValue;

		// let LControl change our mValue to me new value and broadcast a 
		// our message value to our listeners.  we don't broadcast a
		// msg_ClickedButton message, because LTabViewAlternator responds to that
		// by calling our SetValue method, so we'd get into a loop.
	LControl::SetValue(inValue);

	if (oldValue != inValue) {
		Refresh();
	}
}


// ===========================================================================
	PaneIDT
LTab::GetTabViewID() const
{
	return mTabViewID;
}


// ===========================================================================
	void
LTab::SetTabIndex(Int32 inIndex)
{
	mTabIndex = inIndex;
}


// ===========================================================================
	Int32
LTab::GetTabIndex() const
{
	return mTabIndex;
}


// ===========================================================================
	void
LTab::SetDrawLayer(Int16 inDrawLayer)
{
	mDrawLayer = inDrawLayer;
}


// ===========================================================================
	Int16
LTab::GetDrawLayer()
{
	return mDrawLayer;
}


// ===========================================================================
	void
LTab::AdjustSize()
{
		// save the current text settings, which will be restored in 
		// this object's desctructor
	StTextState origTextState;

		// set the text traits to our unselected style
	UTextTraits::SetPortTextTraits(mTextTraitsH ? *mTextTraitsH : nil);
	
		// figure out wide our label is with this style
	short plainLabelWidth = ::StringWidth(mText);

		// remember the plain style and then change it to our selected style
	Int16 plainStyle = (**mTextTraitsH).style;
	(**mTextTraitsH).style ^= mSelectedStyle;

		// set the text traits to our selected style
	UTextTraits::SetPortTextTraits(mTextTraitsH ? *mTextTraitsH : nil);
	
		// figure out wide our label is with this style
	short selectedLabelWidth = ::StringWidth(mText);
	
		// return to our unselected style
	(**mTextTraitsH).style = plainStyle;

	short labelWidth = (plainLabelWidth > selectedLabelWidth) ? plainLabelWidth : selectedLabelWidth;

	short width = mLabelOffsetLeft + labelWidth + mLabelOffsetRight + 2;

		// get the current dimensions of the tab
	SDimension16 currentSize;
	GetFrameSize(currentSize);
	
		// change the width of the tab to accommodate the label, don't
		// change the theight, don't refresh the view
	ResizeFrameTo(width, currentSize.height, false);
}


// ===========================================================================
	void
LTab::SetLayeredRegion(RgnHandle inNewRegionH)
{
	if (inNewRegionH != nil) {
		::CopyRgn(inNewRegionH, mTabRegionLayered);
	}
}


// ===========================================================================
	RgnHandle
LTab::GetLayeredRegion()
{
	return mTabRegionLayered;
}


// ===========================================================================
	RgnHandle
LTab::GetFullRegion()
{
	return mTabRegionFull;
}


