// ===========================================================================
// LViewAlternator.cp
// ===========================================================================
//
// only use Show/Next/Prev/View to change views; hiding and showing views
// manually will mess everything up
//
// stores the pane IDs of the views to be alternated in a resource of type 'VALL'


#include <LCommander.h>

#include "LViewAlternator.h"


const FourCharCode kAlternateViewListResType = 'VALL';


// ===========================================================================
	LViewAlternator*
LViewAlternator::CreateViewAlternatorStream(
	LStream	*inStream)
{
	return (new LViewAlternator(inStream));
}


// ===========================================================================
LViewAlternator::LViewAlternator()
{
	// unused default constructor
}


// ===========================================================================
LViewAlternator::LViewAlternator(LStream* inStream)
	: LView(inStream),
		mViewList(sizeof(LView*))
{
	mCurrentViewIndex = 1;
}


// ===========================================================================
LViewAlternator::~LViewAlternator()
{
}


// ===========================================================================
	void
LViewAlternator::FinishCreateSelf()
{
		// look for a resource containing the resource IDs of the PPobs among
		// which we'll alternate.  the resource should have the same ID as 
		// our pane ID and an exception should not be thrown if StResource 
		// can't allocate enough memory.
	StResource viewPPobListH(kAlternateViewListResType, GetPaneID(), false);

	Assert_(Handle(viewPPobListH));
	
	if (Handle(viewPPobListH) != nil) {
			// we found the resource, so get the number of pane IDs in it.  
			// the resource contains a 32-bit array length followed by 0 or 
			// more Pane IDs (32-bit).  so cast the handle to a pointer to 
			// Pane IDs, to get the number of IDs first.
		PaneIDT* viewPPobListP = (PaneIDT*) *Handle(viewPPobListH);
		Int32 viewCount = viewPPobListP[0];

		Assert_(viewCount);
		
			// point to the first pane ID in the resource, past the list
			// length which is the first thing in the resource
		PaneIDT* viewIDP = &viewPPobListP[1];

		LCommander* defaultCommander = LCommander::GetDefaultCommander();
		
			// reanimate each of the panes whose IDs are in this resource, 
			// adding them to our list of alternate views
		for (Int32 i = 0; i < viewCount; i++) {
				// all the subviews will have the same default commander
			LCommander::SetDefaultCommander(defaultCommander);
			
			LCommanderView* theView = CreateView(*viewIDP++);

			Assert_(theView);
			
				// store a pointer to the view so we can access it later
			mViewList.InsertItemsAt(1, arrayIndex_Last, &theView);
		} // endfor
	}
}


// ===========================================================================
	void
LViewAlternator::ShowView(Int32 inViewIndex)
{
	LCommanderView* newView;
	
	if (mViewList.FetchItemAt(inViewIndex, &newView)) {
		Assert_(newView);
		
			// inViewIndex is a valid index
		LCommanderView* currentView;
		
			// we seem to have to restore the target in the new view before we
			// hide the current one.  if we don't, the current view's tab group 
			// doesn't remember the currently selected edit field, which means
			// it won't be restored the next time we call RestoreTarget on it.
		newView->RestoreTarget();
		
		if (mViewList.FetchItemAt(mCurrentViewIndex, &currentView)) {
				// we have a current view, so hide it
			currentView->Hide();
		}

		newView->Show();
		
			// make the new view the current one
		mCurrentViewIndex = inViewIndex;
	}
}


// ===========================================================================
	void
LViewAlternator::ShowNextView()
{
	Int32 newViewIndex = mCurrentViewIndex + 1;
	
	if (newViewIndex > mViewList.GetCount()) {
			// we've incremented the index beyond the end of the list, so
			// wrap around to the beginning
		newViewIndex = 1;
	}

	LCommanderView* newView = nil;
	
	if (mViewList.FetchItemAt(newViewIndex, &newView)) {
		Assert_(newView);
		
			// we seem to have to restore the target in the new view before we
			// hide the current one.  if we don't, the current view's tab group 
			// doesn't remember the currently selected edit field, which means
			// it won't be restored the next time we call RestoreTarget on it.
		newView->RestoreTarget();
	}
		
	LCommanderView* currentView;
	
	if (mViewList.FetchItemAt(mCurrentViewIndex, &currentView)) {
		Assert_(currentView);
		
			// we have a current view, so hide it
		currentView->Hide();
	}

	if (newView != nil) {
			// show the next view
		newView->Show();
	}
	
	mCurrentViewIndex = newViewIndex;
}


// ===========================================================================
	void
LViewAlternator::ShowPreviousView()
{
	Int32 newViewIndex = mCurrentViewIndex - 1;
	
	if (newViewIndex > mViewList.GetCount()) {
			// we've decremented the index beyond the beginning of the list, so
			// wrap around to the end
		newViewIndex = mViewList.GetCount();
	}

	LCommanderView* newView = nil;
	
	if (mViewList.FetchItemAt(newViewIndex, &newView)) {
		Assert_(newView);
		
			// we seem to have to restore the target in the new view before we
			// hide the current one.  if we don't, the current view's tab group 
			// doesn't remember the currently selected edit field, which means
			// it won't be restored the next time we call RestoreTarget on it.
		newView->RestoreTarget();
	}
		
	LCommanderView* currentView;
	
	if (mViewList.FetchItemAt(mCurrentViewIndex, &currentView)) {
		Assert_(currentView);
		
			// we have a current view, so hide it
		currentView->Hide();
	}

	if (newView != nil) {
			// show the previous view
		newView->Show();
	}
	
	mCurrentViewIndex = newViewIndex;
}


// ===========================================================================
	Int32
LViewAlternator::GetCurrentViewIndex()
{
	return mCurrentViewIndex;
}


// ===========================================================================
	LCommanderView*
LViewAlternator::CreateView(
	PaneIDT inPaneID)
{
	LCommanderView* theView;
	
	LPane::SetDefaultView(this);

		// create the view from the PPob, using us as the default view
	theView = (LCommanderView*) UReanimator::ReadObjects('PPob', inPaneID);
	
	Assert_(theView);
	
	if (theView != nil) {
		theView->FinishCreate();
		
		if (theView->IsVisible()) {
			theView->Refresh();
		}
	}
	
	return theView;
}