// ===========================================================================
//
//	S3TestInfo.cp
//
//	Copyright (C) 1995 Apple Computer, Inc.  All rights reserved.
//
// ===========================================================================


// ===========================================================================
//	Includes
// ===========================================================================

#include <string.h>
#include <UGWorld.h>
#include <UMemoryMgr.h>
#include <UDrawingState.h>
#include <UKeyFilters.h>

#include "RAVE.h"

#include "CSneak3viewApp.h"
#include "CTestInfo.h"
#include "CTestPopupMenu.h"
#include "CTinselTestWindow.h"
#include "CComparisonWindow.h"
#include "CTQAImage.h"
#include "CTQAColorTable.h"
#include "S3TestInfo_sys.h"
#include "S3TestSystem.h"
#include "S3TestInfo.h"
#include "S3TestUtil.h"
#include "S3Utilities.h"
#include "StContinueDialog.h"
#include "UGWorldState.h"
#include "StringConversion.h"


extern "C" {


// ===========================================================================
//	Constants
// ===========================================================================

// ===========================================================================
//	Types
// ===========================================================================

// ===========================================================================
//	Private Prototypes
// ===========================================================================

	TS3Error
S3TestInfo_SetupGDevice(
	S3tTestInfoP			inTestInfo);
	
	TS3Error
S3TestInfo_SetupMemoryDevice(
	S3tTestInfoP			inTestInfo);

	
// ===========================================================================
//	Globals
// ===========================================================================

	// the last RAVE error is stored in a global rather than in the 
	// S3tTestInfo struct because that way, we don't have to change
	// the prefs file version, and there's no real reason to save
	// it between tests
TQAError					gLastRAVEError = 				kQANoErr;

	// the test result buffer is a global for the same reason as 
	// gLastRAVEError.  it can only be accessed inside a test, so there's
	// no reason to save it in the prefs file.
Byte						gTestResultBuffer[kTestResultBufferSize] = { 0 };


// ===========================================================================
//	S3TestInfo_RunTest
// ===========================================================================
	S3tTestStatus
S3TestInfo_RunTest(
	S3tTestInfoP			inTestInfo,
	const char*				inTestName)
{
	if (inTestName != nil) {
			// the caller passed in the name of the test they want to run, so
			// copy it into the testName field of inTestInfo.  if the name is 
			// nil, we'll just use the one that's already there.
		TS3Error			s3Error = 
				S3TestInfo_SetBlock(inTestInfo, kS3TagBlock_TestName, inTestName);
				
		if (s3Error != kS3Error_None) {
			S3Test_Print("*** Error (%d) occurred while setting test name. ***\r", s3Error);
			return kS3Status_Failure;
		}
	}
	
	if (strlen((char*) inTestInfo->testName) == 0) {
			// we can't run the test if we don't have the name of it
		S3Test_Print("*** Name of test is blank. ***\r");

		return kS3Status_NoSuchTest;
	}
		
		// get the info for this test.  we cast the testName to a C string
		// so that CTestPopupMenu will use the C rather than Pascal string 
		// version of GetSharedDataFromString.  
	S3tTestTool 			theTest;
	if (!CTestPopupMenu::GetSharedDataFromString((char*) inTestInfo->testName, &theTest)) {
			// a test with the current name doesn't exist
		S3Test_Print("*** Test named '%s' cannot be found. ***\r", 
				inTestInfo->testName);

		return kS3Status_NoSuchTest;
	}

		// bring the test windows to the front.  this is important to do if
		// a comparison window is open, because that will be frontmost after
		// the last test, when the windows were compared.
	CSneak3viewApp::Instance()->SelectTestWindows();
	
		// print out the date, time and name of the test
	unsigned long seconds = 0;
	::GetDateTime(&seconds);
	
	Str255 dateTimeString;
	
	::TimeString(seconds, true, dateTimeString, nil);
	S3PtoCString(dateTimeString);
	
	S3Test_Print("%s\r%s\r%s on ", 
				 kS3Divider_Hyphen, 
				 theTest.testName, 
				 dateTimeString);

	::DateString(seconds, longDate, dateTimeString, nil);
	S3PtoCString(dateTimeString);
	
	S3Test_Print("%s\r\r", dateTimeString);
		
		// run the test
	S3tTestStatus 			result = (*theTest.testMethod)(inTestInfo);
	
	if (inTestInfo->leftDevice.deviceType == kQADeviceMemory) {
			// we're using a memory device, so we have to copy the pixels from
			// the off-screen memory to the screen
		if (inTestInfo->leftDeviceGWorld != nil) {
				// save the current GWorld and restore it when we leave the block
			StGWorldState();
			
				// prepare to draw into the left window
			CTinselTestWindow::GetLeftWindow()->FocusDraw();
			GrafPtr windowP = CTinselTestWindow::GetLeftWindow()->GetMacPort();
			
				// make sure the colors are set correctly for the CopyBits
			::ForeColor(blackColor);
			::BackColor(whiteColor);
			
				// copy the offscreen GWorld to the window
			inTestInfo->leftDeviceGWorld->CopyImage(windowP, windowP->portRect);

				// unlock the pixels now that we're done with them
			::UnlockPixels(::GetGWorldPixMap(inTestInfo->leftDeviceGWorld->GetMacGWorld()));
			
				// once we've copied the pixels to the screen, we can delete the
				// offscreen GWorld
			delete inTestInfo->leftDeviceGWorld;
			inTestInfo->leftDeviceGWorld = nil;
		}
	}
	
	if (inTestInfo->rightDevice.deviceType == kQADeviceMemory) {
			// we're using a memory device, so we have to copy the pixels from
			// the off-screen memory to the screen
		if (inTestInfo->rightDeviceGWorld != nil) {
				// save the current GWorld and restore it when we leave the block
			StGWorldState();
			
				// prepare to draw into the right window
			CTinselTestWindow::GetRightWindow()->FocusDraw();
			GrafPtr windowP = CTinselTestWindow::GetRightWindow()->GetMacPort();
			
				// make sure the colors are set correctly for the CopyBits
			::ForeColor(blackColor);
			::BackColor(whiteColor);
			
				// copy the offscreen GWorld to the window
			inTestInfo->rightDeviceGWorld->CopyImage(windowP, windowP->portRect);

				// unlock the pixels now that we're done with them
			::UnlockPixels(::GetGWorldPixMap(inTestInfo->rightDeviceGWorld->GetMacGWorld()));
			
				// once we've copied the pixels to the screen, we can delete the
				// offscreen GWorld
			delete inTestInfo->rightDeviceGWorld;
			inTestInfo->rightDeviceGWorld = nil;
		}
	}

		// save the output from the test now, so that if something overlaps
		// part of the test windows, we can redraw them
	CSneak3viewApp::Instance()->SaveTestOutput();
	
	if (result == kS3Status_Failure) {
		S3Test_Print("*** Failure code returned ***\r\r");
	}
	
		// we always call CompareWindows so that if a comparison window
		// is open, it will be updated now that this test has finished. 
		// if no window is open, CompareWindows will do nothing.
	CComparisonWindow::CompareWindows(
		CTinselTestWindow::GetLeftWindow(),
		CTinselTestWindow::GetRightWindow());

	S3Test_Print("\r%s\r", kS3Divider_Hyphen);
		
	return result;
}


// ===========================================================================
//	S3TestInfo_Refresh
// ===========================================================================
	TS3Error
S3TestInfo_Refresh(
	S3tTestInfoP			inTestInfo)
{
	if (strlen((char*) inTestInfo->testName) == 0) {
			// we can't run the test if we don't have the name of it
		S3Test_Print("*** Name of test is blank. ***\r");

		return kS3Error_General;
	}
		
		// get the info for this test
	S3tTestTool 			theTest;
	if (!CTestPopupMenu::GetSharedDataFromString(inTestInfo->testName, &theTest)) {
			// a test with the current name doesn't exist, so convert the 
			// name and print an error message
		char				testNameCString[256];
		S3PtoCString(inTestInfo->testName, testNameCString);
		
		S3Test_Print("*** Test named '%s' cannot be found. ***\r", 
				testNameCString);

		return kS3Error_General;
	}
	
		// in case we've already created draw contexts for the current TestInfo,
		// dispose of any memory it contains
	CTestInfo::DisposeTestInfo(inTestInfo);

	inTestInfo->leftMicroseconds = 0;
	inTestInfo->rightMicroseconds = 0;

		// use the names stored in inTestInfo to get pointers to the engines
	inTestInfo->leftEngine = S3GetEngineByName(inTestInfo->leftEngineName);
	inTestInfo->rightEngine = S3GetEngineByName(inTestInfo->rightEngineName);
	
	if (inTestInfo->leftEngine == nil)  {
			// we can't create a draw context without knowing which engines to use
		S3Test_Print("*** Engine named '%s' cannot be found. ***\r", 
				inTestInfo->leftEngineName);

		return kS3Error_General;
	}
	
	if (inTestInfo->rightEngine == nil)  {
			// we can't create a draw context without knowing which engines to use
		S3Test_Print("*** Engine named '%s' cannot be found. ***\r", 
				inTestInfo->rightEngineName);

		return kS3Error_General;
	}
	
		// make sure we can fulfill the test's prerequisites
	TS3Error				s3Error;
	if (theTest.testPrereqs & kS3Prereq_Texture) {
			// this test requires a texture, so make sure the user has selected one
		if (inTestInfo->textureFSSpec.name[0] > 0) {
				// a texture is selected in the menu, so load the specified PICT 
				// file and create a texture from it for each engine
			s3Error = S3TestInfo_SetTextureFromFileSpec(
					inTestInfo, &inTestInfo->textureFSSpec);
					
			if (s3Error != kS3Error_None) {
					// something went wrong and we couldn't create the textures
				return s3Error;
			}
		} else {
				// the test needs a texture, but the user hasn't selected one,
				// so alert them
			S3Test_Print("*** Test requires a texture, but no texture file is selected. ***\r");
			
			return kS3Error_General;
		}
	}

	if (theTest.testPrereqs & kS3Prereq_Bitmap) {
			// this test requires a bitmap, so make sure the user has selected one
		if (inTestInfo->bitmapFSSpec.name[0] > 0) {
				// a bitmap is selected in the menu, so load the specified PICT 
				// file and create a bitmap from it for each engine
			s3Error = S3TestInfo_SetBitmapFromFileSpec(
					inTestInfo, &inTestInfo->bitmapFSSpec);
					
			if (s3Error != kS3Error_None) {
					// something went wrong and we couldn't create the bitmaps
				return kS3Error_General;
			}
		} else {
				// the test needs a bitmap, but the user hasn't selected one,
				// so alert them
			S3Test_Print("*** Test requires a bitmap, but no bitmap file is selected. ***\r");
			
			return kS3Error_General;
		}
	}

		// set up the devices for the draw contexts
	switch (inTestInfo->deviceMenu) {
		case kS3Device_Window:
			s3Error = S3TestInfo_SetupGDevice(inTestInfo);
			
			if (s3Error != kS3Error_None) {
					// the GDevice couldn't be set up correctly, so we can't
					// run the test
				return s3Error;
			}
			break;
		
		case kS3Device_Memory16:
		case kS3Device_Memory32:
			s3Error = S3TestInfo_SetupMemoryDevice(inTestInfo);
			
			if (s3Error != kS3Error_None) {
					// the memory device couldn't be set up correctly, so we can't
					// run the test
				return s3Error;
			}
			break;
	} // endswitch

		// now make the left clipping region, using local device coords
	inTestInfo->leftClip.clip.clipRgn = 
			CTestInfo::MakeClipRegion(&(inTestInfo->leftRect), 
					inTestInfo->clipRegionMenu);
																				 
		// create the draw context itself, using a clipping region if necessary
	TQAError qaError = QADrawContextNew(
			&inTestInfo->leftDevice, 
			&inTestInfo->leftRect, 
			(inTestInfo->leftClip.clip.clipRgn) ? &inTestInfo->leftClip : nil,
			inTestInfo->leftEngine,
			inTestInfo->contextFlags,
			&inTestInfo->leftContext);
																					 
	if (qaError != kQANoErr || inTestInfo->leftContext == nil) {
			// we couldn't create the left draw context for some reason
		S3Test_Print("*** Left engine could not create a new draw context. ***\r");
		
		gLastRAVEError = qaError;
		
		return kS3Error_RAVE;
	}

		// now make the right clipping region, using local device coords
	inTestInfo->rightClip.clip.clipRgn = 
			CTestInfo::MakeClipRegion(&(inTestInfo->rightRect),
					inTestInfo->clipRegionMenu);	
																				 
		// create the draw context itself, using a clipping region if necessary
	qaError = QADrawContextNew(
			&inTestInfo->rightDevice, 
			&inTestInfo->rightRect, 
			(inTestInfo->rightClip.clip.clipRgn) ? &inTestInfo->rightClip : nil, 
			inTestInfo->rightEngine,
			inTestInfo->contextFlags,
			&inTestInfo->rightContext);

	if (qaError != kQANoErr || inTestInfo->rightContext == nil) {
			// we couldn't create the right draw context for some reason
		S3Test_Print("*** Right engine could not create a new draw context. ***\r");
		
		gLastRAVEError = qaError;
		
		return kS3Error_RAVE;
	}

		// now that the draw contexts have been created, set the state variables
		// for each one
	S3Test_SetInt(inTestInfo, kQATag_ZFunction, inTestInfo->zFunction);

	S3Test_SetFloat(inTestInfo, kQATag_ColorBG_a, inTestInfo->alpha);

	S3Test_SetFloat(inTestInfo, kQATag_ColorBG_r, inTestInfo->red);

	S3Test_SetFloat(inTestInfo, kQATag_ColorBG_g, inTestInfo->green);

	S3Test_SetFloat(inTestInfo, kQATag_ColorBG_b, inTestInfo->blue);

	S3Test_SetFloat(inTestInfo, kQATag_Width, inTestInfo->lineWidth);
	
	S3Test_SetInt(inTestInfo, kQATag_Antialias, inTestInfo->antialiasing);

	S3Test_SetInt(inTestInfo, kQATag_Blend, inTestInfo->blending);

	S3Test_SetInt(inTestInfo, kQATag_PerspectiveZ, inTestInfo->perspectiveZ);

		// if textures have been created, set each draw context's current
		// texture separately
	if (inTestInfo->leftTexture != nil) {
		QASetPtr(inTestInfo->leftContext, kQATag_Texture, inTestInfo->leftTexture);
	}
	
	if (inTestInfo->rightTexture != nil) {
		QASetPtr(inTestInfo->rightContext, kQATag_Texture, inTestInfo->rightTexture);
	}
	
	S3Test_SetInt(inTestInfo, kQATag_TextureFilter, inTestInfo->textureFilter);

	S3Test_SetInt(inTestInfo, kQATag_TextureOp, inTestInfo->textureOp);
	
	if (inTestInfo->resetRandomSeed) {
			// reset the seed used for the S3TestUtil_RandomFloat function
		S3TestUtil_ResetRandomSeed();
	}
	
	return kS3Error_None;
}


// ===========================================================================
//	S3TestInfo_SetupGDevice
// ===========================================================================
	TS3Error
S3TestInfo_SetupGDevice(
	S3tTestInfoP			inTestInfo)
{
	ASSERT(inTestInfo);

	inTestInfo->leftDevice.deviceType = kQADeviceGDevice;
	
		// since the selected device is a screen, make sure that it is the same
		// one on which the test window's upper left corner is located.  the user
		// may have moved the window onto a screen which is not the deepest one,
		// which is where the windows are put by default.
	Boolean 				isOnOneDevice = false;
	inTestInfo->leftDevice.device.gDevice = 
		S3GetDeviceFromRect(&inTestInfo->leftGlobalRect, &isOnOneDevice);
	
	if (inTestInfo->leftDevice.device.gDevice == nil) {
			// for some reason we can't find the device the window is on,
			// so we can't run the test
		S3Test_Print("*** Left device cannot be found. ***\r");
		
		return kS3Error_OS;
	}
	
	TQAError 				qaError = 
			QAEngineCheckDevice(inTestInfo->leftEngine, &inTestInfo->leftDevice);
			
	if (qaError != kQANoErr) {
			// the chosen engine doesn't support the chosen device, so alert the
			// user that we can't run the test
		S3Test_Print("*** Left engine does not support the chosen device. ***\r");
		
		gLastRAVEError = qaError;

		return kS3Error_RAVE;
	}

	inTestInfo->rightDevice.deviceType = kQADeviceGDevice;
	inTestInfo->rightDevice.device.gDevice = 
		S3GetDeviceFromRect(&inTestInfo->rightGlobalRect, &isOnOneDevice);

	if (inTestInfo->rightDevice.device.gDevice == nil) {
			// for some reason we can't find the device the window is on,
			// so we can't run the test
		S3Test_Print("*** Right device cannot be found. ***\r");
		
		return kS3Error_OS;
	}
	
	qaError = QAEngineCheckDevice(inTestInfo->rightEngine, &inTestInfo->rightDevice);
			
	if (qaError != kQANoErr) {
			// the chosen engine doesn't support the chosen device, so alert the
			// user that we can't run the test
		S3Test_Print("*** Right engine does not support the chosen device. ***\r");
		
		gLastRAVEError = qaError;

		return kS3Error_RAVE;
	}
	
		// convert the leftGlobalRect into a TQARect local to the device coords
	Rect 					tempLeftRect = inTestInfo->leftGlobalRect;
	S3GlobalToDeviceLocal(inTestInfo->leftDevice.device.gDevice, &tempLeftRect);
	inTestInfo->leftRect.left = tempLeftRect.left;
	inTestInfo->leftRect.top = tempLeftRect.top;
	inTestInfo->leftRect.right = tempLeftRect.right;
	inTestInfo->leftRect.bottom = tempLeftRect.bottom;

		// convert the rightGlobalRect into a TQARect local to the device coords
	Rect 					tempRightRect = inTestInfo->rightGlobalRect;
	S3GlobalToDeviceLocal(inTestInfo->rightDevice.device.gDevice, &tempRightRect);
	inTestInfo->rightRect.left = tempRightRect.left;
	inTestInfo->rightRect.top = tempRightRect.top;
	inTestInfo->rightRect.right = tempRightRect.right;
	inTestInfo->rightRect.bottom = tempRightRect.bottom;

	return kS3Error_None;
}


// ===========================================================================
//	S3TestInfo_SetupMemoryDevice
// ===========================================================================
	TS3Error
S3TestInfo_SetupMemoryDevice(
	S3tTestInfoP			inTestInfo)
{
	ASSERT(inTestInfo);

	inTestInfo->leftDevice.deviceType = kQADeviceMemory;
	inTestInfo->rightDevice.deviceType = kQADeviceMemory;
	
		// since a memory device is being used, the device rect's
		// origin is at 0,0
	inTestInfo->leftRect.left = 0;
	inTestInfo->leftRect.top = 0;
	inTestInfo->leftRect.right = inTestInfo->leftGlobalRect.right - inTestInfo->leftGlobalRect.left;
	inTestInfo->leftRect.bottom = inTestInfo->leftGlobalRect.bottom - inTestInfo->leftGlobalRect.top;

	inTestInfo->rightRect = inTestInfo->leftRect;
	
	Uint16 bytesPerPixel;
	Uint16 cmpSize;
	Uint16 depth;
	
		// setup the pixel type of the device, based on what kind the user chose
	if (inTestInfo->deviceMenu == kS3Device_Memory16) {
		inTestInfo->leftDevice.device.memoryDevice.pixelType = kQAPixel_ARGB16;
		bytesPerPixel = 2;
		cmpSize = 5;
		depth = 16;
	} else {
		inTestInfo->leftDevice.device.memoryDevice.pixelType = kQAPixel_ARGB32;
		bytesPerPixel = 4;
		cmpSize = 8;
		depth = 32;
	}

	inTestInfo->leftDevice.device.memoryDevice.width = inTestInfo->leftRect.right;
	inTestInfo->leftDevice.device.memoryDevice.height = inTestInfo->leftRect.bottom;

		// copy all these settings to the right device
	inTestInfo->rightDevice.device.memoryDevice = inTestInfo->leftDevice.device.memoryDevice;

	Rect bounds = { 
		inTestInfo->leftRect.top, 
		inTestInfo->leftRect.left, 
		inTestInfo->leftRect.bottom, 
		inTestInfo->leftRect.right
	};
	
	inTestInfo->leftDeviceGWorld = CTestInfo::NewDeviceGWorld(bounds, depth);
	if (inTestInfo->leftDeviceGWorld == nil) {
			// there wasn't enough memory to create a new pixmap, so we can't
			// create the memory device
		S3Test_Print("*** Not enough memory to create left memory device. ***\r");
		
		return kS3Error_NotEnoughMemory;
	}
	
	inTestInfo->rightDeviceGWorld = CTestInfo::NewDeviceGWorld(bounds, depth);
	if (inTestInfo->rightDeviceGWorld == nil) {
			// there wasn't enough memory to create a new pixmap, so we can't
			// create the memory device
		S3Test_Print("*** Not enough memory to create right memory device. ***\r");
		
		return kS3Error_NotEnoughMemory;
	}
	
	PixMapHandle pixMapH = ::GetGWorldPixMap(inTestInfo->leftDeviceGWorld->GetMacGWorld());
	
		// to get the pixmap's actual rowbytes, we have to mask off the top two bits
	inTestInfo->leftDevice.device.memoryDevice.rowBytes = 
		(**pixMapH).rowBytes & 0x3fff;

		// lock the pixel data of the GWorld so that it won't move when the
		// engine draws into it
	::LockPixels(pixMapH);
	
		// point the memory device to the pixels of the GWorld
	inTestInfo->leftDevice.device.memoryDevice.baseAddr = ::GetPixBaseAddr(pixMapH);
	
	pixMapH = ::GetGWorldPixMap(inTestInfo->rightDeviceGWorld->GetMacGWorld());
	
		// to get the pixmap's actual rowbytes, we have to mask off the top two bits
	inTestInfo->rightDevice.device.memoryDevice.rowBytes = 
		(**pixMapH).rowBytes & 0x3fff;

		// lock the pixel data of the GWorld so that it won't move when the
		// engine draws into it
	::LockPixels(pixMapH);
	
		// point the memory device to the pixels of the GWorld
	inTestInfo->rightDevice.device.memoryDevice.baseAddr = ::GetPixBaseAddr(pixMapH);
	
		// if we've gotten here, then we have a valid memory device 
	return kS3Error_None;
}


// ===========================================================================
//	S3TestInfo_AskToContinue
// ===========================================================================
	Boolean
S3TestInfo_AskToContinue(
	S3tTestInfoP			inTestInfo)
{
	#pragma unused(inTestInfo)

		// if the comparison window is open, we want to use that as the 
		// dialog's super commander.  that way, the user can still select
		// the various comparison modes from the menu or with key commands.
	LCommander*				superCommander = CComparisonWindow::GetWindow();
	
	if (superCommander == nil) {
			// the comparison window isn't open, so use the app as the 
			// dialog's super commander
		superCommander = CSneak3viewApp::Instance();
	}
		
		// create a new continue dialog
	StContinueDialog		continueDialog(superCommander);
	
		// loop until the user clicks a button
	for (;;) {
		MessageT			message = continueDialog.DoDialog();
		
		if (message == msg_Continue) {
			return true;
		} else if (message == msg_Stop) {
			S3Test_Print("\r*** Test stopped by user. ***\r\r");
					
			return false;
		}
	} // endfor
	
	return true;
}


// ===========================================================================
//	S3TestInfo_CheckForCancel
// ===========================================================================
	Boolean
S3TestInfo_CheckForCancel(
	S3tTestInfoP			inTestInfo)
{	
	Boolean					foundCancel = false;
	QHdrPtr					eventQueueHead = ::GetEvQHdr();
	EventRecord				translatedEvent;
	
		// loop through each event in the queue
	for (EvQElPtr currentEvent = (EvQEl*) eventQueueHead->qHead; 
			currentEvent; 
			currentEvent = (EvQEl*) currentEvent->qLink) {
		if (currentEvent->evtQWhat == keyDown) {
				// we found a key down event, so convert the event from
				// the queue into a standard EventRecord
			translatedEvent.what = currentEvent->evtQWhat;
			translatedEvent.message = currentEvent->evtQMessage;
			translatedEvent.when = currentEvent->evtQWhen;
			translatedEvent.where = currentEvent->evtQWhere;
			translatedEvent.modifiers = currentEvent->evtQModifiers;
			
				// let PowerPlant decide if this is a command-period
			if (foundCancel = UKeyFilters::IsCmdPeriod(translatedEvent)) {
				S3Test_Print("\r*** Test stopped by user. ***\r\r");
				break;
			}
		}
	} // endfor
	
	return foundCancel;
}


// ===========================================================================
//	S3TestInfo_SetTextureFromFileSpec
// ===========================================================================
	TS3Error
S3TestInfo_SetTextureFromFileSpec(
	S3tTestInfoP			inTestInfo,
	const FSSpec*			inFileSpec)
{
	ASSERT(inTestInfo);
	ASSERT(inFileSpec);
	
		// load the PICT file into a handle
	PicHandle pictH = S3GetPICTFromFile(inFileSpec);
	
		// copy and convert the file's name
	char				fileNameCString[64];
	S3PtoCString(inFileSpec->name, fileNameCString, sizeof(fileNameCString));
	
	if (pictH == nil) {
			// alert the user that we couldn't load that PICT file
		S3Test_Print("*** Could not load PICT file '%s' ***\r", fileNameCString);
		
		return kS3Error_General;
	}
	
		// copy the new file spec into the TestInfo so that the next time
		// S3TestInfo_Refresh is called, it will use the new file.  we wait
		// until after getting the PicHandle from the file to copy the file
		// spec so that we save it only if we know it's good.
	inTestInfo->textureFSSpec = *inFileSpec;
	
	return S3TestInfo_SetTextureFromPICTHandle(inTestInfo, pictH);
}	


// ===========================================================================
//	S3TestInfo_SetTextureFromPICTHandle
// ===========================================================================
	TS3Error
S3TestInfo_SetTextureFromPICTHandle(
	S3tTestInfoP			inTestInfo,
	PicHandle				inPictH)
{
	ASSERT(inTestInfo);
	ASSERT(inPictH);
	
	if (inTestInfo->leftTexture != nil) {
		QATextureDelete(inTestInfo->leftEngine, inTestInfo->leftTexture);
		inTestInfo->leftTexture = nil;
	}
	
	if (inTestInfo->rightTexture != nil) {
		QATextureDelete(inTestInfo->rightEngine, inTestInfo->rightTexture);
		inTestInfo->rightTexture = nil;
	}
	
	if (inTestInfo->textureImage != nil) {
		delete inTestInfo->textureImage;
		inTestInfo->textureImage = nil;
	}

	if (inTestInfo->leftTextureColorTable != nil) {
		QAColorTableDelete(inTestInfo->leftEngine, inTestInfo->leftTextureColorTable);
		inTestInfo->leftTextureColorTable = nil;
	}
	
	if (inTestInfo->rightTextureColorTable != nil) {
		QAColorTableDelete(inTestInfo->rightEngine, inTestInfo->rightTextureColorTable);
		inTestInfo->rightTextureColorTable = nil;
	}

		// create a TQAImage from this PICT, possibly also making all the black
		// pixels transparent and creating a mipmap
	inTestInfo->textureImage = new CTQAImage(inPictH, inTestInfo->textureFormat, 
			inTestInfo->textureMakeBlackTransparent,
			(inTestInfo->textureFlags & kQATexture_Mipmap));

		// once we've created the CTQAImage, we can dispose of the PICT's handle 
	::DisposeHandle((Handle) inPictH);
	
	if (inTestInfo->textureImage == nil ||
			!inTestInfo->textureImage->IsValidTexture()) {
			// alert the user that we couldn't create an image from that PICT
		S3Test_Print("*** Could not create texture from PICT ***\r");
		
		return kS3Error_General;
	}
	
	if ((inTestInfo->textureFlags & kQATexture_Mipmap) && 
			 !inTestInfo->textureImage->IsValidMipMap()) {
			// alert the user that a mipmap for the PICT could not be created 
			// when it should have been
		S3Test_Print("*** Could not create mipmap texture from PICT ***\r");
		
		return kS3Error_General;
	}
	
		// save the size and rowbytes info for this texture
	inTestInfo->textureWidth = inTestInfo->textureImage->GetWidth();
	inTestInfo->textureHeight = inTestInfo->textureImage->GetHeight();
	inTestInfo->textureRowBytes = inTestInfo->textureImage->GetRowBytes();

		// create the texture for the left engine
	TQAError qaError = QATextureNew(inTestInfo->leftEngine,
			inTestInfo->textureFlags, inTestInfo->textureFormat,
			inTestInfo->textureImage->GetImages(), &inTestInfo->leftTexture);

	if (qaError != kQANoErr) {
			// alert the user that the left engine wouldn't create the texture
		S3Test_Print("*** Could not create new texture from ***\r"
					 "*** PICT in left engine.              ***\r");
		
		gLastRAVEError = qaError;
		
		return kS3Error_RAVE;
	}

		// create the texture for the right engine
	qaError = QATextureNew(inTestInfo->rightEngine, 
			inTestInfo->textureFlags, inTestInfo->textureFormat, 
			inTestInfo->textureImage->GetImages(), &inTestInfo->rightTexture);

	if (qaError != kQANoErr) {
			// alert the user that the right engine wouldn't create the texture
		S3Test_Print("*** Could not create new texture from ***\r"
					 "*** PICT in right engine.             ***\r");
		
		gLastRAVEError = qaError;
		
		return kS3Error_RAVE;
	}

	if (inTestInfo->textureFormat == kQAPixel_CL4 || 
			inTestInfo->textureFormat == kQAPixel_CL8) {
			// this texture is in color table lookup format, so create the color
			// table for it.  first get the ColorTable out of the GWorld holding\
			// the texture.
		CTabHandle colorTableH = inTestInfo->textureImage->GetColorTableHandle();
		
		if (colorTableH == nil) {
				// alert the user that we couldn't get the color table fromt the pict
			S3Test_Print("*** Could not extract color table from PICT ***\r");
			
			return kS3Error_General;
		}
		
			// convert the CTabHandle into TQAColorTable format
		CTQAColorTable* colorTableData = 
				new CTQAColorTable(colorTableH, inTestInfo->textureColorTableType);

			// create a new color table in the left engine
		qaError = QAColorTableNew(inTestInfo->leftEngine,
				inTestInfo->textureColorTableType, colorTableData->GetData(), 
				inTestInfo->textureUseTransparentIndex,
				&inTestInfo->leftTextureColorTable);
				
		if (qaError != kQANoErr) {
				// alert the user that the left engine couldn't create a new color table
			S3Test_Print("*** Could not create new color table in left engine ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

			// bind this color table to the left texture
		qaError = QATextureBindColorTable(inTestInfo->leftEngine,
				inTestInfo->leftTexture, inTestInfo->leftTextureColorTable);
		
		if (qaError != kQANoErr) {
				// alert the user that the left engine couldn't bind the color table
			S3Test_Print("*** Could not bind color table to texture ***\r"
						 "*** in left engine.                       ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

			// create a new color table in the right engine
		qaError = QAColorTableNew(inTestInfo->rightEngine, 
				inTestInfo->textureColorTableType, colorTableData->GetData(),
				inTestInfo->textureUseTransparentIndex,
				&inTestInfo->rightTextureColorTable);
				
		if (qaError != kQANoErr) {
				// alert the user that the right engine couldn't create a new color table
			S3Test_Print("*** Could not create new color table in right engine ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

			// bind this color table to the right texture
		qaError = QATextureBindColorTable(inTestInfo->rightEngine,
				inTestInfo->rightTexture, inTestInfo->rightTextureColorTable);
		
		if (qaError != kQANoErr) {
				// alert the user that the right engine couldn't bind the color table
			S3Test_Print("*** Could not bind color table to texture ***\r"
						 "*** in right engine.                      ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}
	}
			
	if (inTestInfo->detachTexture) {
			// IMPORTANT: we can't detach the textures until after we've created
			// the color tables, if any.  this is because after detaching the
			// textures, we delete the image we used to create them, and we need
			// the image to create the color table.
		qaError = QATextureDetach(inTestInfo->leftEngine,
				inTestInfo->leftTexture);
			
		if (qaError != kQANoErr) {
				// alert the user that the left engine wouldn't detach the texture
			S3Test_Print("*** Could not bind detach texture in left engine. ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

		qaError = QATextureDetach(inTestInfo->rightEngine,
				inTestInfo->rightTexture);
			
		if (qaError != kQANoErr) {
				// alert the user that the right engine wouldn't detach the texture
			S3Test_Print("*** Could not bind detach texture in right engine. ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

			// since we've detached the textures, the engines are now responsible for
			// their storage, so we can delete the CTQAImage which is holding them
		delete inTestInfo->textureImage;
	}
	
	return kS3Error_None;
}


// ===========================================================================
//	S3TestInfo_SetBitmapFromFileSpec
// ===========================================================================
	TS3Error
S3TestInfo_SetBitmapFromFileSpec(
	S3tTestInfoP			inTestInfo,
	const FSSpec*			inFileSpec)
{
	ASSERT(inTestInfo);
	ASSERT(inFileSpec);
	
		// load the PICT file into a handle
	PicHandle pictH = S3GetPICTFromFile(inFileSpec);
	
		// copy and convert the file's name
	char				fileNameCString[64];
	S3PtoCString(inFileSpec->name, fileNameCString, sizeof(fileNameCString));
	
	if (pictH == nil) {
			// alert the user that we couldn't load that PICT file
		S3Test_Print("*** Could not load PICT file '%s' ***\r", fileNameCString);
		
		return kS3Error_General;
	}
	
		// copy the new file spec into the TestInfo so that the next time
		// S3TestInfo_Refresh is called, it will use the new file.  we wait
		// until after getting the PicHandle from the file to copy the file
		// spec so that we save it only if we know it's good.
	inTestInfo->bitmapFSSpec = *inFileSpec;
	
	return S3TestInfo_SetBitmapFromPICTHandle(inTestInfo, pictH);
}	


// ===========================================================================
//	S3TestInfo_SetBitmapFromPICTHandle
// ===========================================================================
	TS3Error
S3TestInfo_SetBitmapFromPICTHandle(
	S3tTestInfoP			inTestInfo,
	PicHandle				inPictH)
{
	ASSERT(inTestInfo);
	ASSERT(inPictH);
	
	if (inTestInfo->leftBitmap != nil) {
		QABitmapDelete(inTestInfo->leftEngine, inTestInfo->leftBitmap);
		inTestInfo->leftBitmap = nil;
	}
	
	if (inTestInfo->rightBitmap != nil) {
		QABitmapDelete(inTestInfo->rightEngine, inTestInfo->rightBitmap);
		inTestInfo->rightBitmap = nil;
	}
	
	if (inTestInfo->bitmapImage != nil) {
		delete inTestInfo->bitmapImage;
		inTestInfo->bitmapImage = nil;
	}

	if (inTestInfo->leftBitmapColorTable != nil) {
		QAColorTableDelete(inTestInfo->leftEngine, inTestInfo->leftBitmapColorTable);
		inTestInfo->leftBitmapColorTable = nil;
	}
	
	if (inTestInfo->rightBitmapColorTable != nil) {
		QAColorTableDelete(inTestInfo->rightEngine, inTestInfo->rightBitmapColorTable);
		inTestInfo->rightBitmapColorTable = nil;
	}

		// create a TQAImage from this PICT, possibly also making all the black
		// pixels transparent and creating a mipmap
	inTestInfo->bitmapImage = new CTQAImage(inPictH, inTestInfo->bitmapFormat, 
			inTestInfo->bitmapMakeBlackTransparent);

		// once we've created the CTQAImage, we can dispose of the PICT's handle 
	::DisposeHandle((Handle) inPictH);
	
	if (inTestInfo->bitmapImage == nil ||
			!inTestInfo->bitmapImage->IsValidBitmap()) {
			// alert the user that we couldn't create an image from that PICT
		S3Test_Print("*** Could not create bitmap from PICT ***\r");
		
		return kS3Error_General;
	}
	
		// save the size and rowbytes info for this bitmap
	inTestInfo->bitmapWidth = inTestInfo->bitmapImage->GetWidth();
	inTestInfo->bitmapHeight = inTestInfo->bitmapImage->GetHeight();
	inTestInfo->bitmapRowBytes = inTestInfo->bitmapImage->GetRowBytes();

		// create the bitmap for the left engine
	TQAError qaError = QABitmapNew(inTestInfo->leftEngine,
			inTestInfo->bitmapFlags, inTestInfo->bitmapFormat,
			inTestInfo->bitmapImage->GetImages(), &inTestInfo->leftBitmap);

	if (qaError != kQANoErr) {
			// alert the user that the left engine wouldn't create the bitmap
		S3Test_Print("*** Could not create new bitmap from ***\r"
					 "*** PICT in left engine.             ***\r");
		
		gLastRAVEError = qaError;
		
		return kS3Error_RAVE;
	}

		// create the bitmap for the right engine
	qaError = QABitmapNew(inTestInfo->rightEngine, 
			inTestInfo->bitmapFlags, inTestInfo->bitmapFormat,
			inTestInfo->bitmapImage->GetImages(), &inTestInfo->rightBitmap);

	if (qaError != kQANoErr) {
			// alert the user that the right engine wouldn't create the bitmap
		S3Test_Print("*** Could not create new bitmap from ***\r"
					 "*** PICT in right engine.            ***\r");
		
		gLastRAVEError = qaError;
		
		return kS3Error_RAVE;
	}

	if (inTestInfo->bitmapFormat == kQAPixel_CL4 || 
			inTestInfo->bitmapFormat == kQAPixel_CL8) {
			// this bitmap is in color table lookup format, so create the color
			// table for it.  first get the ColorTable out of the GWorld holding\
			// the bitmap.
		CTabHandle colorTableH = inTestInfo->bitmapImage->GetColorTableHandle();
		
		if (colorTableH == nil) {
				// alert the user that we couldn't get the color table fromt the pict
			S3Test_Print("*** Could not extract color table from PICT ***\r");
			
			return kS3Error_General;
		}
		
			// convert the CTabHandle into TQAColorTable format
		CTQAColorTable* colorTableData = 
				new CTQAColorTable(colorTableH, inTestInfo->bitmapColorTableType);

			// create a new color table in the left engine
		qaError = QAColorTableNew(inTestInfo->leftEngine,
				inTestInfo->bitmapColorTableType, colorTableData->GetData(),
				inTestInfo->bitmapUseTransparentIndex,
				&inTestInfo->leftBitmapColorTable);
				
		if (qaError != kQANoErr) {
				// alert the user that the left engine couldn't create a new color table
			S3Test_Print("*** Could not create new color table in left engine ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

			// bind this color table to the left bitmap
		qaError = QABitmapBindColorTable(inTestInfo->leftEngine,
				inTestInfo->leftBitmap, inTestInfo->leftBitmapColorTable);
		
		if (qaError != kQANoErr) {
				// alert the user that the left engine couldn't bind the color table
			S3Test_Print("*** Could not bind color table to bitmap ***\r"
						 "*** in left engine.                      ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

			// create a new color table in the right engine
		qaError = QAColorTableNew(inTestInfo->rightEngine,
				inTestInfo->bitmapColorTableType, colorTableData->GetData(),
				inTestInfo->bitmapUseTransparentIndex,
				&inTestInfo->rightBitmapColorTable);
				
		if (qaError != kQANoErr) {
				// alert the user that the right engine couldn't create a new color table
			S3Test_Print("*** Could not create new color table in right engine ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

			// bind this color table to the right bitmap
		qaError = QABitmapBindColorTable(inTestInfo->rightEngine,
				inTestInfo->rightBitmap, inTestInfo->rightBitmapColorTable);
		
		if (qaError != kQANoErr) {
				// alert the user that the right engine couldn't bind the color table
			S3Test_Print("*** Could not bind color table to bitmap ***\r"
						 "*** in right engine.                     ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}
	}
			
	if (inTestInfo->detachBitmap) {
			// IMPORTANT: we can't detach the bitmaps until after we've created
			// the color tables, if any.  this is because after detaching the
			// bitmaps, we delete the image we used to create them, and we need
			// the image to create the color table.
		qaError = QABitmapDetach(inTestInfo->leftEngine, inTestInfo->leftBitmap);
			
		if (qaError != kQANoErr) {
				// alert the user that the left engine wouldn't detach the bitmap
			S3Test_Print("*** Could not bind detach bitmap in left engine. ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

		qaError = QABitmapDetach(inTestInfo->rightEngine, inTestInfo->rightBitmap);
			
		if (qaError != kQANoErr) {
				// alert the user that the right engine wouldn't detach the bitmap
			S3Test_Print("*** Could not bind detach bitmap in right engine. ***\r");
			
			gLastRAVEError = qaError;
			
			return kS3Error_RAVE;
		}

			// since we've detached the bitmaps, the engines are now responsible for
			// their storage, so we can delete the CTQAImage which is holding them
		delete inTestInfo->bitmapImage;
	}
	
	return kS3Error_None;
}


// ===========================================================================
//	S3TestInfo_GetLastRAVEError
// ===========================================================================
	TQAError
S3TestInfo_GetLastRAVEError(
	S3tTestInfoP			inTestInfo)
{
	#pragma unused(inTestInfo)
	
	return gLastRAVEError;
}


// ===========================================================================
//	S3TestInfo_GetLong
// ===========================================================================
	TS3Error 
S3TestInfo_GetLong(
	S3tTestInfoP			inTestInfo,
	TS3TagLong				inTag,
	unsigned long*			outLongP)
{
	ASSERT(inTestInfo);
	ASSERT(outLongP);
	
	switch (inTag) {
		case kS3TagLong_DeviceType:
			*outLongP = inTestInfo->deviceMenu;
			return kS3Error_None;
		
		case kS3TagLong_ClipRegion:
			*outLongP = inTestInfo->clipRegionMenu;
			return kS3Error_None;

		case kS3TagLong_ContextFlags:
			*outLongP = inTestInfo->contextFlags;
			return kS3Error_None;
		
		case kS3TagLong_ZFunction:
			*outLongP = inTestInfo->zFunction;
			return kS3Error_None;
		
		case kS3TagLong_Blending:
			*outLongP = inTestInfo->blending;
			return kS3Error_None;
		
		case kS3TagLong_Antialiasing:
			*outLongP = inTestInfo->antialiasing;
			return kS3Error_None;
		
		case kS3TagLong_TextureFormat:
			*outLongP = inTestInfo->textureFormat;
			return kS3Error_None;
		
		case kS3TagLong_TextureFilter:
			*outLongP = inTestInfo->textureFilter;
			return kS3Error_None;
		
		case kS3TagLong_TextureOp:
			*outLongP = inTestInfo->textureOp;
			return kS3Error_None;
		
		case kS3TagLong_TextureFlags:
			*outLongP = inTestInfo->textureFlags;
			return kS3Error_None;
		
		case kS3TagLong_TextureDetach:
			*outLongP = inTestInfo->detachTexture;
			return kS3Error_None;
		
		case kS3TagLong_TextureMakeBlackTransparent:
			*outLongP = inTestInfo->textureMakeBlackTransparent;
			return kS3Error_None;
		
		case kS3TagLong_TextureUseTransparentIndex:
			*outLongP = inTestInfo->textureUseTransparentIndex;
			return kS3Error_None;
		
		case kS3TagLong_BitmapFormat:
			*outLongP = inTestInfo->bitmapFormat;
			return kS3Error_None;
		
		case kS3TagLong_BitmapFlags:
			*outLongP = inTestInfo->bitmapFlags;
			return kS3Error_None;
		
		case kS3TagLong_BitmapDetach:
			*outLongP = inTestInfo->detachBitmap;
			return kS3Error_None;
		
		case kS3TagLong_BitmapMakeBlackTransparent:
			*outLongP = inTestInfo->bitmapMakeBlackTransparent;
			return kS3Error_None;
		
		case kS3TagLong_BitmapUseTransparentIndex:
			*outLongP = inTestInfo->bitmapUseTransparentIndex;
			return kS3Error_None;
	} // endswitch
	
	return kS3Error_Parameter;
}


// ===========================================================================
//	S3TestInfo_SetLong
// ===========================================================================
	TS3Error 
S3TestInfo_SetLong(
	S3tTestInfoP			inTestInfo,
	TS3TagLong				inTag,
	unsigned long			inLong)
{
	ASSERT(inTestInfo);
	
	switch (inTag) {
		case kS3TagLong_DeviceType:
			inTestInfo->deviceMenu = (S3eDevice) inLong;
			return kS3Error_None;

		case kS3TagLong_ClipRegion:
			inTestInfo->clipRegionMenu = (S3eClipRegion) inLong;
			return kS3Error_None;

		case kS3TagLong_ContextFlags:
			inTestInfo->contextFlags = inLong;
			return kS3Error_None;
		
		case kS3TagLong_ZFunction:
			inTestInfo->zFunction = (long) inLong;
			return kS3Error_None;
		
		case kS3TagLong_Blending:
			inTestInfo->blending = (long) inLong;
			return kS3Error_None;
		
		case kS3TagLong_Antialiasing:
			inTestInfo->antialiasing = (long) inLong;
			return kS3Error_None;
		
		case kS3TagLong_TextureFormat:
			inTestInfo->textureFormat = (TQAImagePixelType) inLong;
			
				// we have to keep the texture format and the color table
				// format in sync
			if (inTestInfo->textureFormat == kQAPixel_CL4) {
				inTestInfo->textureColorTableType = kQAColorTable_CL4_RGB32;
			} else if (inTestInfo->textureFormat == kQAPixel_CL8) {
				inTestInfo->textureColorTableType = kQAColorTable_CL8_RGB32;
			}
			return kS3Error_None;
		
		case kS3TagLong_TextureFilter:
			inTestInfo->textureFilter = inLong;
			return kS3Error_None;
		
		case kS3TagLong_TextureOp:
			inTestInfo->textureOp = inLong;
			return kS3Error_None;
		
		case kS3TagLong_TextureFlags:
			inTestInfo->textureFlags = inLong;
			return kS3Error_None;
		
		case kS3TagLong_TextureDetach:
			inTestInfo->detachTexture = (Boolean) inLong;
			return kS3Error_None;
		
		case kS3TagLong_TextureMakeBlackTransparent:
			inTestInfo->textureMakeBlackTransparent = (Boolean) inLong;
			return kS3Error_None;
		
		case kS3TagLong_TextureUseTransparentIndex:
			inTestInfo->textureUseTransparentIndex = (Boolean) inLong;
			return kS3Error_None;
		
		case kS3TagLong_BitmapFormat:
			inTestInfo->bitmapFormat = (TQAImagePixelType) inLong;
			
				// we have to keep the bitmap format and the color table
				// format in sync
			if (inTestInfo->bitmapFormat == kQAPixel_CL4) {
				inTestInfo->bitmapColorTableType = kQAColorTable_CL4_RGB32;
			} else if (inTestInfo->bitmapFormat == kQAPixel_CL8) {
				inTestInfo->bitmapColorTableType = kQAColorTable_CL8_RGB32;
			}
			return kS3Error_None;
		
		case kS3TagLong_BitmapFlags:
			inTestInfo->bitmapFlags = inLong;
			return kS3Error_None;
		
		case kS3TagLong_BitmapDetach:
			inTestInfo->detachBitmap = (Boolean) inLong;
			return kS3Error_None;
		
		case kS3TagLong_BitmapMakeBlackTransparent:
			inTestInfo->bitmapMakeBlackTransparent = (Boolean) inLong;
			return kS3Error_None;
		
		case kS3TagLong_BitmapUseTransparentIndex:
			inTestInfo->bitmapUseTransparentIndex = (Boolean) inLong;
			return kS3Error_None;
	} // endswitch
	
	return kS3Error_Parameter;
}


// ===========================================================================
//	S3TestInfo_GetFloat
// ===========================================================================
	TS3Error 
S3TestInfo_GetFloat(
	S3tTestInfoP			inTestInfo,
	TS3TagFloat				inTag,
	float*					outFloatP)
{
	ASSERT(inTestInfo);	
	ASSERT(outFloatP);	
	
	switch (inTag) {
		case kS3TagFloat_Alpha:
			*outFloatP = inTestInfo->alpha;
			return kS3Error_None;
		
		case kS3TagFloat_Red:
			*outFloatP = inTestInfo->red;
			return kS3Error_None;
		
		case kS3TagFloat_Green:
			*outFloatP = inTestInfo->green;
			return kS3Error_None;
		
		case kS3TagFloat_Blue:
			*outFloatP = inTestInfo->blue;
			return kS3Error_None;
		
		case kS3TagFloat_LineWidth:
			*outFloatP = inTestInfo->lineWidth;
			return kS3Error_None;
	} // endswitch
	
	return kS3Error_Parameter;
}


// ===========================================================================
//	S3TestInfo_SetFloat
// ===========================================================================
	TS3Error 
S3TestInfo_SetFloat(
	S3tTestInfoP			inTestInfo,
	TS3TagFloat				inTag,
	float					inFloat)
{
	ASSERT(inTestInfo);	
	
	switch (inTag) {
		case kS3TagFloat_Alpha:
			inTestInfo->alpha = inFloat;
			return kS3Error_None;
		
		case kS3TagFloat_Red:
			inTestInfo->red = inFloat;
			return kS3Error_None;
		
		case kS3TagFloat_Green:
			inTestInfo->green = inFloat;
			return kS3Error_None;
		
		case kS3TagFloat_Blue:
			inTestInfo->blue = inFloat;
			return kS3Error_None;
		
		case kS3TagFloat_LineWidth:
			inTestInfo->lineWidth = inFloat;
			return kS3Error_None;
	} // endswitch
	
	return kS3Error_Parameter;
}


// ===========================================================================
//	S3TestInfo_GetBlock
// ===========================================================================
	TS3Error 
S3TestInfo_GetBlock(
	S3tTestInfoP			inTestInfo,
	TS3TagBlock				inTag,
	void*					outBlockP,
	TS3TagBlock_Size		inBlockSize)
{
	ASSERT(inTestInfo);	
	ASSERT(outBlockP);	
	
	switch (inTag) {
		case kS3TagBlock_TestName:
			::BlockMoveData(&inTestInfo->testName, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_ArgString:
			::BlockMoveData(&inTestInfo->argString, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_TestResult:
			::BlockMoveData(&gTestResultBuffer, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_LeftEngineName:
			::BlockMoveData(&inTestInfo->leftEngineName, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_RightEngineName:
			::BlockMoveData(&inTestInfo->rightEngineName, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_LeftRect:
			::BlockMoveData(&inTestInfo->leftRect, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_RightRect:
			::BlockMoveData(&inTestInfo->rightRect, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_LeftGlobalRect:
			::BlockMoveData(&inTestInfo->leftGlobalRect, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_RightGlobalRect:
			::BlockMoveData(&inTestInfo->rightGlobalRect, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_TextureFileSpec:
			::BlockMoveData(&inTestInfo->textureFSSpec, outBlockP, inBlockSize);
			return kS3Error_None;
		
		case kS3TagBlock_BitmapFileSpec:
			::BlockMoveData(&inTestInfo->bitmapFSSpec, outBlockP, inBlockSize);
			return kS3Error_None;
	} // endswitch
	
	return kS3Error_Parameter;
}


// ===========================================================================
//	S3TestInfo_SetBlock
// ===========================================================================
	TS3Error 
S3TestInfo_SetBlock(
	S3tTestInfoP			inTestInfo,
	TS3TagBlock				inTag,
	const void*				inBlockP)
{
	ASSERT(inTestInfo);	
	ASSERT(inBlockP);	
	
	switch (inTag) {
		case kS3TagBlock_TestName:
			::BlockMoveData(inBlockP, &inTestInfo->testName, kS3TagBlock_Size_TestName);
			return kS3Error_None;
		
		case kS3TagBlock_ArgString:
			::BlockMoveData(inBlockP, &inTestInfo->argString, kS3TagBlock_Size_ArgString);
			return kS3Error_None;
		
		case kS3TagBlock_TestResult:
			::BlockMoveData(inBlockP, &gTestResultBuffer, kS3TagBlock_Size_TestResult);
			return kS3Error_None;
		
		case kS3TagBlock_LeftEngineName:
			::BlockMoveData(inBlockP, &inTestInfo->leftEngineName, kS3TagBlock_Size_LeftEngineName);
			return kS3Error_None;
		
		case kS3TagBlock_RightEngineName:
			::BlockMoveData(inBlockP, &inTestInfo->rightEngineName, kS3TagBlock_Size_RightEngineName);
			return kS3Error_None;
		
		case kS3TagBlock_LeftRect:
			::BlockMoveData(inBlockP, &inTestInfo->leftRect, kS3TagBlock_Size_LeftRect);
			return kS3Error_None;
		
		case kS3TagBlock_RightRect:
			::BlockMoveData(inBlockP, &inTestInfo->rightRect, kS3TagBlock_Size_RightRect);
			return kS3Error_None;
		
		case kS3TagBlock_LeftGlobalRect:
			::BlockMoveData(inBlockP, &inTestInfo->leftGlobalRect, kS3TagBlock_Size_LeftGlobalRect);
			return kS3Error_None;
		
		case kS3TagBlock_RightGlobalRect:
			::BlockMoveData(inBlockP, &inTestInfo->rightGlobalRect, kS3TagBlock_Size_RightGlobalRect);
			return kS3Error_None;
		
		case kS3TagBlock_TextureFileSpec:
			::BlockMoveData(inBlockP, &inTestInfo->textureFSSpec, kS3TagBlock_Size_TextureFileSpec);
			return kS3Error_None;
		
		case kS3TagBlock_BitmapFileSpec:
			::BlockMoveData(inBlockP, &inTestInfo->bitmapFSSpec, kS3TagBlock_Size_BitmapFileSpec);
			return kS3Error_None;
	} // endswitch
	
	return kS3Error_Parameter;
}


//// ===========================================================================
////	S3TestInfo_GetPointer
//// ===========================================================================
//	TS3Error 
//S3TestInfo_GetPointer(
//	S3tTestInfoP			inTestInfo,
//	TS3TagPointer			inTag,
//	void*					inPointerP);
//{
//	
//}
//
//
//// ===========================================================================
////	S3TestInfo_GetPointer
//// ===========================================================================
//	TS3Error 
//S3TestInfo_SetPointer(
//	S3tTestInfoP			inTestInfo,
//	TS3TagPointer			inTag,
//	const void*				inPointer);
//{
//	
//}


// ===========================================================================
//	S3TestInfo_GetLMicroseconds
// ===========================================================================
unsigned long S3TestInfo_GetLMicroseconds(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->leftMicroseconds;
}

// ===========================================================================
//	S3TestInfo_GetRMicroseconds
// ===========================================================================
unsigned long S3TestInfo_GetRMicroseconds(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->rightMicroseconds;
}

// ===========================================================================
//	S3TestInfo_GetTestName
// ===========================================================================
StringPtr S3TestInfo_GetTestName(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->testName;
}

// ===========================================================================
//	S3TestInfo_GetArgString
// ===========================================================================
char* S3TestInfo_GetArgString(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->argString;
}

// ===========================================================================
//	S3TestInfo_GetPICTList
// ===========================================================================
S3PICTList S3TestInfo_GetPICTList(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->pictList;
}

// ===========================================================================
//	S3TestInfo_GetLDrawContext
// ===========================================================================
TQADrawContext *S3TestInfo_GetLDrawContext(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->leftContext;
}

// ===========================================================================
//	S3TestInfo_GetRDrawContext
// ===========================================================================
TQADrawContext *S3TestInfo_GetRDrawContext(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->rightContext;
}

//// ===========================================================================
////	S3TestInfo_SetLDrawContext
//// ===========================================================================
//void S3TestInfo_SetLDrawContext(
//	S3tTestInfo			*testInfo,
//	TQADrawContext		*newContext)
//{
//	ASSERT(testInfo);
//	ASSERT(newContext);
//	testInfo->leftContext = newContext;
//}
//
//// ===========================================================================
////	S3TestInfo_SetRDrawContext
//// ===========================================================================
//void S3TestInfo_SetRDrawContext(
//	S3tTestInfo			*testInfo,
//	TQADrawContext		*newContext)
//{
//	ASSERT(testInfo);
//	ASSERT(newContext);
//	testInfo->rightContext = newContext;
//}

// ===========================================================================
//	S3TestInfo_GetContextFlags
// ===========================================================================
unsigned long S3TestInfo_GetContextFlags(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->contextFlags;
}

// ===========================================================================
//	S3TestInfo_GetLEngine
// ===========================================================================
TQAEngine *S3TestInfo_GetLEngine(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->leftEngine;
}

// ===========================================================================
//	S3TestInfo_GetREngine
// ===========================================================================
TQAEngine *S3TestInfo_GetREngine(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->rightEngine;
}

// ===========================================================================
//	S3TestInfo_GetWidth
// ===========================================================================
long S3TestInfo_GetWidth(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return (testInfo->leftRect.right - testInfo->leftRect.left);
}

//// ===========================================================================
////	S3TestInfo_SetWidth
//// ===========================================================================
//void S3TestInfo_SetWidth(
//	S3tTestInfo			*testInfo,
//	unsigned long		width)
//{
//	ASSERT(testInfo);
//	testInfo->leftRect.right = testInfo->leftRect.left + width;
//	testInfo->rightRect.right = testInfo->rightRect.left + width;
//}

// ===========================================================================
//	S3TestInfo_GetHeight
// ===========================================================================
long S3TestInfo_GetHeight(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return (testInfo->leftRect.bottom - testInfo->leftRect.top);
}

//// ===========================================================================
////	S3TestInfo_SetHeight
//// ===========================================================================
//void S3TestInfo_SetHeight(
//	S3tTestInfo			*testInfo,
//	unsigned long		height)
//{
//	ASSERT(testInfo);
//	testInfo->leftRect.bottom = testInfo->leftRect.top + height;
//	testInfo->rightRect.bottom = testInfo->rightRect.top + height;
//}

// ===========================================================================
//	S3TestInfo_GetLDevice
// ===========================================================================
TQADevice *S3TestInfo_GetLDevice(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return &testInfo->leftDevice;
}

// ===========================================================================
//	S3TestInfo_GetRDevice
// ===========================================================================
TQADevice *S3TestInfo_GetRDevice(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return &testInfo->rightDevice;
}

// ===========================================================================
//	S3TestInfo_GetTextureImage
// ===========================================================================
S3Image S3TestInfo_GetTextureImage(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return (S3Image) testInfo->textureImage;
}

// ===========================================================================
//	S3TestInfo_GetLTexture
// ===========================================================================
TQATexture *S3TestInfo_GetLTexture(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return testInfo->leftTexture;
}

// ===========================================================================
//	S3TestInfo_GetRTexture
// ===========================================================================
TQATexture *S3TestInfo_GetRTexture(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->rightTexture;
}

// ===========================================================================
//	S3TestInfo_GetTextureWidth
// ===========================================================================
long S3TestInfo_GetTextureWidth(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return testInfo->textureWidth;
}

// ===========================================================================
//	S3TestInfo_GetTextureHeight
// ===========================================================================
long S3TestInfo_GetTextureHeight(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return testInfo->textureHeight;
}

// ===========================================================================
//	S3TestInfo_GetTextureRowBytes
// ===========================================================================
long S3TestInfo_GetTextureRowBytes(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return testInfo->textureRowBytes;
}

// ===========================================================================
//	S3TestInfo_GetTextureFlags
// ===========================================================================
unsigned long S3TestInfo_GetTextureFlags(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->textureFlags;
}

// ===========================================================================
//	S3TestInfo_GetTextureFormat
// ===========================================================================
TQAImagePixelType S3TestInfo_GetTextureFormat(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->textureFormat;
}

// ===========================================================================
//	S3TestInfo_GetDetachTexture
// ===========================================================================
Boolean S3TestInfo_GetDetachTexture(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->detachTexture;
}

// ===========================================================================
//	S3TestInfo_GetTextureMakeBlackTransparent
// ===========================================================================
Boolean S3TestInfo_GetTextureMakeBlackTransparent(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->textureMakeBlackTransparent;
}

// ===========================================================================
//	S3TestInfo_GetLTextureColorTable
// ===========================================================================
TQAColorTable* S3TestInfo_GetLTextureColorTable(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->leftTextureColorTable;
}

// ===========================================================================
//	S3TestInfo_GetRTextureColorTable
// ===========================================================================
TQAColorTable* S3TestInfo_GetRTextureColorTable(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->rightTextureColorTable;
}

// ===========================================================================
//	S3TestInfo_GetTextureColorTableType
// ===========================================================================
TQAColorTableType S3TestInfo_GetTextureColorTableType(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->textureColorTableType;
}

// ===========================================================================
//	S3TestInfo_GetTextureUseTransparentIndex
// ===========================================================================
Boolean S3TestInfo_GetTextureUseTransparentIndex(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->textureUseTransparentIndex;
}

// ===========================================================================
//	S3TestInfo_GetBitmapImage
// ===========================================================================
S3Image S3TestInfo_GetBitmapImage(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return (S3Image) testInfo->bitmapImage;
}

// ===========================================================================
//	S3TestInfo_GetLBitmap
// ===========================================================================
TQABitmap *S3TestInfo_GetLBitmap(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return testInfo->leftBitmap;
}

// ===========================================================================
//	S3TestInfo_GetRBitmap
// ===========================================================================
TQABitmap *S3TestInfo_GetRBitmap(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->rightBitmap;
}

// ===========================================================================
//	S3TestInfo_GetBitmapWidth
// ===========================================================================
long S3TestInfo_GetBitmapWidth(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return testInfo->bitmapWidth;
}

// ===========================================================================
//	S3TestInfo_GetBitmapHeight
// ===========================================================================
long S3TestInfo_GetBitmapHeight(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return testInfo->bitmapHeight;
}

// ===========================================================================
//	S3TestInfo_GetBitmapRowBytes
// ===========================================================================
long S3TestInfo_GetBitmapRowBytes(
	S3tTestInfo		*testInfo)
{
	ASSERT(testInfo);
	return testInfo->bitmapRowBytes;
}

// ===========================================================================
//	S3TestInfo_GetBitmapFlags
// ===========================================================================
unsigned long S3TestInfo_GetBitmapFlags(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->bitmapFlags;
}

// ===========================================================================
//	S3TestInfo_GetBitmapFormat
// ===========================================================================
TQAImagePixelType S3TestInfo_GetBitmapFormat(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->bitmapFormat;
}

// ===========================================================================
//	S3TestInfo_GetDetachBitmap
// ===========================================================================
Boolean S3TestInfo_GetDetachBitmap(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->detachBitmap;
}

// ===========================================================================
//	S3TestInfo_GetBitmapMakeBlackTransparent
// ===========================================================================
Boolean S3TestInfo_GetBitmapMakeBlackTransparent(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->bitmapMakeBlackTransparent;
}

// ===========================================================================
//	S3TestInfo_GetLBitmapColorTable
// ===========================================================================
TQAColorTable* S3TestInfo_GetLBitmapColorTable(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->leftBitmapColorTable;
}

// ===========================================================================
//	S3TestInfo_GetRBitmapColorTable
// ===========================================================================
TQAColorTable* S3TestInfo_GetRBitmapColorTable(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->rightBitmapColorTable;
}

// ===========================================================================
//	S3TestInfo_GetBitmapColorTableType
// ===========================================================================
TQAColorTableType S3TestInfo_GetBitmapColorTableType(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->bitmapColorTableType;
}

// ===========================================================================
//	S3TestInfo_GetBitmapUseTransparentIndex
// ===========================================================================
Boolean S3TestInfo_GetBitmapUseTransparentIndex(
	S3tTestInfo			*testInfo)
{
	ASSERT(testInfo);
	return testInfo->bitmapUseTransparentIndex;
}


} // extern "C"
