/*****************************************************************************\

	S3TestSystem.c

	Interfaces for a test tool to the Sneak3View test system registration,
	and to the RAVE.h wrapper routines which route QA calls to the 
	mirrored engines.

	Copyright (C) 1995 Apple Computer, Inc.  All rights reserved.

\*****************************************************************************/

/*=========================================================================*\
	Includes
\*=========================================================================*/
#include <string.h>
#include <stdarg.h>
#include <stdio.h>

#include "RAVE.h"
#include "S3TestInfo_sys.h"
#include "S3TestSystem.h"
#include "S3TestUtil.h"
#include "S3TestC++Interface.h"
#include "S3Utilities.h"

/*=========================================================================*\
	Constants
\*=========================================================================*/
		
/*=========================================================================*\
	Types
\*=========================================================================*/

/*=========================================================================*\
	Macros
\*=========================================================================*/
#define ASSERT_TestInfoDC(ti)	\
	ASSERT(	\
		(ti)->leftContext != NULL &&	\
		(ti)->rightContext != NULL);

/*=========================================================================*\
	Test System Routines
	
	Perhaps add a "test flag" to indicate how a test behaves? 
	(NeedCleanDrawContext, etc.)
\*=========================================================================*/

/*=========================================================================*\
	S3Test_Register
\*=========================================================================*/
void S3Test_Register(
	const char				*testName,
	S3tTestPrereqs			testPrereqs,
	S3tTestMethod			testMethod)
{
	CSettingsDoc_RegisterTest(testName, testPrereqs, testMethod);
}

/*=========================================================================*\
	S3Test_Print
\*=========================================================================*/
void S3Test_Print(
	const char				*formatStr,
	...)
{
	char		aBigAssedBuffer[1024];
	va_list		va;
	
	va_start(va, formatStr);
	vsprintf(aBigAssedBuffer, formatStr, va);
	va_end(va);
	
	CSettingsDoc_S3Print(aBigAssedBuffer, strlen(aBigAssedBuffer));
}

/*=========================================================================*\
	RAVE.h Wrapper routines 
	(use these to perform an action to both windows.)
\*=========================================================================*/

/*=========================================================================*\
	S3Test_SetFloat
\*=========================================================================*/
void S3Test_SetFloat (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagFloat				tag,				/* Tag of variable to set */
	float					newValue)			/* New value for variable */
{
	unsigned long microseconds = 0;
	
	ASSERT_TestInfoDC(testInfo);
	
	microseconds = S3MicrosecondCount();
	QASetFloat(testInfo->leftContext, tag, newValue);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);
	
	microseconds = S3MicrosecondCount();
	QASetFloat(testInfo->rightContext, tag, newValue);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
	/* Add verification? QAGetFloat to see if it sticked? */
}

/*=========================================================================*\
	S3Test_SetInt
\*=========================================================================*/
void S3Test_SetInt (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagInt				tag,				/* Tag of variable to set */
	unsigned long			newValue)			/* New value for variable */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QASetInt(testInfo->leftContext, tag, newValue);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QASetInt(testInfo->rightContext, tag, newValue);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
	/* Add verification? QAGetInt to see if it sticked? */
}

/*=========================================================================*\
	S3Test_SetPtr
\*=========================================================================*/
void S3Test_SetPtr (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagPtr				tag,				/* Tag of variable to set */
	const void				*newValue)			/* New value for variable */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QASetPtr(testInfo->leftContext, tag, newValue);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QASetPtr(testInfo->rightContext, tag, newValue);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
	/* Add verification? QAGetPtr to see if it sticked? */
}

/*=========================================================================*\
	S3Test_GetLFloat
\*=========================================================================*/
float S3Test_GetLFloat (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagFloat				tag)				/* Tag of variable to get */
{
	unsigned long microseconds;
	float value;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	value = QAGetFloat(testInfo->leftContext, tag);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	return value;
}

/*=========================================================================*\
	S3Test_GetLInt
\*=========================================================================*/
unsigned long S3Test_GetLInt (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagInt				tag)				/* Tag of variable to get */
{
	unsigned long microseconds;
	int value;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	value = QAGetInt(testInfo->leftContext, tag);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	return value;
}

/*=========================================================================*\
	S3Test_GetLPtr
\*=========================================================================*/
void *S3Test_GetLPtr(
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagPtr				tag)				/* Tag of variable to get */
{
	unsigned long microseconds;
	void* value;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	value = QAGetPtr(testInfo->leftContext, tag);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	return value;
}

/*=========================================================================*\
	S3Test_GetRFloat
\*=========================================================================*/
float S3Test_GetRFloat (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagFloat				tag)				/* Tag of variable to get */
{
	unsigned long microseconds;
	float value;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	value = QAGetFloat(testInfo->rightContext, tag);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);

	return value;
}

/*=========================================================================*\
	S3Test_GetRInt
\*=========================================================================*/
unsigned long S3Test_GetRInt (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagInt				tag)				/* Tag of variable to get */
{
	unsigned long microseconds;
	int value;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	value = QAGetInt(testInfo->rightContext, tag);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);

	return value;
}

/*=========================================================================*\
	S3Test_GetRPtr
\*=========================================================================*/
void *S3Test_GetRPtr(
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagPtr				tag)				/* Tag of variable to get */
{
	unsigned long microseconds;
	void* value;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	value = QAGetPtr(testInfo->rightContext, tag);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);

	return value;
}

/*=========================================================================*\
	S3Test_DrawPoint
\*=========================================================================*/
void S3Test_DrawPoint (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVGouraud		*v)					/* Vertex */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QADrawPoint(testInfo->leftContext, v);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QADrawPoint(testInfo->rightContext, v);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_DrawLine
\*=========================================================================*/
void S3Test_DrawLine (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVGouraud 		*v0,				/* Vertex 0 */
	const TQAVGouraud 		*v1)				/* Vertex 1 */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QADrawLine(testInfo->leftContext, v0, v1);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QADrawLine(testInfo->rightContext, v0, v1);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_DrawTriGouraud
\*=========================================================================*/
void S3Test_DrawTriGouraud (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVGouraud 		*v0,				/* Vertex 0 */
	const TQAVGouraud 		*v1,				/* Vertex 1 */
	const TQAVGouraud 		*v2,				/* Vertex 2 */
	unsigned long			flags)				/* Mask of kQATriFlags_xxx flags */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QADrawTriGouraud(testInfo->leftContext, v0, v1, v2, flags);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QADrawTriGouraud(testInfo->rightContext, v0, v1, v2, flags);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_DrawTriTexture
\*=========================================================================*/
void S3Test_DrawTriTexture (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVTexture 		*v0,				/* Vertex 0 */
	const TQAVTexture 		*v1,				/* Vertex 1 */
	const TQAVTexture 		*v2,				/* Vertex 2 */
	unsigned long			flags)				/* Mask of kQATriFlags_xxx flags */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QADrawTriTexture(testInfo->leftContext, v0, v1, v2, flags);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QADrawTriTexture(testInfo->rightContext, v0, v1, v2, flags);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_DrawVGouraud
\*=========================================================================*/
void S3Test_DrawVGouraud (
	S3tTestInfo				*testInfo,			/* Test environment */
	unsigned long			nVertices,			/* Number of vertices */
	TQAVertexMode			vertexMode,			/* One of kQAVertexMode_xxx enumerated values */
	const TQAVGouraud 		vertices[],			/* Array of vertices */
	const unsigned long		flags[])			/* Array of per-triangle flags (or NULL) */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QADrawVGouraud(
		testInfo->leftContext, nVertices, vertexMode, vertices, flags);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QADrawVGouraud(
		testInfo->rightContext, nVertices, vertexMode, vertices, flags);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_DrawVTexture
\*=========================================================================*/
void S3Test_DrawVTexture (
	S3tTestInfo				*testInfo,			/* Test environment */
	unsigned long			nVertices,			/* Number of vertices */
	TQAVertexMode			vertexMode,			/* One of kQAVertexMode_xxx enumerated values */
	const TQAVTexture 		vertices[],			/* Array of vertices */
	const unsigned long		flags[])			/* Array of per-triangle flags (or NULL) */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QADrawVTexture(
		testInfo->leftContext, nVertices, vertexMode, vertices, flags);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QADrawVTexture(
		testInfo->rightContext, nVertices, vertexMode, vertices, flags);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_DrawBitmap
\*=========================================================================*/
void S3Test_DrawBitmap (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVGouraud 		*v,					/* xyz, and (if a 1 bit/pixel bitmap) argb */
	TQABitmap				*leftBitmap,		/* Previously allocated by QABitmapNew() or nil to use bitmap stored in testInfo */
	TQABitmap				*rightBitmap)		/* Previously allocated by QABitmapNew() or nil to use bitmap stored in testInfo */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);
	
	if (leftBitmap == nil) {
		ASSERT(testInfo->leftBitmap);
		leftBitmap = testInfo->leftBitmap;
	}
	
	if (rightBitmap == nil) {
		ASSERT(testInfo->rightBitmap);
		rightBitmap = testInfo->rightBitmap;
	}
	
	microseconds = S3MicrosecondCount();
	QADrawBitmap(testInfo->leftContext, v, leftBitmap);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QADrawBitmap(testInfo->rightContext, v, rightBitmap);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_SubmitVerticesGouraud
\*=========================================================================*/
void S3Test_SubmitVerticesGouraud (
	S3tTestInfo				*testInfo,			/* Test environment */
	unsigned long			vertexCount,		/* Number of vertices being submitted */
	const TQAVGouraud 		*vertices)			/* Array of Gouraud vertices */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);
	ASSERT(vertices);
	
	microseconds = S3MicrosecondCount();
	QASubmitVerticesGouraud(testInfo->leftContext, vertexCount, vertices);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);
	
	microseconds = S3MicrosecondCount();
	QASubmitVerticesGouraud(testInfo->rightContext, vertexCount, vertices);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_DrawTriMeshGouraud
\*=========================================================================*/
void S3Test_DrawTriMeshGouraud (
	S3tTestInfo					*testInfo,			/* Test environment */
	unsigned long				triangleCount,		/* Number of triangles being drawn */
	const TQAIndexedTriangle 	*triangles)			/* Array of vertex indexes */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);
	ASSERT(triangles);
	
	microseconds = S3MicrosecondCount();
	QADrawTriMeshGouraud(testInfo->leftContext, triangleCount, triangles);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);
	
	microseconds = S3MicrosecondCount();
	QADrawTriMeshGouraud(testInfo->rightContext, triangleCount, triangles);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_SubmitVerticesTexture
\*=========================================================================*/
void S3Test_SubmitVerticesTexture (
	S3tTestInfo				*testInfo,			/* Test environment */
	unsigned long			vertexCount,		/* Number of vertices being submitted */
	const TQAVTexture 		*vertices)			/* Array of Texture vertices */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);
	ASSERT(vertices);
	
	microseconds = S3MicrosecondCount();
	QASubmitVerticesTexture(testInfo->leftContext, vertexCount, vertices);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);
	
	microseconds = S3MicrosecondCount();
	QASubmitVerticesTexture(testInfo->rightContext, vertexCount, vertices);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_DrawTriMeshTexture
\*=========================================================================*/
void S3Test_DrawTriMeshTexture (
	S3tTestInfo					*testInfo,			/* Test environment */
	unsigned long				triangleCount,		/* Number of triangles being drawn */
	const TQAIndexedTriangle 	*triangles)			/* Array of vertex indexes */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);
	ASSERT(triangles);
	
	microseconds = S3MicrosecondCount();
	QADrawTriMeshTexture(testInfo->leftContext, triangleCount, triangles);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);
	
	microseconds = S3MicrosecondCount();
	QADrawTriMeshTexture(testInfo->rightContext, triangleCount, triangles);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_RenderStart
\*=========================================================================*/
void S3Test_RenderStart (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQARect			*dirtyRect)		/* Initial background image (or NULL) */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QARenderStart(testInfo->leftContext, dirtyRect, NULL);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QARenderStart(testInfo->rightContext, dirtyRect, NULL);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_RenderEnd
\*=========================================================================*/
TQAError S3Test_RenderEnd (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQARect			*modifiedRect)		/* Minimum area to swap; NULL means whole buffer */
{
	TQAError	lerr, rerr;
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	lerr = QARenderEnd(testInfo->leftContext, modifiedRect);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	rerr = QARenderEnd(testInfo->rightContext, modifiedRect);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
	
	/* 	
		I'm not sure if this is the way to handle this....
		for example, perhaps we should only return errors for the engine
		they care about (always left, or something)?
	*/
		
	if (lerr == kQANoErr)
		return rerr;
	if ((rerr != kQANoErr) && (rerr != lerr)) 
	{
		S3Test_Print(
			"S3Test_RenderEnd: 2 errors\n\tL: %s\n\tR: %s\n",
			S3TestUtil_TQAErrorToString(lerr),
			S3TestUtil_TQAErrorToString(rerr));
	}
	return lerr;
}

/*=========================================================================*\
	S3Test_RenderAbort
\*=========================================================================*/
void S3Test_RenderAbort (
	S3tTestInfo				*testInfo)			/* Test environment */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QARenderAbort(testInfo->leftContext);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QARenderAbort(testInfo->rightContext);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_Flush
\*=========================================================================*/
void S3Test_Flush (
	S3tTestInfo				*testInfo)			/* Test environment */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QAFlush(testInfo->leftContext);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QAFlush(testInfo->rightContext);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}

/*=========================================================================*\
	S3Test_Sync
\*=========================================================================*/
void S3Test_Sync (
	S3tTestInfo				*testInfo)			/* Test environment */
{
	unsigned long microseconds;

	ASSERT_TestInfoDC(testInfo);

	microseconds = S3MicrosecondCount();
	QASync(testInfo->leftContext);
	testInfo->leftMicroseconds += S3ElapsedMicroseconds(microseconds);

	microseconds = S3MicrosecondCount();
	QASync(testInfo->rightContext);
	testInfo->rightMicroseconds += S3ElapsedMicroseconds(microseconds);
}
