// ===========================================================================
// S3Utilities.c
// ===========================================================================


#include <string.h>
#include <limits.h>
#include <Timer.h>
#include <QuickDraw.h>
#include <Files.h>
#include <Processes.h>
#include <Types.h>

#include "S3Utilities.h"
#include "S3TestSystem.h"


// ===========================================================================
// Finds a TinselTown drawing engine given its name.  Will not find engines
// whose ASCII names are longer than 255 characters.
//
// inEngineName: C-string containing the desired engine's name
//
// RETURNS: pointer to the engine with inEngineName as its name, or nil if
//		the engine wasn't found.
//

	TQAEngine* 
S3GetEngineByName(
	char* inEngineName)
{
	TQAEngine* engine;
	char currentEngineName[kMaxEngineNameLength];
	long currentEngineNameLength = 0;
	long targetEngineNameLength = strlen(inEngineName);
	
		// loop through every registered engine by passing nil for the device
	for (engine = QADeviceGetFirstEngine(nil); engine; 
			 engine = QADeviceGetNextEngine(nil, engine)) {
		if (QAEngineGestalt(engine, kQAGestalt_ASCIINameLength, 
												&currentEngineNameLength) != kQANoErr) {
			continue;
		}
		
		if (currentEngineNameLength != targetEngineNameLength ||
				currentEngineNameLength > kMaxEngineNameLength - 1) {
				// the engine name can't be the same as inEngineName if it's not
				// the same length, so look at the next one.  or it's longer than
				// the longest string we can handle, so just ignore it.
			continue;
		}
		
		if (QAEngineGestalt(engine, kQAGestalt_ASCIIName, 
												&currentEngineName) != kQANoErr) {
				// we got an error, so just go on to the next engine
			continue;
		}
		
		if (strcmp(inEngineName, currentEngineName) == 0) {
				// we found an engine with the right name, so return it
			return engine;
		}
	} // endfor
	
		// we couldn't find an engine with the right name
	return nil;
}


// ===========================================================================
// Gets the name of an engine, given a pointer to the engine.  Assumes that
// inEngine is a pointer to a valid engine.
//
// inEngine: pointer to the engine; must be a valid pointer
// outEngineName: pointer to C-string into which the engine's name will be copied
// inEngineNameLength: the maximum length of the outEngineName string, 
//		including the null byte.
//
// RETURNS: true if the engine's name was found and outEngineName was long
//		enough to hold it, false otherwise 
//

	Boolean
S3GetEngineName(
	const TQAEngine* inEngine, 
	char* outEngineName, 
	long inEngineNameLength)
{
	long nameLength = 0;
	
	ASSERT(inEngine);
	
	if (inEngineNameLength < 2) {
			// the engine name length has to be at least 2: 1 char and the 
			// null byte.  since it's not, we can't do anything useful.
		return false;
	}
	
		// get the engine name's actual length
	QAEngineGestalt(inEngine, kQAGestalt_ASCIINameLength, &nameLength);
	
	if (nameLength > inEngineNameLength - 1) {
			// the actual length of the engine's name is longer than the
			// string (minus the null byte) that the caller gave us.
		return false;
	}
	
		// the string buffer is big enough, so fetch the name into it
	QAEngineGestalt(inEngine, kQAGestalt_ASCIIName, outEngineName);
	
		// everything worked okay
	return true;
}


// ===========================================================================
// Checks whether an engine will support a particular device.  
//
// inEngine: the engine to be tested.  If this is nil, the function always 
//		returns false.
// inDevice: the TQADevice to be tested.  If nil, the function always returns
//		true.
//
// RETURNS: true if the engine can handle the device, false otherwise
//

	Boolean
S3EngineSupportsDevice(const TQAEngine* inEngine, const TQADevice* inDevice)
{
	TQAEngine* engine;
	
		// loop through every registered engine that can handle this device
	for (engine = QADeviceGetFirstEngine(inDevice); 
			 engine; 
			 engine = QADeviceGetNextEngine(inDevice, engine)) {
		if (engine == inEngine) {
				// since the engine was returned by QADeviceGetNextEngine, we know
				// it supports the device in inDevice, so return true
			return true;
		}
	}
	
		// inEngine was never returned, so we know it doesn't support inDevice
	return false;
}


// ===========================================================================
// Finds the device which encloses the upper-left corner of a Rect, and
// optionally returns whether the Rect is completely enclosed by the device.
//
// inRect: the rectangle bounds to be tested
// outIsOnOneDevice: on return, is true if the device's bounds completely
// 		enclose inRect, false otherwise.  if outIsOnOneDevice is nil, 
// 		it's ignored 
//
// RETURNS: the handle to the device, nil if no device is found, which means
//		the Rect's top-left corner is hanging off the screen.
//

	GDHandle
S3GetDeviceFromRect(const Rect* inRect, Boolean* outIsOnOneDevice)
{
	GDHandle theDevice = nil;

	if (outIsOnOneDevice != nil) {
			// we set this flag to false now because if it turns out to be true,
			// it will get set inside the loop, just before we return.
		*outIsOnOneDevice = false;
	}
	
		// loop through every available GDevice, looking for the one that
		// encloses inRect
	for (theDevice = GetDeviceList(); theDevice;
			 theDevice = GetNextDevice(theDevice)) {
		if ((**theDevice).gdFlags & screenDevice) {
				// this is a screen, not an off-screen gWorld, so check if its rect
				// encloses inRect
			if ((**theDevice).gdRect.left 	<= inRect->left		&&
					(**theDevice).gdRect.right 	>= inRect->left	 	&&
					(**theDevice).gdRect.top 		<= inRect->top 		&&
					(**theDevice).gdRect.bottom >= inRect->top) {
					// we've found the device on which this rect's top, left corner
					// is located.  now check if the entire rect fits on this device
				if (outIsOnOneDevice != nil) {
					*outIsOnOneDevice =
						(Boolean) ((**theDevice).gdRect.right >= inRect->right &&
											 (**theDevice).gdRect.bottom >= inRect->bottom);
				}
				
				return theDevice;
			}
		}
	}

		// if we got here, we couldn't find a device which encloses the rect's
		// upper left corner
	return nil;
}


// ===========================================================================
// Finds the first screen with the largest greatest depth.
//
// RETURNS: handle to the screen device
//

	GDHandle
S3GetDeepestScreen(void)
{
	GDHandle colorGDevice = nil, deepestGDevice = nil;
	long deviceDepth = 0, maxDepth = 0;

	for (colorGDevice = GetDeviceList(); colorGDevice;
			 colorGDevice = GetNextDevice(colorGDevice)) {
		if ((**colorGDevice).gdFlags & screenDevice) {
				// this is a screen, not an off-screen gWorld, so check its depth
			deviceDepth = (*(**colorGDevice).gdPMap)->pixelSize;
			
			if (deviceDepth > maxDepth) {
				maxDepth = deviceDepth;
				deepestGDevice = colorGDevice;
			}
		}
	}

	return deepestGDevice;
}


// ===========================================================================
// Converts a global rect into one local to a particular device.
//
// inDevice: the device for which the lcoal rect is calculated
// ioRect: on entry, pointer to the global rect; on exit, the same rect
//		contains the local rect
//
// RETURNS: void
//

	void
S3GlobalToDeviceLocal(
	const GDHandle inDevice,
	Rect* ioRect)
{
	Rect deviceRect;
	
	ASSERT(inDevice);
	ASSERT(ioRect);

	deviceRect = (**inDevice).gdRect;
	
	ioRect->left = ioRect->left - deviceRect.left;
	ioRect->top = ioRect->top - deviceRect.top;
	ioRect->right = ioRect->right - deviceRect.left;
	ioRect->bottom = ioRect->bottom - deviceRect.top;
}


// ===========================================================================
// Returns the low word of the number returned by Microseconds().  Getting
// just a single long rather than a struct makes it easier to deal with.  
// Use this with S3ElapsedMicroseconds to time drawing operations.
//
// RETURNS: the lo slot of the UnsignedWide containing the microseconds
//		since startup.
//

	unsigned long
S3MicrosecondCount(void)
{
	UnsignedWide currentCount;
	
	Microseconds(&currentCount);
	
	return currentCount.lo;
}


// ===========================================================================
// Calculates the elapsed time from inLastCount to the current microseconds
// count.  Use this along with S3MicrosecondCount to time operations.
//
// inLastCount: the value returned by a previous call of S3MicrosecondCount
//
// RETURNS: the difference between the current microseconds and inLastCount
//

	unsigned long
S3ElapsedMicroseconds(
	unsigned long inLastCount)
{
	UnsignedWide currentCount;
	unsigned long elapsedMicroseconds = 0;
	
	Microseconds(&currentCount);

	if (currentCount.lo < inLastCount) {
			// the lo doubleword has overflowed and so is now less than inLastCount.  
			// we get the actual elapsed time by figuring out how far inLastCount
			// was from overflowing and adding that to currentCount.lo.  of course,
			// this only works if there's been only one overflow, but it's pretty
			// unlikely any tests are going to be longer than 4.2 billion microseconds
		elapsedMicroseconds = currentCount.lo + (ULONG_MAX - inLastCount);
	} else {
			// there's been no overflow, so just subtract to get elapsed time
		elapsedMicroseconds = currentCount.lo - inLastCount;
	}

	return elapsedMicroseconds;
}


// ===========================================================================
// Determines whether a number is a power of 2.
//
// inNumber: the number to check
// outPower: if not nil, on exit it points to the base-2 log of inNumber
//
// RETURNS: true if inNumber is a power of 2, false otherwise
//

	Boolean
S3IsPowerOf2(
	unsigned long inNumber,
	unsigned short* outPower)
{
	unsigned short power = 0;
	Boolean isPowerOf2;
	
	if (inNumber == 0) {
			// zero's not a power of 2 and the loop below won't end if the number
			// is zero, so bail here
		return false;
	}
	
		// loop until we find a 1 in the lowest bit
	while ((inNumber & 1) == 0) {
		inNumber >>= 1;
		power++;
	};
	
		// get rid of the 1 we just found
	inNumber >>= 1;
	
		// if inNumber is a power of 2, it should now be 0.  if it has any
		// higher bits set, it won't be 0 and it won't be a power of 2
	isPowerOf2 = (inNumber == 0);
	
	if (isPowerOf2 && (outPower != nil)) {
		*outPower = power;
	}
	
	return isPowerOf2;
}


#define kARGB16mask_alpha		(1 << 15)
#define kARGB16mask_RGB			~(1 << 15)


// ===========================================================================
// Sets the alpha value of each pixel in a pixmap to 0 if its color is black.
// Otherwise, it makes the pixel opaque.  This is useful for creating an
// alpha channel mask for a texture.
//
// inWidth: width of the pixmap
// inHeight: height of the pixmap
// inRowBytes: bytes per each row of the pixmap
// inPixMap: pointer to the pixmap's base address
// inPixelType: the type of pixel in the pixmap, either kQAPixel_ARGB16 or
//		kQAPixel_ARGB32.  the function has no effect if any other value
//		is passed here.
//
// RETURNS: void
//

	void
S3MakeBlackTransparent(
	long 							inWidth,
	long 							inHeight,
	long 							inRowBytes,
	void* 						inPixMap,
	TQAImagePixelType inPixelType)
{
	if (inPixelType == kQAPixel_ARGB16) {
		Uint16* pixMap = inPixMap;
		Uint16* scanline;
		long pixelsPerRow = inRowBytes / 2;
		long x, y;

		for (y = 0; y < inHeight; y++) {
				// find the beginning of the next scanline
			scanline = pixMap + (y * pixelsPerRow);
			
			for (x = 0; x < inWidth; x++) {
				if ((scanline[x] & kARGB16mask_RGB) == 0) {
					 // the color of this pixel is black, so make it transparent.
					 // and'ing the pixel with the inverse of kARGB16mask_alpha 
					 // turns off the high bit.
					scanline[x] &= ~kARGB16mask_alpha;
				} else {
					 // the color of this pixel is not black, so make it opaque.
					 // or'ing the pixel with kARGB16mask_alpha turns on the high bit.
					scanline[x] |= kARGB16mask_alpha;
				} 
			} // endfor
		} // endfor
	} else if (inPixelType == kQAPixel_ARGB32) {
		typedef struct {
			unsigned char		a;		/* D31:24 */
			unsigned char		r;		/* D23:16 */
			unsigned char		g;		/* D15:8 */
			unsigned char		b;		/* D7:0 */
		} Pixel32;
		
		Pixel32* scanline;
		Pixel32* pixMap = inPixMap;
		long pixelsPerRow = inRowBytes / 4;
		long x, y;
		
		for (y = 0; y < inHeight; y++) {
				// find the beginning of the next scanline
			scanline = pixMap + (y * pixelsPerRow);
			
			for (x = 0; x < inWidth; x++) {
				if ((scanline[x].r == 0) && (scanline[x].g == 0) && (scanline[x].b == 0)) {
					 // the color of this pixel is black, so make it transparent (alpha = 0)
					scanline[x].a = 0;
				} else {
					 // the color of this pixel is not black, so make it opaque (alpha = 255)
					scanline[x].a = ~0;
				} 
			} // endfor
		} // endfor
	}
}


// ===========================================================================
// Prompts the user for a PICT file with the standard get file dialog box.
//
// outFSSpec: on exit, the FSSpec for the file the user selected
//
// RETURNS: true if the user selected a file and clicked OK, false otherwise
//

	Boolean
S3GetPICTFileSpec(FSSpec* outFSSpec)
{
		// we have to initialize all the fields in this struct to 0 so that 
		// the caller gets back a clean FSSpec, with no garbage following the
		// file name
	StandardFileReply	macFileReply = { 0, 0, 0, { 0, 0, { 0 } }, 0, 0, 0, 0, 0, 0 };
	SFTypeList			typeList;
	
	InitCursor();

	typeList[0] = 'PICT';

	StandardGetFile(nil, 1, typeList, &macFileReply);

	if (macFileReply.sfGood) {
		*outFSSpec = macFileReply.sfFile;

		return true;
	} 

	return false;
}


// ===========================================================================
// Reads the PICT data out of the data fork of a file.  The PICT header is
// ignored.
//
// inFSSpec: the FSSpec for the PICT to be read
//
// RETURNS: a handle to the PICT data, which the caller is responsible for
// 		disposing of.
//

	PicHandle
S3GetPICTFromFile(const FSSpec* inFSSpec)
{
	OSErr	err;
	PicHandle pictH;
	Int32	pictFileLength, bytesRead;
	short refNum;
	
		// open the file's data fork so we can read it
	err = FSpOpenDF(inFSSpec, fsRdPerm, &refNum);
	if (err != noErr) { return nil; }
	
		// get the length of the PICT file
	err = GetEOF(refNum, &pictFileLength);
	if (err != noErr) { return nil; }
	
		// there is a 512-byte header on PICT files, so subtract that amount
		// from the file length to get the file's actual length
	pictFileLength -= kPICTFileHeaderSize;
	
		// we create a PicHandle big enough to store the file, then move it 
		// high and lock it, so we can read the file into it
	pictH = (PicHandle) NewHandle(pictFileLength);
	if (pictH == nil) { return nil; }
	HLockHi((Handle) pictH);
	
		// set the file marker past the PICT header 
	err = SetFPos(refNum, fsFromStart, kPICTFileHeaderSize);
	if (err != noErr) { return nil; }

		// then read the rest of the file into the PicHandle
	bytesRead = pictFileLength;
	err = FSRead(refNum, &bytesRead, *pictH);
	HUnlock((Handle) pictH);
	if (err != noErr) { return nil; }
	ASSERT(bytesRead == pictFileLength);
	
	return pictH;
}


// ===========================================================================
// A "nearby" PICT file is in the same folder as the Sneak3view app or any
// of that folder's subfolders.  These are the same files which are listed
// in the texture and bitmap popup menus in the settings window.  This is
// useful for hard-coding particular textures into a test without requiring
// them to be in a particular spot.
//
// inTestInfo: pointer to a S3tTestInfo struct.  used to recover the
//			S3PICTList which contains the PICT file specs.
// inFileName: a C string file name
//
// RETURNS: a handle to the PICT data, which the caller is responsible for
// 		disposing of.
//

	PicHandle
S3GetNearbyPICT(S3tTestInfo* inTestInfo, char* inFileName)
{
	FSSpec pictFileSpec;
	S3PICTList pictList;
	
	ASSERT(inTestInfo);
	ASSERT(inFileName);
	
	pictList = S3TestInfo_GetPICTList(inTestInfo);
	ASSERT(pictList);
	
	if (S3PICTList_GetFileSpec(pictList, inFileName, &pictFileSpec) == false) {
			// a file with this name wasn't found
		return nil;
	}
	
		// return a handle to the PICT inside this file
	return S3GetPICTFromFile(&pictFileSpec);
}


// ===========================================================================
// Finds the FSSpec of the Sneak3view app.  This is useful if you want to find
// out where it was launched from, to read other files from its folder, 
// for instance.
//
// outFSSpec: on exit, points to the app's FSSpec if the function succeeded.
//			if not, the FSSpec's fields will all be 0.  outFSSpec cannot be nil.
//
// RETURNS: void
//

	void
S3GetAppFileSpec(FSSpec* outFSSpec)
{
	ProcessSerialNumber processNum;
	ProcessInfoRec processInfo;
	OSErr err;
	
	ASSERT(outFSSpec);
	
		// zero out the fields of the FSSpec so that the caller will know if
		// the function failed 
	outFSSpec->vRefNum = 0;
	outFSSpec->parID = 0;
	outFSSpec->name[0] = 0;
	
		// find out the app's process number
	err = GetCurrentProcess(&processNum);
	ASSERT(err == noErr);
	
		// set up ProcessInfoRec to get the app's FSSpec and put it in inFSSpec
	processInfo.processInfoLength = sizeof(ProcessInfoRec);
	processInfo.processName = nil;
	processInfo.processAppSpec = outFSSpec;
	
		// get our app's ProcessInfoRec so we can find out where 
		// the app was launched from
	err = GetProcessInformation(&processNum, &processInfo);
	ASSERT(err == noErr);
}

