// ===========================================================================
//	UWindowState.cp	  1995 J. Rodden, DD/MF & Associates. All rights reserved
// ===========================================================================
// Utilities for saving and restoring a window's state (position, size, and
// zoom state) in a file. Typically usefull for document windows as per
// Apple's User Interface Guidelines. Use when opening and saving a document.
//
// All rights reserved. You may use this code in any application, recognition
// would be appreciated but is not required.

#include "UWindowState.h"

#include <LFile.h>
#include <LWindow.h>
#include <UWindows.h>
#include <Displays.h>
#include <QuickDraw.h>
#include <stdlib.h>
#include <limits.h>

const ResType kWindowStateRsrc = 'WSET';

struct WindowStateData {
	Rect	userBounds;
	Boolean	isZoomed;
};

// ===========================================================================

void
UWindowState::SaveWindowState( LFile* inFile, LWindow* inWindow, short inResID)
{
		// make sure our parameters are okay
	if (inFile == nil || inWindow == nil) {
		return;
	}
	
	// Make file a resource fork if it doesn't already have one
	FSSpec	theFileSpec;
	FInfo	theFileInfo;
	
	inFile->GetSpecifier(theFileSpec);
	
	::FSpGetFInfo( &theFileSpec, &theFileInfo);
	::FSpCreateResFile( &theFileSpec, theFileInfo.fdCreator, theFileInfo.fdType, nil);


	// -----------------------------------------
	// Get the window state
	Rect			dummyRect;
	WindowStateData	theWindowState;
	
	inWindow->CalcPortFrameRect(theWindowState.userBounds);
	inWindow->PortToGlobalPoint(topLeft(theWindowState.userBounds));
	inWindow->PortToGlobalPoint(botRight(theWindowState.userBounds));
	
	theWindowState.isZoomed = inWindow->CalcStandardBounds(dummyRect);


	// -----------------------------------------
	// Store the window state
  ::UseResFile(inFile->OpenResourceFork(fsWrPerm));
	
	WindowStateData	**theHandle = 
					(WindowStateData**) ::Get1Resource( kWindowStateRsrc, inResID);
	
	if ( theHandle != nil )	{
		// a state data resource already exists -- update it
		**theHandle = theWindowState;
	  ::ChangedResource(Handle(theHandle));
	} else {
		// no state data has yet been saved -- add state data resource
		theHandle = (WindowStateData**) (NewHandle(sizeof(theWindowState)));
		if ( theHandle != nil ) {
			**theHandle = theWindowState;
		  ::AddResource( Handle(theHandle), kWindowStateRsrc, inResID, nil);
		}
	}
	
	if ( theHandle != nil ) {
	  ::UpdateResFile(inFile->GetResourceForkRefNum());
	  ::ReleaseResource(Handle(theHandle));
	}
	
	// inFile->CloseResourceFork();		// Let destructor close it
}


// ===========================================================================

void
UWindowState::RestoreWindowState( LFile* inFile, LWindow* inWindow, short inResID)
{
	//WStateData	theWindowState;
	WindowStateData	theWindowState;
	WindowStateData	**theHandle;
	
	// -----------------------------------------
	// Get saved window state
  ::UseResFile(inFile->OpenResourceFork(fsRdPerm));
		
	theHandle = (WindowStateData**) ::Get1Resource( kWindowStateRsrc, inResID);
		
	// -----------------------------------------
	// Restore window state
	if ( theHandle != nil )	{	// handle to data succeeded -- retrieve saved user state
		theWindowState = **theHandle;
		
		PositionWindowOnScreen(inWindow, &theWindowState.userBounds);
		
		inWindow->DoSetBounds(theWindowState.userBounds);
		
			// make sure this window is zoomable before we try to zoom it
		if (inWindow->HasAttribute(windAttr_Zoomable)) {
			inWindow->DoSetZoom(theWindowState.isZoomed);
		}
	}	
	
  ::ReleaseResource(Handle(theHandle));
  
	// inFile->CloseResourceFork();		// Let destructor close it
}


// ===========================================================================
	void
UWindowState::PositionWindowOnScreen(
	LWindow*				inWindow,
	Rect*					ioUserBounds)
{
	Rect					structureRect;
	structureRect = UWindows::GetWindowStructureRect((WindowPtr) inWindow->GetMacPort());
	
	Rect					contentRect;
	contentRect = UWindows::GetWindowContentRect((WindowPtr) inWindow->GetMacPort());
	
		// extensions like Stretch (and presumably Copland) cause a frame to be
		// drawn around the window content, which makes the bottom of the 
		// structure rect lower than the bottom of the content rect.  we're 
		// only interested in the area from the top of the structure to the top
		// of the content, which should include just the title bar.
	short 					titleBarHeight = contentRect.top - structureRect.top;
	
		// the top of the title bar will be titleBarHeight pixels above the
		// window's current user bounds.  we have to make the title bar rect
		// relative to ioUserBounds rather than contentRect because contentRect
		// and structureRect are in terms of where the window is right now
		// (its default position), rather than the position saved in the 
		// window state resource.
	Rect					titleBarRect = *ioUserBounds;
	titleBarRect.top = ioUserBounds->top - titleBarHeight;
	titleBarRect.bottom = ioUserBounds->top;

#if 0
	RgnHandle				desktopRegion = nil;
	DMGetDeskRegion(&desktopRegion);
	
	Point					topLeft = { titleBarRect.top, titleBarRect.left };
	Point					topRight = { titleBarRect.top, titleBarRect.right };
	Point					bottomLeft = { titleBarRect.bottom, titleBarRect.left };
	Point					bottomRight = { titleBarRect.bottom, titleBarRect.right };
	
	Boolean					topLeftIsOnADisplay = PtInRgn(topLeft, desktopRegion);
	Boolean					topRightIsOnADisplay = PtInRgn(topRight, desktopRegion);
	Boolean					bottomLeftIsOnADisplay = PtInRgn(bottomLeft, desktopRegion);
	Boolean					bottomRightIsOnADisplay = PtInRgn(bottomRight, desktopRegion);
	
	if (topLeftIsOnADisplay && topRightIsOnADisplay && 
			bottomLeftIsOnADisplay && bottomRightIsOnADisplay) {
			// the title bar is completely on screen, so we don't have to 
			// move the window before showing it
		return;
	}

		// set the minimum horizontal and vertical shifts to a big number, so
		// that the first shift we find will always be smaller than it
	short 					minHorizontalShift = SHRT_MAX;
	short 					minVerticalShift = SHRT_MAX;
	long					minShift = ULONG_MAX;
	
		// this counter ensures that we don't go forever in the do loop below,
		// which might happen if the title bar is wider or taller than the 
		// screen (which is actually pretty unlikely)
	short 					loopRepetitions = 0;

	do {
		for (GDHandle currentGDevice = DMGetFirstScreenDevice(dmOnlyActiveDisplays);
				currentGDevice != nil; 
				currentGDevice = DMGetNextScreenDevice(currentGDevice, dmOnlyActiveDisplays)) {
				// if an edge of the title bar hasn't already been found within at
				// least one of the displays, check if it is within this 
				// display.  if it is, check to see if the amount we need to shift 
				// it (either horizontally or vertically) is less than the smallest 
				// shift amount that we've already found.  if it is, save it as 
				// the new, minimum shift amount.
			Rect						currentGDRect;
			currentGDRect = (**currentGDevice).gdRect;
			
			if (currentGDevice == ::GetMainDevice()) {
					// this is the main device, so subtract the height of the 
					// menu bar from its rect, since we don't want the windows
					// getting covered by the menu bar.
				currentGDRect.top += ::GetMBarHeight();
			}
			
			if (!topLeftIsOnADisplay) {
				short		newHorizontalShift, newVerticalShift;
				
				if (titleBarRect.left < currentGDRect.left) {
					newHorizontalShift = currentGDRect.left - titleBarRect.left;
				} else if (titleBarRect.left > currentGDRect.right) {
					newHorizontalShift = currentGDRect.right - titleBarRect.left;
				} else {
					newHorizontalShift = 0;
				}
				
				if (titleBarRect.top < currentGDRect.top) {
					newVerticalShift = currentGDRect.top - titleBarRect.top;
				} else if (titleBarRect.top > currentGDRect.bottom) {
					newVerticalShift = currentGDRect.bottom - titleBarRect.top;
				} else {
					newVerticalShift = 0;
				}
				
				long 		newMinShift = 
						(newHorizontalShift * newHorizontalShift) + 
						(newVerticalShift * newVerticalShift);
						
				if (newMinShift < minShift) {
					minHorizontalShift = newHorizontalShift;
					minVerticalShift = newVerticalShift;
					minShift = newMinShift;
				}
			}
					
			if (!topRightIsOnADisplay) {
				short		newHorizontalShift, newVerticalShift;
				
				if (titleBarRect.right < currentGDRect.left) {
					newHorizontalShift = currentGDRect.left - titleBarRect.right;
				} else if (titleBarRect.right > currentGDRect.right) {
					newHorizontalShift = currentGDRect.right - titleBarRect.right;
				} else {
					newHorizontalShift = 0;
				}
				
				if (titleBarRect.top < currentGDRect.top) {
					newVerticalShift = currentGDRect.top - titleBarRect.top;
				} else if (titleBarRect.top > currentGDRect.bottom) {
					newVerticalShift = currentGDRect.bottom - titleBarRect.top;
				} else {
					newVerticalShift = 0;
				}
				
				long 		newMinShift = 
						(newHorizontalShift * newHorizontalShift) + 
						(newVerticalShift * newVerticalShift);
						
				if (newMinShift < minShift) {
					minHorizontalShift = newHorizontalShift;
					minVerticalShift = newVerticalShift;
					minShift = newMinShift;
				}
			}
					
			if (!bottomLeftIsOnADisplay) {
				short		newHorizontalShift, newVerticalShift;
				
				if (titleBarRect.left < currentGDRect.left) {
					newHorizontalShift = currentGDRect.left - titleBarRect.left;
				} else if (titleBarRect.right > currentGDRect.right) {
					newHorizontalShift = currentGDRect.right - titleBarRect.left;
				} else {
					newHorizontalShift = 0;
				}
				
				if (titleBarRect.bottom < currentGDRect.top) {
					newVerticalShift = currentGDRect.top - titleBarRect.bottom;
				} else if (titleBarRect.bottom > currentGDRect.bottom) {
					newVerticalShift = currentGDRect.bottom - titleBarRect.bottom;
				} else {
					newVerticalShift = 0;
				}
				
				long 		newMinShift = 
						(newHorizontalShift * newHorizontalShift) + 
						(newVerticalShift * newVerticalShift);
						
				if (newMinShift < minShift) {
					minHorizontalShift = newHorizontalShift;
					minVerticalShift = newVerticalShift;
					minShift = newMinShift;
				}
			}
					
			if (!bottomRightIsOnADisplay) {
				short		newHorizontalShift, newVerticalShift;
				
				if (titleBarRect.right < currentGDRect.left) {
					newHorizontalShift = currentGDRect.left - titleBarRect.right;
				} else if (titleBarRect.right > currentGDRect.right) {
					newHorizontalShift = currentGDRect.right - titleBarRect.right;
				} else {
					newHorizontalShift = 0;
				}
				
				if (titleBarRect.bottom < currentGDRect.top) {
					newVerticalShift = currentGDRect.top - titleBarRect.bottom;
				} else if (titleBarRect.bottom > currentGDRect.bottom) {
					newVerticalShift = currentGDRect.bottom - titleBarRect.bottom;
				} else {
					newVerticalShift = 0;
				}
				
				long 		newMinShift = 
						(newHorizontalShift * newHorizontalShift) + 
						(newVerticalShift * newVerticalShift);
						
				if (newMinShift < minShift) {
					minHorizontalShift = newHorizontalShift;
					minVerticalShift = newVerticalShift;
					minShift = newMinShift;
				}
			}
		} // endfor
		
		titleBarRect.left += minHorizontalShift;
		titleBarRect.right += minHorizontalShift;
		titleBarRect.top += minVerticalShift;
		titleBarRect.bottom += minVerticalShift;
		
		topLeftIsOnADisplay = PtInRgn(topLeft, desktopRegion);
		topRightIsOnADisplay = PtInRgn(topRight, desktopRegion);
		bottomLeftIsOnADisplay = PtInRgn(bottomLeft, desktopRegion);
		bottomRightIsOnADisplay = PtInRgn(bottomRight, desktopRegion);
		
		loopRepetitions++;
	} while ((!(topLeftIsOnADisplay && topRightIsOnADisplay && 
			bottomLeftIsOnADisplay && bottomRightIsOnADisplay)) &&
			(loopRepetitions < 5));
	
#else 

	Boolean					leftEdgeIsOnADisplay = false;
	Boolean					rightEdgeIsOnADisplay = false;
	Boolean					topEdgeIsOnADisplay = false;
	Boolean					bottomEdgeIsOnADisplay = false;
	
		// set the minimum horizontal and vertical shifts to a big number, so
		// that the first shift we find will always be smaller than it
	short 					minHorizontalShift = SHRT_MAX;
	short 					minVerticalShift = SHRT_MAX;
	short 					newShift;
	
		// this counter ensures that we don't go forever in the do loop below,
		// which might happen if the title bar is wider or taller than the 
		// screen (which is actually pretty unlikely)
	short 					loopRepetitions = 0;

	do {
		for (GDHandle currentGDevice = DMGetFirstScreenDevice(dmOnlyActiveDisplays);
				currentGDevice != nil; 
				currentGDevice = DMGetNextScreenDevice(currentGDevice, dmOnlyActiveDisplays)) {
				// if an edge of the title bar hasn't already been found within at
				// least one of the displays, check if it is within this 
				// display.  if it is, check to see if the amount we need to shift 
				// it (either horizontally or vertically) is less than the smallest 
				// shift amount that we've already found.  if it is, save it as 
				// the new, minimum shift amount.
			Rect						currentGDRect;
			currentGDRect = (**currentGDevice).gdRect;
			
			if (currentGDevice == ::GetMainDevice()) {
					// this is the main device, so subtract the height of the 
					// menu bar from its rect, since we don't want the windows
					// getting covered by the menu bar.
				currentGDRect.top += ::GetMBarHeight();
			}
			
			if (!leftEdgeIsOnADisplay) {
				if (titleBarRect.left < currentGDRect.left) {
					newShift = currentGDRect.left - titleBarRect.left;
					minHorizontalShift = (abs(newShift) < abs(minHorizontalShift)) ? 
							newShift : minHorizontalShift;
				} else if (titleBarRect.left > currentGDRect.right) {
					newShift = currentGDRect.right - titleBarRect.left;
					minHorizontalShift = (abs(newShift) < abs(minHorizontalShift)) ? 
							newShift : minHorizontalShift;
				} else {
					leftEdgeIsOnADisplay = true;
				}
			}
					
			if (!rightEdgeIsOnADisplay) {
				if (titleBarRect.right < currentGDRect.left) {
					newShift = currentGDRect.left - titleBarRect.right;
					minHorizontalShift = (abs(newShift) < abs(minHorizontalShift)) ? 
							newShift : minHorizontalShift;
				} else if (titleBarRect.right > currentGDRect.right) {
					newShift = currentGDRect.right - titleBarRect.right;
					minHorizontalShift = (abs(newShift) < abs(minHorizontalShift)) ? 
							newShift : minHorizontalShift;
				} else {
					rightEdgeIsOnADisplay = true;
				}
			}
					
			if (!topEdgeIsOnADisplay) {
				if (titleBarRect.top < currentGDRect.top) {
					newShift = currentGDRect.top - titleBarRect.top;
					minVerticalShift = (abs(newShift) < abs(minVerticalShift)) ? 
							newShift : minVerticalShift;
				} else if (titleBarRect.top > currentGDRect.bottom) {
					newShift = currentGDRect.bottom - titleBarRect.top;
					minVerticalShift = (abs(newShift) < abs(minVerticalShift)) ? 
							newShift : minVerticalShift;
				} else {
					topEdgeIsOnADisplay = true;
				}
			}
					
			if (!bottomEdgeIsOnADisplay) {
				if (titleBarRect.bottom < currentGDRect.top) {
					newShift = currentGDRect.top - titleBarRect.bottom;
					minVerticalShift = (abs(newShift) < abs(minVerticalShift)) ? 
							newShift : minVerticalShift;
				} else if (titleBarRect.bottom > currentGDRect.bottom) {
					newShift = currentGDRect.bottom - titleBarRect.bottom;
					minVerticalShift = (abs(newShift) < abs(minVerticalShift)) ? 
							newShift : minVerticalShift;
				} else {
					bottomEdgeIsOnADisplay = true;
				}
			}
		} // endfor
		
		if (!(leftEdgeIsOnADisplay && rightEdgeIsOnADisplay)) {
			titleBarRect.left += minHorizontalShift;
			titleBarRect.right += minHorizontalShift;
		}
		
		if (!(topEdgeIsOnADisplay && bottomEdgeIsOnADisplay)) {
			titleBarRect.top += minVerticalShift;
			titleBarRect.bottom += minVerticalShift;
		}
		
		loopRepetitions++;
	} while ((!(leftEdgeIsOnADisplay && rightEdgeIsOnADisplay && 
			topEdgeIsOnADisplay && bottomEdgeIsOnADisplay)) &&
			(loopRepetitions < 5));
#endif

		// before we move ioUserBounds, we have to record its height
	short					ioUserBoundsHeight = ioUserBounds->bottom - ioUserBounds->top;
	
		// move ioUserBounds directly underneath the position to which the
		// loop above moved the title bar.
	ioUserBounds->left = titleBarRect.left;
	ioUserBounds->right = titleBarRect.right;
	ioUserBounds->top = titleBarRect.bottom;
	ioUserBounds->bottom = titleBarRect.bottom + ioUserBoundsHeight;
}
