// ===========================================================================
// CBroadcastEditField.cp
// ===========================================================================


#include <UKeyFilters.h>

#include <strstream.h>
#include <stdio.h>
#include <string.h>

#include "CBroadcastEditField.h"
#include "StringConversion.h"


// ===========================================================================
//		 UnsignedFloatField
// ===========================================================================
//	Key Filter for Unsigned Float characters
//
//		> Identify delete and cursor keys
//		> Accept numbers (0 to 9) and decimal point
//		> Reject all other printing characters
//		> PassUp all other characters

	EKeyStatus
CBroadcastEditField::UnsignedFloatField(const EventRecord &inKeyEvent)
{
	EKeyStatus	theKeyStatus = keyStatus_PassUp;
	Char16		theKey = inKeyEvent.message;
	Char16		theChar = theKey & charCodeMask;
	
	if (UKeyFilters::IsTEDeleteKey(theKey)) {
		theKeyStatus = keyStatus_TEDelete;
	} else if (UKeyFilters::IsTECursorKey(theKey)) {
		theKeyStatus = keyStatus_TECursor;
	} else if (UKeyFilters::IsExtraEditKey(theKey)) {
		theKeyStatus = keyStatus_ExtraEdit;
	} else if (UKeyFilters::IsPrintingChar(theChar)) {
		if (UKeyFilters::IsNumberChar(theChar)  ||  theChar == '.') {
			theKeyStatus = keyStatus_Input;
		} else {
			theKeyStatus = keyStatus_Reject;
		}
	}
	
	return theKeyStatus;
}


// ===========================================================================
//		 SignedIntField
// ===========================================================================
//	Key Filter for Signed Int characters
//
//		> Identify delete and cursor keys
//		> Accept numbers (0 to 9) and plus and minus sign
//		> Reject all other printing characters
//		> PassUp all other characters

	EKeyStatus
CBroadcastEditField::SignedIntField(const EventRecord &inKeyEvent)
{
	EKeyStatus	theKeyStatus = keyStatus_PassUp;
	Char16		theKey = inKeyEvent.message;
	Char16		theChar = theKey & charCodeMask;
	
	if (UKeyFilters::IsTEDeleteKey(theKey)) {
		theKeyStatus = keyStatus_TEDelete;
	} else if (UKeyFilters::IsTECursorKey(theKey)) {
		theKeyStatus = keyStatus_TECursor;
	} else if (UKeyFilters::IsExtraEditKey(theKey)) {
		theKeyStatus = keyStatus_ExtraEdit;
	} else if (UKeyFilters::IsPrintingChar(theChar)) {
		if (UKeyFilters::IsNumberChar(theChar)  ||  theChar == '-'  ||  theChar == '+') {
			theKeyStatus = keyStatus_Input;
		} else {
			theKeyStatus = keyStatus_Reject;
		}
	}
	
	return theKeyStatus;
}


// ===========================================================================
//		 CreateBroadcastFieldStream
// ===========================================================================
//	Create a new EditField object from the data in a Stream

	CBroadcastEditField*
CBroadcastEditField::CreateBroadcastEditFieldStream(LStream	*inStream)
{
	return (new CBroadcastEditField(inStream));
}


// ===========================================================================
//		 LEditField
// ===========================================================================
//	Default Contructor

CBroadcastEditField::CBroadcastEditField()
	: LEditField()
{
}

// ===========================================================================
//		 LEditField(LEditField&)
// ===========================================================================
//	Copy Contructor

CBroadcastEditField::CBroadcastEditField(const LEditField	&inOriginal)
	: LEditField(inOriginal)
{
}


// ===========================================================================
//		 LEditField
// ===========================================================================
//	Contruct from input parameters

CBroadcastEditField::CBroadcastEditField(
	const SPaneInfo	&inPaneInfo,
	Str255			inString,
	ResIDT			inTextTraitsID,
	Int16			inMaxChars,
	Boolean			inHasBox,
	Boolean			inHasWordWrap,
	KeyFilterFunc	inKeyFilter,
	LCommander		*inSuper)
	: LEditField(inPaneInfo, inString, inTextTraitsID, inMaxChars, inHasBox,
							 inHasWordWrap, inKeyFilter, inSuper)
{
}							 


// ===========================================================================
//		 LEditField(LStream*)
// ===========================================================================
CBroadcastEditField::CBroadcastEditField(
	const SPaneInfo	&inPaneInfo,
	Str255			inString,
	ResIDT			inTextTraitsID,
	Int16			inMaxChars,
	Uint8			inAttributes,
	KeyFilterFunc	inKeyFilter,
	LCommander		*inSuper)
	: LEditField(inPaneInfo, inString, inTextTraitsID, inMaxChars, inAttributes, 
  						 inKeyFilter, inSuper)
{
}							 


// ===========================================================================
//		 LEditField(LStream*)
// ===========================================================================
//	Contruct an EditField from the data in a Stream

CBroadcastEditField::CBroadcastEditField(LStream *inStream)
	: LEditField(inStream)
{
}


// ===========================================================================
//		 UserChangedText
// ===========================================================================
//	Text of EditField has changed as a result of user action
//
//	Override to validate field and/or dynamically update as the user
//	types. This function is not called by SetDescriptor, which is typically
//	used to programatically change the text.

	void
CBroadcastEditField::UserChangedText()
{
		// we need to call this to update the menus since the user has dirtied 
		// the window and the Save command should now be enabled
	SetUpdateCommandStatus(true);

		// just use our Pane ID as the message value	
	MessageT message = (MessageT) GetPaneID();

// should we broadcast a pointer to the text?  a TE handle?
// an int?  a float?
		
	BroadcastMessage(message, (void*) this);
}


// ===========================================================================
	void
CBroadcastEditField::SetValueAsFloat(float inFloatValue)
{
	Str255 descriptorString;
	
	sprintf((char*) &descriptorString[1], "%g", inFloatValue);
	
		// set the length byte
	descriptorString[0] = strlen((char*) &descriptorString[1]);
	
		// set the edit field's descriptor to this string representation
	SetDescriptor(descriptorString);
}


// ===========================================================================
	float
CBroadcastEditField::GetValueAsFloat()
{
//	char descriptorCstring[sizeof(Str255)];
	Str255 descriptorString;

//	GetDescriptor((Str255) descriptorCstring);
	GetDescriptor(descriptorString);
	
//	p2cstr((Str255) descriptorCstring, descriptorCstring);

// maybe use stdio funcs instead of string streams?

	istrstream descriptorStream((char*) &descriptorString[1], descriptorString[0]);
	float theFloatValue = 0.0;
	descriptorStream >> theFloatValue;
	
	return theFloatValue;
}


//// ===========================================================================
//	void
//CBroadcastEditField::SetDescriptor(const char* inDescriptor)
//{
//	Str255 pascalString;
//	
//	S3CtoPString(inDescriptor, pascalString);
//	
//	LEditField::SetDescriptor(pascalString);
//}
//
//
//// ===========================================================================
//	char*
//CBroadcastEditField::GetDescriptor(char* outDescriptor) const
//{
//	Str255 pascalString;
//	
//	LEditField::GetDescriptor(pascalString);
//	
//	S3PtoCString(pascalString, outDescriptor);
//	
//	return outDescriptor;
//}
