// ===========================================================================
//	CPvalPopupMenu.cp
// ===========================================================================
//
//  Much of the following code was taken from the PowerPlant LMenu class.
//
//	Each item in this type of popup menu can have an arbitrary value 
//  associated with it which is used instead of the item number when 
//  getting, setting and broadcasting the menu's value.  The values are
//  stored in a mPvals resource which has the same ID as the popup's
//  MENU ID.  The mPvals values do not have to be distinct, unlike Mcmd
//  numbers.
//
//
//	Manages a popup menu. Maintains a mapping between Menu items and
//	control values.
//
//	Resources:
//		'MENU'		Standard Mac resource for a Menu
//		'Pval'		control values for each item in the Menu
//
//	 Value Numbers
//	Values are 32-bit, signed integer values.
//	PowerPlant reserves command numbers -999 to +999.
//
//		 Synthetic Value Numbers
//		A synthetic command number has the Menu ID in the high 16 bits
//		and the item number in the low 16 bits, with the result negatated.
//			syntheticCmd = - (MenuID << 16) - ItemNumber
//		A synthetic command is the negative of the value returned by
//		the Toolbox traps MenuSelect and MenuKey.
//		You can extract the components as follows:
//			MenuID = HiWord(-syntheticCmd);
//			ItemNumber = LoWord(-syntheticCmd);
//		Alternatively, the LValueer class has an IsSyntheticValue
//		static function that extracts this information.
//
//		You must use Menu ID numbers between 1 and 32767. Therefore, a
//		command number is synthetic if it is negative and the hi word
//		of its negation is not zero (which would correspond to a zero
//		Menu ID).
//
//		Synthetic commands numbers range from $80000000 (Menu ID 32767)
//		to $FFFF0000 (Menu ID 1).
//
//		 Negative command numbers
//		Negative command numbers from $FFFF0001 to $FFFFFFFF (-65535 to -1)
//		are not valid synthetic command numbers. Programs can use command
//		numbers in this range. (As mentioned above PowerPlant reserves
//		-999 to +999). However, the PowerPlant Application class treats
//		positive and negative command numbers differently.
//
//		The Application class updates the enabled state of menu items with
//		positive command numbers just before letting the user make a menu
//		selection. It does not update the state of menu items with negative
//		command numbers. You might want to use a negative command number
//		for a menu item that is always enabled (except when its entire
//		menu is disabled; disabling an entire menu overrides the settings
//		of individual menu items).
//
//	 Value Table
//	The command table (mValueNums) is a handle of command numbers. The
//	table always starts with the first menu item. For example, if
//	mNumValues is 6, then the commands in the table map to items 1 to 6.
//
//	 ValueFromIndex
//	ValueFromIndex returns the command for a particular menu item. It
//	returns a synthetic command number in two cases:
//		1) The menu item is greater than the size of the command table.
//		This happens when you add menu items at runtime (for example, by
//		calling AppendResMenu for a Font menu or maintaining a Windows menu).
//
//		2) The item has a command of cmd_UseMenuItem (value of -1).
//
//	 Using the Toolbox Menu Manager
//	You should use direct Toolbox traps for all menu operations except
//	adding and removing items that have command numbers. You can get the
//	MenuHandle for a Menu object from the GetMacMenuH function.
//
//	Note that is OK to add and remove items that don't have command numbers.
//	For example, the Windows menu in some programs have some fixed items
//	at the top, then a variable list of the names of open windows. Using
//	LMenu, only the fixed items would have command numbers. The items for
//	the open windows would not have command numbers, so you can call
//	the Toolbox traps InsertMenuItem and DeleteMenuItem for those items.
//	For items with command numbers, you must use the LMenu functions
//	InsertValue, RemoveValue, and RemoveItem.


// *** WARNING ***
// the value changing functions (if I've even written them) probably don't work yet

// jad
// we should test the command and item insertion funcs, as well as
// mUsePvalAsValue stuff


#ifdef PowerPlant_PCH
#include PowerPlant_PCH
#endif


#include <PP_Messages.h>

#include "CPvalPopupMenu.h"


// ===========================================================================
//  CPvalPopupMenu											   CPvalPopupMenu 
// ===========================================================================
	CPvalPopupMenu*
CPvalPopupMenu::CreatePvalPopupMenuStream(LStream	*inStream)
{
	return (new CPvalPopupMenu(inStream));
}


// ===========================================================================
//		 CPvalPopupMenu
// ===========================================================================
//	Construct from input parameters

CPvalPopupMenu::CPvalPopupMenu(
	const SPaneInfo	&inPaneInfo,
	MessageT		inValueMessage,
	Int16			inTitleOptions,
	ResIDT			inMENUid,
	Int16			inTitleWidth,
	Int16			inPopupVariation,
	ResIDT			inTextTraitsID,
	Str255			inTitle,
	OSType			inResTypeMENU,
	Int16			inInitialMenuItem)
		: LStdPopupMenu(inPaneInfo, inValueMessage, inTitleOptions,
				inMENUid, inTitleWidth, inPopupVariation, inTextTraitsID,
				inTitle, inResTypeMENU, inInitialMenuItem)
{
	InitPvalPopupMenu();
}


// ===========================================================================
//		 CPvalPopupMenu
// ===========================================================================
//	Construct from input parameters and an existing ControlHandle

CPvalPopupMenu::CPvalPopupMenu(
	const SPaneInfo	&inPaneInfo,
	MessageT		inValueMessage,
	Int32			inMaxValue,
	ResIDT			inTextTraitsID,
	ControlHandle	inMacControlH)
		: LStdPopupMenu(inPaneInfo, inValueMessage, inMaxValue,
						inTextTraitsID, inMacControlH)
{
	InitPvalPopupMenu();
}


// ===========================================================================
//		 CPvalPopupMenu(LStream*)
// ===========================================================================
//	Construct from data in a Stream

CPvalPopupMenu::CPvalPopupMenu(
	LStream	*inStream)
		: LStdPopupMenu(inStream)
{	
	InitPvalPopupMenu();
}


// ===========================================================================
CPvalPopupMenu::~CPvalPopupMenu()
{
	if (mPvals != nil) {
		::DisposeHandle((Handle) mPvals);
	}
}


// ===========================================================================
//		 InitPvalPopupMenu
// ===========================================================================
//	Private initializer
//
// LStdPopupMenu::InitStrPopupMenu() should handle setting the initial,
// min, and max values

	void
CPvalPopupMenu::InitPvalPopupMenu()
{
	mMacMenuH = GetMacMenuH();
	mMENUid = (**mMacMenuH).menuID;
	
	mBroadcastPval = true;
	
	mNextMenu = nil;
	mNumPvals = 0;
	mPvals = nil;

	ReadPvals();
}


//// ===========================================================================
////		 CPvalPopupMenu
//// ===========================================================================
////	Construct Menu from a 'MENU' resource
////
////	This is equivalent to the Toolbox GetMenu() call
//
//CPvalPopupMenu::CPvalPopupMenu(
//	ResIDT	inMENUid)
//{
//	mMacMenuH = ::GetMenu(inMENUid);	// Even though Inside Mac warns
//										//  about calling GetMenu more than
//										//	once for a Menu, TechNote
//										//	TB 550 - Menu Mgr Q&As says
//										//	that it's OK for the Mac Plus
//										//	or any more recent machine.
//	ThrowIfNil_(mMacMenuH);
//	
//	mMENUid = inMENUid;
//		
//		// PowerPlant requires that the MENU resource ID be the same
//		// as the Menu ID stored in the resource data. Signal an error
//		// if they are different.
//	
//	SignalIf_(inMENUid != (**mMacMenuH).menuID);
//	
//	mNextMenu = nil;
//	mNumPvals = 0;
//	mPvals = nil;
//	mIsInstalled = false;
//	mUsed = false;
//
//	if (mMacMenuH != nil) {
//		ReadPvals();
//	}
//}


// ===========================================================================
//		 ~CPvalPopupMenu
// ===========================================================================
//	Destructor

//CPvalPopupMenu::~CPvalPopupMenu()
//{
//	if (mMacMenuH != nil) {
//	
//		// Error to delete Menu without first de-installing it.
//		// De-installing is *not* the same as removing the Menu from
//		// the Mac MenuBar. An installed MenuBar is under the control
//		// of a client (usually some kind of MenuBar class), and can
//		// not be deleted until that client has relinquished control.
//		// Since the Menu object has no knowledge of its clients, we
//		// can't notify them. All we can do is signal an error.
//		// +++ Come up with a better name than "installed"
//		
//		SignalIf_(mIsInstalled);
//		
//		::DeleteMenu(mMENUid);			// Remove Menu from Mac MenuBar
//		
//			// Free Mac MenuHandle. At this point, we don't know if
//			// the menu was created using GetMenu or NewMenu, which
//			// require different ways to delete the MenuHandle.
//			// Thus, we call ReleaseResource first, which will fail,
//			// but not crash, if the MenuHandle is not from a resource.
//			// If ReleaseResource fails with a resNotFound error, the
//			// MenuHandle is not a resource, so we call DisposeMenu.
//		
//		::ReleaseResource((Handle) mMacMenuH);
//		if (ResError() == resNotFound) {
//			::DisposeMenu(mMacMenuH);
//		}
//	}
//
//	if (mPvals != nil) {
//		::DisposeHandle((Handle) mPvals);
//	}
//}


// ===========================================================================
	void
CPvalPopupMenu::BroadcastValueMessage()
{
	if (mValueMessage != cmd_Nothing) {
		Int32	value = 0;

		if (mBroadcastPval) {
			value = GetPval();
		} else {
			value = GetValue();
		}

		BroadcastMessage(mValueMessage, (void*) &value);
	}
}


// ===========================================================================
	Int32
CPvalPopupMenu::GetPval() const
{	
	return PvalFromItem((Int16) GetValue());
}


// ===========================================================================
	void
CPvalPopupMenu::SetValueByPval(Int32 inPval)
{	
	SetValue(ItemFromPval(inPval));
}


// ===========================================================================
//		 ReadPvals
// ===========================================================================
//	Get value from the 'Pval' resource
//
//	A 'Pval" resource with the same ID as the 'MENU' resource contains a
//	list of values. The format of a 'Pval' resource is:
//		Number of Values: n		(2 bytes)
//		Value Num for Item 1		(4 bytes)
//		   ...
//		Value Num for Item n		(4 bytes)

	void
CPvalPopupMenu::ReadPvals()
{
	Int16**	thePvalH = (Int16**) ::GetResource('Pval', mMENUid);

	Assert_(thePvalH);
	
	if (thePvalH != nil) {
		mNumPvals = (*thePvalH)[0];
		if (mNumPvals > 0) {
				
				// Our values list is the same as the 'Pval'
				// resource without the 2-byte count at the top. So we
				// can reuse the 'Pval' resource Handle by detaching it,
				// shifting the values down by 2-bytes, and
				// resizing it.
		
			::DetachResource((Handle) thePvalH);
			mPvals = (Int32**) thePvalH;
			Int32	valuesSize = mNumPvals * sizeof(Int32);
			::BlockMoveData(*thePvalH+1, *mPvals, valuesSize);
			::SetHandleSize((Handle) mPvals, valuesSize);
		}
	}
}


// ===========================================================================
//		 PvalFromItem
// ===========================================================================
//	Return the value for a particular Menu item

	Int32
CPvalPopupMenu::PvalFromItem(Int16 inIndex) const
{
	Int32	theValue = 0;

		// Table lookup if index is in range
	if (inIndex <= mNumPvals) {
		theValue = (*mPvals)[inIndex-1];
	}
	
// we need some way to create new values for items without values
// in the Pval resource
//	if (theValue == 0) {
//	
//			// Value number is a special flag, either because it
//			// was set that way in the value table or because the
//			// index was greater than the value table size.
//			
//			// In this case, we return a synthetic value number,
//			// which has the MENU id and item number embedded
//			
//		theValue = SyntheticPvalFromItem(inIndex);
//	}
	
	return theValue;
}


// ===========================================================================
//		 ItemFromPval
// ===========================================================================
//	Return the Menu item index number for a particular value
//		Return 0 if the value number is not used for this Menu

	Int16
CPvalPopupMenu::ItemFromPval(Int32 inValue) const
{
		// Search for value in value list for this Menu
		
	Int32	*valuesP = *mPvals;
	
	for (Int16 theIndex = 0; theIndex < mNumPvals; theIndex++) {
		if (inValue == *valuesP++) {
			return theIndex+1;
		}
	}
	
		// value not found in values list
		// If it is a synthetic value, extract the index number from
		//   the low 16 bits of the value number
		
//	if (((Int16) ((-inValue) >> 16)) > 0) {
//		return (Int16) (-inValue);
//	}
	
	return 0;				// value not found
}


// ===========================================================================
//		 SyntheticPvalFromItem
// ===========================================================================
//	Return the synthetic value for a particular Menu item
//
//	A synthetic value has the MENU id in the high 16 bits and
//	the item number in the low 16 bits, and the resulting 32-bit number
//	is negated (to distinguish it from regular values). The
//	synthetic value is the negative of the value that would be
//	returned by the Toolbox trap MenuSelect for the menu item.

	Int32
CPvalPopupMenu::SyntheticPvalFromItem(Int16 inIndex) const
{
	return (-(((Int32) mMENUid) << 16) - inIndex);
}


// ===========================================================================
//		 FindNextValue
// ===========================================================================
//	Pass back the next value in the Menu
//		On entry, ioIndex is the position of an item (0 is allowed)
//		On exit, ioIndex is the position of the next item
//			If ioIndex is greater than the number of values in the Menu
//				function returns false (and outValue is unchanged).
//			Otherwise, function returns true and returns the value
//				of the item after the input index
//
//	Use this function to iterate over all values in a Menu:
//		Int16	index = 0;
//		Int32	value;
//		while (theMenu->FindNextValue(index, value)) {
//			// Do something with value
//		}

	Boolean
CPvalPopupMenu::FindNextValue(
	Int16		&ioIndex,
	Int32	&outValue) const
{
	if (ioIndex < 0) {					// Safety check for negative items
		ioIndex = 0;					// Set to zero so index is valid
	}

	Boolean	cmdFound = false;
	if (ioIndex < mNumPvals) {		// Index is in range
										// Lookup value
		outValue = (*mPvals)[ioIndex++];
		cmdFound = true;
	}
	
	return cmdFound;
}


// ===========================================================================
//		 SetItemValue
// ===========================================================================
//	Set the value for a Menu item

	void
CPvalPopupMenu::SetItemValue(
	Int16		inIndex,
	Int32	inValue)
{
									// Do nothing if inIndex is out of range
	Int16	menuLength = ::CountMItems(mMacMenuH);
	if ((inIndex > 0)   &&  (inIndex <= menuLength)) {
		if (inIndex > mNumPvals) {
									// Setting value for item beyond end
									//   end of value list
			
									// Grow value list
			if (mPvals == nil) {
				mPvals = (Int32**) ::NewHandle(
											inIndex * sizeof(Int32));
				ThrowIfMemFail_(mPvals);
			} else {
				::SetHandleSize((Handle) mPvals,
								inIndex * sizeof(Int32));
				ThrowIfMemError_();
			}
			
			Int16	origNumPvals = mNumPvals;
			mNumPvals = inIndex;
			
			if (inIndex > origNumPvals + 1) {
			
				// There are items between the last original value
				// and the one to set. We must set the value for
				// these items to cmd_UseMenuItem.
				// Example:
				//		3 items in value list originally
				//		SetValue for item 6
				//		Items 4, 5, 6 previously had no entry in value list
				//			(value list was shorter than number of items
				//			in the Menu)
				//		Set value for items 4 and 5 to cmd_UseMenuItem
				
				Int16		newDefaults = inIndex - origNumPvals - 1;
				Int32	*cp = *mPvals + origNumPvals;
				do {
					*cp++ = cmd_UseMenuItem;
				} while (--newDefaults);
			}
		}
									// Store value for this item
		(*mPvals)[inIndex - 1] = inValue;
	}
}


// ===========================================================================
//		 InsertValue
// ===========================================================================
//	Insert an item with the specified text and value after
//	particular item
//
//	NOTE: This function does not support adding multiple menu items
//		using "Return" or "Semicolon" characters within inItemText

	void
CPvalPopupMenu::InsertValue(
	Str255		inItemText,
	Int32	inValue,
	Int16		inAfterItem)
{
									// Determine insertion index. It's
									//   usually one plus the after index.
	Int16	menuLength = ::CountMItems(mMacMenuH);
	Int16	itemToInsert = inAfterItem + 1;
	if (itemToInsert < 1) {	
		itemToInsert = 1;			// Insert at beginning of Menu
	} else if (itemToInsert > menuLength) {
									// Append to end of Menu
		itemToInsert = menuLength + 1;
	}
									// Put item in Mac Menu
	::InsertMenuItem(mMacMenuH, inItemText, itemToInsert - 1);

	if (itemToInsert <= mNumPvals) {
									// Inserting into middle of list
									// Value count increases by one
									//   and value table grows
		SetHandleSize((Handle) mPvals,
			(mNumPvals + 1) * sizeof(Int32));
		ThrowIfMemError_();
		mNumPvals++;
									// Shift up values beyond
									//    insertion point
		::BlockMoveData(*mPvals + itemToInsert - 1,
				  *mPvals + itemToInsert,
				  (mNumPvals - itemToInsert) * sizeof(Int32));
	}
									// Store value for inserted item
	SetItemValue(itemToInsert, inValue);
}


// ===========================================================================
//		 RemoveValue
// ===========================================================================
//	Remove the Menu item with the specified value

	void
CPvalPopupMenu::RemoveValue(Int32 inValue)
{
										// Lookup index for value
	Int16	itemToRemove = ItemFromPval(inValue);
	if (itemToRemove != 0) {			// Value found
		RemoveItem(itemToRemove);		// Remove associated menu item
	}
}


// ===========================================================================
//		 RemoveItem
// ===========================================================================
//	Remove the Menu item at the specified position

	void
CPvalPopupMenu::RemoveItem(Int16 inItemToRemove)
{
	SignalIf_(inItemToRemove <= 0);
	
										// Remove item from Mac Menu
	::DeleteMenuItem(mMacMenuH, inItemToRemove);
		
	if (inItemToRemove <= mNumPvals) {
										// Item to remove has a value
										// Shift down values above the
										//   item to remove
		::BlockMoveData(*mPvals + inItemToRemove,
				  *mPvals + inItemToRemove - 1,
				  (mNumPvals - inItemToRemove) * sizeof(Int32));

										// Decrease count of values and
										//   shrink value table
		::SetHandleSize((Handle) mPvals,
						--mNumPvals * sizeof(Int32));
	}
}


// ===========================================================================
//		 ItemIsEnabled
// ===========================================================================
//	Return whether an item at the specified position is enabled

	Boolean
CPvalPopupMenu::ItemIsEnabled(Int16 inIndex) const	// One-based index
{
	Boolean	enabled = true;
	if (inIndex < 32) {
		enabled = ((**mMacMenuH).enableFlags & (1L << inIndex)) != 0;
	}
	return enabled;
}


