// ===========================================================================
//	CTinselTestWindow.cp
// ===========================================================================


#include <QuickDraw.h>
#include <LDocument.h>

#include "RAVE.h"
#include "CTestInfo.h"
#include "CTinselTestWindow.h"
#include "S3Utilities.h"
#include "S3Alert.h"


const OSType TestWindow_FileType = 'PICT';
const OSType TestWindow_Creator  = 'ttxt';
const ResIDT	STRx_SavePICTAs		= 600;


CTinselTestWindow* CTinselTestWindow::sLeftWindow = nil;
CTinselTestWindow* CTinselTestWindow::sRightWindow = nil;

	
// ===========================================================================
//		 CreateTinselTestWindow
// ===========================================================================
//	Return a newly created Window object initialized from a PPob resource

	CTinselTestWindow*
CTinselTestWindow::CreateTinselTestWindow(
	ResIDT		inWindowID,
	LCommander	*inSuperCommander)
{
	SetDefaultCommander(inSuperCommander);

	CTinselTestWindow	*theWindow =
			(CTinselTestWindow*) UReanimator::ReadObjects('PPob', inWindowID);

	Assert_(theWindow);
	
	theWindow->FinishCreate();

	if (theWindow->HasAttribute(windAttr_ShowNew)) {
		theWindow->Show();
	}
	
	return theWindow;
}


// ===========================================================================
//		 CreateTinselTestWindowStream
// ===========================================================================
//	Return a newly created Window object initialized with data from a Stream

	CTinselTestWindow*
CTinselTestWindow::CreateTinselTestWindowStream(
	LStream	*inStream)
{
	return (new CTinselTestWindow(inStream));
}


// ===========================================================================
// Calculates the default positions for the test windows on a particular
// screen.
//
// inDevice: the device to position the windows on; if nil, the screen
//		with the deepest color is used
// outLeftRect: if not nil, the left window's position is returned here
// outRightRect: if not nil, the right window's position is returned here
//
// RETURNS: void
//

	void
CTinselTestWindow::GetDefaultGlobalRects(
	GDHandle inDevice, 
	Rect* outLeftRect, 
	Rect* outRightRect)
{
	Rect deviceRect;
	long deviceWidth, deviceHeight;
	long windowSize, windowSpacingH, windowSpacingV;
	
	if (inDevice == nil) {
			// the caller wants the window positions to be figured for the screen
			// with the deepest available bitdepth
		inDevice = S3GetDeepestScreen();
	}

	deviceRect = (**inDevice).gdRect;
	
	deviceWidth = (deviceRect.right - deviceRect.left);
	deviceHeight = (deviceRect.bottom - deviceRect.top);
	
		// calc the amount of screen covered by the two windows and divide
		// that in 2 to get the size of each window
	windowSize = deviceWidth * kWindowScreenPercentage / 2;
	
		// there are three empty areas horizontally (empty-window-empty-window-empty)
		// and two vertically (empty-window-empty)
	windowSpacingH = (deviceWidth - 2 * windowSize) / 3;
	windowSpacingV = (deviceHeight - windowSize) / 2;
	
	if (windowSpacingV < 0) {
			// the screen is much wider than it is high (unlikely), so we have to
			// make the windows smaller so that they fit vertically.
		windowSize = deviceHeight * kWindowScreenPercentage;

		windowSpacingH = (deviceWidth - 2 * windowSize) / 3;
		windowSpacingV = windowSize / 2;
	}
	
		// fill in the Rects with the bounds for the left and right windows
	if (outLeftRect != nil) {
		outLeftRect->left = deviceRect.left + windowSpacingH;
		outLeftRect->right = outLeftRect->left + windowSize;
		outLeftRect->top = deviceRect.top + windowSpacingV;
		outLeftRect->bottom = outLeftRect->top + windowSize;
	}
	
	if (outRightRect != nil) {
		outRightRect->right = deviceRect.right - windowSpacingH;
		outRightRect->left = outRightRect->right - windowSize;
		outRightRect->top = deviceRect.top + windowSpacingV;
		outRightRect->bottom = outRightRect->top + windowSize;
	}
}


// ===========================================================================
//		 CTinselTestWindow
// ===========================================================================
//	Default Constructor

CTinselTestWindow::CTinselTestWindow()
	: LWindow()
{
	InitTestWindow();
}


// ===========================================================================
//		 CTinselTestWindow(SWindowInfo&)
// ===========================================================================
//	Construct Window from the data in a struct

CTinselTestWindow::CTinselTestWindow(
	const SWindowInfo	&inWindowInfo)
	: LWindow(inWindowInfo)
{
	InitTestWindow();
}


// ===========================================================================
//		 CTinselTestWindow(ResIDT, Uint32, LCommander*)
// ===========================================================================
//	Construct a Window from a WIND Resource with the specified attributes
//	and SuperCommander

CTinselTestWindow::CTinselTestWindow(
	ResIDT		inWINDid,
	Uint32		inAttributes,
	LCommander	*inSuper)
		: LWindow(inWINDid, inAttributes, inSuper)
{
	InitTestWindow();
}


// ===========================================================================
//		 CTinselTestWindow(LStream*)
// ===========================================================================
//	Construct Window from the data in a stream

CTinselTestWindow::CTinselTestWindow(
	LStream	*inStream)
	: LWindow(inStream)
{
	InitTestWindow();
}


// ===========================================================================
	void
CTinselTestWindow::InitTestWindow()
{
		// save pointers to the left and right windows
	if (GetPaneID() == paneID_TinselTestWindow_Left) {
		sLeftWindow = this;
	} else if (GetPaneID() == paneID_TinselTestWindow_Right) {
		sRightWindow = this;
	}
	
	mGWorld = nil;
}


// ===========================================================================
//		 ListenToMessage
// ===========================================================================
	void
CTinselTestWindow::ListenToMessage(MessageT inMessage, void* ioParam)
{
	if (inMessage == msg_TestWindow_Move + GetUserCon()) {
			// we know that the message was addressed to this window because our
			// user constant (TestWindow_Left or TestWindow_Right) was added to it.
			// so save our old width and height before we move to the new position.
		long oldWidth = mUserBounds.right - mUserBounds.left;
		long oldHeight = mUserBounds.bottom - mUserBounds.top;
		
			// the CTestInfo sends the new bounds in the ioParam
		Rect newRect = *(Rect*) ioParam;
		
		if (BlockCompare(&newRect, &mUserBounds, sizeof(Rect)) == 0) {
				// our bounds haven't changed, so don't do anything
			return;
		}
		
			// move the window to the new bounds
		DoSetBounds(newRect);
		
		if (mUserBounds.right - mUserBounds.left != oldWidth ||
				mUserBounds.bottom - mUserBounds.top != oldHeight) {
					// the height and/or width of the window changed, so we want
					// to redraw the entire window, not just the part that got the
					// update event because otherwise the window might show
					// part of the old-sized GWorld and part of the new one.
			Draw(nil);
		}
	} else if (inMessage == msg_TestWindow_SetTitle + GetUserCon()) {
		Str255 currentWindowName;
		GetDescriptor(currentWindowName);
		
		if (!EqualString((ConstStr255Param) ioParam, currentWindowName, true, false)) {
				// our title has changed, so set it to the string sent by the CTestInfo
			SetDescriptor((ConstStr255Param) ioParam);
		}
	}
}


// ===========================================================================
//		 ClickInDrag
// ===========================================================================
//	Respond to commands
	void
CTinselTestWindow::ClickInDrag(
	const EventRecord	&inMacEvent)
{
		// remember where we are now
	Rect oldUserBounds = mUserBounds;
	
		// let LWindow handle the dragging
	LWindow::ClickInDrag(inMacEvent);

	if (!::EqualRect(&oldUserBounds, &mUserBounds)) {
			// we've been dragged by the user, so broadcast a message to the 
			// CTestInfo to let it know.  the message we use is msg_TestWindow_Drag +
			// our user constant, which lets the CTestInfo know which window was moved.
			// our user bounds is the value.
		BroadcastMessage(msg_TestWindow_Drag + GetUserCon(), (void*) &mUserBounds);
	}
}


// ===========================================================================
	void
CTinselTestWindow::ClickInGrow(
	const EventRecord	&inMacEvent)
{
		// remember where we are now
	Rect oldUserBounds = mUserBounds;
	
		// let LWindow handle the growing
	LWindow::ClickInGrow(inMacEvent);

	if (!::EqualRect(&oldUserBounds, &mUserBounds)) {
			// we've been resized by the user, so broadcast a message to the 
			// CTestInfo to let it know.  the same message gets sent for either
			// window because they both end up being the same size.  for the 
			// value, send a rect containing the offsets from the old 
			// bounds to the new.  the top and left values are 0 because they
			// haven't changed (although that's not true if you have something 
			// like Stretch installed, where you can drag the left and top sides
			// of a window, but that's not really worth worrying about).
		Rect offsetBounds = { 0, 0, 
													mUserBounds.bottom - oldUserBounds.bottom,
													mUserBounds.right - oldUserBounds.right };

			// since we've been resized, we need to redraw the entire window,
			// because the test output pixmap is always scaled to cover the 
			// whole thing
		Draw(nil);
		
		BroadcastMessage(msg_TestWindow_Grow, (void*) &offsetBounds);
	}
}


// ---------------------------------------------------------------------------
//		 DrawSizeBox
// ---------------------------------------------------------------------------
	void
CTinselTestWindow::DrawSizeBox()
{
	// we override this so that the size box isn't drawn
}


// ===========================================================================
//		 ObeyCommand
// ===========================================================================
//	Respond to commands
	Boolean
CTinselTestWindow::ObeyCommand(
	CommandT	inCommand,
	void		*ioParam)
{
	Boolean	cmdHandled = true;
	
	switch (inCommand) {
		case cmd_Save:
			AskSaveAs();
			break;

		default:
			cmdHandled = LWindow::ObeyCommand(inCommand, ioParam);
			break;
	}
	
	return cmdHandled;
}


// ---------------------------------------------------------------------------
//		 FindCommandStatus
// ---------------------------------------------------------------------------
//	Pass back status of a (menu) command
	void
CTinselTestWindow::FindCommandStatus(
	CommandT	inCommand,
	Boolean		&outEnabled,
	Boolean		&outUsesMark,
	Char16		&outMark,
	Str255		outName)
{
	outUsesMark = false;
	
	switch (inCommand) {
		case cmd_Print:
		case cmd_PrintOne:
		case cmd_PageSetup:
				// we don't handle printing
			outEnabled = false;
			break;
			
	  case cmd_Save:
				// the test window is not a document, but we can save its contents
				// to a file, so enable the Save command.  Save As... would be more
				// appropriate, since every time Save is selected, the user has to
				// give a new filename, but it's more convenient to have a command-key.
				// so enable Save, as long as our GWorld isn't empty.
			if (mGWorld != nil) {
				outEnabled = true;
			} else {
				outEnabled = false;
			}
			break;
			
	  case cmd_SaveAs:
				// for the test windows, Save and Save As... are equivalent, so 
				// disable Save As... and just leave Save.
			outEnabled = false;
			break;
			
		default:
			LWindow::FindCommandStatus(inCommand, outEnabled, outUsesMark,
								outMark, outName);
			break;
	}
}


// ===========================================================================
//		 AskSaveAs
// ===========================================================================
//	Ask the user to save a Document and give it a name
//
//	Returns false if the user cancels the operation
	Boolean
CTinselTestWindow::AskSaveAs()
{
		// set up the Save As prompt
	Str255 saveAsPrompt;
	::GetIndString(saveAsPrompt, STRx_SavePICTAs, 1);
	
		// create the default name for the PICT file by concatenating the
		// date, time, and Left|Right
	unsigned long seconds = 0;
	::GetDateTime(&seconds);
	
	LStr255 dateString;
	::DateString(seconds, shortDate, dateString, nil);
	
	LStr255 timeString;
	::TimeString(seconds, true, timeString, nil);

		// replace all the colons in the timeString with periods, since colons
		// are illegal in file names
	Uint8 colonPosition = 1;
	while ((colonPosition = timeString.Find(':', colonPosition)) != 0) {
		timeString.Replace(colonPosition, 1, '.');
	} // endwhile
	
	dateString += "\p ";
	dateString += timeString;
	
		// also add "Left" or "Right", depending on which window it is
	if (GetPaneID() == paneID_TinselTestWindow_Left) {
		dateString += "\p - Left";
	} else {
		dateString += "\p - Right";
	}

		// open the file dialog and get the user's reply
	StandardFileReply	reply;
	::UDesktop::Deactivate();
	::StandardPutFile(saveAsPrompt, dateString, &reply);
	::UDesktop::Activate();
	
	if (reply.sfGood) {
			// the user didn't cancel, so save the window's contents as a PICT file
		DoAESave(reply.sfFile, TestWindow_FileType);
	}
	
	return reply.sfGood;
}


// ===========================================================================
	void
CTinselTestWindow::DoAESave(
	FSSpec&	inFileSpec,
	OSType	inFileType)
{
	if (mGWorld == nil) {
			// our GWorld is empty, so we have nothing to save into the file
		return;
	}

		// get the window's current bounds
	Rect bounds;
	CalcLocalFrameRect(bounds);
	
	if (mGWorld->BeginDrawing() == false) {
			// the GWorld's memory can't be locked down, so we can't create
			// a picture to save to the file
		S3StopAlert(ALRT_StopAlert, AlertString_CantSavePICT, nil);
		return;
	}
	
		// get a handle to the GWorld
	GWorldPtr gWorldPtr = mGWorld->GetMacGWorld();
	
		// get a handle to the GWorld's pixmap
	PixMapHandle pixMapH = ::GetGWorldPixMap(gWorldPtr);

		// make sure the colors are set correctly for the CopyBits
	::ForeColor(blackColor);
	::BackColor(whiteColor);

		// open the PICT into which we're going to draw our GWorld
	PicHandle pictH = ::OpenPicture(&bounds);
	
		// copy the pixels our GWorld back into our GWorld, just so the data
		// will get saved into the PICT
	::CopyBits((BitMap*) *pixMapH, 
						 (BitMap*) *pixMapH,
						 &(gWorldPtr->portRect),
						 &(gWorldPtr->portRect),
						 srcCopy, nil);
						 
		// make sure we succeeded in copying
	Boolean success = (QDError() == noErr);
	
		// we've finished drawing into the PICT
	::ClosePicture();
	
		// unlock the GWorld's memory and restore the previous port
	mGWorld->EndDrawing();
	
	if (!success) {
			// we didn't copy the pixels correctly, so we can't save the PICT
		S3StopAlert(ALRT_StopAlert, AlertString_CantSavePICT, nil);
		return;
	}

		// Make a new file object
	LFileStream theFile(inFileSpec);		
	
		// Find proper file type
	OSType fileType = TestWindow_FileType;			
	
	if (inFileType != fileType_Default) {
		fileType = inFileType;
	}

		// Make a new file on disk and open it for writing
	theFile.CreateNewDataFile(TestWindow_Creator, fileType, 0);
	theFile.OpenDataFork(fsRdWrPerm);

// check for file opening errors here	
// should do try/catch around all of this

		// create an empty handle for the PICT comment header
	StClearHandleBlock pictHeaderH(kPICTFileHeaderSize);
	StHandleLocker(Handle(pictHeaderH));

// check for writing the right number of bits

		// write out the header
	theFile.WriteData(*pictHeaderH, kPICTFileHeaderSize);

	StHandleLocker(Handle(pictH));
	
		// write out the PICT
	theFile.WriteData(*pictH, ::GetHandleSize((Handle) pictH));

		// deallocate the PICT
	::KillPicture(pictH);
}


// ===========================================================================
	void
CTinselTestWindow::Clear()
{
		// save the current port and its attributes on the stack so
		// that they'll be restored when we exit the function
	CGrafPtr oldGrafPtr;
	GDHandle oldGDHandle;
	::GetGWorld(&oldGrafPtr, &oldGDHandle);
	StColorPortState savedPortState((GrafPtr) oldGrafPtr);
	
		// get our GrafPtr
	CGrafPtr theGrafPtr = (CGrafPtr) GetMacPort();
	
		// set the port to this window and then erase the rect to 
		// make sure the window is cleared before the next test
	::SetGWorld(theGrafPtr, ::GetGWorldDevice(theGrafPtr));
	::EraseRect(&(theGrafPtr->portRect));
	
		// we delete our GWorld so that it won't get redrawn the next time we
		// get an update event
	delete mGWorld;
	mGWorld = nil;
}


// ===========================================================================
	void
CTinselTestWindow::SaveTestOutput()
{
		// make sure this window is the current port
	FocusDraw();
	
		// get this window's CGrafPort and the device it's on
	CGrafPtr windowGrafPtr = (CGrafPtr) GetMacPort();
	GDHandle deviceH = S3GetDeviceFromRect(&mUserBounds, nil);

		// get rid of the current GWorld, since we're about to replace it
		// with a new image
	delete mGWorld;
	mGWorld = nil;
	
		// we want the GWorld to have the same depth as this window.  we have
		// to get the device from the window's bounds and then check the device's
		// pixmap to get the color depth because for some reason, just checking
		// the window's pixmap doesn't work.
	Int16 pixelDepth = (**((**deviceH).gdPMap)).pixelSize;
	
	try {
			// try to create a new offscreen GWorld in which to copy the window's 
			// current contents.  
		mGWorld = new LGWorld(windowGrafPtr->portRect, pixelDepth);
	}

		// we catch all errors here because that's the only way it seems to work
		// with the exceptions thrown by LGWorld
	catch(...) {
		try {
			if (mGWorld == nil) {
					// there isn't enough memory to allocate the GWorld in our heap, so
					// try putting it in temporary memory
				mGWorld = new LGWorld(windowGrafPtr->portRect, pixelDepth, useTempMem);
			}
		}
		
		catch(...) {
			if (mGWorld == nil) {
					// we still don't have enough memory for the GWorld, so just return,
					// since we can't save the test output without it
				return;
			}
		} 
	}
	
	GWorldPtr gWorldPtr = mGWorld->GetMacGWorld();
	
	if (mGWorld->BeginDrawing() == false) {
			// the GWorld's memory can't be locked down, so set the GWorld back
			// to whatever was saved in the mGWorld (when BeginDrawing was
			// called) and then delete it
		mGWorld->EndDrawing();
		delete mGWorld;
		mGWorld = nil;
		
		return;
	}
	
		// make sure the colors are set correctly for the CopyBits
	::ForeColor(blackColor);
	::BackColor(whiteColor);

		// copy the pixels inside our window to the offscreen GWorld
	::CopyBits((BitMap*) *(::GetGWorldPixMap(windowGrafPtr)), 
						 (BitMap*) *(::GetGWorldPixMap(gWorldPtr)),
						 &(windowGrafPtr->portRect),
						 &(gWorldPtr->portRect),
						 srcCopy, nil);
						 
		// make sure we succeeded in copying
	Boolean success = (QDError() == noErr);
	
		// unlock the GWorld's memory and restore the previous port
	mGWorld->EndDrawing();
	
	if (!success) {
			// we didn't copy the pixels correctly, so get rid of the GWorld
		delete mGWorld;
		mGWorld = nil;
	}
}


// ===========================================================================
	void
CTinselTestWindow::DrawSelf()
{
		// let the window do any setup it needs to before we draw into it
	LWindow::DrawSelf();

	if (mGWorld != nil) {
		Rect bounds;
		
		CalcLocalFrameRect(bounds);
		
			// make sure the colors are set correctly for the CopyBits
		::ForeColor(blackColor);
		::BackColor(whiteColor);
		
			// dither the output if we're on an 8-bit screen
		mGWorld->CopyImage(GetMacPort(), bounds, srcCopy + ditherCopy);
	}
}


