/*****************************************************************************\

	S3TestSystem.h

	Interfaces for a test tool to the Sneak3View test system registration,
	and to the Drive3D.h wrapper routines which route QA calls to the 
	mirrored engines.

	Copyright (C) 1995 Apple Computer, Inc.  All rights reserved.

\*****************************************************************************/
#pragma once

/*=========================================================================*\
	Includes
\*=========================================================================*/

#include "RAVE.h"
#include "S3TestInfo.h"

#ifdef __cplusplus
extern "C" {
#endif

/*=========================================================================*\
	Constants
\*=========================================================================*/

typedef enum {
	kS3Prereq_None = 0,
	kS3Prereq_Texture = 1,
	kS3Prereq_Bitmap = 2
} S3tTestPrereqs;

/*=========================================================================*\
	Types
\*=========================================================================*/

typedef S3tTestStatus (*S3tTestMethod)(
	S3tTestInfo				*testInfo);

/*=========================================================================*\
	Test System Routines
	
	Perhaps add a "test flag" to indicate how a test behaves? 
	(NeedCleanDrawContext, etc.)
\*=========================================================================*/

void S3Test_Init(
	void);

void S3Test_Register(
	const char				*testName,
	S3tTestPrereqs			testPrereqs,
	S3tTestMethod			testMethod);

void S3Test_Print(
	const char				*formatStr,
	...);

void S3Test_Exit(
	void);

#ifdef Debug_Sneak3view

#define ASSERT(x)	\
	((x) ? ((void) 0) : _S3AssertFunc(#x, __FILE__, __LINE__))

#else

#define ASSERT(x)

#endif

/* this function is defined in S3Debugging.cp */
void _S3AssertFunc(
	char*		inMessage,
	char*		inFile,
	long		inLine);

/*=========================================================================*\
	RAVE.h Wrapper routines 
	(use these to perform an action to both windows.)
\*=========================================================================*/

void S3Test_SetFloat (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagFloat				tag,				/* Tag of variable to set */
	float					newValue);			/* New value for variable */

void S3Test_SetInt (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagInt				tag,				/* Tag of variable to set */
	unsigned long			newValue);			/* New value for variable */

void S3Test_SetPtr (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagPtr				tag,				/* Tag of variable to set */
	const void				*newValue);			/* New value for variable */

float S3Test_GetLFloat (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagFloat				tag);				/* Tag of variable to get */

unsigned long S3Test_GetLInt (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagInt				tag);				/* Tag of variable to get */

void *S3Test_GetLPtr (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagPtr				tag);				/* Tag of variable to get */

float S3Test_GetRFloat (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagFloat				tag);				/* Tag of variable to get */

unsigned long S3Test_GetRInt (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagInt				tag);				/* Tag of variable to get */

void *S3Test_GetRPtr (
	S3tTestInfo				*testInfo,			/* Test environment */
	TQATagPtr				tag);				/* Tag of variable to get */

void S3Test_DrawPoint (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVGouraud		*v);				/* Vertex */

void S3Test_DrawLine (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVGouraud 		*v0,				/* Vertex 0 */
	const TQAVGouraud 		*v1);				/* Vertex 1 */

void S3Test_DrawTriGouraud (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVGouraud 		*v0,				/* Vertex 0 */
	const TQAVGouraud 		*v1,				/* Vertex 1 */
	const TQAVGouraud 		*v2,				/* Vertex 2 */
	unsigned long			flags);				/* Mask of kQATriFlags_xxx flags */

void S3Test_DrawTriTexture (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVTexture 		*v0,				/* Vertex 0 */
	const TQAVTexture 		*v1,				/* Vertex 1 */
	const TQAVTexture 		*v2,				/* Vertex 2 */
	unsigned long			flags);				/* Mask of kQATriFlags_xxx flags */

void S3Test_DrawVGouraud (
	S3tTestInfo				*testInfo,			/* Test environment */
	unsigned long			nVertices,			/* Number of vertices */
	TQAVertexMode			vertexMode,			/* One of kQAVertexMode_xxx enumerated values */
	const TQAVGouraud 		vertices[],			/* Array of vertices */
	const unsigned long		flags[]);			/* Array of per-triangle flags (or NULL) */

void S3Test_DrawVTexture (
	S3tTestInfo				*testInfo,			/* Test environment */
	unsigned long			nVertices,			/* Number of vertices */
	TQAVertexMode			vertexMode,			/* One of kQAVertexMode_xxx enumerated values */
	const TQAVTexture 		vertices[],			/* Array of vertices */
	const unsigned long		flags[]);			/* Array of per-triangle flags (or NULL) */

void S3Test_DrawBitmap (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQAVGouraud 		*v,					/* xyz, and (if a 1 bit/pixel bitmap) argb */
	TQABitmap				*leftBitmap,		/* Previously allocated by QABitmapNew() or nil to use bitmap stored in testInfo */
	TQABitmap				*rightBitmap);		/* Previously allocated by QABitmapNew() or nil to use bitmap stored in testInfo */

void S3Test_SubmitVerticesGouraud (
	S3tTestInfo				*testInfo,			/* Test environment */
	unsigned long			vertexCount,		/* Number of vertices being submitted */
	const TQAVGouraud 		*vertices);			/* Array of Gouraud vertices */

void S3Test_DrawTriMeshGouraud (
	S3tTestInfo					*testInfo,			/* Test environment */
	unsigned long				triangleCount,		/* Number of triangles being drawn */
	const TQAIndexedTriangle 	*triangles);		/* Array of vertex indexes */

void S3Test_SubmitVerticesTexture (
	S3tTestInfo				*testInfo,			/* Test environment */
	unsigned long			vertexCount,		/* Number of vertices being submitted */
	const TQAVTexture 		*vertices);			/* Array of Texture vertices */

void S3Test_DrawTriMeshTexture (
	S3tTestInfo					*testInfo,			/* Test environment */
	unsigned long				triangleCount,		/* Number of triangles being drawn */
	const TQAIndexedTriangle 	*triangles);		/* Array of vertex indexes */

void S3Test_RenderStart (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQARect			*dirtyRect);		/* Initial background image (or NULL) */

TQAError S3Test_RenderEnd (
	S3tTestInfo				*testInfo,			/* Test environment */
	const TQARect			*modifiedRect);		/* Minimum area to swap; NULL means whole buffer */

void S3Test_RenderAbort (
	S3tTestInfo				*testInfo);			/* Test environment */

void S3Test_Flush (
	S3tTestInfo				*testInfo);			/* Test environment */

void S3Test_Sync (
	S3tTestInfo				*testInfo);			/* Test environment */

#ifdef __cplusplus
}
#endif


