/*****************************************************************************\

	S3GameTest.c

	A game-like test scene
	
	Copyright (C) 1995 Apple Computer, Inc.  All rights reserved.

\*****************************************************************************/

/*=========================================================================*\
	Includes
\*=========================================================================*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <QuickDraw.h>

#include "S3TestSystem.h"
#include "S3TestInfo.h"
#include "S3TestUtil.h"
#include "S3Utilities.h"
#include "S3ColorTable.h"
#include "S3Alert.h"


/*=========================================================================*\
	Constants
\*=========================================================================*/
#define kTextureTriCount	2
#define kTextureTriGroupCount	10

#define kClipping_front		0.0
#define kClipping_back		10.0

#define kBackgroundZ		0.9
#define kBackgroundInvW		(1.0 / 20.0)
#define kWallsZ_front		0.1
#define kWallsZ_back		0.8
#define kWallsZeye_front	1.0
#define kWallsZeye_back		6.0
#define kWallsInvW_front	(1.0 / kWallsZeye_front)
#define kWallsInvW_back		(1.0 / kWallsZeye_back)

#define kBackground_U		4.0
#define kBackground_V		4.0
#define kWalls_U			5.0
#define kWalls_V			1.0
#define kBackWall_U			1.0
#define kBackWall_V			1.0
#define kFloor_U			1.0
#define kFloor_V			5.0

#define kColumn_left		0.25
#define kColumn_right		0.5
#define kColumn_top			-0.5
#define kColumn_bottom		0.5
#define kColumnZ_front		2.0
#define kColumnZ_back		2.25

#define kColumnFrontU_min	0.0
#define kColumnFrontU_max	0.25
#define kColumnFrontV_min	0.0
#define kColumnFrontV_max	1.0
#define kColumnSideU_min	0.5
#define kColumnSideU_max	0.75
#define kColumnSideV_min	0.0
#define kColumnSideV_max	1.0

#define kBox1_left			-0.5
#define kBox1_right			-0.1
#define kBox1_top			0.1
#define kBox1_bottom		0.5
#define kBox1_front			1.1
#define kBox1_back			1.9
#define kBox1_triCount		100

#define kTexture_background	0
#define kTexture_leftWall	1
#define kTexture_rightWall	2
#define kTexture_backWall	3
#define kTexture_floor		4
#define kTexture_column		5
#define kTexture_greenLava	6
#define kTexture_redLava	7
#define kTextureCount		8

#define kGunZ				0.001
#define kIntestines_Zeye	3.5
#define kIntestines_Yeye	0.5
#define kBitmap_gun			0
#define kBitmap_intestines	1
#define kBitmapCount		2


/*=========================================================================*\
	Structures
\*=========================================================================*/
typedef struct S3tTextureTri {
	TQAVTexture				v1;
	TQAVTexture				v2;
	TQAVTexture				v3;
	unsigned long			flags;
	short					texture;
} S3tTextureTri;

typedef struct S3tTextureTriGroup {
	S3tTextureTri					tris[10];
	struct S3tTextureTriGroup*		next;
} S3tTextureTriGroup;

typedef struct S3tTextureTriList {
	S3tTextureTriGroup*		head;
	S3tTextureTriGroup*		tail;
	Uint8					triIndex;
} S3tTextureTriList;

typedef struct S3tTexturePtrs {
	TQATexture*				left;
	TQATexture*				right;
} S3tTexturePtrs;

typedef struct S3tColorTablePtrs {
	TQAColorTable*			left;
	TQAColorTable*			right;
} S3tColorTablePtrs;

typedef struct S3tBitmapPtrs {
	TQABitmap*				left;
	TQABitmap*				right;
} S3tBitmapPtrs;


/*=========================================================================*\
	Globals
\*=========================================================================*/
S3tTestInfo*				gTestInfo;
float						gWindowWidth;
float						gWindowHeight;

unsigned long				gTriFlags;
S3tTextureTriList 			gTextureTriList;

S3tTexturePtrs				gTextures[kTextureCount];
S3Image						gTextureImages[kTextureCount];
S3tColorTablePtrs			gTextureColorTables[kTextureCount];
char*						gTextureFileNames[kTextureCount] = {
								"Stars.16.pict",
								"Dark brown wall.16.pict",
								"Green brown oval wall.16.pict",
								"Marathon door.16.pict",
								"Grey floor.16.pict",
								"Yellow black stripes.16.pict",
								"Green lava.16.pict",
								"Red lava.16.pict"
							};

S3tBitmapPtrs				gBitmaps[kBitmapCount];
S3Image						gBitmapImages[kBitmapCount];
S3tColorTablePtrs			gBitmapColorTables[kBitmapCount];
char*						gBitmapFileNames[kBitmapCount] = {
								"Assault rifle.cl8.pict",
								"Intestines.cl8.pict"
							};


/*=========================================================================*\
	Prototypes
\*=========================================================================*/
void S3GameTest_Register(void);
S3tTestStatus S3GameSceneEveryFormat(
	S3tTestInfo*			inTestInfo);
S3tTestStatus S3TestGameScene(
	S3tTestInfo				*testInfo);
Boolean S3LoadTextures(
	S3tTestInfo*			inTestInfo);
void S3UnloadTextures(
	S3tTestInfo*			inTestInfo);
Boolean S3LoadBitmaps(
	S3tTestInfo*			inTestInfo);
void S3UnloadBitmaps(
	S3tTestInfo*			inTestInfo);
void S3DrawBitmaps();

void S3BoundedRandomVTexture(
	TQAVTexture*			inVertex,
	float					inLeft,
	float					inRight,
	float					inTop,
	float					inBottom,
	float					inFront,
	float					inBack);
void S3RandomRGBDefaultHighlight(
	TQAVTexture*			inVertex);
float S3GetScreenX(
	float					inX,
	float					inZ);
float S3GetScreenY(
	float					inY,
	float					inZ);
float S3GetScreenZ(
	float					inZ);
	
void S3StoreBackground(
	float 					inWidth,
	float 					inHeight,
	float 					inDepth);
void S3StoreColumn();
void S3StoreWalls(
	float 					inBackWidth,
	float 					inBackHeight,
	float 					inFrontZ,
	float 					inBackZ);
void S3StoreRandomTris();
	
Boolean S3TextureTriList_New(
	S3tTextureTriList*		inTextureTriList);
void S3TextureTriList_Delete(
	S3tTextureTriList*		inTextureTriList);
Boolean S3TextureTriList_IsValid(
	S3tTextureTriList*		inTextureTriList);
Boolean S3TextureTriList_NewTextureTriGroup(
	S3tTextureTriList*		inTextureTriList);
Boolean S3TextureTriList_StoreTri(
	S3tTextureTriList*		inTextureTriList,
	TQAVTexture*			inVert1,
	TQAVTexture*			inVert2,
	TQAVTexture*			inVert3,
	unsigned long			inFlags,
	short					inTexture);
Boolean S3TextureTriList_StoreQuad(
	S3tTextureTriList*		inTextureTriList,
	TQAVTexture*			inVert1,
	TQAVTexture*			inVert2,
	TQAVTexture*			inVert3,
	TQAVTexture*			inVert4,
	unsigned long			inFlags,
	short					inTexture);
void S3TextureTriList_Draw(
	S3tTextureTriList*		inTextureTriList,
	S3tTestInfo*			inTestInfo);


/*=========================================================================*\
	S3GameTest_Register
\*=========================================================================*/
	void 
S3GameTest_Register(void)
{
	S3Test_Register("Game Scene",					kS3Prereq_None, S3TestGameScene);
	S3Test_Register("Game Scene (Every Format)",	kS3Prereq_None, S3GameSceneEveryFormat);
}

/*=========================================================================*\
	S3GameSceneEveryFormat
\*=========================================================================*/
	S3tTestStatus 
S3GameSceneEveryFormat(
	S3tTestInfo*			inTestInfo)
{
	TQAImagePixelType		format;
	TS3Error				s3Error;
	S3tTestStatus			testResult = kS3Status_Completed;
	
		// we want the textures to have an opaque alpha channel.  setting
		// make black transparent will force most of the textures to be 
		// opaque, since they don't have any black in them.  the striped
		// wall and the stars, however, will be transparent.
	s3Error = S3TestInfo_SetLong(inTestInfo, 
			kS3TagLong_TextureMakeBlackTransparent, true);
	
	if (s3Error != kS3Error_None) {
		S3Test_Print("*** Error (%d) occurred. ***\r", s3Error);
		return kS3Status_Failure;
	}
	
		// we don't want the color table textures to be transparent
	s3Error = S3TestInfo_SetLong(inTestInfo, 
			kS3TagLong_TextureUseTransparentIndex, false);
	
	if (s3Error != kS3Error_None) {
		S3Test_Print("*** Error (%d) occurred. ***\r", s3Error);
		return kS3Status_Failure;
	}
	
		// the bitmaps always have a transparent portion, so make them
		// transparent both when we're using color tables and also when we're not
	s3Error = S3TestInfo_SetLong(inTestInfo, 
			kS3TagLong_BitmapMakeBlackTransparent, true);
	
	if (s3Error != kS3Error_None) {
		S3Test_Print("*** Error (%d) occurred. ***\r", s3Error);
		return kS3Status_Failure;
	}
	
	s3Error = S3TestInfo_SetLong(inTestInfo, 
			kS3TagLong_BitmapUseTransparentIndex, true);
	
	if (s3Error != kS3Error_None) {
		S3Test_Print("*** Error (%d) occurred. ***\r", s3Error);
		return kS3Status_Failure;
	}
	
	for (format = kQAPixel_RGB16; format <= kQAPixel_CL8; format++) {
			// set the texture and bitmap formats to the same type
		s3Error = S3TestInfo_SetLong(inTestInfo, kS3TagLong_TextureFormat, format);
		
		if (s3Error != kS3Error_None) {
			S3Test_Print("*** Error (%d) occurred. ***\r", s3Error);
			return kS3Status_Failure;
		}
		
		s3Error = S3TestInfo_SetLong(inTestInfo, kS3TagLong_BitmapFormat, format);
		
		if (s3Error != kS3Error_None) {
			S3Test_Print("*** Error (%d) occurred. ***\r", s3Error);
			return kS3Status_Failure;
		}

			// let the user know which format we're currently using
		S3Test_Print("Texture Format: %d\rBitmap Format:  %d\r\r",
				format, format);
				
			// now run the Game Scene using this format
		testResult = S3TestInfo_RunTest(inTestInfo, "Game Scene");
		
		S3Test_Print("Test Result: %d\r\r\r", testResult);

			// wait for user input before going on to next test
		if (S3TestInfo_AskToContinue(inTestInfo) == false) {
			return kS3Status_Cancelled;
		}
	} // endfor

		// now set the texture and bitmap formats to RGB32 for the texture
		// and Alpha1 for the bitmap.  this is the only combination of 
		// formats supported by the Apple hardware card.
	s3Error = S3TestInfo_SetLong(inTestInfo, kS3TagLong_TextureFormat, 
			kQAPixel_RGB32);
	
	if (s3Error != kS3Error_None) {
		S3Test_Print("*** Error (%d) occurred. ***\r", s3Error);
		return kS3Status_Failure;
	}
	
	s3Error = S3TestInfo_SetLong(inTestInfo, kS3TagLong_BitmapFormat, 
			kQAPixel_Alpha1);
	
	if (s3Error != kS3Error_None) {
		S3Test_Print("*** Error (%d) occurred. ***\r", s3Error);
		return kS3Status_Failure;
	}

		// let the user know which format we're currently using
	S3Test_Print("Texture Format: %d\rBitmap Format:  %d\r\r",
			kQAPixel_RGB32, kQAPixel_Alpha1);
			
		// now run the Game Scene using this format
	testResult = S3TestInfo_RunTest(inTestInfo, "Game Scene");
	
	S3Test_Print("Test Result: %d\r\r\r", testResult);

	return kS3Status_Completed;
}

/*=========================================================================*\
	S3TestGameScene
\*=========================================================================*/
	S3tTestStatus 
S3TestGameScene(
	S3tTestInfo*			testInfo)
{
	char					drawBitmapsFlagString[4] = "";
	Boolean					drawBitmaps = true;
	
	if (S3TextureTriList_New(&gTextureTriList) == false) {
			// there's not enough memory to store the tris, so we can't
			// run the test.  print an error message and return failure.
		S3Test_Print("*** There is not enough memory to run this test.   ***\r"
					 "*** Try increasing Sneak3view's memory allocation. ***\r\r");
					 
		return kS3Status_Failure;
	}
	
		// check to see if "bitmaps:off" has been passed in the arg string
	sscanf(S3TestInfo_GetArgString(testInfo), "bitmaps:%3s", &drawBitmapsFlagString);
	
		// we'll draw the bitmaps as long as the %s above didn't match "off"
	drawBitmaps = (strcmp(drawBitmapsFlagString, "off") != 0);
	
	gTestInfo = testInfo;
	
		// all the triangles are front-facing, so use "none"
		// for the flags parameter
	gTriFlags = kQATriFlags_None;
	
		// set up the window size globals
	gWindowWidth = S3TestInfo_GetWidth(testInfo);
	gWindowHeight = S3TestInfo_GetHeight(testInfo);

	if (S3LoadTextures(testInfo) == false) {
			// there was a problem while loading the textures, so dispose
			// of them, print an error message and return failure
		S3UnloadTextures(testInfo);

		S3Test_Print("*** An error occurred while loading the     ***\r"
					 "*** textures, so the test could not be run. ***\r\r");

		return kS3Status_Failure;
	}
	
	if (drawBitmaps && S3LoadBitmaps(testInfo) == false) {
			// there was a problem while loading the bitmaps, so dispose
			// of them, print an error message and return failure
		S3UnloadBitmaps(testInfo);

		S3Test_Print("*** An error occurred while loading the    ***\r"
					 "*** bitmaps, so the test could not be run. ***\r\r");

		return kS3Status_Failure;
	}
	
	S3StoreBackground(gWindowWidth, gWindowHeight, kBackgroundZ);
	
	S3StoreWalls(
		gWindowWidth * (1.0 / 6.0), gWindowHeight * (1.0 / 6.0), 
		kWallsZ_front, kWallsZ_back);

	S3StoreColumn();
	
	S3StoreRandomTris();
		
	if (!S3TextureTriList_IsValid(&gTextureTriList)) {
			// at some point we ran out of memory in which to store new
			// tris, so we can't run the test.  deallocate any existing
			// memory, print an error message and return failure.
		S3TextureTriList_Delete(&gTextureTriList);
		S3UnloadTextures(testInfo);

		S3Test_Print("*** There is not enough memory to run this test.   ***\r"
					 "*** Try increasing Sneak3view's memory allocation. ***\r\r");

		return kS3Status_Failure;
	}
	
		// render all the triangles we've stored in the tri list
	S3Test_RenderStart(testInfo, NULL);

	S3TextureTriList_Draw(&gTextureTriList, testInfo);

	if (drawBitmaps) {
		S3DrawBitmaps();
	}
	
	S3Test_RenderEnd(testInfo, NULL);
	
		// wait for the engines to catch up and draw everything
	S3Test_Sync(testInfo);
		
		// clean up
	S3TextureTriList_Delete(&gTextureTriList);
	S3UnloadTextures(testInfo);
	
	if (drawBitmaps) {
		S3UnloadBitmaps(testInfo);
	}
	
	return kS3Status_Completed;
}


/*=========================================================================*\
	S3LoadTextures
\*=========================================================================*/
	Boolean
S3LoadTextures(
	S3tTestInfo*			inTestInfo)
{
	TQAEngine* 				leftEngine = S3TestInfo_GetLEngine(inTestInfo);
	TQAEngine* 				rightEngine = S3TestInfo_GetREngine(inTestInfo);
	Boolean 				makeMipMap = (S3TestInfo_GetTextureFlags(inTestInfo) & kQATexture_Mipmap);
	Boolean 				makeBlackTransparent = S3TestInfo_GetTextureMakeBlackTransparent(inTestInfo);
	Boolean 				detachTextures = S3TestInfo_GetDetachTexture(inTestInfo);
	unsigned long			textureFlags = S3TestInfo_GetTextureFlags(inTestInfo);
	TQAImagePixelType		textureFormat = S3TestInfo_GetTextureFormat(inTestInfo);
	TQAColorTableType		colorTableType = S3TestInfo_GetTextureColorTableType(inTestInfo);
	unsigned long			useTransparentIndex = S3TestInfo_GetTextureUseTransparentIndex(inTestInfo);
	PicHandle				pictH = nil;
	Boolean					foundError = false;
	TQAError				qaError = kQANoErr;
	short					i;
	
	for (i = 0; i < kTextureCount; i++) {
		if (pictH != nil) {
				// S3GetNearbyPICT returns a handle which we are responsible
				// for disposing of when we're done
			KillPicture(pictH);
			pictH = nil;
		}
		
		pictH = S3GetNearbyPICT(inTestInfo, gTextureFileNames[i]);
		
		if (pictH == nil) {
				// we couldn't find the PICT file, so record an error and
				// then try the next one
			foundError = true;
			
			S3Test_Print("*** Could not find PICT file \"%s\" ***\r", gTextureFileNames[i]);
			
			continue;
		}
	
		gTextureImages[i] = S3Image_New(pictH, textureFormat, makeBlackTransparent, makeMipMap);
		
		if (gTextureImages[i] == nil) {
				// for some reason we couldn't create a S3Image (probably
				// not enough memory), so print an error and break out
				// of the loop.  we break instead of continuing since if
				// there's not enough memory, there's no point going on.
			foundError = true;
			
			S3Test_Print("*** Could not load texture from \"%s\" ***\r", gTextureFileNames[i]);
			
			break;
		}
		
		if (makeMipMap && !S3Image_IsValidMipMap(gTextureImages[i])) {
				// the texture is supposed to be mipmapped, but S3Image couldn't
				// mip map it for some reason, so print a message and break,
				// since if we can't mip map this one, we probably can't do
				// any of them
			foundError = true;
			
			S3Test_Print("*** Could not mip map \"%s\" ***\r", gTextureFileNames[i]);
			
			break;
		}
		
			// try to create a texture for the left engine
		qaError = QATextureNew(
			leftEngine,
			textureFlags,
			textureFormat,
			S3Image_GetImages(gTextureImages[i]),
			&gTextures[i].left);
			
		if (qaError != kQANoErr) {
				// for some reason the engine couldn't create a texture (probably
				// not enough memory), so print an error and break out
				// of the loop.  we break instead of continuing since if
				// there's not enough memory, there's no point going on.
			foundError = true;
			
			S3Test_Print("*** Left engine could not create texture from \"%s\" ***\r", gTextureFileNames[i]);
			
			break;
		}
		
			// try to create a texture for the right engine
		qaError = QATextureNew(
			rightEngine,
			textureFlags,
			textureFormat,
			S3Image_GetImages(gTextureImages[i]),
			&gTextures[i].right);
			
		if (qaError != kQANoErr) {
				// for some reason the engine couldn't create a texture (probably
				// not enough memory), so print an error and break out
				// of the loop.  we break instead of continuing since if
				// there's not enough memory, there's no point going on.
			foundError = true;
			
			S3Test_Print("*** Right engine could not create texture from \"%s\" ***\r", gTextureFileNames[i]);
			
			break;
		}
		
		if (textureFormat == kQAPixel_CL4 || 
				textureFormat == kQAPixel_CL8) {
			CTabHandle				colorTableH = nil;
			S3tColorTableP			colorTableP = nil;
			
				// extract the CTabHandle from the PICT the texture came from
			colorTableH = S3Image_GetColorTableHandle(gTextureImages[i]);
					
			if (colorTableH == nil) {
					// for some reason we couldn't extract the color table
					// handle from the S3Image, so we can't create a new
					// color table for the texture, which means we won't be
					// able to draw it.  so flag an error and break the loop.
				foundError = true;
				
				S3Test_Print("*** Could not extract color table from texture \"%s\" ***\r", gTextureFileNames[i]);
			
				break;
			}
			
				// turn the CTabHandle into a TQAColorTable
			colorTableP = S3ColorTable_New(colorTableH, colorTableType);
					
			if (colorTableP == nil) {
					// for some reason we couldn't create a new color table,
					// which means we won't be able to draw it.  so flag an 
					// error and break the loop.
				foundError = true;
				
				S3Test_Print("*** Could not create color table for texture \"%s\" ***\r", gTextureFileNames[i]);
			
				break;
			}
			
				// create a new color table in the left engine
			qaError = QAColorTableNew(
					leftEngine, colorTableType, 
					S3ColorTable_GetData(colorTableP),
					useTransparentIndex, &gTextureColorTables[i].left);
					
			if (qaError != kQANoErr) {
					// we couldn't create a color table in this engine (perhaps
					// it doesn't support CL textures), so print an error 
					// message and break out of the loop, since we can't continue
					// if one of the engines can't create color tables.
				foundError = true;
				
				S3Test_Print("*** Left engine could not create a color table for \"%s\" ***\r", gTextureFileNames[i]);
				
				break;
			}
	
				// bind this color table to the left texture
			qaError = QATextureBindColorTable(
					leftEngine, gTextures[i].left, gTextureColorTables[i].left);
			
			if (qaError != kQANoErr) {
					// we couldn't bind this color table to the texture in 
					// this engine, so print an error message and break out of 
					// the loop, since we won't be able to draw this texture
				foundError = true;
				
				S3Test_Print("*** Left engine could not bind a color table for \"%s\" ***\r", gTextureFileNames[i]);
				
				break;
			}
			
				// create a new color table in the right engine
			qaError = QAColorTableNew(
					rightEngine, colorTableType, 
					S3ColorTable_GetData(colorTableP),
					useTransparentIndex, &gTextureColorTables[i].right);
					
			if (qaError != kQANoErr) {
					// we couldn't create a color table in this engine (perhaps
					// it doesn't support CL textures), so print an error 
					// message and break out of the loop, since we can't continue
					// if one of the engines can't create color tables.
				foundError = true;
				
				S3Test_Print("*** Right engine could not create a color table for \"%s\" ***\r", gTextureFileNames[i]);
				
				break;
			}
	
				// bind this color table to the right texture
			qaError = QATextureBindColorTable(
					rightEngine, gTextures[i].right, gTextureColorTables[i].right);
			
			if (qaError != kQANoErr) {
					// we couldn't bind this color table to the texture in 
					// this engine, so print an error message and break out of 
					// the loop, since we won't be able to draw this texture
				foundError = true;
				
				S3Test_Print("*** Right engine could not bind a color table for \"%s\" ***\r", gTextureFileNames[i]);
				
				break;
			}
		}
		
		if (detachTextures) {
			Boolean					leftTextureDetached = false;
			
				// try to detach texture from the left engine
			qaError = QATextureDetach(
					leftEngine,
					gTextures[i].left);

			if (qaError != kQANoErr) {
					// the engine couldn't detach this texture (probably
					// not enough memory), so print a message.  we don't 
					// really care if it gets detached or not, so we don't
					// break out of the loop or flag an error.
				S3Test_Print("*** Left engine could not detach texture \"%s\" ***\r", gTextureFileNames[i]);
			} else {
					// remember that the texture was successfully detached.
					// when the right texture gets detached, it will check 
					// this flag to make sure it's okay to dispose of the
					// image data which the two textures share.
				leftTextureDetached = true;
			}

				// try to detach texture from the right engine
			qaError = QATextureDetach(
					rightEngine,
					gTextures[i].right);

			if (qaError != kQANoErr) {
					// the engine couldn't detach this texture (probably
					// not enough memory), so print a message.  we don't 
					// really care if it gets detached or not, so we don't
					// break out of the loop or flag an error.
				S3Test_Print("*** Right engine could not detach texture \"%s\" ***\r", gTextureFileNames[i]);
			} else if (leftTextureDetached) {
					// both the left and right textures were successfully 
					// detached, so we can safely dispose of the image 
					// data we used to create them
				S3Image_Delete(gTextureImages[i]);
				gTextureImages[i] = nil;
			}
		}
	} // endfor

		// clean up our allocated memory
	if (pictH != nil) {
		KillPicture(pictH);
	}
	
	return (foundError == false);
}


/*=========================================================================*\
	S3UnloadTextures
\*=========================================================================*/
	void
S3UnloadTextures(
	S3tTestInfo*			inTestInfo)
{
	TQAEngine* 				leftEngine = S3TestInfo_GetLEngine(inTestInfo);
	TQAEngine* 				rightEngine = S3TestInfo_GetREngine(inTestInfo);
	short					i;

	ASSERT(leftEngine);
	ASSERT(rightEngine);
	
	for (i = 0; i < kTextureCount; i++) {
			// dispose of the left and right textures
		if (gTextures[i].left != nil) {
				// dispose of the texture in this slot
			QATextureDelete(leftEngine, gTextures[i].left);
			gTextures[i].left = nil;
		}

		if (gTextures[i].right != nil) {
				// dispose of the texture in this slot
			QATextureDelete(rightEngine, gTextures[i].right);
			gTextures[i].right = nil;
		}
		
			// dispose of the left and right color tables
		if (gTextureColorTables[i].left != nil) {
			QAColorTableDelete(leftEngine, gTextureColorTables[i].left);
			gTextureColorTables[i].left = nil;
		}

		if (gTextureColorTables[i].right != nil) {
			QAColorTableDelete(rightEngine, gTextureColorTables[i].right);
			gTextureColorTables[i].right = nil;
		}
		
			// dispose of the image which is referenced by the texture
		if (gTextureImages[i] != nil) {
			S3Image_Delete(gTextureImages[i]);
			gTextureImages[i] = nil;
		}
	}
}


/*=========================================================================*\
	S3LoadBitmaps
\*=========================================================================*/
	Boolean
S3LoadBitmaps(
	S3tTestInfo*			inTestInfo)
{
	TQAEngine* 				leftEngine = S3TestInfo_GetLEngine(inTestInfo);
	TQAEngine* 				rightEngine = S3TestInfo_GetREngine(inTestInfo);
	Boolean 				makeBlackTransparent = S3TestInfo_GetBitmapMakeBlackTransparent(inTestInfo);
	Boolean 				detachBitmaps = S3TestInfo_GetDetachBitmap(inTestInfo);
	TQAImagePixelType		bitmapFormat = S3TestInfo_GetBitmapFormat(inTestInfo);
	unsigned long			bitmapFlags = S3TestInfo_GetBitmapFlags(inTestInfo);
	TQAColorTableType		colorTableType = S3TestInfo_GetBitmapColorTableType(inTestInfo);
	unsigned long			useTransparentIndex = S3TestInfo_GetBitmapUseTransparentIndex(inTestInfo);
	PicHandle				pictH = nil;
	Boolean					foundError = false;
	TQAError				qaError = kQANoErr;
	short					i;
	
	for (i = 0; i < kBitmapCount; i++) {
		if (pictH != nil) {
				// S3GetNearbyPICT returns a handle which we are responsible
				// for disposing of when we're done
			KillPicture(pictH);
			pictH = nil;
		}
		
		pictH = S3GetNearbyPICT(inTestInfo, gBitmapFileNames[i]);
		
		if (pictH == nil) {
				// we couldn't find the PICT file, so record an error and
				// then try the next one
			foundError = true;
			
			S3Test_Print("*** Could not find PICT file \"%s\" ***\r", gBitmapFileNames[i]);
			
			continue;
		}
	
			// the false is for the makeMipMap parameter.  bitmaps are not mipmapped.
		gBitmapImages[i] = S3Image_New(pictH, bitmapFormat, makeBlackTransparent, false);
		
		if (gBitmapImages[i] == nil) {
				// for some reason we couldn't create a S3Image (probably
				// not enough memory), so print an error and break out
				// of the loop.  we break instead of continuing since if
				// there's not enough memory, there's no point going on.
			foundError = true;
			
			S3Test_Print("*** Could not load bitmap from \"%s\" ***\r", gBitmapFileNames[i]);
			
			break;
		}
		
			// try to create a bitmap for the left engine
		qaError = QABitmapNew(
			leftEngine,
			bitmapFlags,
			bitmapFormat,
			S3Image_GetImages(gBitmapImages[i]),
			&gBitmaps[i].left);
			
		if (qaError != kQANoErr) {
				// for some reason the engine couldn't create a bitmap (probably
				// not enough memory), so print an error and break out
				// of the loop.  we break instead of continuing since if
				// there's not enough memory, there's no point going on.
			foundError = true;
			
			S3Test_Print("*** Left engine could not create bitmap from \"%s\" ***\r", gBitmapFileNames[i]);
			
			break;
		}
		
			// try to create a Bitmap for the right engine
		qaError = QABitmapNew(
			rightEngine,
			bitmapFlags,
			bitmapFormat,
			S3Image_GetImages(gBitmapImages[i]),
			&gBitmaps[i].right);
			
		if (qaError != kQANoErr) {
				// for some reason the engine couldn't create a bitmap (probably
				// not enough memory), so print an error and break out
				// of the loop.  we break instead of continuing since if
				// there's not enough memory, there's no point going on.
			foundError = true;
			
			S3Test_Print("*** Right engine could not create bitmap from \"%s\" ***\r", gBitmapFileNames[i]);
			
			break;
		}
		
		if (bitmapFormat == kQAPixel_CL4 || 
				bitmapFormat == kQAPixel_CL8) {
			CTabHandle				colorTableH = nil;
			S3tColorTableP			colorTableP = nil;
			
				// extract the CTabHandle from the PICT the bitmap came from
			colorTableH = S3Image_GetColorTableHandle(gBitmapImages[i]);
					
			if (colorTableH == nil) {
					// for some reason we couldn't extract the color table
					// handle from the S3Image, so we can't create a new
					// color table for the bitmap, which means we won't be
					// able to draw it.  so flag an error and break the loop.
				foundError = true;
				
				S3Test_Print("*** Could not extract color table from bitmap \"%s\" ***\r", gBitmapFileNames[i]);
			
				break;
			}
			
				// turn the CTabHandle into a TQAColorTable
			colorTableP = S3ColorTable_New(colorTableH, colorTableType);
					
			if (colorTableP == nil) {
					// for some reason we couldn't create a new color table,
					// which means we won't be able to draw it.  so flag an 
					// error and break the loop.
				foundError = true;
				
				S3Test_Print("*** Could not create color table for bitmap \"%s\" ***\r", gBitmapFileNames[i]);
			
				break;
			}
			
				// create a new color table in the left engine
			qaError = QAColorTableNew(
					leftEngine, colorTableType, 
					S3ColorTable_GetData(colorTableP),
					useTransparentIndex, &gBitmapColorTables[i].left);
					
			if (qaError != kQANoErr) {
					// we couldn't create a color table in this engine (perhaps
					// it doesn't support CL bitmaps), so print an error 
					// message and break out of the loop, since we can't continue
					// if one of the engines can't create color tables.
				foundError = true;
				
				S3Test_Print("*** Left engine could not create a color table for \"%s\" ***\r", gBitmapFileNames[i]);
				
				break;
			}
	
				// bind this color table to the left bitmap
			qaError = QABitmapBindColorTable(
					leftEngine, gBitmaps[i].left, gBitmapColorTables[i].left);
			
			if (qaError != kQANoErr) {
					// we couldn't bind this color table to the bitmap in 
					// this engine, so print an error message and break out of 
					// the loop, since we won't be able to draw this bitmap
				foundError = true;
				
				S3Test_Print("*** Left engine could not bind a color table for \"%s\" ***\r", gBitmapFileNames[i]);
				
				break;
			}
			
				// create a new color table in the right engine
			qaError = QAColorTableNew(
					rightEngine, colorTableType, 
					S3ColorTable_GetData(colorTableP),
					useTransparentIndex, &gBitmapColorTables[i].right);
					
			if (qaError != kQANoErr) {
					// we couldn't create a color table in this engine (perhaps
					// it doesn't support CL bitmaps), so print an error 
					// message and break out of the loop, since we can't continue
					// if one of the engines can't create color tables.
				foundError = true;
				
				S3Test_Print("*** Right engine could not create a color table for \"%s\" ***\r", gBitmapFileNames[i]);
				
				break;
			}
	
				// bind this color table to the right bitmap
			qaError = QABitmapBindColorTable(
					rightEngine, gBitmaps[i].right, gBitmapColorTables[i].right);
			
			if (qaError != kQANoErr) {
					// we couldn't bind this color table to the bitmap in 
					// this engine, so print an error message and break out of 
					// the loop, since we won't be able to draw this bitmap
				foundError = true;
				
				S3Test_Print("*** Right engine could not bind a color table for \"%s\" ***\r", gBitmapFileNames[i]);
				
				break;
			}
		}
		
		if (detachBitmaps) {
				// now that both bitmaps have been sucessfully created, we can
				// try to detach them
			Boolean					leftBitmapDetached = false;
			
				// try to detach Bitmap from the left engine
			qaError = QABitmapDetach(
				leftEngine,
				gBitmaps[i].left);

			if (qaError != kQANoErr) {
					// the engine couldn't detach this Bitmap (probably
					// not enough memory), so print a message.  we don't 
					// really care if it gets detached or not, so we don't
					// break out of the loop or flag an error.
				S3Test_Print("*** Left engine could not detach bitmap \"%s\" ***\r", gBitmapFileNames[i]);
			} else {
					// remember that the bitmap was successfully detached.
					// when the right bitmap gets detached, it will check 
					// this flag to make sure it's okay to dispose of the
					// image data which the two bitmaps share.
				leftBitmapDetached = true;
			}

				// try to detach Bitmap from the right engine
			qaError = QABitmapDetach(
				rightEngine,
				gBitmaps[i].right);

			if (qaError != kQANoErr) {
					// the engine couldn't detach this Bitmap (probably
					// not enough memory), so print a message.  we don't 
					// really care if it gets detached or not, so we don't
					// break out of the loop or flag an error.
				S3Test_Print("*** Right engine could not detach bitmap \"%s\" ***\r", gBitmapFileNames[i]);
			} else if (leftBitmapDetached) {
					// both the left and right bitmaps were successfully 
					// detached, so we can safely dispose of the image 
					// data we used to create them
				S3Image_Delete(gBitmapImages[i]);
				gBitmapImages[i] = nil;
			}
		}
	} // endfor

		// clean up our allocated memory
	if (pictH != nil) {
		KillPicture(pictH);
	}
	
	return (foundError == false);
}


/*=========================================================================*\
	S3UnloadBitmaps
\*=========================================================================*/
	void
S3UnloadBitmaps(
	S3tTestInfo*			inTestInfo)
{
	TQAEngine* 				leftEngine = S3TestInfo_GetLEngine(inTestInfo);
	TQAEngine* 				rightEngine = S3TestInfo_GetREngine(inTestInfo);
	short					i;

	ASSERT(leftEngine);
	ASSERT(rightEngine);
	
	for (i = 0; i < kBitmapCount; i++) {
			// dispose of the left and right bitmaps
		if (gBitmaps[i].left != nil) {
			QABitmapDelete(leftEngine, gBitmaps[i].left);
			gBitmaps[i].left = nil;
		}

		if (gBitmaps[i].right != nil) {
			QABitmapDelete(rightEngine, gBitmaps[i].right);
			gBitmaps[i].right = nil;
		}
		
			// dispose of the left and right color tables
		if (gBitmapColorTables[i].left != nil) {
			QAColorTableDelete(leftEngine, gBitmapColorTables[i].left);
			gBitmapColorTables[i].left = nil;
		}

		if (gBitmapColorTables[i].right != nil) {
			QAColorTableDelete(rightEngine, gBitmapColorTables[i].right);
			gBitmapColorTables[i].right = nil;
		}
		
			// dispose of the image which is referenced by the Bitmap
		if (gBitmapImages[i] != nil) {
			S3Image_Delete(gBitmapImages[i]);
			gBitmapImages[i] = nil;
		}
	}
}


/*=========================================================================*\
	S3DrawBitmaps
\*=========================================================================*/
	void 
S3DrawBitmaps()
{
		// eyeZ / eyeZrange == screenZ / screenZrange
		// solve for screenZ to get a Z value between 0 and 1 for intestines
	float					eyeZrange = kWallsZeye_back - kWallsZeye_front;
	float					screenZrange = kWallsZ_back - kWallsZ_front;
	float					intestinesZ = (kIntestines_Zeye / eyeZrange) * screenZrange;
	TQAVGouraud				v;
	
		// set the color and alpha to white
	v.r = v.g = v.b = v.a = 1.0;
	
		// center the gun bitmap horizontally and place it at the bottom of
		// the window
	v.x = (gWindowWidth - S3Image_GetWidth(gBitmapImages[kBitmap_gun])) / 2.0;
	v.y = gWindowHeight - S3Image_GetHeight(gBitmapImages[kBitmap_gun]);
	v.z = kGunZ;
	v.invW = 1.0 / kGunZ;
	
		// draw the bitmap in both contexts
	S3Test_DrawBitmap(gTestInfo, &v, gBitmaps[kBitmap_gun].left, gBitmaps[kBitmap_gun].right);
	
		// center the intestines bitmap horizontally and place it so its 
		// bottom touches the floor at its Zeye coord, which affects its
		// vertical coord on the screen.
	v.x = (gWindowWidth - S3Image_GetWidth(gBitmapImages[kBitmap_intestines])) / 2.0;
	v.y = S3GetScreenY(kIntestines_Yeye, kIntestines_Zeye) - S3Image_GetHeight(gBitmapImages[kBitmap_intestines]);
	v.z = intestinesZ;
	v.invW = 1.0 / intestinesZ;
	
		// draw the bitmap in both contexts
	S3Test_DrawBitmap(gTestInfo, &v, gBitmaps[kBitmap_intestines].left, gBitmaps[kBitmap_intestines].right);
}


/*=========================================================================*\
	S3GetScreenX
\*=========================================================================*/
	float 
S3GetScreenX(
	float					inX,
	float					inZ)
{
	return (gWindowWidth * (0.5 + (inX / inZ)));
}


/*=========================================================================*\
	S3GetScreenY
\*=========================================================================*/
	float 
S3GetScreenY(
	float					inY,
	float					inZ)
{
		// because screen Y values diminish towards the top of the screen,
		// eye Y (inY) values are assumed to be positive towards the bottom 
		// of the screen and negative towards the top.
	return (gWindowHeight * (0.5 + (inY / inZ)));
}


/*=========================================================================*\
	S3GetScreenZ
\*=========================================================================*/
	float 
S3GetScreenZ(
	float					inZ)
{	
		// inZ is assumed to be less than kClipping_back.  this maps the
		// the eye-space Z to a value between 0 and 1
	return (inZ / kClipping_back);
}


/*=========================================================================*\
	S3BoundedRandomVTexture
\*=========================================================================*/
	void 
S3BoundedRandomVTexture(
	TQAVTexture*			inVertex,
	float					inLeft,
	float					inRight,
	float					inTop,
	float					inBottom,
	float					inFront,
	float					inBack)
{
	float					rangeX = inRight - inLeft;
	float					rangeY = inBottom - inTop;
	float					rangeZ = inBack - inFront;
	
		// first random values in eye space
	inVertex->x = inLeft + rangeX * S3TestUtil_RandomFloat();
	inVertex->y = inTop + rangeY * S3TestUtil_RandomFloat();
	inVertex->z = inFront + rangeZ * S3TestUtil_RandomFloat();
	
		// invert the eye-space Z coord to get invW
	inVertex->invW = 1.0 / inVertex->z;
	
		// now convert these to screen space
	inVertex->x = S3GetScreenX(inVertex->x, inVertex->z);
	inVertex->y = S3GetScreenY(inVertex->y, inVertex->z);
	inVertex->z = S3GetScreenZ(inVertex->z);
	
		// make random U and V between 0 and 2
	inVertex->uOverW = 2.0 * S3TestUtil_RandomFloat() * inVertex->invW;
	inVertex->vOverW = 2.0 * S3TestUtil_RandomFloat() * inVertex->invW;
	
		// get random RGB values
	inVertex->a = 1.0;
	inVertex->r = S3TestUtil_RandomFloat();
	inVertex->g = S3TestUtil_RandomFloat();
	inVertex->b = S3TestUtil_RandomFloat();
	
	inVertex->kd_r = 1.0;
	inVertex->kd_g = 1.0;
	inVertex->kd_b = 1.0;
	
	inVertex->ks_r = 0.0;
	inVertex->ks_g = 0.0;
	inVertex->ks_b = 0.0;
}


/*=========================================================================*\
	S3RandomRGBDefaultHighlight
\*=========================================================================*/
	void 
S3RandomRGBDefaultHighlight(
	TQAVTexture*			inVertex)
{
		// make the vertex's RGB values random
	inVertex->r = S3TestUtil_RandomFloat();
	inVertex->g = S3TestUtil_RandomFloat();
	inVertex->b = S3TestUtil_RandomFloat();

		// make the alpha and highlight values "default"
	inVertex->a = 1.0;
	
	inVertex->kd_r = 1.0;
	inVertex->kd_g = 1.0;
	inVertex->kd_b = 1.0;
	
	inVertex->ks_r = 0.0;
	inVertex->ks_g = 0.0;
	inVertex->ks_b = 0.0;
}


/*=========================================================================*\
	S3StoreBackground
\*=========================================================================*/
	void 
S3StoreBackground(
	float 					inWidth,
	float 					inHeight,
	float 					inDepth)
{
	TQAVTexture				verts[4];

		// we subtract a little bit from the width and height values so that 
		// they're fully inside the screen pixels
	inWidth -= 0.1;
	inHeight -= 0.1;
	
		// the background tris all have the same depth and invW
	verts[0].z = verts[1].z = verts[2].z = verts[3].z = inDepth;
	verts[0].invW = verts[1].invW = verts[2].invW = verts[3].invW = kBackgroundInvW;

		// all tri vertices will have random RGB values, default highlight
	S3RandomRGBDefaultHighlight(&verts[0]);
	S3RandomRGBDefaultHighlight(&verts[1]);
	S3RandomRGBDefaultHighlight(&verts[2]);
	S3RandomRGBDefaultHighlight(&verts[3]);

	verts[0].x = 0.0;
	verts[0].y = 0.0;
	verts[0].uOverW = 0.0;
	verts[0].vOverW = kBackground_V * verts[0].invW;
		
	verts[1].x = 0.0;
	verts[1].y = inHeight;
	verts[1].uOverW = 0.0;
	verts[1].vOverW = 0.0;
		
	verts[2].x = inWidth;
	verts[2].y = inHeight;
	verts[2].uOverW = kBackground_U * verts[2].invW;
	verts[2].vOverW = 0.0;

	verts[3].x = inWidth;
	verts[3].y = 0.0;
	verts[3].uOverW = kBackground_U * verts[3].invW;
	verts[3].vOverW = kBackground_V * verts[3].invW;

		// store the background rectangle
	S3TextureTriList_StoreQuad(
		&gTextureTriList, 
		&verts[0], 
		&verts[1], 
		&verts[2], 
		&verts[3], 
		gTriFlags,
		kTexture_background);
}


/*=========================================================================*\
	S3StoreColumn
\*=========================================================================*/
	void
S3StoreColumn()
{
	TQAVTexture				verts[4];
	
		// eyeZ / eyeZrange == screenZ / screenZrange
		// solve for screenZ to get a Z value between 0 and 1 for the front
		// and side of the column
	float					eyeZrange = kWallsZeye_back - kWallsZeye_front;
	float					screenZrange = kWallsZ_back - kWallsZ_front;
	float					columnScreenZ_front = (kColumnZ_front / eyeZrange) * screenZrange;
	float					columnScreenZ_back = (kColumnZ_back / eyeZrange) * screenZrange;
	
		// all tri vertices will have random RGB values, default highlight
	S3RandomRGBDefaultHighlight(&verts[0]);
	S3RandomRGBDefaultHighlight(&verts[1]);
	S3RandomRGBDefaultHighlight(&verts[2]);
	S3RandomRGBDefaultHighlight(&verts[3]);

		// front of column
	verts[0].x = S3GetScreenX(kColumn_left, kColumnZ_front);
	verts[0].y = S3GetScreenY(kColumn_top, kColumnZ_front);
	verts[0].z = columnScreenZ_front;
	verts[0].invW = 1.0 / kColumnZ_front;
	verts[0].uOverW = kColumnFrontU_min * verts[0].invW;
	verts[0].vOverW = kColumnFrontV_max * verts[0].invW;
		
	verts[1].x = S3GetScreenX(kColumn_left, kColumnZ_front);
	verts[1].y = S3GetScreenY(kColumn_bottom, kColumnZ_front);
	verts[1].z = columnScreenZ_front;
	verts[1].invW = 1.0 / kColumnZ_front;
	verts[1].uOverW = kColumnFrontU_min * verts[1].invW;
	verts[1].vOverW = kColumnFrontV_min * verts[1].invW;
		
	verts[2].x = S3GetScreenX(kColumn_right, kColumnZ_front);
	verts[2].y = S3GetScreenY(kColumn_bottom, kColumnZ_front);
	verts[2].z = columnScreenZ_front;
	verts[2].invW = 1.0 / kColumnZ_front;
	verts[2].uOverW = kColumnFrontU_max * verts[2].invW;
	verts[2].vOverW = kColumnFrontV_min * verts[2].invW;
		
	verts[3].x = S3GetScreenX(kColumn_right, kColumnZ_front);
	verts[3].y = S3GetScreenY(kColumn_top, kColumnZ_front);
	verts[3].z = columnScreenZ_front;
	verts[3].invW = 1.0 / kColumnZ_front;
	verts[3].uOverW = kColumnFrontU_max * verts[3].invW;
	verts[3].vOverW = kColumnFrontV_max * verts[3].invW;
	
	S3TextureTriList_StoreQuad(
		&gTextureTriList, 
		&verts[0], 
		&verts[1], 
		&verts[2], 
		&verts[3], 
		gTriFlags,
		kTexture_column);

		// side of column
	verts[0].x = S3GetScreenX(kColumn_left, kColumnZ_back);
	verts[0].y = S3GetScreenY(kColumn_top, kColumnZ_back);
	verts[0].z = columnScreenZ_back;
	verts[0].invW = 1.0 / kColumnZ_back;
	verts[0].uOverW = kColumnSideU_min * verts[0].invW;
	verts[0].vOverW = kColumnSideV_max * verts[0].invW;
		
	verts[1].x = S3GetScreenX(kColumn_left, kColumnZ_back);
	verts[1].y = S3GetScreenY(kColumn_bottom, kColumnZ_back);
	verts[1].z = columnScreenZ_front;
	verts[1].invW = 1.0 / kColumnZ_back;
	verts[1].uOverW = kColumnSideU_min * verts[1].invW;
	verts[1].vOverW = kColumnSideV_min * verts[1].invW;
		
	verts[2].x = S3GetScreenX(kColumn_left, kColumnZ_front);
	verts[2].y = S3GetScreenY(kColumn_bottom, kColumnZ_front);
	verts[2].z = columnScreenZ_front;
	verts[2].invW = 1.0 / kColumnZ_front;
	verts[2].uOverW = kColumnSideU_max * verts[2].invW;
	verts[2].vOverW = kColumnSideV_min * verts[2].invW;
		
	verts[3].x = S3GetScreenX(kColumn_left, kColumnZ_front);
	verts[3].y = S3GetScreenY(kColumn_top, kColumnZ_front);
	verts[3].z = columnScreenZ_front;
	verts[3].invW = 1.0 / kColumnZ_front;
	verts[3].uOverW = kColumnSideU_max * verts[3].invW;
	verts[3].vOverW = kColumnSideV_max * verts[3].invW;
	
	S3TextureTriList_StoreQuad(
		&gTextureTriList, 
		&verts[0], 
		&verts[1], 
		&verts[2], 
		&verts[3], 
		gTriFlags,
		kTexture_column);
}


/*=========================================================================*\
	S3StoreWalls
\*=========================================================================*/
	void
S3StoreWalls(
	float 					inBackWidth,
	float 					inBackHeight,
	float 					inFrontZ,
	float 					inBackZ)
{
	float					windowWidth = gWindowWidth - 0.1;
	float					windowHeight = gWindowHeight - 0.1;
	float					wallWidth = (windowWidth - inBackWidth) / 2.0;
	float					floorDepth = (windowHeight - inBackHeight) / 2.0;
	float					maxUV = 5.0;
	TQAVTexture				verts[4];
	
		// all tri vertices will have random RGB values, default highlight
	S3RandomRGBDefaultHighlight(&verts[0]);
	S3RandomRGBDefaultHighlight(&verts[1]);
	S3RandomRGBDefaultHighlight(&verts[2]);
	S3RandomRGBDefaultHighlight(&verts[3]);

		// the wall and floor u and v values go from 0 to maxUV, so uOverW 
		// and vOverW == maxUV * invW == 1 / W, and W == z
		
		// left wall
	verts[0].x = 0.0;
	verts[0].y = 0.0;
	verts[0].z = inFrontZ;
	verts[0].invW = kWallsInvW_front;
	verts[0].uOverW = 0.0;
	verts[0].vOverW = kWalls_V * verts[0].invW;
		
	verts[1].x = 0.0;
	verts[1].y = windowHeight;
	verts[1].z = inFrontZ;
	verts[1].invW = kWallsInvW_front;
	verts[1].uOverW = 0.0;
	verts[1].vOverW = 0.0;
		
	verts[2].x = wallWidth;
	verts[2].y = floorDepth + inBackHeight;
	verts[2].z = inBackZ;
	verts[2].invW = kWallsInvW_back;
	verts[2].uOverW = kWalls_U * verts[2].invW;
	verts[2].vOverW = 0.0;
		
	verts[3].x = wallWidth;
	verts[3].y = floorDepth;
	verts[3].z = inBackZ;
	verts[3].invW = kWallsInvW_back;
	verts[3].uOverW = kWalls_U * verts[3].invW;
	verts[3].vOverW = kWalls_V * verts[3].invW;
		
	S3TextureTriList_StoreQuad(
		&gTextureTriList, 
		&verts[0], 
		&verts[1], 
		&verts[2], 
		&verts[3], 
		gTriFlags,
		kTexture_leftWall);
		
		// back wall
	verts[0].x = wallWidth;
	verts[0].y = floorDepth;
	verts[0].z = inBackZ;
	verts[0].invW = kWallsInvW_back;
	verts[0].uOverW = 0.0;
	verts[0].vOverW = kBackWall_V * verts[0].invW;
		
	verts[1].x = wallWidth;
	verts[1].y = floorDepth + inBackHeight;
	verts[1].z = inBackZ;
	verts[1].invW = kWallsInvW_back;
	verts[1].uOverW = 0.0;
	verts[1].vOverW = 0.0;
		
	verts[2].x = wallWidth + inBackWidth;
	verts[2].y = floorDepth + inBackHeight;
	verts[2].z = inBackZ;
	verts[2].invW = kWallsInvW_back;
	verts[2].uOverW = kBackWall_U * verts[2].invW;
	verts[2].vOverW = 0.0;
		
	verts[3].x = wallWidth + inBackWidth;
	verts[3].y = floorDepth;
	verts[3].z = inBackZ;
	verts[3].invW = kWallsInvW_back;
	verts[3].uOverW = kBackWall_U * verts[3].invW;
	verts[3].vOverW = kBackWall_V * verts[3].invW;
		
	S3TextureTriList_StoreQuad(
		&gTextureTriList, 
		&verts[0], 
		&verts[1], 
		&verts[2], 
		&verts[3], 
		gTriFlags,
		kTexture_backWall);
		
		// right wall
	verts[0].x = wallWidth + inBackWidth;
	verts[0].y = floorDepth;
	verts[0].z = inBackZ;
	verts[0].invW = kWallsInvW_back;
	verts[0].uOverW = 0.0;
	verts[0].vOverW = kWalls_V * verts[0].invW;
		
	verts[1].x = wallWidth + inBackWidth;
	verts[1].y = floorDepth + inBackHeight;
	verts[1].z = inBackZ;
	verts[1].invW = kWallsInvW_back;
	verts[1].uOverW = 0.0;
	verts[1].vOverW = 0.0;
		
	verts[2].x = windowWidth;
	verts[2].y = windowHeight;
	verts[2].z = inFrontZ;
	verts[2].invW = kWallsInvW_front;
	verts[2].uOverW = kWalls_U * verts[2].invW;
	verts[2].vOverW = 0.0;
		
	verts[3].x = windowWidth;
	verts[3].y = 0.0;
	verts[3].z = inFrontZ;
	verts[3].invW = kWallsInvW_front;
	verts[3].uOverW = kWalls_U * verts[3].invW;
	verts[3].vOverW = kWalls_V * verts[3].invW;
		
	S3TextureTriList_StoreQuad(
		&gTextureTriList, 
		&verts[3], 
		&verts[2], 
		&verts[1], 
		&verts[0], 
		gTriFlags,
		kTexture_rightWall);
		
//	S3TextureTriList_StoreQuad(
//		&gTextureTriList, 
//		&verts[0], 
//		&verts[1], 
//		&verts[2], 
//		&verts[3], 
//		gTriFlags,
//		kTexture_rightWall);
		
		// floor
	verts[0].x = wallWidth;
	verts[0].y = floorDepth + inBackHeight;
	verts[0].z = inBackZ;
	verts[0].invW = kWallsInvW_back;
	verts[0].uOverW = 0.0;
	verts[0].vOverW = kFloor_V * verts[0].invW;
		
	verts[1].x = 0.0;
	verts[1].y = windowHeight;
	verts[1].z = inFrontZ;
	verts[1].invW = kWallsInvW_front;
	verts[1].uOverW = 0.0;
	verts[1].vOverW = 0.0;
		
	verts[2].x = windowWidth;
	verts[2].y = windowHeight;
	verts[2].z = inFrontZ;
	verts[2].invW = kWallsInvW_front;
	verts[2].uOverW = kFloor_U * verts[2].invW;
	verts[2].vOverW = 0.0;
		
	verts[3].x = wallWidth + inBackWidth;
	verts[3].y = floorDepth + inBackHeight;
	verts[3].z = inBackZ;
	verts[3].invW = kWallsInvW_back;
	verts[3].uOverW = kFloor_U * verts[3].invW;
	verts[3].vOverW = kFloor_V * verts[3].invW;

//	verts[1].kd_r = verts[2].kd_r = 2.0;
//	verts[1].kd_g = verts[2].kd_g = 2.0;
//	verts[1].kd_b = verts[2].kd_b = 2.0;
	
//	verts[0].kd_r = verts[3].kd_r = 0.5;
//	verts[0].kd_g = verts[3].kd_g = 0.5;
//	verts[0].kd_b = verts[3].kd_b = 0.5;
	
//	inVertex->ks_r = 0.0;
//	inVertex->ks_g = 0.0;
//	inVertex->ks_b = 0.0;
		
	S3TextureTriList_StoreQuad(
		&gTextureTriList, 
		&verts[0], 
		&verts[1], 
		&verts[2], 
		&verts[3], 
		gTriFlags,
		kTexture_floor);
}


/*=========================================================================*\
	S3StoreRandomTris
\*=========================================================================*/
	void
S3StoreRandomTris()
{
	TQAVTexture				verts[3];
	short					texture, i;

	for (i = 0; i < kBox1_triCount; i++) {
			// get a random vertex which is inside box1
		S3BoundedRandomVTexture(
				&verts[0], kBox1_left, kBox1_right, kBox1_top, 
				kBox1_bottom, kBox1_front, kBox1_back);
		S3BoundedRandomVTexture(
				&verts[1], kBox1_left, kBox1_right, kBox1_top, 
				kBox1_bottom, kBox1_front, kBox1_back);
		S3BoundedRandomVTexture(
				&verts[2], kBox1_left, kBox1_right, kBox1_top, 
				kBox1_bottom, kBox1_front, kBox1_back);
				
			// use either the green or red lava textures.  this simply adds
			// 0 or 1 to kTexture_greenLava.  we subtract 0.01 just to make
			// sure we never get exactly 2.0.
		texture = kTexture_greenLava + (int) (S3TestUtil_RandomFloat() * 2.0 - 0.01);
		
		S3TextureTriList_StoreTri(
				&gTextureTriList, &verts[0], &verts[1], &verts[2], gTriFlags, texture);
	} // endfor
}


/*=========================================================================*\
	S3TextureTriList_New
\*=========================================================================*/
	Boolean 
S3TextureTriList_New(
	S3tTextureTriList*		inTextureTriList)
{
	ASSERT(inTextureTriList);
	
		// we assume inTextureTriList doesn't already point to anything
	inTextureTriList->head = nil;
	inTextureTriList->tail = nil;
	inTextureTriList->triIndex = 0;

		// allocate a new S3tTextureTriGroup for the head of the list
	inTextureTriList->head = (S3tTextureTriGroup*) malloc(sizeof(S3tTextureTriGroup));
	inTextureTriList->tail = inTextureTriList->head;
	
	if (inTextureTriList->head == nil) {
			// we don't have enough memory to store a S3tTextureTriGroup,
			// so let the caller know.  the fact that we don't have enough
			// memory is indicated by the nil tail pointer.
		return false;
	}
	
	inTextureTriList->head->next = nil;
	
	return true;
}


/*=========================================================================*\
	S3TextureTriList_Delete
\*=========================================================================*/
	void 
S3TextureTriList_Delete(
	S3tTextureTriList*		inTextureTriList)
{
	S3tTextureTriGroup*		currentGroup;
	S3tTextureTriGroup*		nextGroup;
	
	ASSERT(inTextureTriList);
	
	for (currentGroup = inTextureTriList->head, nextGroup = nil; 
		 currentGroup;
		 currentGroup = nextGroup) {
		 	// save a pointer to the next group, before we free the current one
		nextGroup = currentGroup->next;
		
		free(currentGroup);
	} // endfor
	
	inTextureTriList->head = nil;
	inTextureTriList->tail = nil;
	inTextureTriList->triIndex = 0;
}


/*=========================================================================*\
	S3TextureTriList_IsValid
\*=========================================================================*/
	Boolean 
S3TextureTriList_IsValid(
	S3tTextureTriList*		inTextureTriList)
{
	ASSERT(inTextureTriList);
	
	return (inTextureTriList->tail != nil);
}


/*=========================================================================*\
	S3TextureTriList_NewTextureTriGroup
\*=========================================================================*/
	Boolean 
S3TextureTriList_NewTextureTriGroup(
	S3tTextureTriList*		inTextureTriList)
{
	S3tTextureTriGroup*		newGroup = nil;

	ASSERT(inTextureTriList);
	
	if (inTextureTriList->tail == nil) {
			// the tail should never be nil.  if it is, it means we were unable
			// to create a new S3tTextureTriGroup at some point.  so let the
			// caller know that.
		return false;
	}
	
	if (inTextureTriList->triIndex < kTextureTriGroupCount) {
			// we don't need a new S3tTextureTriGroup because the current
			// one isn't full yet
		return true;
	}
	
		// allocate a new S3tTextureTriGroup at the end of the list
	newGroup = (S3tTextureTriGroup*) malloc(sizeof(S3tTextureTriGroup));
	
	if (newGroup == nil) {
			// there's not enough memory to add a new group, so indicate that
			// in the list and let the caller know
		inTextureTriList->tail = nil;
		return false;
	}
	
		// attach the new group to the end of the list
	newGroup->next = nil;
	inTextureTriList->tail->next = newGroup;
	inTextureTriList->tail = newGroup;
		
		// reset the index
	inTextureTriList->triIndex = 0;
	
	return true;
}


/*=========================================================================*\
	S3TextureTriList_StoreTri
\*=========================================================================*/
	Boolean 
S3TextureTriList_StoreTri(
	S3tTextureTriList*		inTextureTriList,
	TQAVTexture*			inVert1,
	TQAVTexture*			inVert2,
	TQAVTexture*			inVert3,
	unsigned long			inFlags,
	short					inTexture)
{
	ASSERT(inTextureTriList);
	ASSERT(inVert1);
	ASSERT(inVert2);
	ASSERT(inVert3);
	
		// allocate a new S3tTextureTriGroup if necessary
	if (S3TextureTriList_NewTextureTriGroup(inTextureTriList) == false) {
			// there wasn't enough memory
		return false;
	}
	
		// copy the three vertices into the current S3tTextureTri
	inTextureTriList->tail->tris[inTextureTriList->triIndex].v1 = *inVert1;
	inTextureTriList->tail->tris[inTextureTriList->triIndex].v2 = *inVert2;
	inTextureTriList->tail->tris[inTextureTriList->triIndex].v3 = *inVert3;

		// copy the flags and texture index into the current S3tTextureTri
	inTextureTriList->tail->tris[inTextureTriList->triIndex].flags = inFlags;
	inTextureTriList->tail->tris[inTextureTriList->triIndex].texture = inTexture;
		
		// we just added a S3tTextureTri, so increment the index
	inTextureTriList->triIndex++;
		
	return true;
}


/*=========================================================================*\
	S3TextureTriList_StoreQuad
\*=========================================================================*/
	Boolean 
S3TextureTriList_StoreQuad(
	S3tTextureTriList*		inTextureTriList,
	TQAVTexture*			inVert1,
	TQAVTexture*			inVert2,
	TQAVTexture*			inVert3,
	TQAVTexture*			inVert4,
	unsigned long			inFlags,
	short					inTexture)
{
	Boolean result = false;
	
	ASSERT(inTextureTriList);
	ASSERT(inVert1);
	ASSERT(inVert2);
	ASSERT(inVert3);
	ASSERT(inVert4);

		// first triangle
	result = S3TextureTriList_StoreTri(
		inTextureTriList,
		inVert1,
		inVert2,
		inVert3,
		inFlags,
		inTexture);

	if (result == false) {
		return result;
	}
	
		// second triangle
	result = S3TextureTriList_StoreTri(
		inTextureTriList,
		inVert1,
		inVert3,
		inVert4,
		inFlags,
		inTexture);
	
	return result;
}


/*=========================================================================*\
	S3TextureTriList_Draw
\*=========================================================================*/
	void 
S3TextureTriList_Draw(
	S3tTextureTriList*		inTextureTriList,
	S3tTestInfo*			inTestInfo)
{
	TQADrawContext* 		leftContext = S3TestInfo_GetLDrawContext(inTestInfo);
	TQADrawContext* 		rightContext = S3TestInfo_GetRDrawContext(inTestInfo);
	S3tTextureTriGroup*		currentGroup;
	short					maxTriIndex, i;
	
	ASSERT(inTextureTriList);
	
	for (currentGroup = inTextureTriList->head; 
		 currentGroup;
		 currentGroup = currentGroup->next) {
		if (currentGroup->next == nil) {
				// we're on the last group.  the index of the last tri in
				// this group is stored in inTextureTriList->triIndex
			maxTriIndex = inTextureTriList->triIndex;
		} else {
				// the current group is full, so draw the full tri count
			maxTriIndex = kTextureTriGroupCount;
		}
		
		for (i = 0; i < maxTriIndex; i++) {
			ASSERT(currentGroup->tris[i].texture < kTextureCount);
			
				// set the left draw context's texture
			QASetPtr(
				leftContext, 
				kQATag_Texture, 
				gTextures[currentGroup->tris[i].texture].left);

				// set the right draw context's texture
			QASetPtr(
				rightContext, 
				kQATag_Texture, 
				gTextures[currentGroup->tris[i].texture].right);

			S3Test_DrawTriTexture(
				inTestInfo, 
				&currentGroup->tris[i].v1,
				&currentGroup->tris[i].v2,
				&currentGroup->tris[i].v3,
				currentGroup->tris[i].flags);
		} // endfor
	} // endfor
}