/*****************************************************************************\

	S3MoreTests.c

	Copyright (C) 1995 Apple Computer, Inc.  All rights reserved.

\*****************************************************************************/

/*===========================================================================*\
	Includes
\*===========================================================================*/
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <Files.h>
#include <Memory.h>
#include <Quickdraw.h>

#include "Drive3D.h"
#include "S3Image.h"
#include "S3TestSystem.h"
#include "S3TestInfo.h"
#include "S3TestUtil.h"
#include "S3Utilities.h"

/*===========================================================================*\
	Defines
\*===========================================================================*/

#define kNPoints					100
#define kNLines						100
#define kNFlushes					5

#define DIVDEC1		(1.0)		/* Used to generate the vertex increment					*/
#define DIVDEC2		(3.0)		/* along 2 edges of the window.								*/
#define INCREMENT	(0.5)		/* Number of degrees from one vertex to	another.			*/
#define RADIUS		(100.0)		/* Radius of circle.										*/
#define STEP		(10.0)		/* Steps in degrees between the first vertex and the next.	*/
#define NUMTRIS		(1000)		/* Number of triangles to be drawn.							*/
#define PI	 		((float)3.1415926535898)

/*===========================================================================*\
	Prototypes
\*===========================================================================*/

void S3MoreTests_Register(void);
S3tTestStatus GetTriTextures(S3Image *theImages, S3tTestInfo *testInfo);
S3tTestStatus DrawTriTexture(S3Image *theImages, short index, TQAVTexture *vertices, short i,
	short j, short k, S3tTestInfo *testInfo);
void GetVertices(TQAVGouraud vertices[], float width, float height, float triwidth,
	float triheight);

S3tTestStatus S3Test100Points(
	S3tTestInfo				*testInfo);
S3tTestStatus S3Test100Lines(
	S3tTestInfo				*testInfo);
S3tTestStatus S3CheckTouchTris(S3tTestInfo *testInfo);
S3tTestStatus S3CheckTouchTrisReverse(S3tTestInfo *testInfo);
S3tTestStatus S3SkinnyHorizontalTris(S3tTestInfo *testInfo);
S3tTestStatus S3SkinnyVerticalTris(S3tTestInfo *testInfo);
S3tTestStatus S3MultipleModeTris(S3tTestInfo *testInfo);
S3tTestStatus S3Pie(S3tTestInfo *testInfo);
S3tTestStatus S3LineBorder(S3tTestInfo *testInfo);
S3tTestStatus S3Coverage(S3tTestInfo *testInfo);
S3tTestStatus S3ColoredHK(S3tTestInfo *testInfo);
S3tTestStatus S3SideRect(S3tTestInfo *testInfo);
S3tTestStatus S3SlowTris(S3tTestInfo *testInfo);
S3tTestStatus S3FastestTris25(S3tTestInfo *testInfo);
S3tTestStatus S3FasterTris25(S3tTestInfo *testInfo);
S3tTestStatus S3FastestTris50(S3tTestInfo *testInfo);
S3tTestStatus S3FasterTris50(S3tTestInfo *testInfo);
S3tTestStatus S3Random25Tris(S3tTestInfo *testInfo);
S3tTestStatus S3Random49Tris(S3tTestInfo *testInfo);


/*=========================================================================*\
	S3MoreTests_Register
\*=========================================================================*/
void S3MoreTests_Register(void)
{
	S3Test_Register("100 Points",					kS3Prereq_None, S3Test100Points);
	S3Test_Register("100 Lines",					kS3Prereq_None, S3Test100Lines);
	S3Test_Register("Touching Tris",				kS3Prereq_None, S3CheckTouchTris);
	S3Test_Register("Touching Tris Reverse",		kS3Prereq_None, S3CheckTouchTrisReverse);
	S3Test_Register("Skinny Wide Tris",				kS3Prereq_None, S3SkinnyHorizontalTris);
	S3Test_Register("Skinny Tall Tris",				kS3Prereq_None, S3SkinnyVerticalTris);
	S3Test_Register("Gouraud, Texture, Decal Tris",	kS3Prereq_None, S3MultipleModeTris);
	S3Test_Register("Pie",							kS3Prereq_None, S3Pie);
	S3Test_Register("Border Lines",					kS3Prereq_None, S3LineBorder);
	S3Test_Register("Coverage",						kS3Prereq_None, S3Coverage);
	S3Test_Register("Colored HK Texture",			kS3Prereq_Texture, S3ColoredHK);
	S3Test_Register("Sideways Rect - Negative UV",	kS3Prereq_Texture, S3SideRect);
	S3Test_Register("Slow Drawing Tris",			kS3Prereq_Texture, S3SlowTris);

	S3Test_Register("Fastest Drawing Tris - 25 pixels",	kS3Prereq_None, S3FastestTris25);
	S3Test_Register("Faster Drawing Tris - 25 pixels",	kS3Prereq_None, S3FasterTris25);
	S3Test_Register("Fastest Drawing Tris - 50 pixels",	kS3Prereq_None, S3FastestTris50);
	S3Test_Register("Faster Drawing Tris - 50 pixels",	kS3Prereq_None, S3FasterTris50);
	
	S3Test_Register("Random 25-pixel Tris",			kS3Prereq_None, S3Random25Tris);
	S3Test_Register("Random 49-pixel Tris",			kS3Prereq_None, S3Random49Tris);
}

/*=========================================================================*\
	S3Test100Points
\*=========================================================================*/
S3tTestStatus S3Test100Points(
	S3tTestInfo				*testInfo)
{
	long			i, flushCount;
	TQAVGouraud		vGouraud;
	
	S3Test_RenderStart(testInfo, NULL);

	for (i = 0, flushCount = kNPoints / kNFlushes; i++ < kNPoints; /* Nothing */)
	{
		S3TestUtil_RandomVGouraud(testInfo, &vGouraud);
		
		S3Test_DrawPoint(testInfo, &vGouraud);

		if ((i % flushCount) == 0)
		{
			S3Test_Flush(testInfo);
		}
	}

	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}

/*=========================================================================*\
	S3Test100Lines
\*=========================================================================*/
S3tTestStatus S3Test100Lines(
	S3tTestInfo				*testInfo)
{
	long		i, flushCount;
	TQAVGouraud	vGouraud[2];
	
	S3Test_RenderStart(testInfo, NULL);

	for (i = 0, flushCount = kNLines / kNFlushes; i++ < kNLines; /* Nothing */)
	{
		S3TestUtil_RandomVGouraud(testInfo, &vGouraud[0]);
		S3TestUtil_RandomVGouraud(testInfo, &vGouraud[1]);

		S3Test_DrawLine(testInfo, &vGouraud[0], &vGouraud[1]);
		
		if ((i % flushCount) == 0)
		{
			S3Test_Flush(testInfo);
		}
	}

	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}
					
/*===========================================================================*\
	GetTriTexture
	
	Get the images from the files.
\*===========================================================================*/

S3tTestStatus GetTriTextures(S3Image *theImages, S3tTestInfo *testInfo) {
	FSSpec filespec;
	PicHandle thePict;
	Boolean mipmap = (S3TestInfo_GetTextureFlags(testInfo) & kQATexture_Mipmap);
	
	if (FSMakeFSSpec(0, 0, "\p:Textures:Red-blue-black checker.pict", &filespec) != noErr) {
		S3Test_Print("Unable to find texture!");
		return kS3Status_Failure;
	}
	thePict = S3GetPICTFromFile(&filespec);
	theImages[0] = S3Image_New(thePict, S3TestInfo_GetTextureFormat(testInfo),
		S3TestInfo_GetTextureMakeBlackTransparent(testInfo), mipmap);
	DisposeHandle((Handle) thePict);

	if (FSMakeFSSpec(0, 0, "\p:Textures:Black-white check.black A.pict", &filespec) != noErr) {
		S3Test_Print("Unable to find texture!");
		S3Image_Delete(theImages[0]);
		return kS3Status_Failure;
	}
	thePict = S3GetPICTFromFile(&filespec);
	theImages[1] = S3Image_New(thePict, S3TestInfo_GetTextureFormat(testInfo),
		S3TestInfo_GetTextureMakeBlackTransparent(testInfo), mipmap);
	DisposeHandle((Handle) thePict);

	if (FSMakeFSSpec(0, 0, "\p:Textures:Red alien.32.pict", &filespec) != noErr) {
		S3Test_Print("Unable to find texture!");
		S3Image_Delete(theImages[0]);
		S3Image_Delete(theImages[1]);
		return kS3Status_Failure;
	}
	thePict = S3GetPICTFromFile(&filespec);
	theImages[2] = S3Image_New(thePict, S3TestInfo_GetTextureFormat(testInfo),
		S3TestInfo_GetTextureMakeBlackTransparent(testInfo), mipmap);
	DisposeHandle((Handle) thePict);

	return kS3Status_Completed;
}


/*===========================================================================*\
	DrawTriTexture
	
	Draws the triangle with vertices given by the indices, i, j and k, using
	the texture in theImages array with index, index.
\*===========================================================================*/

S3tTestStatus DrawTriTexture(S3Image *theImages, short index, TQAVTexture *vertices, short i,
short j, short k, S3tTestInfo *testInfo) {
	TQAImage *theQAImage = S3Image_GetImages(theImages[index]);
	TQATexture *theLTexture, *theRTexture;
	TQAEngine *theLEngine = S3TestInfo_GetLEngine(testInfo);
	TQAEngine *theREngine = S3TestInfo_GetREngine(testInfo);
	TQADrawContext *theLContext = S3TestInfo_GetLDrawContext(testInfo);
	TQADrawContext *theRContext = S3TestInfo_GetRDrawContext(testInfo);
	Boolean error = FALSE;
	Boolean mipmap = (S3TestInfo_GetTextureFlags(testInfo) & kQATexture_Mipmap);
	
	if (QATextureNew(theLEngine, mipmap, S3TestInfo_GetTextureFormat(testInfo), theQAImage,
		&theLTexture) != kQANoErr) {
		S3Test_Print("Could not get the left texture!");
		error = TRUE;
	} else if (QATextureNew(theREngine, mipmap, S3TestInfo_GetTextureFormat(testInfo),
		theQAImage, &theRTexture) != kQANoErr) {
		S3Test_Print("Could not get the right texture!");
		QATextureDelete(theLEngine, theLTexture);
		error = TRUE;
	}
	if (!error) {
		QASetPtr(theLContext, kQATag_Texture, theLTexture);
		QASetPtr(theRContext, kQATag_Texture, theRTexture);
		S3Test_DrawTriTexture(testInfo, &vertices[i], &vertices[j], &vertices[k],
			kQATriFlags_None);
		QATextureDelete(theLEngine, theLTexture);
		QATextureDelete(theREngine, theRTexture);
		return kS3Status_Completed;
	} else {
		S3Test_RenderEnd(testInfo, NULL);
		return kS3Status_Failure;
	}
}


/*===========================================================================*\
	GetVertices
	
	Generates a right triangle with width, triwidth, and height, triheight.
	Randomly generates the color and the z value at the vertices. Randomly
	rotates the triangle by 0, pi, 2pi or 3pi radians and randomly displaces
	the triangle. Randomly flips the triheight edge of the triangle about the
	triwidth edge. Makes sure that the triangle is within the boundaries of
	the window given by width and height.
\*===========================================================================*/

void GetVertices(TQAVGouraud vertices[], float width, float height, float triwidth,
float triheight) {
	int i, flip;
	Boolean inWindow = FALSE;
	float theta, dx, dy, costheta, sintheta;
	
	flip = ((rand() % 2) == 1) ? 1 : -1;
	for (i = 0; i < 3; i++) {
		vertices[i].r = (float) rand() / RAND_MAX;
		vertices[i].g = (float) rand() / RAND_MAX;
		vertices[i].b = (float) rand() / RAND_MAX;
		vertices[i].z = (float) rand() / RAND_MAX;
		vertices[i].invW = 1.0 / vertices[i].z;
	}
	while (!inWindow) {
		theta = (float) ((rand() % 4) * PI / 2.0);
		dx = (float) ((int) (((float) rand() / RAND_MAX) * width));
		dy = (float) ((int) (((float) rand() / RAND_MAX) * height));
		costheta = cos((double)theta);
		sintheta = sin((double)theta);
		vertices[0].x = dx;
		vertices[0].y = dy;
		vertices[1].x = triwidth * costheta + dx;
		vertices[1].y = triwidth * sintheta + dy;
		vertices[2].x = - triheight * flip * sintheta + dx;
		vertices[2].y = triheight * flip * costheta + dy;
		inWindow = TRUE;
		for (i = 0; i < 3; i++) {
			if (vertices[i].x < 0.0 || vertices[i].x > width) inWindow = FALSE;
			if (vertices[i].y < 0.0 || vertices[i].y > height) inWindow = FALSE;
		}
	}
}

/*===========================================================================*\
	S3CheckTouchTris
	
	Draws 8 triangles around the whole window each with one vertex in the
	middle of the screen. This checks that all pixels are covered and there
	are no problems around the edges of the window.
\*===========================================================================*/

S3tTestStatus S3CheckTouchTris(S3tTestInfo *testInfo) {
	float xmid, xfar;
	float ymid, yfar;
	TQAVGouraud vertices[9] = {
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
	};
	
	xmid = (float) S3TestInfo_GetWidth(testInfo) / 2.0;
	xfar = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	ymid = (float) S3TestInfo_GetHeight(testInfo) / 2.0;
	yfar = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	
	vertices[1].x = xmid;
	vertices[2].x = xfar;
	
	vertices[3].y = vertices[4].y = vertices[5].y = ymid;
	vertices[4].x = xmid;
	vertices[5].x = xfar;

	vertices[6].y = vertices[7].y = vertices[8].y = yfar;
	vertices[7].x = xmid;
	vertices[8].x = xfar;
	
	S3Test_RenderStart(testInfo, NULL);
	
	vertices[0].r = vertices[1].r = vertices[4].r = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[4], kQATriFlags_None);
	
	vertices[0].r = vertices[1].r = vertices[4].r = 0.0;
	vertices[1].g = vertices[2].g = vertices[4].g = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[1], &vertices[2], &vertices[4], kQATriFlags_None);
	
	vertices[1].g = vertices[2].g = vertices[4].g = 0.0;
	vertices[2].b = vertices[4].b = vertices[5].b = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[2], &vertices[4], &vertices[5], kQATriFlags_None);
	
	vertices[2].b = vertices[4].b = vertices[5].b = 0.0;
	vertices[4].r = vertices[5].r = vertices[8].r = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[4], &vertices[5], &vertices[8], kQATriFlags_None);
	
	vertices[4].r = vertices[5].r = vertices[8].r = 0.0;
	vertices[4].g = vertices[7].g = vertices[8].g = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[4], &vertices[7], &vertices[8], kQATriFlags_None);
	
	vertices[4].g = vertices[7].g = vertices[8].g = 0.0;
	vertices[4].b = vertices[6].b = vertices[7].b = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[4], &vertices[6], &vertices[7], kQATriFlags_None);
	
	vertices[4].b = vertices[6].b = vertices[7].b = 0.0;
	vertices[3].r = vertices[4].r = vertices[6].r = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[3], &vertices[4], &vertices[6], kQATriFlags_None);
	
	vertices[3].r = vertices[4].r = vertices[6].r = 0.0;
	vertices[0].g = vertices[3].g = vertices[4].g = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[3], &vertices[4], kQATriFlags_None);
	
	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3CheckTouchTrisReverse
	
	Similar to S3CheckTouchTris except that the triangles are drawn in the
	reverse order. This ensures that adjacent triangles only draw to a boundary
	pixel once. If any of the pixels flip color between this test and the
	previous one, there is a problem with the coverage of the pixel.
\*===========================================================================*/

S3tTestStatus S3CheckTouchTrisReverse(S3tTestInfo *testInfo) {
	float xmid, xfar;
	float ymid, yfar;
	TQAVGouraud vertices[9] = {
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 0.0, 1.0 },
	};
	
	xmid = (float) S3TestInfo_GetWidth(testInfo) / 2.0;
	xfar = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	ymid = (float) S3TestInfo_GetHeight(testInfo) / 2.0;
	yfar = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	
	vertices[1].x = xmid;
	vertices[2].x = xfar;
	
	vertices[3].y = vertices[4].y = vertices[5].y = ymid;
	vertices[4].x = xmid;
	vertices[5].x = xfar;

	vertices[6].y = vertices[7].y = vertices[8].y = yfar;
	vertices[7].x = xmid;
	vertices[8].x = xfar;
	
	S3Test_RenderStart(testInfo, NULL);
	
	vertices[0].g = vertices[3].g = vertices[4].g = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[3], &vertices[4], kQATriFlags_None);

	vertices[0].g = vertices[3].g = vertices[4].g = 0.0;
	vertices[3].r = vertices[4].r = vertices[6].r = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[3], &vertices[4], &vertices[6], kQATriFlags_None);
	
	vertices[3].r = vertices[4].r = vertices[6].r = 0.0;
	vertices[4].b = vertices[6].b = vertices[7].b = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[4], &vertices[6], &vertices[7], kQATriFlags_None);
	
	vertices[4].b = vertices[46].b = vertices[7].b = 0.0;
	vertices[4].g = vertices[7].g = vertices[8].g = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[4], &vertices[7], &vertices[8], kQATriFlags_None);
	
	vertices[4].g = vertices[7].g = vertices[8].g = 0.0;
	vertices[4].r = vertices[5].r = vertices[8].r = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[4], &vertices[5], &vertices[8], kQATriFlags_None);
	
	vertices[4].r = vertices[5].r = vertices[8].r = 0.0;
	vertices[2].b = vertices[4].b = vertices[5].b = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[2], &vertices[4], &vertices[5], kQATriFlags_None);
	
	vertices[2].b = vertices[4].b = vertices[5].b = 0.0;
	vertices[1].g = vertices[2].g = vertices[4].g = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[1], &vertices[2], &vertices[4], kQATriFlags_None);
	
	vertices[1].g = vertices[2].g = vertices[4].g = 0.0;
	vertices[0].r = vertices[1].r = vertices[4].r = 1.0;
	S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[4], kQATriFlags_None);
	
	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3SkinnyHorizontalTris
	
	Draws a whole series of very narrow almost horizontal triangles which fill
	up almost the whole window. The step size of the vertices along the left
	and right edge of the window differs.
\*===========================================================================*/

S3tTestStatus S3SkinnyHorizontalTris(S3tTestInfo *testInfo) {
	TQAVGouraud *vertices;
	unsigned long numVerts;
	unsigned long i;
	float leftStep, rightStep;
	float width =  (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float check;
	
	if (height <= DIVDEC1 || height <= DIVDEC2) {
		S3Test_Print("The window is too short for this test!");
		return kS3Status_Failure;
	}
	leftStep = height / (height - DIVDEC1);
	rightStep = height / (height - DIVDEC2);
	numVerts = (unsigned long) ((height / leftStep) + (height / rightStep)) + 2;
	vertices = (TQAVGouraud *) calloc(sizeof(TQAVGouraud), numVerts);
	for (i = 0; i < numVerts; i++) {
		if ((i % 2) == 0) {
			check = leftStep * (i / 2);
			if (check > height) {
				numVerts = i;
				break;
			} else {
				vertices[i].x = 0.0;
				vertices[i].y = check;
			}
		} else {
			check = rightStep * ((i - 1) / 2);
			if (check > height) {
				numVerts = i;
				break;
			} else {
				vertices[i].x = width;
				vertices[i].y = check;
			}
		}
		vertices[i].z = 0.1;
		vertices[i].invW = 10.0;
		vertices[i].r = vertices[i].g = vertices[i].b = vertices[i].a = 1.0;
	}
	
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < numVerts - 2; i++) {
		S3Test_DrawTriGouraud(testInfo, &vertices[i], &vertices[i + 1], &vertices[i + 2],
			kQATriFlags_None);
	}
	S3Test_RenderEnd(testInfo, NULL);
	
	free(vertices);
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3SkinnyVerticalTris
	
	Similar to S3SkinnyHorizontalTris except that the triangles are now almost
	vertical.
\*===========================================================================*/

S3tTestStatus S3SkinnyVerticalTris(S3tTestInfo *testInfo) {
	TQAVGouraud *vertices;
	unsigned long numVerts;
	unsigned long i;
	float bottomStep, topStep;
	float width =  (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float check;
	
	if (width <= DIVDEC1 || width <= DIVDEC2) {
		S3Test_Print("The window is too narrow for this test!");
		return kS3Status_Failure;
	}
	bottomStep = width / (width - DIVDEC1);
	topStep = width / (width - DIVDEC2);
	numVerts = (unsigned long) ((width / bottomStep) + (width / topStep)) + 2;
	vertices = (TQAVGouraud *) calloc(sizeof(TQAVGouraud), numVerts);
	for (i = 0; i < numVerts; i++) {
		if ((i % 2) == 0) {
			check = bottomStep * (i / 2);
			if (check > width) {
				numVerts = i;
				break;
			} else {
				vertices[i].x = check;
				vertices[i].y = height;
			}
		} else {
			check = topStep * ((i - 1) / 2);
			if (check > width) {
				numVerts = i;
				break;
			} else {
				vertices[i].x = check;
				vertices[i].y = 0.0;
			}
		}
		vertices[i].z = 0.1;
		vertices[i].invW = 10.0;
		vertices[i].r = vertices[i].g = vertices[i].b = vertices[i].a = 1.0;
	}
	
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < numVerts - 2; i++) {
		S3Test_DrawTriGouraud(testInfo, &vertices[i], &vertices[i + 1], &vertices[i + 2],
			kQATriFlags_None);
	}
	S3Test_RenderEnd(testInfo, NULL);
	
	free(vertices);
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3MultipleModeTris
	
	Draws various Gouraud, white textured and decal triangles at various z
	values. If the z-compare mode is set to true, the order of the triangles
	should be reversed.
\*===========================================================================*/

S3tTestStatus S3MultipleModeTris(S3tTestInfo *testInfo) {
	short i;
	S3Image theImages[3];
	float width = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float widthStep = width / 11.0;
	float ymid = height / 2.0;
	TQAVGouraud GVertices[12] = {
		{ 0.0, 0.0, 0.05, 1.0 / 0.05,
		  1.0, 0.0, 0.0, 1.0 },
		{ 2.0, 0.0, 0.05, 1.0 / 0.05,
		  0.0, 1.0, 0.0, 1.0 },
		{ 2.0, 0.0, 0.05, 1.0 / 0.05,
		  0.0, 0.0, 1.0, 1.0 },
		  
		{ 4.0, 0.0, 0.45, 1.0 / 0.45,
		  0.0, 1.0, 0.0, 1.0 },
		{ 6.0, 0.0, 0.45, 1.0 / 0.45,
		  0.0, 0.0, 1.0, 1.0 },
		{ 6.0, 0.0, 0.45, 1.0 / 0.45,
		  1.0, 0.0, 0.0, 1.0 },
		  
		{ 7.0, 0.0, 0.75, 1.0 / 0.75,
		  0.0, 0.0, 1.0, 1.0 },
		{ 9.0, 0.0, 0.75, 1.0 / 0.75,
		  1.0, 0.0, 0.0, 1.0 },
		{ 9.0, 0.0, 0.75, 1.0 / 0.75,
		  0.0, 1.0, 0.0, 1.0 },
		  
		{ 9.0, 0.0, 0.95, 1.0 / 0.95,
		  1.0, 0.0, 0.0, 1.0 },
		{ 11.0, 0.0, 0.95, 1.0 / 0.95,
		  0.0, 1.0, 0.0, 1.0 },
		{ 11.0, 0.0, 0.95, 1.0 / 0.95,
		  0.0, 0.0, 1.0, 1.0 },
	};
	TQAVTexture TVertices[18] = {
		{ 1.0, 0.0, 0.15, 1.0 / 0.15,
		  1.0, 0.0, 0.0, 1.0,
		  0.5 / 0.15, 1.0 / 0.15,
		  0.5, 0.5, 0.5,
		  0.0, 0.0, 0.0 },
		{ 3.0, 0.0, 0.15, 1.0 / 0.15,
		  0.0, 1.0, 0.0, 1.0,
		  1.0 / 0.15, 0.0,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 3.0, 0.0, 0.15, 1.0 / 0.15,
		  0.0, 0.0, 1.0, 1.0,
		  0.0, 1.0 / 0.15,
		  0.0, 0.0, 0.0,
		  0.0, 0.0, 0.0 },
		  
		{ 2.0, 0.0, 0.25, 1.0 / 0.25,
		  0.0, 1.0, 0.0, 1.0,
		  0.0, 1.0 / 0.25,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 4.0, 0.0, 0.25, 1.0 / 0.25,
		  0.0, 0.0, 1.0, 1.0,
		  0.5 / 0.25, 1.0 / 0.25,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 4.0, 0.0, 0.25, 1.0 / 0.25,
		  1.0, 0.0, 0.0, 1.0,
		  1.0 / 0.25, 0.0,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		  
		{ 3.0, 0.0, 0.35, 1.0 / 0.35,
		  0.0, 0.0, 1.0, 1.0,
		  1.0 / 0.35, 0.0,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 5.0, 0.0, 0.35, 1.0 / 0.35,
		  1.0, 0.0, 0.0, 1.0,
		  0.0, 1.0 / 0.35,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 5.0, 0.0, 0.35, 1.0 / 0.35,
		  0.0, 1.0, 0.0, 1.0,
		  0.5 / 0.35, 1.0 / 0.35,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		  
		{ 5.0, 0.0, 0.55, 1.0 / 0.55,
		  1.0, 0.0, 0.0, 1.0,
		  0.5 / 0.55, 1.0 / 0.55,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 7.0, 0.0, 0.55, 1.0 / 0.55,
		  0.0, 1.0, 0.0, 1.0,
		  1.0 / 0.55, 0.0,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 7.0, 0.0, 0.55, 1.0 / 0.55,
		  0.0, 0.0, 1.0, 1.0,
		  0.0, 1.0 / 0.55,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		  
		{ 6.0, 0.0, 0.65, 1.0 / 0.65,
		  0.0, 1.0, 0.0, 1.0,
		  0.0, 1.0 / 0.65,
		  0.3, 0.3, 0.3,
		  0.1, 0.1, 0.1 },
		{ 8.0, 0.0, 0.65, 1.0 / 0.65,
		  0.0, 0.0, 1.0, 1.0,
		  0.5 / 0.65, 1.0 / 0.65,
		  0.5, 0.5, 0.5,
		  0.0, 0.0, 0.0 },
		{ 8.0, 0.0, 0.65, 1.0 / 0.65,
		  1.0, 0.0, 0.0, 1.0,
		  1.0 / 0.65, 0.0,
		  0.0, 0.0, 0.0,
		  1.0, 1.0, 1.0 },
		  
		{ 8.0, 0.0, 0.85, 1.0 / 0.85,
		  0.0, 0.0, 1.0, 1.0,
		  1.0 / 0.85, 0.0,
		  1.0, 0.4, 0.3,
		  0.7, 0.4, 0.1 },
		{ 10.0, 0.0, 0.85, 1.0 / 0.85,
		  1.0, 0.0, 0.0, 1.0,
		  0.0, 1.0 / 0.85,
		  0.2, 0.5, 1.0,
		  0.0, 0.1, 0.5 },
		{ 10.0, 0.0, 0.85, 1.0 / 0.85,
		  0.0, 1.0, 0.0, 1.0,
		  0.5 / 0.85, 1.0 / 0.85,
		  0.0, 0.5, 0.0,
		  0.9, 0.5, 0.1 },
	};
	
	for (i = 0; i < 12; i += 3) {
		GVertices[i].y = ymid;
		GVertices[i + 2].y = height;
	}
	for (i = 1; i < 12; i++) GVertices[i].x *= widthStep;
	
	for (i = 0; i < 18; i += 3) {
		TVertices[i].y = ymid;
		TVertices[i + 2].y = height;
	}
	for (i = 0; i < 18; i++) TVertices[i].x *= widthStep;
	
	if (GetTriTextures(theImages, testInfo) == kS3Status_Failure) return kS3Status_Failure;
	
	S3Test_RenderStart(testInfo, NULL);
	S3Test_DrawTriGouraud(testInfo, &GVertices[0], &GVertices[1], &GVertices[2],
		kQATriFlags_None);
	
	S3Test_SetInt(testInfo, kQATag_TextureOp, kQATextureOp_Modulate);
	if (DrawTriTexture(theImages, 0, TVertices, 0, 1, 2, testInfo) == kS3Status_Failure) {
		S3Test_RenderEnd(testInfo, NULL);
		for (i = 0; i < 3; i++) S3Image_Delete(theImages[i]);
		return kS3Status_Failure;
	}
	
	S3Test_SetInt(testInfo, kQATag_TextureOp, kQATextureOp_Decal);
	if (DrawTriTexture(theImages, 1, TVertices, 3, 4, 5, testInfo) == kS3Status_Failure) {
		S3Test_RenderEnd(testInfo, NULL);
		for (i = 0; i < 3; i++) S3Image_Delete(theImages[i]);
		return kS3Status_Failure;
	}
	
	S3Test_SetInt(testInfo, kQATag_TextureOp, kQATextureOp_None);
	if (DrawTriTexture(theImages, 2, TVertices, 6, 7, 8, testInfo) == kS3Status_Failure) {
		S3Test_RenderEnd(testInfo, NULL);
		for (i = 0; i < 3; i++) S3Image_Delete(theImages[i]);
		return kS3Status_Failure;
	}
	
	S3Test_DrawTriGouraud(testInfo, &GVertices[3], &GVertices[4], &GVertices[5],
		kQATriFlags_None);
	
	if (DrawTriTexture(theImages, 1, TVertices, 9, 10, 11, testInfo) == kS3Status_Failure) {
		S3Test_RenderEnd(testInfo, NULL);
		for (i = 0; i < 3; i++) S3Image_Delete(theImages[i]);
		return kS3Status_Failure;
	}
	
	S3Test_SetInt(testInfo, kQATag_TextureOp, kQATextureOp_Modulate | kQATextureOp_Highlight);
	if (DrawTriTexture(theImages, 0, TVertices, 12, 13, 14, testInfo) == kS3Status_Failure) {
		S3Test_RenderEnd(testInfo, NULL);
		for (i = 0; i < 3; i++) S3Image_Delete(theImages[i]);
		return kS3Status_Failure;
	}
	
	S3Test_DrawTriGouraud(testInfo, &GVertices[6], &GVertices[7], &GVertices[8],
		kQATriFlags_None);
	
	S3Test_SetInt(testInfo, kQATag_TextureOp, kQATextureOp_Decal | kQATextureOp_Modulate |
		kQATextureOp_Highlight);
	if (DrawTriTexture(theImages, 2, TVertices, 15, 16, 17, testInfo) == kS3Status_Failure) {
		S3Test_RenderEnd(testInfo, NULL);
		for (i = 0; i < 3; i++) S3Image_Delete(theImages[i]);
		return kS3Status_Failure;
	}
	
	S3Test_DrawTriGouraud(testInfo, &GVertices[9], &GVertices[10], &GVertices[11],
		kQATriFlags_None);
	
	S3Test_RenderEnd(testInfo, NULL);

	for (i = 0; i < 3; i++) S3Image_Delete(theImages[i]);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3Pie
	
	Reads in the vertices of triangles which form a circle and draws them to
	the window if the window is big enough. Checks that there are no holes in
	the circle to ensure correct coverage.
\*===========================================================================*/

S3tTestStatus S3Pie(S3tTestInfo *testInfo) {
	FILE *theFile;
	short i;
	TQAVGouraud vertices[600];
	float width = (float) S3TestInfo_GetWidth(testInfo);
	float height = (float) S3TestInfo_GetHeight(testInfo);
	
	theFile = fopen(":pie.tri", "r");
	if (!theFile) {
		S3Test_Print("Unable to open file, pie.tri!");
		return kS3Status_Failure;
	}
	for (i = 0; i < 600; i += 3) {
		vertices[i].z = vertices[i + 1].z = vertices[i + 2].z = 0.1;
		vertices[i].invW = vertices[i + 1].invW = vertices[i + 2].invW = 10.0;
		vertices[i].r = vertices[i].g = vertices[i].b = vertices[i].a = 1.0;
		vertices[i + 1].r = vertices[i + 1].g = vertices[i + 1].b = vertices[i + 1].a = 1.0;
		vertices[i + 2].r = vertices[i + 2].g = vertices[i + 2].b = vertices[i + 2].a = 1.0;
		if (fscanf(theFile, "%f %f %f %f %f %f\n", &(vertices[i].x), &(vertices[i].y),
			&(vertices[i + 1].x), &(vertices[i + 1].y), &(vertices[i + 2].x),
			&(vertices[i + 2].y)) != 6) {
			S3Test_Print("File is in incorrect format: x1 y1 x2 y2 x3 y3<return>");
			return kS3Status_Failure;
		}
		if (vertices[i].x >= width || vertices[i + 1].x >= width || vertices[i + 2].x
			>= width) {
			S3Test_Print("Window is too narrow!");
			return kS3Status_Failure;
		}
		if (vertices[i].y >= height || vertices[i + 1].y >= height || vertices[i + 2].y
			>= height) {
			S3Test_Print("Window is too short!");
			return kS3Status_Failure;
		}
	}
	fclose(theFile);
	
	S3Test_RenderStart(testInfo, NULL);
	
	for (i = 0; i < 600; i += 3) {
		S3Test_DrawTriGouraud(testInfo, &vertices[i], &vertices[i + 1], &vertices[i + 2],
			kQATriFlags_None);
	}
	
	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3LineBorder
	
	Draws 4 lines around the border of the window and a line diagonally across
	the center of the window.
\*===========================================================================*/

S3tTestStatus S3LineBorder(S3tTestInfo *testInfo) {
	short i;
	float width = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	TQAVGouraud vertices[6] = {
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 0.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 1.0, 0.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.0, 1.0,
		  1.0, 1.0, 1.0, 1.0 },
	};
	
	vertices[1].x = width;
	vertices[2].x = width;
	vertices[2].y = height;
	vertices[3].y = height;
	vertices[4].x = width / 3;
	vertices[4].y = height / 3;
	vertices[5].x = 2 * width / 3;
	vertices[5].y = 2 * height / 3;
	
	S3Test_RenderStart(testInfo, NULL);
	
	for (i = 0; i < 4; i++) S3Test_DrawLine(testInfo, &vertices[i], &vertices[(i + 1) % 4]);
	S3Test_DrawLine(testInfo, &vertices[4], &vertices[5]);

	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3Coverage
	
	Draws really thin triangles in a circle such that parts of the triangle
	start to disappear which indicates aliasing.
\*===========================================================================*/

S3tTestStatus S3Coverage(S3tTestInfo *testInfo) {
	float width = (float) S3TestInfo_GetWidth(testInfo);
	float height = (float) S3TestInfo_GetHeight(testInfo);
	float xmid = width / 2.0;
	float ymid = height / 2.0;
	float degree;
	TQAVGouraud vertices[3] = {
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
	};
	
	if ((xmid - RADIUS) < 0 || (xmid + RADIUS) >= width || (ymid - RADIUS) < 0 ||
		(ymid + RADIUS) >= height) {
		S3Test_Print("The window is too small!");
		return kS3Status_Failure;
	}
	vertices[0].x = xmid;
	vertices[0].y = ymid;
	S3Test_RenderStart(testInfo, NULL);
	for (degree = 0; degree < 360; degree += STEP) {
		vertices[1].x = xmid + cos((double) degree * PI / 180.0) * RADIUS;
		vertices[1].y = ymid + sin((double) degree * PI / 180.0) * RADIUS;
		vertices[2].x = xmid + cos((double) (degree + INCREMENT) * PI / 180.0) * RADIUS;
		vertices[2].y = ymid + sin((double) (degree + INCREMENT) * PI / 180.0) * RADIUS;
		S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[2],
			kQATriFlags_None);
	}
	S3Test_RenderEnd(testInfo, NULL);
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3ColoredHK
	
	Generates triangles similar to those generated in S3TouchingTris except
	these are color textured and have vertices at varying depths.
\*===========================================================================*/

S3tTestStatus S3ColoredHK(S3tTestInfo *testInfo) {
	float xmid, xfar;
	float ymid, yfar;
	TQAVTexture vertices[9] = {
		{ 0.0, 0.0, 0.1, 1.0 / 0.1,
		  0.0, 0.0, 0.0, 1.0,
		  0.0, 1.0 / 0.1,
		  0.0, 0.5, 0.0,
		  0.0, 0.0, 0.0 },
		{ 0.0, 0.0, 0.5, 1.0 / 0.5,
		  0.0, 0.0, 0.0, 1.0,
		  0.5 / 0.5, 1.0 / 0.5,
		  0.5, 0.5, 0.0,
		  0.0, 0.5, 0.0 },
		{ 0.0, 0.0, 0.3, 1.0 / 0.3,
		  0.0, 0.0, 0.0, 1.0,
		  1.0 / 0.3, 1.0 / 0.3,
		  1.0, 0.5, 0.0,
		  0.0, 1.0, 0.0 },
		{ 0.0, 0.0, 0.9, 1.0 / 0.9,
		  0.0, 0.0, 0.0, 1.0,
		  0.0, 0.5 / 0.9,
		  0.0, 0.5, 0.5,
		  0.0, 0.5, 0.0 },
		{ 0.0, 0.0, 0.7, 1.0 / 0.7,
		  0.0, 0.0, 0.0, 1.0,
		  0.5 / 0.7, 0.5 / 0.7,
		  0.5, 0.5, 0.5,
		  0.0, 0.75, 0.0 },
		{ 0.0, 0.0, 0.2, 1.0 / 0.2,
		  0.0, 0.0, 0.0, 1.0,
		  1.0 / 0.2, 0.5 / 0.2,
		  1.0, 0.5, 0.5,
		  0.0, 1.0, 0.0 },
		{ 0.0, 0.0, 0.6, 1.0 / 0.6,
		  0.0, 0.0, 0.0, 1.0,
		  0.0, 0.0,
		  0.0, 0.5, 1.0,
		  0.0, 1.0, 0.0 },
		{ 0.0, 0.0, 0.4, 1.0 / 0.4,
		  0.0, 0.0, 0.0, 1.0,
		  0.5 / 0.4, 0.0,
		  0.5, 0.5, 1.0,
		  0.0, 1.0, 0.0 },
		{ 0.0, 0.0, 0.8, 1.0 / 0.8,
		  0.0, 0.0, 0.0, 1.0,
		  1.0 / 0.8, 0.0,
		  1.0, 0.5, 1.0,
		  0.0, 1.0, 0.0 },
	};
	
	xmid = (float) S3TestInfo_GetWidth(testInfo) / 2.0;
	xfar = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	ymid = (float) S3TestInfo_GetHeight(testInfo) / 2.0;
	yfar = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	
	vertices[1].x = xmid;
	vertices[2].x = xfar;
	
	vertices[3].y = vertices[4].y = vertices[5].y = ymid;
	vertices[4].x = xmid;
	vertices[5].x = xfar;

	vertices[6].y = vertices[7].y = vertices[8].y = yfar;
	vertices[7].x = xmid;
	vertices[8].x = xfar;
	
	S3Test_RenderStart(testInfo, NULL);
	
	S3Test_DrawTriTexture(testInfo, &vertices[0], &vertices[1], &vertices[4], kQATriFlags_None);
	
	S3Test_DrawTriTexture(testInfo, &vertices[1], &vertices[2], &vertices[4], kQATriFlags_None);
	
	S3Test_DrawTriTexture(testInfo, &vertices[2], &vertices[4], &vertices[5], kQATriFlags_None);
	
	S3Test_DrawTriTexture(testInfo, &vertices[4], &vertices[5], &vertices[8], kQATriFlags_None);
	
	S3Test_DrawTriTexture(testInfo, &vertices[4], &vertices[7], &vertices[8], kQATriFlags_None);
	
	S3Test_DrawTriTexture(testInfo, &vertices[4], &vertices[6], &vertices[7], kQATriFlags_None);
	
	S3Test_DrawTriTexture(testInfo, &vertices[3], &vertices[4], &vertices[6], kQATriFlags_None);
	
	S3Test_DrawTriTexture(testInfo, &vertices[0], &vertices[3], &vertices[4], kQATriFlags_None);
	
	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3SideRect
	
	Generates two white textured triangles which have a coincident vertical
	edge. The uv's are negative to test texture wrapping.
\*===========================================================================*/

S3tTestStatus S3SideRect(S3tTestInfo *testInfo) {
	float xmax = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float ymax = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float xmid = xmax / 2.0;
	float ymid = ymax / 2.0;
	TQAVTexture vertices[4] = {
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 0.0, 0.0, 1.0,
		  0.0, 0.0,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 1.0, 0.0, 1.0,
		  0.0, -1.0 * 10.0,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  0.0, 0.0, 1.0, 1.0,
		  -1.0 * 10.0, -1.0 * 10.0,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0,
		  -1.0 * 10.0, 0.0,
		  1.0, 1.0, 1.0,
		  0.0, 0.0, 0.0 },
	};
	
	vertices[0].y = vertices[2].y = ymid;
	vertices[3].y = ymax;
	vertices[1].x = vertices[3].x = xmid;
	vertices[2].x = xmax;
	
	S3Test_RenderStart(testInfo, NULL);
	
	S3Test_DrawTriTexture(testInfo, &vertices[0], &vertices[1], & vertices[3], kQATriFlags_None);
	S3Test_DrawTriTexture(testInfo, &vertices[1], &vertices[2], & vertices[3], kQATriFlags_None);

	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3SlowTris
	
	Similar to S3SkinnyVerticalTris except that the triangles are white
	textured. This will probably give an indication of the slowest speed of
	the rasterizer.
\*===========================================================================*/

S3tTestStatus S3SlowTris(S3tTestInfo *testInfo) {
	TQAVTexture *vertices;
	unsigned long numVerts;
	unsigned long i;
	float bottomStep, topStep;
	float width =  (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float check;
	
	if (width <= DIVDEC1 || width <= DIVDEC2) {
		S3Test_Print("The window is too narrow for this test!");
		return kS3Status_Failure;
	}
	bottomStep = width / (width - DIVDEC1);
	topStep = width / (width - DIVDEC2);
	numVerts = (unsigned long) ((width / bottomStep) + (width / topStep)) + 2;
	vertices = (TQAVTexture *) calloc(sizeof(TQAVTexture), numVerts);
	for (i = 0; i < numVerts; i++) {
		if ((i % 2) == 0) {
			check = bottomStep * (i / 2);
			if (check > width) {
				numVerts = i;
				break;
			} else {
				vertices[i].x = check;
				vertices[i].y = height;
			}
		} else {
			check = topStep * ((i - 1) / 2);
			if (check > width) {
				numVerts = i;
				break;
			} else {
				vertices[i].x = check;
				vertices[i].y = 0.0;
			}
		}
		vertices[i].z = 0.1;
		vertices[i].invW = 10.0;
		vertices[i].uOverW = vertices[i].invW * vertices[i].y / height;
		vertices[i].vOverW = vertices[i].invW * vertices[i].x / width;
	}
	
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < numVerts - 2; i++) {
		S3Test_DrawTriTexture(testInfo, &vertices[i], &vertices[i + 1], &vertices[i + 2],
			kQATriFlags_None);
	}
	S3Test_RenderEnd(testInfo, NULL);
	
	free(vertices);
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3FastestTris25
	
	A single line triangle with 25 pixels with the frontmost drawn first. This
	will probably give the fastest drawing speed of the rasterizer.
\*===========================================================================*/

S3tTestStatus S3FastestTris25(S3tTestInfo *testInfo) {
	TQAVGouraud vertices[3] = {
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 }
	};
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float width = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float zincr;
	int i;
	
	vertices[0].y = (height / 2.0) - 0.5;
	vertices[1].y = (height / 2.0) + 0.5;
	vertices[2].y = height / 2.0;
	vertices[2].x = 4.0 * width / 5.0;
	zincr = 0.9 / NUMTRIS;
	
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < NUMTRIS; i++) {
		S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[2],
			kQATriFlags_None);
		vertices[0].z = vertices[1].z = vertices[2].z = vertices[0].z + zincr;
		vertices[0].invW = vertices[1].invW = vertices[2].invW = 1.0 / vertices[0].z;
	}
	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3FasterTris25
	
	Similar to S3FastestTris25 except th triangles are drawn in back to front
	order. This also gives an indication of drawing speed.
\*===========================================================================*/

S3tTestStatus S3FasterTris25(S3tTestInfo *testInfo) {
	TQAVGouraud vertices[3] = {
		{ 0.0, 0.0, 1.0, 1.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 1.0, 1.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 1.0, 1.0,
		  1.0, 1.0, 1.0, 1.0 }
	};
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float width = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float zincr;
	int i;
	
	vertices[0].y = (height / 2.0) - 0.5;
	vertices[1].y = (height / 2.0) + 0.5;
	vertices[2].y = height / 2.0;
	vertices[2].x = 4.0 * width / 5.0;
	zincr = 0.9 / NUMTRIS;
	
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < NUMTRIS; i++) {
		S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[2],
			kQATriFlags_None);
		vertices[0].z = vertices[1].z = vertices[2].z = vertices[0].z - zincr;
		vertices[0].invW = vertices[1].invW = vertices[2].invW = 1.0 / vertices[0].z;
	}
	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3FastestTris50
	
	Similar to S3FastestTris255 except 50 pixel triangles are drawn.
\*===========================================================================*/

S3tTestStatus S3FastestTris50(S3tTestInfo *testInfo) {
	TQAVGouraud vertices[3] = {
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 0.1, 10.0,
		  1.0, 1.0, 1.0, 1.0 }
	};
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float width = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float zincr;
	int i;
	
	vertices[0].y = (height / 2.0) - 0.55;
	vertices[1].y = (height / 2.0) + 0.55;
	vertices[2].y = height / 2.0;
	vertices[2].x = width / 7.3;
	zincr = 0.9 / NUMTRIS;
	
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < NUMTRIS; i++) {
		S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[2],
			kQATriFlags_None);
		vertices[0].z = vertices[1].z = vertices[2].z = vertices[0].z + zincr;
		vertices[0].invW = vertices[1].invW = vertices[2].invW = 1.0 / vertices[0].z;
	}
	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3FasterTris50
	
	Similar to S3FasterTris25 except 50 pixel triangles are drawn.
\*===========================================================================*/

S3tTestStatus S3FasterTris50(S3tTestInfo *testInfo) {
	TQAVGouraud vertices[3] = {
		{ 0.0, 0.0, 1.0, 1.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 1.0, 1.0,
		  1.0, 1.0, 1.0, 1.0 },
		{ 0.0, 0.0, 1.0, 1.0,
		  1.0, 1.0, 1.0, 1.0 }
	};
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	float width = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float zincr;
	int i;
	
	vertices[0].y = (height / 2.0) - 0.55;
	vertices[1].y = (height / 2.0) + 0.55;
	vertices[2].y = height / 2.0;
	vertices[2].x = width / 7.3;
	zincr = 0.9 / NUMTRIS;
	
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < NUMTRIS; i++) {
		S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[2],
			kQATriFlags_None);
		vertices[0].z = vertices[1].z = vertices[2].z = vertices[0].z - zincr;
		vertices[0].invW = vertices[1].invW = vertices[2].invW = 1.0 / vertices[0].z;
	}
	S3Test_RenderEnd(testInfo, NULL);
	
	return kS3Status_Completed;
}


/*===========================================================================*\
	S3Random25Tris
	
	25 pixel triangles drawn random around screen. Gives an indication of
	25-pixel rasterization rate.
\*===========================================================================*/

S3tTestStatus S3Random25Tris(S3tTestInfo *testInfo) {
	float width = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	TQAVGouraud vertices[3];
	int i;
	
	vertices[0].a = vertices[1].a = vertices[2].a = 1.0;
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < NUMTRIS; i++) {
		GetVertices(vertices, width, height, 10.0, 5.0);
		S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[2],
			kQATriFlags_None);
	}
	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}


/*===========================================================================*\
	S3Random49Tris
	
	Similar to S3Random25Tris except 49 pixel triangles are used.
\*===========================================================================*/

S3tTestStatus S3Random49Tris(S3tTestInfo *testInfo) {
	float width = (float) S3TestInfo_GetWidth(testInfo) - 0.1;
	float height = (float) S3TestInfo_GetHeight(testInfo) - 0.1;
	TQAVGouraud vertices[3];
	int i;
	
	vertices[0].a = vertices[1].a = vertices[2].a = 1.0;
	S3Test_RenderStart(testInfo, NULL);
	for (i = 0; i < NUMTRIS; i++) {
		GetVertices(vertices, width, height, 14.0, 7.0);
		S3Test_DrawTriGouraud(testInfo, &vertices[0], &vertices[1], &vertices[2],
			kQATriFlags_None);
	}
	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}
