/*****************************************************************************\

	S3TextureBitmapTests.c

	Texture, Bitmap and Color Lookup Table Tests
	
	Copyright (C) 1995 Apple Computer, Inc.  All rights reserved.

\*****************************************************************************/


/*=========================================================================*\
	Includes
\*=========================================================================*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "S3TestSystem.h"
#include "S3TestInfo.h"
#include "S3TestUtil.h"
#include "S3Utilities.h"
#include "S3ColorTable.h"
#include "S3Alert.h"


/*=========================================================================*\
	Constants
\*=========================================================================*/

#define kNPoints					10000
#define kNLines						1000
#define kNTris						100
#define kNFlushes					5


/*=========================================================================*\
	Prototypes
\*=========================================================================*/

void S3TextureBitmapTests_Register(void);
void S3SetupOneTexturedRectangle(
	S3tTestInfo				*testInfo,
	TQAVTexture				inTextureVerts[]);
S3tTestStatus S3OneRectangle(
	S3tTestInfo				*testInfo);
S3tTestStatus S3OneBitmap(
	S3tTestInfo				*testInfo);
S3tTestStatus S3OneAlphaRectangle(
	S3tTestInfo				*testInfo);
S3tTestStatus S3AnimatedTextureColorTable(
	S3tTestInfo				*testInfo);
S3tTestStatus S3NineBitmaps(
	S3tTestInfo				*testInfo);
S3tTestStatus S3ShowAllMipMaps(
	S3tTestInfo				*testInfo);
void S3DrawBitmapsAtZ(
	S3tTestInfo				*testInfo,
	float					inZ);
S3tTestStatus S3BitmapsBehindTriGouraud(
	S3tTestInfo				*testInfo);
S3tTestStatus S3BitmapsInFrontTriGouraud(
	S3tTestInfo				*testInfo);
S3tTestStatus S3TestShowColorTable(
	S3tTestInfo				*testInfo,
	TQAImagePixelType		inPixelType,
	TQAColorTableType		inColorTableType,
	TQAColorTable*			inLeftColorTable,
	TQAColorTable*			inRightColorTable,
	Boolean					inUseLeftEngine,
	Boolean					inUseRightEngine);
S3tTestStatus S3TestShowTextureColorTable(
	S3tTestInfo				*testInfo);
S3tTestStatus S3TestShowBitmapColorTable(
	S3tTestInfo				*testInfo);
S3tTestStatus S3TestShowTextureAndColorTable(
	S3tTestInfo				*testInfo);
S3tTestStatus S3TestShowBitmapAndColorTable(
	S3tTestInfo				*testInfo);
S3tTestStatus S3AnimatedBitmapColorTable(
	S3tTestInfo				*testInfo);
S3tTestStatus S3TestColorTableTexture(
	S3tTestInfo				*testInfo);

	
/*=========================================================================*\
	S3TextureBitmapTests_Register
\*=========================================================================*/
void S3TextureBitmapTests_Register(void)
{
	S3Test_Register("One Textured Rectangle",		kS3Prereq_Texture, S3OneRectangle);
	S3Test_Register("One Alpha Textured Rectangle",	kS3Prereq_Texture, S3OneAlphaRectangle);
	S3Test_Register("Show All MipMaps",				kS3Prereq_Texture, S3ShowAllMipMaps);
	S3Test_Register("Show Texture Color Table",		kS3Prereq_Texture, S3TestShowTextureColorTable);
	S3Test_Register("Show Texture And Color Table",	kS3Prereq_Texture, S3TestShowTextureAndColorTable);
	S3Test_Register("Animated Texture Color Table",	kS3Prereq_Texture, S3AnimatedTextureColorTable);

		// add a divider to the list of tests by passing nil
	S3Test_Register("",								kS3Prereq_None, nil);
	
	S3Test_Register("One Bitmap",					kS3Prereq_Bitmap, S3OneBitmap);
	S3Test_Register("Nine Bitmaps",					kS3Prereq_Bitmap, S3NineBitmaps);
	S3Test_Register("Bitmaps Behind Tris",			kS3Prereq_Bitmap, S3BitmapsBehindTriGouraud);
	S3Test_Register("Bitmaps in Front of Tris",		kS3Prereq_Bitmap, S3BitmapsInFrontTriGouraud);
	S3Test_Register("Show Bitmap Color Table",		kS3Prereq_Bitmap, S3TestShowBitmapColorTable);
	S3Test_Register("Show Bitmap And Color Table",	kS3Prereq_Bitmap, S3TestShowBitmapAndColorTable);
	S3Test_Register("Animated Bitmap Color Table",	kS3Prereq_Bitmap, S3AnimatedBitmapColorTable);
}

/*=========================================================================*\
	S3SetupOneTexturedRectangle
\*=========================================================================*/
void S3SetupOneTexturedRectangle(
	S3tTestInfo				*testInfo,
	TQAVTexture				inTextureVerts[])
{
	float left = (float) S3TestInfo_GetWidth(testInfo) * 0.1;
	float top = (float) S3TestInfo_GetHeight(testInfo) * 0.1;
	float right = (float) S3TestInfo_GetWidth(testInfo) * 0.9;
	float bottom = (float) S3TestInfo_GetHeight(testInfo) * 0.9;
	float z = 0.1, invW = 1.0 / z;
	
	inTextureVerts[0].x = left;
	inTextureVerts[0].y = bottom;
	inTextureVerts[0].z = z;
	inTextureVerts[0].invW = invW;
	inTextureVerts[0].r = 1.0;
	inTextureVerts[0].g = 0.0;
	inTextureVerts[0].b = 0.0;
	inTextureVerts[0].a = 1.0;
	inTextureVerts[0].uOverW = 0.0;
	inTextureVerts[0].vOverW = 0.0;
	inTextureVerts[0].kd_r = 1.0;
	inTextureVerts[0].kd_g = 1.0;
	inTextureVerts[0].kd_b = 1.0;
	inTextureVerts[0].ks_r = 0.0;
	inTextureVerts[0].ks_g = 0.0;
	inTextureVerts[0].ks_b = 0.0;

	inTextureVerts[1].x = left;
	inTextureVerts[1].y = top;
	inTextureVerts[1].z = z;
	inTextureVerts[1].invW = invW;
	inTextureVerts[1].r = 0.0;
	inTextureVerts[1].g = 1.0;
	inTextureVerts[1].b = 0.0;
	inTextureVerts[1].a = 1.0;
	inTextureVerts[1].uOverW = 0.0;
	inTextureVerts[1].vOverW = 10.0;
	inTextureVerts[1].kd_r = 1.0;
	inTextureVerts[1].kd_g = 1.0;
	inTextureVerts[1].kd_b = 1.0;
	inTextureVerts[1].ks_r = 0.0;
	inTextureVerts[1].ks_g = 0.0;
	inTextureVerts[1].ks_b = 0.0;

	inTextureVerts[2].x = right;
	inTextureVerts[2].y = bottom;
	inTextureVerts[2].z = z;
	inTextureVerts[2].invW = invW;
	inTextureVerts[2].r = 0.0;
	inTextureVerts[2].g = 0.0;
	inTextureVerts[2].b = 1.0;
	inTextureVerts[2].a = 1.0;
	inTextureVerts[2].uOverW = 10.0;
	inTextureVerts[2].vOverW = 0.0;
	inTextureVerts[2].kd_r = 1.0;
	inTextureVerts[2].kd_g = 1.0;
	inTextureVerts[2].kd_b = 1.0;
	inTextureVerts[2].ks_r = 0.0;
	inTextureVerts[2].ks_g = 0.0;
	inTextureVerts[2].ks_b = 0.0;

	inTextureVerts[3].x = right;
	inTextureVerts[3].y = top;
	inTextureVerts[3].z = z;
	inTextureVerts[3].invW = invW;
	inTextureVerts[3].r = 1.0;
	inTextureVerts[3].g = 1.0;
	inTextureVerts[3].b = 1.0;
	inTextureVerts[3].a = 1.0;
	inTextureVerts[3].uOverW = 10.0;
	inTextureVerts[3].vOverW = 10.0;
	inTextureVerts[3].kd_r = 1.0;
	inTextureVerts[3].kd_g = 1.0;
	inTextureVerts[3].kd_b = 1.0;
	inTextureVerts[3].ks_r = 0.0;
	inTextureVerts[3].ks_g = 0.0;
	inTextureVerts[3].ks_b = 0.0;
}	
	
/*=========================================================================*\
	S3OneRectangle
\*=========================================================================*/
S3tTestStatus S3OneRectangle(
	S3tTestInfo				*testInfo)
{
	TQAVTexture	vTexture[4];
	
	S3SetupOneTexturedRectangle(testInfo, vTexture);
	
	S3Test_RenderStart(testInfo, NULL);
	
	S3Test_DrawTriTexture(
		testInfo,
		&vTexture[0],
		&vTexture[1],
		&vTexture[2],
		kQATriFlags_None);
	
	S3Test_DrawTriTexture(
		testInfo,
		&vTexture[1],
		&vTexture[2],
		&vTexture[3],
		kQATriFlags_None);

	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}			
	
/*=========================================================================*\
	S3OneAlphaRectangle
\*=========================================================================*/
S3tTestStatus S3OneAlphaRectangle(
	S3tTestInfo				*testInfo)
{
	TQAVTexture				vTexture[4];
	float					alpha = 0.5;
	char* 					argStringP = S3TestInfo_GetArgString(testInfo);

		// scan the argument string to get the number of points to draw
	sscanf(argStringP, "%f", &alpha);

		// clamp the entered alpha to a legal value
	if (alpha > 1.0) {
		alpha = 1.0;
	} else if (alpha < 0.0) {
		alpha = 0.0;
	}
	
	S3SetupOneTexturedRectangle(testInfo, vTexture);

		// set the alpha channel of the 4 points of the rectangle to whatever
		// the user entered, if anything
	vTexture[0].a = alpha;
	vTexture[1].a = alpha;
	vTexture[2].a = alpha;
	vTexture[3].a = alpha;
	
	S3Test_RenderStart(testInfo, NULL);
	
	S3Test_DrawTriTexture(
		testInfo,
		&vTexture[0],
		&vTexture[1],
		&vTexture[2],
		kQATriFlags_None);
	
	S3Test_DrawTriTexture(
		testInfo,
		&vTexture[1],
		&vTexture[2],
		&vTexture[3],
		kQATriFlags_None);

	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}			
	
/*=========================================================================*\
	S3OneBitmap
\*=========================================================================*/
S3tTestStatus S3OneBitmap(
	S3tTestInfo				*testInfo)
{
	TQAVGouraud v = {
		0.0, 0.0, 0.5, 2.0,
		1.0, 1.0, 1.0, 1.0,
	};
	
	v.x = (S3TestInfo_GetWidth(testInfo) - S3TestInfo_GetBitmapWidth(testInfo)) / 2.0; 
	v.y = (S3TestInfo_GetHeight(testInfo) - S3TestInfo_GetBitmapHeight(testInfo)) / 2.0; 

	S3Test_RenderStart(testInfo, NULL);
			
		// passing nil for the bitmap pointers tells S3Test_DrawBitmap
		// to use the bitmaps stored in testInfo
	S3Test_DrawBitmap(testInfo, &v, nil, nil);
	
	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}			

/*=========================================================================*\
	S3NineBitmaps
\*=========================================================================*/
S3tTestStatus S3NineBitmaps(
	S3tTestInfo				*testInfo)
{
	TQAVGouraud v = {
		0.0, 0.0, 0.5, 2.0,
		1.0, 1.0, 1.0, 1.0,
	};
	long windowWidthHalf = S3TestInfo_GetWidth(testInfo) / 2;
	long windowHeightHalf = S3TestInfo_GetHeight(testInfo) / 2;
	long bitmapWidthHalf = S3TestInfo_GetBitmapWidth(testInfo) / 2;
	long bitmapHeightHalf = S3TestInfo_GetBitmapHeight(testInfo) / 2;
	long x, y;
	short i, j;

	S3Test_RenderStart(testInfo, NULL);
	
	for (y = 0, j = 0; j < 3; y += windowHeightHalf, j++) {
		for (x = 0, i = 0; i < 3; x += windowWidthHalf, i++) {
			v.x = x - bitmapWidthHalf;
			v.y = y - bitmapHeightHalf;
			
				// passing nil for the bitmap pointers tells S3Test_DrawBitmap
				// to use the bitmaps stored in testInfo
			S3Test_DrawBitmap(testInfo, &v, nil, nil);
		} // endfor
	} // endfor
	
	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}			

/*=========================================================================*\
	S3ShowAllMipMaps
\*=========================================================================*/
S3tTestStatus S3ShowAllMipMaps(
	S3tTestInfo				*testInfo)
{
	long width = S3TestInfo_GetTextureWidth(testInfo);
	long height = S3TestInfo_GetTextureHeight(testInfo);
	long windowWidth = S3TestInfo_GetWidth(testInfo);
	long windowHeight = S3TestInfo_GetHeight(testInfo);
	long* longerSideP = (width >= height) ? &width : &height;
	long* shorterSideP = (width < height) ? &width : &height;
	long x, y;
	TQAVTexture	vTexture[4] = {
		{ 
			0.0, 0.0, 0.1, 10.0,
			1.0, 0.0, 0.0, 1.0,
			0.0, 0.0,
			1.0, 1.0, 1.0,
			0.0, 0.0, 0.0
		},
		{ 
			0.0, 0.0, 0.1, 10.0,
			0.0, 1.0, 0.0, 1.0,
			0.0, 10.0,
			1.0, 1.0, 1.0,
			0.0, 0.0, 0.0
		},
		{ 
			0.0, 0.0, 0.1, 10.0,
			0.0, 0.0, 1.0, 1.0,
			10.0, 0.0,
			1.0, 1.0, 1.0,
			0.0, 0.0, 0.0
		},
		{ 
			0.0, 0.0, 0.1, 10.0,
			1.0, 1.0, 1.0, 1.0,
			10.0, 10.0,
			1.0, 1.0, 1.0,
			0.0, 0.0, 0.0
		},
	};

		// make sure the current window is big enough to show all the
		// mipmaps and return a failure if not.  we add 1 to the window 
		// size because we want to be able to draw a triangle from, say, 
		// 0 to 128 if the texture is 128 wide, but 128.0 is outside the
		// window if it's 128 pixels wide.  we can only draw to 128.0
		// if the window is at least 129 pixels wide.
	if ((windowWidth < (width + (width >> 1) + 1)) || 
		(windowHeight < (height + 1))) {
			// the window's not big enough, so tell the user to enlarge it
		Str255 messageString;
		
		sprintf((char*) &messageString[1],
				"The window is not large enough to show all of the mipmaps. "
				"It must be at least %d X %d.", 
				(width + (width >> 1) + 1), height + 1);
				
			// set the length of the Pascal string to be the length of the
			// C-string starting from messageString[1]
		messageString[0] = strlen((char*) &messageString[1]);
		
			// alert the user
		S3StopAlert(0, 0, messageString);
		
		return kS3Status_Failure;
	}

	if (!(S3TestInfo_GetTextureFlags(testInfo) & kQATexture_Mipmap)) {
		S3StopAlert(0, 0, "\pThe 'Create Mipmap' option has not been "
					"selected, so the texture has no mipmap data. The "
					"test output is therefore not valid.");
	}
	
	vTexture[0].x = 0;
	vTexture[0].y = height;

	vTexture[1].x = 0;
	vTexture[1].y = 0;
	
	vTexture[2].x = width;
	vTexture[2].y = height;
	
	vTexture[3].x = width;
	vTexture[3].y = 0;
	
	S3Test_RenderStart(testInfo, NULL);
	
		// draw the full-size texture map
	S3Test_DrawTriTexture(
		testInfo,
		&vTexture[0],
		&vTexture[1],
		&vTexture[2],
		kQATriFlags_None);
	
	S3Test_DrawTriTexture(
		testInfo,
		&vTexture[1],
		&vTexture[2],
		&vTexture[3],
		kQATriFlags_None);

		// the remaining mipmaps will be drawn to the right of the first one,
		// one below the other, starting at the top of the window
	x = width;
	y = 0;
	
		// draw all the remaining mipmaps, down to the 1 X 1 size
	while (*longerSideP > 0) {
			// halve the longer side
		*longerSideP >>= 1;
		
		if (*shorterSideP > 0) {
				// as long as the shorter side is at least 1 pixel long, halve it
			*shorterSideP >>= 1;
		}
		
		vTexture[0].x = x;
		vTexture[0].y = y + height;
	
		vTexture[1].x = x;
		vTexture[1].y = y;
		
		vTexture[2].x = x + width;
		vTexture[2].y = y + height;
		
		vTexture[3].x = x + width;
		vTexture[3].y = y;
		
		S3Test_DrawTriTexture(
			testInfo,
			&vTexture[0],
			&vTexture[1],
			&vTexture[2],
			kQATriFlags_None);
		
		S3Test_DrawTriTexture(
			testInfo,
			&vTexture[1],
			&vTexture[2],
			&vTexture[3],
			kQATriFlags_None);
	
			// draw the next mipmap below this one
		y += height;
	} // endwhile
	
	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}

/*=========================================================================*\
	S3DrawBitmapsAtZ
\*=========================================================================*/
void S3DrawBitmapsAtZ(
	S3tTestInfo				*testInfo,
	float					inZ)
{
	TQAVGouraud v = {
		0.0, 0.0, 0.0, 1.0,
		1.0, 0.0, 0.0, 1.0,
	};
		// get the width and height of the window n bitmap
	long windowWidth = S3TestInfo_GetWidth(testInfo);
	long windowHeight = S3TestInfo_GetHeight(testInfo);
	long bitmapWidth = S3TestInfo_GetBitmapWidth(testInfo);
	long bitmapHeight = S3TestInfo_GetBitmapHeight(testInfo);
	long x, y;

		// set the bitmap's Z value
	v.z = inZ;
	
		// cover the entire window with copies of the bitmap
	for (y = 0; y < windowHeight; y += bitmapHeight) {
		for (x = 0; x < windowWidth; x += bitmapWidth) {
			v.x = x;
			v.y = y;
			
				// passing nil for the bitmap pointers tells S3Test_DrawBitmap
				// to use the bitmaps stored in testInfo
			S3Test_DrawBitmap(testInfo, &v, nil, nil);
		} // endfor
	} // endfor
}

/*=========================================================================*\
	S3BitmapsBehindTriGouraud
\*=========================================================================*/
S3tTestStatus S3BitmapsBehindTriGouraud(
	S3tTestInfo				*testInfo)
{
	long		i, flushCount;
		
	S3Test_RenderStart(testInfo, NULL);
	
		// draw random Gouraud tris
	for (i = 0, flushCount = kNTris / kNFlushes; i++ < kNTris; /* Nothing */)
	{
		S3TestUtil_DrawRandomTriGouraud(testInfo);
		
		if ((i % flushCount) == 0)
		{
			S3Test_Flush(testInfo);
		}
	}

		// draw the bitmap at the furthest Z value.  we draw them after
		// the tris to test whether Z-buffering is working correctly.  if
		// it is not, the bitmaps will be drawn on top of the tris, rather
		// than behind them.
	S3DrawBitmapsAtZ(testInfo, 1.0);

	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}

/*=========================================================================*\
	S3BitmapsInFrontTriGouraud
\*=========================================================================*/
S3tTestStatus S3BitmapsInFrontTriGouraud(
	S3tTestInfo				*testInfo)
{
	long		i, flushCount;
		
	S3Test_RenderStart(testInfo, NULL);

		// draw the bitmap at the closest Z value.  we draw them before
		// the tris to test whether Z-buffering is working correctly.  if
		// it is not, the tris will be drawn on top of the bitmaps, rather
		// than behind them.  if the bitmap has an alpha channel, the tris
		// should also show through it.
	S3DrawBitmapsAtZ(testInfo, 0.0);
	
		// draw random Gouraud tris
	for (i = 0, flushCount = kNTris / kNFlushes; i++ < kNTris; /* Nothing */)
	{
		S3TestUtil_DrawRandomTriGouraud(testInfo);
		
		if ((i % flushCount) == 0)
		{
			S3Test_Flush(testInfo);
		}
	}

	S3Test_RenderEnd(testInfo, NULL);

	return kS3Status_Completed;
}

/*=========================================================================*\
	S3TestShowColorTable
\*=========================================================================*/
S3tTestStatus S3TestShowColorTable(
	S3tTestInfo				*testInfo,
	TQAImagePixelType		inPixelType,
	TQAColorTableType		inColorTableType,
	TQAColorTable*			inLeftColorTable,
	TQAColorTable*			inRightColorTable,
	Boolean					inUseLeftEngine,
	Boolean					inUseRightEngine)
{
	Byte*					textureData;
	Byte*					textureRow;
	long					textureNibbleCount = 0;
	TQAVTexture				textureVerts[4];
	TQAImage				textureImage;
	TQATexture*				leftColorTableTexture = nil;
	TQATexture*				rightColorTableTexture = nil;
	TQAEngine*				leftEngine = S3TestInfo_GetLEngine(testInfo);
	TQAEngine*				rightEngine = S3TestInfo_GetREngine(testInfo);
	TQADrawContext*			leftContext = S3TestInfo_GetLDrawContext(testInfo);
	TQADrawContext*			rightContext = S3TestInfo_GetRDrawContext(testInfo);
	TQAError				qaError;
	short					x, y;
	unsigned short			elementNumber;

	#define 				kCL8_ElementCount	 			256
	#define					kCL8_ElementsPerSide 			16
	#define					kCL8_ElementPixelSize 			1
//	#define					kCL8_ElementPixelSize 			4
	#define 				kCL8_PixelNibbleCount 			2
	#define 				kCL4_ElementCount 				16
	#define					kCL4_ElementsPerSide 			4
	#define					kCL4_ElementPixelSize 			1
//	#define					kCL4_ElementPixelSize 			16
	#define 				kCL4_PixelNibbleCount 			1

		// fill in the data for 4 verts which are the 4 corners of a rectangle
		// 80% of the width and height of the window, at z == 0.1
	S3SetupOneTexturedRectangle(testInfo, textureVerts);
	
		// figure out how much space we need to store the texture.  we count
		// nibbles instead of bytes because CL4 textures have 2 pixels per byte.
	switch (inColorTableType) {
//		case kQACL8_RGB16:
//			textureNibbleCount = kCL8_ElementCount * kCL8_PixelNibbleCount * 
//					kCL8_ElementPixelSize * kCL8_ElementPixelSize;
//			break;
		
		case kQAColorTable_CL8_RGB32:
			textureNibbleCount = kCL8_ElementCount * kCL8_PixelNibbleCount * 
					kCL8_ElementPixelSize * kCL8_ElementPixelSize;
			break;
		
//		case kQACL4_RGB16:
//			textureNibbleCount = kCL4_ElementCount * kCL4_PixelNibbleCount * 
//					kCL4_ElementPixelSize * kCL4_ElementPixelSize;
//			break;
		
		case kQAColorTable_CL4_RGB32:
			textureNibbleCount = kCL4_ElementCount * kCL4_PixelNibbleCount * 
					kCL4_ElementPixelSize * kCL4_ElementPixelSize;
			break;
	} // endswitch
	
	textureData = (Byte*) malloc(textureNibbleCount / 2);
	
	if (textureData == nil) {
		S3Test_Print("*** Not enough memory to create texture ***");
		
		return kS3Status_Failure;
	}

	elementNumber = 0;
	
	switch (inColorTableType) {
//		case kQACL8_RGB16:
		case kQAColorTable_CL8_RGB32:
			textureImage.width = kCL8_ElementsPerSide;
			textureImage.height = kCL8_ElementsPerSide;
			textureImage.rowBytes = kCL8_ElementsPerSide;
			textureRow = textureData;
			
				// set each pixel in the texture to one of the color table
				// element indexes.  this ensures we see the entire color table.
			for (y = 0; y < kCL8_ElementsPerSide; y++) {
				for (x = 0; x < kCL8_ElementsPerSide; x++) {
					textureRow[x] = elementNumber++;
				} // endfor
				
				textureRow += textureImage.rowBytes;
			} // endfor
			break;
		
//		case kQACL4_RGB16:
		case kQAColorTable_CL4_RGB32:
			textureImage.width = kCL4_ElementsPerSide;
			textureImage.height = kCL4_ElementsPerSide;
			textureImage.rowBytes = kCL4_ElementsPerSide;
			textureRow = textureData;
			
				// set each pixel in the texture to one of the color table
				// element indexes.  this ensures we see the entire color table.
			for (y = 0; y < kCL4_ElementsPerSide; y++) {
				for (x = 0; x < (kCL4_ElementsPerSide / 2); x++) {
					textureRow[x] = ((elementNumber++ << 4) | elementNumber++);
				} // endfor
				
				textureRow += textureImage.rowBytes;
			} // endfor
			break;
	} // endswitch	
	
	textureImage.pixmap = textureData;
	
		// for each engine that the caller wants us to, create a new texture
		// using the pixel data we just filled in.  bind the color table the 
		// user chose in the color table popup menu to each of these new 
		// textures.  set the texture pointer to this new texture.  then draw
		// the texture on two triangles.
	if (inUseLeftEngine) {
		qaError = QATextureNew(leftEngine, kQATexture_None, inPixelType,
				&textureImage, &leftColorTableTexture);
				
		if (qaError != kQANoErr) goto failure;
		
		qaError = QATextureBindColorTable(leftEngine, leftColorTableTexture, 
				inLeftColorTable);
		
		if (qaError != kQANoErr) goto failure;

		QASetPtr(leftContext, kQATag_Texture, leftColorTableTexture);
		
		QARenderStart(leftContext, nil, nil);
		
		QADrawTriTexture(leftContext, &textureVerts[0], &textureVerts[1], 
				&textureVerts[2], kQATriFlags_None);
		
		QADrawTriTexture(leftContext, &textureVerts[1], &textureVerts[2], 
				&textureVerts[3], kQATriFlags_None);
				
		QARenderEnd(leftContext, nil);
	}
	
	if (inUseRightEngine) {
		qaError = QATextureNew(rightEngine, kQATexture_None, inPixelType,
				&textureImage, &rightColorTableTexture);
		
		if (qaError != kQANoErr) goto failure;
		
		qaError = QATextureBindColorTable(rightEngine, rightColorTableTexture, 
				inRightColorTable);
		
		if (qaError != kQANoErr) goto failure;
	
		QASetPtr(rightContext, kQATag_Texture, rightColorTableTexture);
		
		QARenderStart(rightContext, nil, nil);
		
		QADrawTriTexture(rightContext, &textureVerts[0], &textureVerts[1], 
				&textureVerts[2], kQATriFlags_None);
		
		QADrawTriTexture(rightContext, &textureVerts[1], &textureVerts[2], 
				&textureVerts[3], kQATriFlags_None);
				
		QARenderEnd(rightContext, nil);
	}

		// clean up
	if (textureData != nil) free(textureData);
	if (leftColorTableTexture != nil) QATextureDelete(leftEngine, leftColorTableTexture);
	if (rightColorTableTexture != nil) QATextureDelete(rightEngine, rightColorTableTexture);

	return kS3Status_Completed;
	
failure:
	if (textureData != nil) free(textureData);
	if (leftColorTableTexture != nil) QATextureDelete(leftEngine, leftColorTableTexture);
	if (rightColorTableTexture != nil) QATextureDelete(rightEngine, rightColorTableTexture);
	
	return kS3Status_Failure;
}

/*=========================================================================*\
	S3TestShowTextureColorTable
\*=========================================================================*/
S3tTestStatus S3TestShowTextureColorTable(
	S3tTestInfo				*testInfo)
{
	TQAImagePixelType		textureFormat = S3TestInfo_GetTextureFormat(testInfo);

	if (textureFormat != kQAPixel_CL8 && textureFormat != kQAPixel_CL4) {
			// alert the user that this test can only run when the texture is
			// CL_4 or CL_8
		S3StopAlert(0, 0, "\pTo run this test, the texture must be in 4-bit or "
						  "8-bit color lookup table format.");
		
		return kS3Status_Failure;
	}
	
	return S3TestShowColorTable(testInfo, 
			S3TestInfo_GetTextureFormat(testInfo),
			S3TestInfo_GetTextureColorTableType(testInfo),
			S3TestInfo_GetLTextureColorTable(testInfo),
			S3TestInfo_GetRTextureColorTable(testInfo),
			true, true);
}

/*=========================================================================*\
	S3TestShowBitmapColorTable
\*=========================================================================*/
S3tTestStatus S3TestShowBitmapColorTable(
	S3tTestInfo				*testInfo)
{
	TQAImagePixelType		bitmapFormat = S3TestInfo_GetBitmapFormat(testInfo);

	if (bitmapFormat != kQAPixel_CL8 && bitmapFormat != kQAPixel_CL4) {
			// alert the user that this test can only run when the bitmap is
			// CL_4 or CL_8
		S3StopAlert(0, 0, "\pTo run this test, the bitmap must be in 4-bit or "
						  "8-bit color lookup table format.");
		
		return kS3Status_Failure;
	}
	
	return S3TestShowColorTable(testInfo, 
			S3TestInfo_GetBitmapFormat(testInfo),
			S3TestInfo_GetBitmapColorTableType(testInfo),
			S3TestInfo_GetLBitmapColorTable(testInfo),
			S3TestInfo_GetRBitmapColorTable(testInfo),
			true, true);
}

/*=========================================================================*\
	S3TestShowTextureAndColorTable
\*=========================================================================*/
S3tTestStatus S3TestShowTextureAndColorTable(
	S3tTestInfo				*testInfo)
{
	TQAEngine*				leftEngine = S3TestInfo_GetLEngine(testInfo);
	TQAEngine*				rightEngine = S3TestInfo_GetREngine(testInfo);
	TQADrawContext*			leftContext = S3TestInfo_GetLDrawContext(testInfo);
	TQAVTexture				textureVerts[4];
	TQAImagePixelType		textureFormat = S3TestInfo_GetTextureFormat(testInfo);

	if (textureFormat != kQAPixel_CL8 && textureFormat != kQAPixel_CL4) {
			// alert the user that this test can only run when the texture is
			// CL_4 or CL_8
		S3StopAlert(0, 0, "\pTo run this test, the texture must be in 4-bit or "
						  "8-bit color lookup table format.");
		
		return kS3Status_Failure;
	}
	
	if (leftEngine != rightEngine) {
			// alert the user that it makes sense to run this test only if
			// both windows display the same engine
		S3StopAlert(0, 0, "\pTo run this test, both windows must use the same engine.");
		
		return kS3Status_Failure;
	}
	
		// position the four points in a rectangle, centered in the window
	S3SetupOneTexturedRectangle(testInfo, textureVerts);
		
		// draw the texture in the left window
	QARenderStart(leftContext, nil, nil);
	
	QADrawTriTexture(leftContext, &textureVerts[0], &textureVerts[1], 
			&textureVerts[2], kQATriFlags_None);
	
	QADrawTriTexture(leftContext, &textureVerts[1], &textureVerts[2], 
			&textureVerts[3], kQATriFlags_None);
			
	QARenderEnd(leftContext, nil);

		// draw the color table in the right window
	if (S3TestShowColorTable(testInfo, 
			S3TestInfo_GetTextureFormat(testInfo),
			S3TestInfo_GetTextureColorTableType(testInfo),
			S3TestInfo_GetLTextureColorTable(testInfo),
			S3TestInfo_GetRTextureColorTable(testInfo),
			false, true) != kS3Status_Completed) {
		return kS3Status_Failure;
	}
	
	return kS3Status_Completed;
}

/*=========================================================================*\
	S3TestShowBitmapAndColorTable
\*=========================================================================*/
S3tTestStatus S3TestShowBitmapAndColorTable(
	S3tTestInfo				*testInfo)
{
	TQAEngine*				leftEngine = S3TestInfo_GetLEngine(testInfo);
	TQAEngine*				rightEngine = S3TestInfo_GetREngine(testInfo);
	TQADrawContext*			leftContext = S3TestInfo_GetLDrawContext(testInfo);
	TQAImagePixelType		bitmapFormat = S3TestInfo_GetBitmapFormat(testInfo);
	TQAVGouraud v = {
		0.0, 0.0, 0.5,
		1.0, 1.0, 0.0, 0.0,
	};
	
	if (bitmapFormat != kQAPixel_CL8 && bitmapFormat != kQAPixel_CL4) {
			// alert the user that this test can only run when the bitmap is
			// CL_4 or CL_8
		S3StopAlert(0, 0, "\pTo run this test, the bitmap must be in 4-bit or "
						  "8-bit color lookup table format.");
		
		return kS3Status_Failure;
	}
	
	if (leftEngine != rightEngine) {
			// alert the user that it makes sense to run this test only if
			// both windows display the same engine
		S3StopAlert(0, 0, "\pTo run this test, both windows must use the same engine.");
		
		return kS3Status_Failure;
	}
	
		// center the bitmap in the window
	v.x = (S3TestInfo_GetWidth(testInfo) - S3TestInfo_GetBitmapWidth(testInfo)) / 2.0, 
	v.y = (S3TestInfo_GetHeight(testInfo) - S3TestInfo_GetBitmapHeight(testInfo)) / 2.0, 

		// draw the bitmap in the left window
	QARenderStart(leftContext, nil, nil);
	
	QADrawBitmap(leftContext, &v, S3TestInfo_GetLBitmap(testInfo));
	
	QARenderEnd(leftContext, nil);

		// draw the color table in the right window
	if (S3TestShowColorTable(testInfo, 
			S3TestInfo_GetBitmapFormat(testInfo),
			S3TestInfo_GetBitmapColorTableType(testInfo),
			S3TestInfo_GetLBitmapColorTable(testInfo),
			S3TestInfo_GetRBitmapColorTable(testInfo),
			false, true) != kS3Status_Completed) {
		return kS3Status_Failure;
	}
	
	return kS3Status_Completed;
}	

/*=========================================================================*\
	S3AnimatedTextureColorTable
\*=========================================================================*/
S3tTestStatus S3AnimatedTextureColorTable(
	S3tTestInfo				*testInfo)
{
	#define 				kTextureFrameCount 5
	#define 				kTextureBrightnessDelta (2.0 / (kTextureFrameCount + 1))
	S3Image					image = S3TestInfo_GetTextureImage(testInfo);
	CTabHandle				colorTableH = S3Image_GetColorTableHandle(image);
	S3tColorTableP			colorTable = nil;
	TQAColorTableType		colorTableType = S3TestInfo_GetTextureColorTableType(testInfo);
	TQAColorTable*			leftColorTables[kTextureFrameCount] = { 0, 0, 0, 0, 0 };
	TQAColorTable*			rightColorTables[kTextureFrameCount] = { 0, 0, 0, 0, 0 };
	TQATexture*				leftTexture = S3TestInfo_GetLTexture(testInfo);
	TQATexture*				rightTexture = S3TestInfo_GetRTexture(testInfo);
	TQAImagePixelType		textureFormat = S3TestInfo_GetTextureFormat(testInfo);
	TQAVTexture				vTexture[4];
	TQAEngine*				leftEngine = S3TestInfo_GetLEngine(testInfo);
	TQAEngine*				rightEngine = S3TestInfo_GetREngine(testInfo);
	unsigned long			useTransparentIndex = S3TestInfo_GetTextureUseTransparentIndex(testInfo);
	TQAError				qaError;
	S3tTestStatus			status = kS3Status_Completed;
	short					i;
	
	if (textureFormat != kQAPixel_CL8 && textureFormat != kQAPixel_CL4) {
			// alert the user that this test can only run when the texture is
			// CL_4 or CL_8
		S3StopAlert(0, 0, "\pTo run this test, the texture must be in 4-bit or "
						  "8-bit color lookup table format.");
		
		return kS3Status_Failure;
	}
	
	S3SetupOneTexturedRectangle(testInfo, vTexture);
	
	for (i = 0; i < kTextureFrameCount; i++) {
		colorTable = S3ColorTable_NewChangeBrightness(colorTableH, 
				colorTableType, kTextureBrightnessDelta * (i + 1));
				
		if (colorTable == nil) {
			S3Test_Print("*** A new color table could not be created. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		qaError = QAColorTableNew(leftEngine, colorTableType, 
				S3ColorTable_GetData(colorTable), useTransparentIndex,
				&leftColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The left engine could not create a new color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		qaError = QAColorTableNew(rightEngine, colorTableType, 
				S3ColorTable_GetData(colorTable), useTransparentIndex,
				&rightColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The right engine could not create a new color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
			// we've created the TQAColorTables, so we can now dispose of the
			// data which was used to create them
		S3ColorTable_Delete(colorTable);
		colorTable = nil;
	} // endfor
	
		// redraw the rectangle kTextureFrameCount times, each time with a brighter
		// color table
	for (i = 0; i < kTextureFrameCount; i++) {
		qaError = QATextureBindColorTable(leftEngine, leftTexture,
				leftColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The left engine could not bind the color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		qaError = QATextureBindColorTable(rightEngine, rightTexture,
				rightColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The right engine could not bind the color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		S3Test_RenderStart(testInfo, NULL);
		
		S3Test_DrawTriTexture(
			testInfo,
			&vTexture[0],
			&vTexture[1],
			&vTexture[2],
			kQATriFlags_None);
		
		S3Test_DrawTriTexture(
			testInfo,
			&vTexture[1],
			&vTexture[2],
			&vTexture[3],
			kQATriFlags_None);
	
		S3Test_RenderEnd(testInfo, NULL);
	}

		// now draw each color table again, from brightest to darkest
	for (i = kTextureFrameCount - 1; i >= 0; i--) {
		qaError = QATextureBindColorTable(leftEngine, leftTexture,
				leftColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The left engine could not bind the color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		qaError = QATextureBindColorTable(rightEngine, rightTexture,
				rightColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The right engine could not bind the color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		S3Test_RenderStart(testInfo, NULL);
		
		S3Test_DrawTriTexture(
			testInfo,
			&vTexture[0],
			&vTexture[1],
			&vTexture[2],
			kQATriFlags_None);
		
		S3Test_DrawTriTexture(
			testInfo,
			&vTexture[1],
			&vTexture[2],
			&vTexture[3],
			kQATriFlags_None);
	
		S3Test_RenderEnd(testInfo, NULL);
	}

cleanup:
	for (i = 0; i < kTextureFrameCount; i++) {
		if (leftColorTables[i] != nil) {
			QAColorTableDelete(leftEngine, leftColorTables[i]);
		}

		if (rightColorTables[i] != nil) {
			QAColorTableDelete(rightEngine, rightColorTables[i]);
		}
		
		if (colorTable != nil) {
			S3ColorTable_Delete(colorTable);
		}
	} // endfor
	
	return status;
}			

/*=========================================================================*\
	S3AnimatedBitmapColorTable
\*=========================================================================*/
S3tTestStatus S3AnimatedBitmapColorTable(
	S3tTestInfo				*testInfo)
{
	#define 				kBitmapFrameCount 10
	#define 				kBitmapBrightnessDelta (2.0 / (kBitmapFrameCount + 1))
	S3Image					image = S3TestInfo_GetBitmapImage(testInfo);
	CTabHandle				colorTableH = S3Image_GetColorTableHandle(image);
	S3tColorTableP			colorTable = nil;
	TQAColorTableType		colorTableType = S3TestInfo_GetBitmapColorTableType(testInfo);
	TQAColorTable*			leftColorTables[kBitmapFrameCount] = { 0, 0, 0, 0, 0 };
	TQAColorTable*			rightColorTables[kBitmapFrameCount] = { 0, 0, 0, 0, 0 };
	TQABitmap*				leftBitmap = S3TestInfo_GetLBitmap(testInfo);
	TQABitmap*				rightBitmap = S3TestInfo_GetRBitmap(testInfo);
	TQAVGouraud 			v = { 0.0, 0.0, 0.5, 1.0, 1.0, 0.0, 0.0 };
	TQAImagePixelType		bitmapFormat = S3TestInfo_GetBitmapFormat(testInfo);
	TQAEngine*				leftEngine = S3TestInfo_GetLEngine(testInfo);
	TQAEngine*				rightEngine = S3TestInfo_GetREngine(testInfo);
	unsigned long			useTransparentIndex = S3TestInfo_GetBitmapUseTransparentIndex(testInfo);
	TQAError				qaError;
	S3tTestStatus			status = kS3Status_Completed;
	short					i;
	
	v.x = (S3TestInfo_GetWidth(testInfo) - S3TestInfo_GetBitmapWidth(testInfo)) / 2.0;
	v.y = (S3TestInfo_GetHeight(testInfo) - S3TestInfo_GetBitmapHeight(testInfo)) / 2.0;
	
	if (bitmapFormat != kQAPixel_CL8 && bitmapFormat != kQAPixel_CL4) {
			// alert the user that this test can only run when the bitmap is
			// CL_4 or CL_8
		S3StopAlert(0, 0, "\pTo run this test, the bitmap must be in 4-bit or "
						  "8-bit color lookup table format.");
		
		return kS3Status_Failure;
	}
	
	for (i = 0; i < kBitmapFrameCount; i++) {
		colorTable = S3ColorTable_NewChangeBrightness(colorTableH, 
				colorTableType, kBitmapBrightnessDelta * (i + 1));
				
		if (colorTable == nil) {
			S3Test_Print("*** A new color table could not be created. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		qaError = QAColorTableNew(leftEngine, colorTableType, 
				S3ColorTable_GetData(colorTable), useTransparentIndex,
				&leftColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The left engine could not create a new color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		qaError = QAColorTableNew(rightEngine, colorTableType, 
				S3ColorTable_GetData(colorTable), useTransparentIndex,
				&rightColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The right engine could not create a new color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
			// we've created the TQAColorTables, so we can now dispose of the
			// data which was used to create them
		S3ColorTable_Delete(colorTable);
		colorTable = nil;
	} // endfor
	
		// redraw the rectangle kBitmapFrameCount times, each time with a brighter
		// color table
	for (i = 0; i < kBitmapFrameCount; i++) {
		qaError = QABitmapBindColorTable(leftEngine, leftBitmap,
				leftColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The left engine could not bind the color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		qaError = QABitmapBindColorTable(rightEngine, rightBitmap,
				rightColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The right engine could not bind the color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		S3Test_RenderStart(testInfo, NULL);
		
		S3Test_DrawBitmap(testInfo, &v, nil, nil);
	
		S3Test_RenderEnd(testInfo, NULL);
	}

		// now draw each color table again, from brightest to darkest
	for (i = kBitmapFrameCount - 1; i >= 0; i--) {
		qaError = QABitmapBindColorTable(leftEngine, leftBitmap,
				leftColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The left engine could not bind the color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		qaError = QABitmapBindColorTable(rightEngine, rightBitmap,
				rightColorTables[i]);
				
		if (qaError != kQANoErr) {
			S3Test_Print("*** The right engine could not bind the color table. ***\r");

			status = kS3Status_Failure;
			goto cleanup;
		}
		
		S3Test_RenderStart(testInfo, NULL);
		
		S3Test_DrawBitmap(testInfo, &v, nil, nil);
	
		S3Test_RenderEnd(testInfo, NULL);
	}

cleanup:
	for (i = 0; i < kBitmapFrameCount; i++) {
		if (leftColorTables[i] != nil) {
			QAColorTableDelete(leftEngine, leftColorTables[i]);
		}

		if (rightColorTables[i] != nil) {
			QAColorTableDelete(rightEngine, rightColorTables[i]);
		}
		
		if (colorTable != nil) {
			S3ColorTable_Delete(colorTable);
		}
	} // endfor
	
	return status;
}			

/*=========================================================================*\
	S3TestColorTableTexture
\*=========================================================================*/
S3tTestStatus S3TestColorTableTexture(
	S3tTestInfo				*testInfo)
{

	FSSpec					colorTableFSSpec;
	StringPtr				colorTableFileName;
	short					colorTableRefNum;
	short					savedResFile = CurResFile();
	Handle					colorTableH = nil;
	short*					colorTableDataP;
	TQAColorTableType		colorTableType;
	TQAColorTable*			leftColorTable = nil;
	TQAColorTable*			rightColorTable = nil;
	TQAEngine*				leftEngine = S3TestInfo_GetLEngine(testInfo);
	TQAEngine*				rightEngine = S3TestInfo_GetREngine(testInfo);
	TQADrawContext*			leftContext = S3TestInfo_GetLDrawContext(testInfo);
	TQAError				qaError;
	
	float left = (float) S3TestInfo_GetWidth(testInfo) * 0.1;
	float top = (float) S3TestInfo_GetHeight(testInfo) * 0.1;
	float right = (float) S3TestInfo_GetWidth(testInfo) * 0.9;
	float bottom = (float) S3TestInfo_GetHeight(testInfo) * 0.9;
	float z = 0.1;
	
	TQAVTexture	vTexture[4] = {
		{ 
			0.0, 0.0, 0.0, 10.0,
			1.0, 0.0, 0.0, 1.0,
			0.0, 0.0,
			1.0, 1.0, 1.0,
			0.0, 0.0, 0.0
		},
		{ 
			0.0, 0.0, 0.0, 10.0,
			0.0, 1.0, 0.0, 1.0,
			0.0, 10.0,
			1.0, 1.0, 1.0,
			0.0, 0.0, 0.0
		},
		{ 
			0.0, 0.0, 0.0, 10.0,
			0.0, 0.0, 1.0, 1.0,
			10.0, 0.0,
			1.0, 1.0, 1.0,
			0.0, 0.0, 0.0
		},
		{ 
			0.0, 0.0, 0.0, 10.0,
			1.0, 1.0, 1.0, 1.0,
			10.0, 10.0,
			1.0, 1.0, 1.0,
			0.0, 0.0, 0.0
		},
	};

	vTexture[0].x = left;
	vTexture[0].y = bottom;
	vTexture[0].z = z;

	vTexture[1].x = left;
	vTexture[1].y = top;
	vTexture[1].z = z;
	
	vTexture[2].x = right;
	vTexture[2].y = bottom;
	vTexture[2].z = z;
	
	vTexture[3].x = right;
	vTexture[3].y = top;
	vTexture[3].z = z;
	
	S3GetAppFileSpec(&colorTableFSSpec);
	
		// change the name in the file spec to "q3ctDefs.rsrc"
	colorTableFileName = "\pq3ctDefs.rsrc";
	BlockMoveData(colorTableFileName, &colorTableFSSpec.name, colorTableFileName[0] + 1);
	
	colorTableRefNum = FSpOpenResFile(&colorTableFSSpec, fsRdPerm);
	
	if (ResError() != noErr) {
		goto failure;
	}
	
	colorTableH = Get1NamedResource('q3ct', "\pTest Color Table");
	
	if (colorTableH == nil) {
		goto failure;
	}
	
	colorTableType = kQAColorTable_CL4_RGB32;
//	colorTableType = *(TQAImagePixelType*) *colorTableH;
	
//	if (S3TestInfo_GetTextureFormat(testInfo) != colorTableType) {
//		goto failure;
//	}
	
	HLock(colorTableH);
	colorTableDataP = (short*) *colorTableH;
	colorTableDataP++;
//	((short*) colorTableDataP)++;
	
	qaError = QAColorTableNew(leftEngine, colorTableType, colorTableDataP, -1, &leftColorTable);
	
	HUnlock(colorTableH);
	ReleaseResource(colorTableH);
	
	if (qaError != kQANoErr) {
		goto failure;
	}
	
	qaError = QATextureBindColorTable(leftEngine, 
			S3TestInfo_GetLTexture(testInfo), leftColorTable);
	
	if (qaError != kQANoErr) {
		goto failure;
	}
	
	QARenderStart(leftContext, nil, nil);
	
	QADrawTriTexture(leftContext, &vTexture[0], &vTexture[1], 
			&vTexture[2], kQATriFlags_None);
	
	QADrawTriTexture(leftContext, &vTexture[1], &vTexture[2], 
			&vTexture[3], kQATriFlags_None);
	
	QARenderEnd(leftContext, nil);
	
	QAColorTableDelete(leftEngine, leftColorTable);
	
	UseResFile(savedResFile);

	return kS3Status_Completed;
	
failure:
	UseResFile(savedResFile);

	return kS3Status_Failure;
}
