/*****************************************************************************\

	S3TestBasic.c

	Basic timing tests.
	
	Copyright (C) 1995 Apple Computer, Inc.  All rights reserved.

\*****************************************************************************/


/*=========================================================================*\
	Includes
\*=========================================================================*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <Displays.h>

#include "S3TestSystem.h"
#include "S3TestInfo.h"
#include "S3TestUtil.h"
#include "S3Utilities.h"
#include "S3SceneFile.h"
#include "S3Alert.h"


/*=========================================================================*\
	Constants
\*=========================================================================*/

#define kNPoints					10000
#define kNLines						1000
#define kNTris						100
#define kNFlushes					5
#define kDefaultFrameCount 			50
#define kS3SuppressNumber 			1


/*=========================================================================*\
	Prototypes
\*=========================================================================*/

void S3TimingTests_Register(void);
S3tTestStatus S3TestSceneFile(
	S3tTestInfo				*testInfo);
S3tTestStatus S3Test1000PixelGouraudTris(
	S3tTestInfo				*testInfo);
S3tTestStatus S3Test50PixelGouraudTris(
	S3tTestInfo				*testInfo);
S3tTestStatus S3Test25PixelGouraudTris(
	S3tTestInfo				*testInfo);
S3tTestStatus S3TestNPixelGouraudTris(
	S3tTestInfo				*testInfo);
S3tTestStatus DrawGouraudTris(
	S3tTestInfo				*testInfo,
	unsigned long 			inSquareSize,
	unsigned long 			inFrameCount);
S3tTestStatus S3Test1000PixelTextureTris(
	S3tTestInfo				*testInfo);
S3tTestStatus S3Test50PixelTextureTris(
	S3tTestInfo				*testInfo);
S3tTestStatus S3Test25PixelTextureTris(
	S3tTestInfo				*testInfo);
S3tTestStatus S3TestNPixelTextureTris(
	S3tTestInfo				*testInfo);
S3tTestStatus DrawTextureTris(
	S3tTestInfo				*testInfo,
	unsigned long 			inSquareSize,
	unsigned long 			inFrameCount);
S3tTestStatus S3TestEngineGestalt(
	S3tTestInfo				*testInfo);
void S3PrintDisplayInfo(
	TQADevice*				inDevice);


/*=========================================================================*\
	S3TimingTests_Register
\*=========================================================================*/
void S3TimingTests_Register(void)
{
	S3Test_Register("25-pixel Gouraud Tris",		kS3Prereq_None, S3Test25PixelGouraudTris);
	S3Test_Register("50-pixel Gouraud Tris",		kS3Prereq_None, S3Test50PixelGouraudTris);
	S3Test_Register("1000-pixel Gouraud Tris",		kS3Prereq_None, S3Test1000PixelGouraudTris);
	S3Test_Register("N-pixel Gouraud Tris",			kS3Prereq_None, S3TestNPixelGouraudTris);
	S3Test_Register("25-pixel Texture Tris",		kS3Prereq_Texture, S3Test25PixelTextureTris);
	S3Test_Register("50-pixel Texture Tris",		kS3Prereq_Texture, S3Test50PixelTextureTris);
	S3Test_Register("1000-pixel Texture Tris",		kS3Prereq_Texture, S3Test1000PixelTextureTris);
	S3Test_Register("N-pixel Texture Tris",			kS3Prereq_Texture, S3TestNPixelTextureTris);

		// add a divider to the list of tests by passing nil
	S3Test_Register("",								kS3Prereq_None, nil);

	S3Test_Register("Engine Gestalt",				kS3Prereq_None, S3TestEngineGestalt);
}

/*=========================================================================*\
	S3Test1000PixelGouraudTris
\*=========================================================================*/
S3tTestStatus S3Test1000PixelGouraudTris(
	S3tTestInfo				*testInfo)
{
	unsigned long frameCount = kDefaultFrameCount;
	char* argString = S3TestInfo_GetArgString(testInfo);
	
	sscanf(argString, "%lu", &frameCount);
	
	return DrawGouraudTris(testInfo, 45, frameCount);
}
	
/*=========================================================================*\
	S3Test50PixelGouraudTris
\*=========================================================================*/
S3tTestStatus S3Test50PixelGouraudTris(
	S3tTestInfo				*testInfo)
{
	unsigned long frameCount = kDefaultFrameCount;
	char* argString = S3TestInfo_GetArgString(testInfo);
	
	sscanf(argString, "%lu", &frameCount);
	
	return DrawGouraudTris(testInfo, 10, frameCount);
}

/*=========================================================================*\
	S3Test25PixelGouraudTris
\*=========================================================================*/
S3tTestStatus S3Test25PixelGouraudTris(
	S3tTestInfo				*testInfo)
{
	unsigned long frameCount = kDefaultFrameCount;
	char* argString = S3TestInfo_GetArgString(testInfo);
	
	sscanf(argString, "%lu", &frameCount);
	
	return DrawGouraudTris(testInfo, 7, frameCount);
}
	
/*=========================================================================*\
	S3TestNPixelGouraudTris
\*=========================================================================*/
S3tTestStatus S3TestNPixelGouraudTris(
	S3tTestInfo				*testInfo)
{
	unsigned long frameCount = kDefaultFrameCount;
	unsigned long pixelsPerTri = 50;
	unsigned long squareSize = 10;
	char* argString = S3TestInfo_GetArgString(testInfo);
	
	sscanf(argString, "%lu %lu", &pixelsPerTri, &frameCount);
	
	squareSize = (unsigned long) sqrt((double) 2 * pixelsPerTri);
	
	return DrawGouraudTris(testInfo, squareSize, frameCount);
}
	
/*=========================================================================*\
	DrawGouraudTris
\*=========================================================================*/
S3tTestStatus DrawGouraudTris(
	S3tTestInfo				*testInfo,
	unsigned long 			inSquareSize,
	unsigned long 			inFrameCount)
{
	TQADrawContext *drawContext;
	TQAVGouraud *vertices, *vertex;
	unsigned short squareCountH, squareCountV;
	unsigned long i, j, k, microseconds, verticesPerLayer;
	unsigned long totalTriangles, totalPixels;
	float z, invW, totalSeconds;
	char string255[kMaxEngineNameLength];
	
		// get the number of inSquareSize X inSquareSize pixel squares 
		// which will fit in the window
	squareCountH = S3TestInfo_GetWidth(testInfo) / inSquareSize;
	squareCountV = S3TestInfo_GetHeight(testInfo) / inSquareSize;
 	
		// the total number of tris we create is the number of inSquareSize X inSquareSize rects
		// that fill the window (squareCountH * squareCountV) times 2 (2 tris
		// per square) * 2 (each tri is drawn twice) times the number of frames
	totalTriangles = squareCountH * squareCountV * 2 * 2 * inFrameCount;

		// the total number of pixels is the number of pixels in each 
		// triangle (half the square's area) * the total number of triangles
	totalPixels = ((inSquareSize * inSquareSize) / 2) * totalTriangles;

	S3Test_Print("%-40s%d\r%-40s%d\r%-40s%d X %d\r\r", 
				 "Pixels per triangle:", (inSquareSize * inSquareSize) / 2,
				 "Number of triangles:", totalTriangles,
				 "Window width X height:", S3TestInfo_GetWidth(testInfo),
				 S3TestInfo_GetHeight(testInfo));
	
		// allocate an array of Gouraud vertices.  there are 4 vertices
		// per square.
	verticesPerLayer = squareCountH * squareCountV * 4;
	vertices = (TQAVGouraud*) calloc(sizeof(TQAVGouraud), verticesPerLayer);
	
	if (vertices == nil) {
			// there's not enough memory to store the vertices, so we can't
			// run the test.  print an error message and return failure.
		S3Test_Print("*** There is not enough memory to run this test. ***\r"
					 "*** Try reducing the window size or increasing   ***\r"
					 "*** Sneak3view's memory allocation.              ***\r\r");
					 
		return kS3Status_Failure;
	}
	
	vertex = vertices;
		
	for (j = 0; j < squareCountV; j++) {
		for (k = 0; k < squareCountH; k++) {
				// top left
			vertex->x = k * inSquareSize;
			vertex->y = j * inSquareSize;
			vertex->r = S3TestUtil_RandomFloat();
			vertex->g = S3TestUtil_RandomFloat();
			vertex->b = S3TestUtil_RandomFloat();
			vertex->a = 1.0;
			vertex++;
			
				// top right
				// we multiply by 0.99 so that the right and bottom sides
				// of the square are < the window's width and height
			vertex->x = k * inSquareSize + inSquareSize - 0.01;
			vertex->y = j * inSquareSize;
			vertex->r = S3TestUtil_RandomFloat();
			vertex->g = S3TestUtil_RandomFloat();
			vertex->b = S3TestUtil_RandomFloat();
			vertex->a = 1.0;
			vertex++;
			
				// bottom left
			vertex->x = k * inSquareSize;
			vertex->y = j * inSquareSize + inSquareSize - 0.01;
			vertex->r = S3TestUtil_RandomFloat();
			vertex->g = S3TestUtil_RandomFloat();
			vertex->b = S3TestUtil_RandomFloat();
			vertex->a = 1.0;
			vertex++;
			
				// bottom right
			vertex->x = k * inSquareSize + inSquareSize - 0.01;
			vertex->y = j * inSquareSize + inSquareSize - 0.01;
			vertex->r = S3TestUtil_RandomFloat();
			vertex->g = S3TestUtil_RandomFloat();
			vertex->b = S3TestUtil_RandomFloat();
			vertex->a = 1.0;
			vertex++;
		} // endfor
	} // endfor

		// first run the test with the left draw context
	drawContext = S3TestInfo_GetLDrawContext(testInfo);
	ASSERT(drawContext);

	S3GetEngineName(S3TestInfo_GetLEngine(testInfo), string255, 
					sizeof(string255));
					
	microseconds = S3MicrosecondCount();
	
		// draw the same frame inFrameCount times
	for (i = 0; i < inFrameCount; i++) {
		QARenderStart(drawContext, nil, nil);
		
			// we run through this loop twice, once when the Z value
			// of the triangles is .1, then again when it's .6.
		for (z = 0.1; z < 1.0; z += 0.5) {
			invW = 1.0 / z;
			
			for (j = 0; j < verticesPerLayer; j += 4) {
					// set the z values for the four vertices which 
					// form the current square
				vertices[j].z = z;
				vertices[j].invW = invW;
				vertices[j + 1].z = z;
				vertices[j + 1].invW = invW;
				vertices[j + 2].z = z;
				vertices[j + 2].invW = invW;
				vertices[j + 3].z = z;
				vertices[j + 3].invW = invW;
	
					// draw the two triangles which fill the inSquareSize X inSquareSize rect.
				QADrawTriGouraud(
					drawContext,
					&vertices[j],			// top left
					&vertices[j + 2],		// bottom left
					&vertices[j + 3],		// bottom right
					kQATriFlags_None);
	
				QADrawTriGouraud(
					drawContext,
					&vertices[j],			// top left
					&vertices[j + 1],		// top right
					&vertices[j + 3],		// bottom right
					kQATriFlags_None);
			} // endfor
		} // endfor
		
		QARenderEnd(drawContext, nil);
	} // endfor
	
		// wait for the engine to finish drawing the queued triangles
	QASync(drawContext);
	
	microseconds = S3ElapsedMicroseconds(microseconds);

	S3Test_Print("Timing info for %s:\r", string255);
	
	totalSeconds = microseconds / 1000000.0;

	S3PrintDisplayInfo(S3TestInfo_GetLDevice(testInfo));
	
	sprintf(string255, "  Time to draw %d frames (seconds):", inFrameCount);
	
	S3Test_Print("%-40s%.4f\r", string255, totalSeconds);
	
	S3Test_Print("%-40s%.4f\r", "  Time per frame (seconds):",
				 totalSeconds / inFrameCount);
	
	S3Test_Print("%-40s%.4f\r", "  Pixels per second:",
				 totalPixels / totalSeconds);
	
	S3Test_Print("%-40s%.4f\r\r", "  Triangles per second:", 
				 totalTriangles / totalSeconds);
	
		// now repeat the test for the right draw context
	drawContext = S3TestInfo_GetRDrawContext(testInfo);
	ASSERT(drawContext);

	S3GetEngineName(S3TestInfo_GetREngine(testInfo), string255, 
					sizeof(string255));
					
	microseconds = S3MicrosecondCount();
	
		// draw the same frame inFrameCount times
	for (i = 0; i < inFrameCount; i++) {
		QARenderStart(drawContext, nil, nil);
		
			// we run through this loop twice, once when the Z value
			// of the triangles is .1, then again when it's .6.
		for (z = 0.1; z < 1.0; z += 0.5) {
			invW = 1.0 / z;
			
			for (j = 0; j < verticesPerLayer; j += 4) {
					// set the z values for the four vertices which 
					// form the current square
				vertices[j].z = z;
				vertices[j].invW = invW;
				vertices[j + 1].z = z;
				vertices[j + 1].invW = invW;
				vertices[j + 2].z = z;
				vertices[j + 2].invW = invW;
				vertices[j + 3].z = z;
				vertices[j + 3].invW = invW;
	
					// draw the two triangles which fill the inSquareSize X inSquareSize rect.
				QADrawTriGouraud(
					drawContext,
					&vertices[j],			// top left
					&vertices[j + 2],		// bottom left
					&vertices[j + 3],		// bottom right
					kQATriFlags_None);
	
				QADrawTriGouraud(
					drawContext,
					&vertices[j],			// top left
					&vertices[j + 1],		// top right
					&vertices[j + 3],		// bottom right
					kQATriFlags_None);
			} // endfor
		} // endfor
		
		QARenderEnd(drawContext, nil);
	} // endfor
	
		// wait for the engine to finish drawing the queued triangles
	QASync(drawContext);
	
	microseconds = S3ElapsedMicroseconds(microseconds);

		// we're done with the allocated vertices
 	free(vertices);
	
	S3Test_Print("Timing info for %s:\r", string255);
	
	totalSeconds = microseconds / 1000000.0;

	S3PrintDisplayInfo(S3TestInfo_GetRDevice(testInfo));
	
	sprintf(string255, "  Time to draw %d frames (seconds):", inFrameCount);
	
	S3Test_Print("%-40s%.4f\r", string255, totalSeconds);
	
	S3Test_Print("%-40s%.4f\r", "  Time per frame (seconds):",
				 totalSeconds / inFrameCount);
	
	S3Test_Print("%-40s%.4f\r", "  Pixels per second:",
				 totalPixels / totalSeconds);
	
	S3Test_Print("%-40s%.4f\r\r", "  Triangles per second:", 
				 totalTriangles / totalSeconds);
	
	S3Test_Print("\rIgnore the following timing information:\r");
	
	return kS3Status_Completed;
}

/*=========================================================================*\
	S3Test1000PixelTextureTris
\*=========================================================================*/
S3tTestStatus S3Test1000PixelTextureTris(
	S3tTestInfo				*testInfo)
{
	unsigned long frameCount = kDefaultFrameCount;
	char* argString = S3TestInfo_GetArgString(testInfo);
	
	sscanf(argString, "%lu", &frameCount);
	
	return DrawTextureTris(testInfo, 45, frameCount);
}

/*=========================================================================*\
	S3Test50PixelTextureTris
\*=========================================================================*/
S3tTestStatus S3Test50PixelTextureTris(
	S3tTestInfo				*testInfo)
{
	unsigned long frameCount = kDefaultFrameCount;
	char* argString = S3TestInfo_GetArgString(testInfo);
	
	sscanf(argString, "%lu", &frameCount);
	
	return DrawTextureTris(testInfo, 10, frameCount);
}

/*=========================================================================*\
	S3Test25PixelTextureTris
\*=========================================================================*/
S3tTestStatus S3Test25PixelTextureTris(
	S3tTestInfo				*testInfo)
{
	unsigned long frameCount = kDefaultFrameCount;
	char* argString = S3TestInfo_GetArgString(testInfo);
	
	sscanf(argString, "%lu", &frameCount);
	
	return DrawTextureTris(testInfo, 7, frameCount);
}
	
/*=========================================================================*\
	S3TestNPixelTextureTris
\*=========================================================================*/
S3tTestStatus S3TestNPixelTextureTris(
	S3tTestInfo				*testInfo)
{
	unsigned long frameCount = kDefaultFrameCount;
	unsigned long pixelsPerTri = 50;
	unsigned long squareSize = 10;
	char* argString = S3TestInfo_GetArgString(testInfo);
	
	sscanf(argString, "%lu %lu", &pixelsPerTri, &frameCount);
	
	squareSize = (unsigned long) sqrt((double) 2 * pixelsPerTri);
	
	return DrawTextureTris(testInfo, squareSize, frameCount);
}
	
/*=========================================================================*\
	DrawTextureTris
\*=========================================================================*/
S3tTestStatus DrawTextureTris(
	S3tTestInfo				*testInfo,
	unsigned long 			inSquareSize,
	unsigned long 			inFrameCount)
{
	TQADrawContext *drawContext;
	TQAVTexture *vertices, *vertex;
	unsigned short squareCountH, squareCountV;
	unsigned long i, j, k, microseconds, verticesPerLayer;
	unsigned long totalTriangles, totalPixels;
	float z, invW, totalSeconds;
	char string255[kMaxEngineNameLength];
	
		// get the number of inSquareSize X inSquareSize pixel squares 
		// which will fit in the window
	squareCountH = S3TestInfo_GetWidth(testInfo) / inSquareSize;
	squareCountV = S3TestInfo_GetHeight(testInfo) / inSquareSize;
 	
		// the total number of tris we create is the number of inSquareSize X inSquareSize rects
		// that fill the window (squareCountH * squareCountV) times 2 (2 tris
		// per square) * 2 (each tri is drawn twice) times the number of frames
	totalTriangles = squareCountH * squareCountV * 2 * 2 * inFrameCount;
	
		// the total number of pixels is the number of pixels in each 
		// triangle (half the square's area) * the total number of triangles
	totalPixels = ((inSquareSize * inSquareSize) / 2) * totalTriangles;
	
	S3Test_Print("%-40s%d\r%-40s%d\r%-40s%d X %d\r\r", 
				 "Pixels per triangle:", (inSquareSize * inSquareSize) / 2,
				 "Number of triangles:", totalTriangles,
				 "Window width X height:", S3TestInfo_GetWidth(testInfo),
				 S3TestInfo_GetHeight(testInfo));
	
		// allocate an array of texture vertices.  there are 4 vertices
		// per square.
	verticesPerLayer = squareCountH * squareCountV * 4;
	vertices = (TQAVTexture*) calloc(sizeof(TQAVTexture), verticesPerLayer);
	
	if (vertices == nil) {
			// there's not enough memory to store the vertices, so we can't
			// run the test.  print an error message and return failure.
		S3Test_Print("*** There is not enough memory to run this test. ***\r"
					 "*** Try reducing the window size or increasing   ***\r"
					 "*** Sneak3view's memory allocation.              ***\r\r");
					 
		return kS3Status_Failure;
	}
	
	vertex = vertices;
		
	for (j = 0; j < squareCountV; j++) {
		for (k = 0; k < squareCountH; k++) {
				// top left
			vertex->x = k * inSquareSize;
			vertex->y = j * inSquareSize;
			vertex->r = S3TestUtil_RandomFloat();
			vertex->g = S3TestUtil_RandomFloat();
			vertex->b = S3TestUtil_RandomFloat();
			vertex->a = 1.0;
			vertex->uOverW = 0.0;
			vertex->vOverW = 1.0;
			vertex->kd_r = 1.0;
			vertex->kd_g = 1.0;
			vertex->kd_b = 1.0;
			vertex->ks_r = 0.0;
			vertex->ks_g = 0.0;
			vertex->ks_b = 0.0;
			vertex++;
			
				// top right
				// we multiply by 0.99 so that the right and bottom sides
				// of the square are < the window's width and height
			vertex->x = k * inSquareSize + inSquareSize - 0.01;
			vertex->y = j * inSquareSize;
			vertex->r = S3TestUtil_RandomFloat();
			vertex->g = S3TestUtil_RandomFloat();
			vertex->b = S3TestUtil_RandomFloat();
			vertex->a = 1.0;
			vertex->uOverW = 1.0;
			vertex->vOverW = 1.0;
			vertex->kd_r = 1.0;
			vertex->kd_g = 1.0;
			vertex->kd_b = 1.0;
			vertex->ks_r = 0.0;
			vertex->ks_g = 0.0;
			vertex->ks_b = 0.0;
			vertex++;
			
				// bottom left
			vertex->x = k * inSquareSize;
			vertex->y = j * inSquareSize + inSquareSize - 0.01;
			vertex->r = S3TestUtil_RandomFloat();
			vertex->g = S3TestUtil_RandomFloat();
			vertex->b = S3TestUtil_RandomFloat();
			vertex->a = 1.0;
			vertex->uOverW = 0.0;
			vertex->vOverW = 0.0;
			vertex->kd_r = 1.0;
			vertex->kd_g = 1.0;
			vertex->kd_b = 1.0;
			vertex->ks_r = 0.0;
			vertex->ks_g = 0.0;
			vertex->ks_b = 0.0;
			vertex++;
			
				// bottom right
			vertex->x = k * inSquareSize + inSquareSize - 0.01;
			vertex->y = j * inSquareSize + inSquareSize - 0.01;
			vertex->r = S3TestUtil_RandomFloat();
			vertex->g = S3TestUtil_RandomFloat();
			vertex->b = S3TestUtil_RandomFloat();
			vertex->a = 1.0;
			vertex->uOverW = 1.0;
			vertex->vOverW = 0.0;
			vertex->kd_r = 1.0;
			vertex->kd_g = 1.0;
			vertex->kd_b = 1.0;
			vertex->ks_r = 0.0;
			vertex->ks_g = 0.0;
			vertex->ks_b = 0.0;
			vertex++;
		} // endfor
	} // endfor

		// first run the test with the left draw context
	drawContext = S3TestInfo_GetLDrawContext(testInfo);
	ASSERT(drawContext);

	S3GetEngineName(S3TestInfo_GetLEngine(testInfo), string255, 
					sizeof(string255));
					
	microseconds = S3MicrosecondCount();
	
		// draw the same frame inFrameCount times
	for (i = 0; i < inFrameCount; i++) {
		QARenderStart(drawContext, nil, nil);
		
			// we run through this loop twice, once when the Z value
			// of the triangles is .1, then again when it's .6.
		for (z = 0.1; z < 1.0; z += 0.5) {
			invW = 1.0 / z;
			
			for (j = 0; j < verticesPerLayer; j += 4) {
					// we set the z and invW values for the four vertices 
					// which form the current square.  we also set the 
					// uOverW and vOverW values to invW for the vertices 
					// which have 1.0 in those slots.  this is because the
					// u,v parameters have to be divided by z.  but since
					// the u,v values are either 0 or 1.0, we can just use
					// invW (1.0 / z), which we calculated above.

					// top left
				vertices[j].z = z;
				vertices[j].invW = invW;
				vertices[j].vOverW = invW;
				
					// top right
				vertices[j + 1].z = z;
				vertices[j + 1].invW = invW;
				vertices[j + 1].uOverW = invW;
				vertices[j + 1].vOverW = invW;
					
					// bottom left
				vertices[j + 2].z = z;
				vertices[j + 2].invW = invW;
				
					// bottom right
				vertices[j + 3].z = z;
				vertices[j + 3].invW = invW;
				vertices[j + 3].uOverW = invW;
	
					// draw the two triangles which fill the inSquareSize X inSquareSize rect.
				QADrawTriTexture(
					drawContext,
					&vertices[j],			// top left
					&vertices[j + 2],		// bottom left
					&vertices[j + 3],		// bottom right
					kQATriFlags_None);
	
				QADrawTriTexture(
					drawContext,
					&vertices[j],			// top left
					&vertices[j + 1],		// top right
					&vertices[j + 3],		// bottom right
					kQATriFlags_None);
			} // endfor
		} // endfor
		
		QARenderEnd(drawContext, nil);
	} // endfor
	
		// wait for the engine to finish drawing the queued triangles
	QASync(drawContext);
	
	microseconds = S3ElapsedMicroseconds(microseconds);

	S3Test_Print("Timing info for %s:\r", string255);
	
	totalSeconds = microseconds / 1000000.0;

	S3PrintDisplayInfo(S3TestInfo_GetLDevice(testInfo));
	
	sprintf(string255, "  Time to draw %d frames (seconds):", inFrameCount);
	
	S3Test_Print("%-40s%.4f\r", string255, totalSeconds);
	
	S3Test_Print("%-40s%.4f\r", "  Time per frame (seconds):",
				 totalSeconds / inFrameCount);
	
	S3Test_Print("%-40s%.4f\r", "  Pixels per second:",
				 totalPixels / totalSeconds);
	
	S3Test_Print("%-40s%.4f\r\r", "  Triangles per second:", 
				 totalTriangles / totalSeconds);
	
		// now repeat the test for the right draw context
	drawContext = S3TestInfo_GetRDrawContext(testInfo);
	ASSERT(drawContext);

	S3GetEngineName(S3TestInfo_GetREngine(testInfo), string255, 
					sizeof(string255));
					
	microseconds = S3MicrosecondCount();
	
		// draw the same frame inFrameCount times
	for (i = 0; i < inFrameCount; i++) {
		QARenderStart(drawContext, nil, nil);
		
			// we run through this loop twice, once when the Z value
			// of the triangles is .1, then again when it's .6.
		for (z = 0.1; z < 1.0; z += 0.5) {
			invW = 1.0 / z;
			
			for (j = 0; j < verticesPerLayer; j += 4) {
					// we set the z and invW values for the four vertices 
					// which form the current square.  we also set the 
					// uOverW and vOverW values to invW for the vertices 
					// which have 1.0 in those slots.  this is because the
					// u,v parameters have to be divided by z.  but since
					// the u,v values are either 0 or 1.0, we can just use
					// invW (1.0 / z), which we calculated above.

					// top left
				vertices[j].z = z;
				vertices[j].invW = invW;
				vertices[j].vOverW = invW;
				
					// top right
				vertices[j + 1].z = z;
				vertices[j + 1].invW = invW;
				vertices[j + 1].uOverW = invW;
				vertices[j + 1].vOverW = invW;
					
					// bottom left
				vertices[j + 2].z = z;
				vertices[j + 2].invW = invW;
				
					// bottom right
				vertices[j + 3].z = z;
				vertices[j + 3].invW = invW;
				vertices[j + 3].uOverW = invW;
	
					// draw the two triangles which fill the inSquareSize X inSquareSize rect.
				QADrawTriTexture(
					drawContext,
					&vertices[j],			// top left
					&vertices[j + 2],		// bottom left
					&vertices[j + 3],		// bottom right
					kQATriFlags_None);
	
				QADrawTriTexture(
					drawContext,
					&vertices[j],			// top left
					&vertices[j + 1],		// top right
					&vertices[j + 3],		// bottom right
					kQATriFlags_None);
			} // endfor
		} // endfor
		
		QARenderEnd(drawContext, nil);
	} // endfor
	
		// wait for the engine to finish drawing the queued triangles
	QASync(drawContext);
	
	microseconds = S3ElapsedMicroseconds(microseconds);

		// we're done with the allocated vertices
 	free(vertices);
	
	S3Test_Print("Timing info for %s:\r", string255);
	
	totalSeconds = microseconds / 1000000.0;

	S3PrintDisplayInfo(S3TestInfo_GetRDevice(testInfo));
	
	sprintf(string255, "  Time to draw %d frames (seconds):", inFrameCount);
	
	S3Test_Print("%-40s%.4f\r", string255, totalSeconds);
	
	S3Test_Print("%-40s%.4f\r", "  Time per frame (seconds):",
				 totalSeconds / inFrameCount);
	
	S3Test_Print("%-40s%.4f\r", "  Pixels per second:",
				 totalPixels / totalSeconds);
	
	S3Test_Print("%-40s%.4f\r\r", "  Triangles per second:", 
				 totalTriangles / totalSeconds);
	
	S3Test_Print("\rIgnore the following timing information:\r");
	
	return kS3Status_Completed;
}

/*=========================================================================*\
	S3PrintDisplayInfo
\*=========================================================================*/
void S3PrintDisplayInfo(
	TQADevice*				inDevice)
{
	if (inDevice->deviceType == kQADeviceGDevice) {
		GDevice**			device = inDevice->device.gDevice;
		DisplayIDType		displayID;
		Str255				displayName;
		OSErr 				osError;
		
		osError = DMGetDisplayIDByGDevice(device, &displayID, false);
		ASSERT(osError == noErr);	
		
		osError = DMGetNameByAVID(displayID, 0, (Str255*) displayName);
		ASSERT(osError == noErr);	
	
			// convert the string to a C string
		displayName[displayName[0] + 1] = nil;
		
			// print the name as a C string
		S3Test_Print("%-40s%s\r", "  Device - Name:", (char*) &displayName[1]);
		
		S3Test_Print("%-40s%d\r", "  Device - Bits per pixel:", 
				(**(**device).gdPMap).pixelSize);
		S3Test_Print("%-40s%d X %d\r", "  Device - Width X Height:", 
				(**device).gdRect.right - (**device).gdRect.left,
				(**device).gdRect.bottom - (**device).gdRect.top);
	} else if (inDevice->deviceType == kQADeviceMemory) {
		short				bitdepth = 0;
		
		S3Test_Print("%-40sMemory\r", "  Device - Name:");
		
		switch (inDevice->device.memoryDevice.pixelType) {
			case kQAPixel_ARGB32:
			case kQAPixel_RGB32:
				bitdepth = 32;
				break;
			
			case kQAPixel_ARGB16:
			case kQAPixel_RGB16:
				bitdepth = 16;
				break;
		} // endswitch
		
		S3Test_Print("%-40s%d\r", "  Device - Bits per pixel:", bitdepth);
		S3Test_Print("%-40s%d X %d\r", "  Device - Width X Height:", 
				inDevice->device.memoryDevice.width,
				inDevice->device.memoryDevice.height);
	}
}


/*=========================================================================*\
	S3TestSceneFile
\*=========================================================================*/
S3tTestStatus S3TestSceneFile(
	S3tTestInfo				*testInfo)
{
	S3tSceneFile			sceneFile;
	
	if (S3SceneFile_Open(&sceneFile, "Hoover test")) {
//	if (S3SceneFile_Open(&sceneFile, "Hoover tris")) {
		if (S3SceneFile_ReadAndDraw(sceneFile, testInfo) == false) {
			S3Test_Print("*** An error occurred while reading the file. ***\r\r");
		}
		
		S3SceneFile_Close(sceneFile);
	}
	
	return kS3Status_Completed;
}

/*=========================================================================*\
	S3TestEngineGestalt
\*=========================================================================*/
S3tTestStatus S3TestEngineGestalt(
	S3tTestInfo				*testInfo)
{
	typedef struct {
		char*			name;
		unsigned long	feature;
	} S3tFeatureTable;
	
	static S3tFeatureTable	optionalFeatureTable[] = {
		{"DeepZ"				, kQAOptional_DeepZ			},
		{"Texture"				, kQAOptional_Texture		},
		{"TextureHQ"			, kQAOptional_TextureHQ		},
		{"TextureColor"			, kQAOptional_TextureColor	},
		{"Blend"				, kQAOptional_Blend			},
		{"BlendAlpha"			, kQAOptional_BlendAlpha	},
		{"Antialias"			, kQAOptional_Antialias		},
		{"ZSorted"				, kQAOptional_ZSorted		},
		{"PerspectiveZ"			, kQAOptional_PerspectiveZ	},
		{"OpenGL"				, kQAOptional_OpenGL		},
		{"NoClear"				, kQAOptional_NoClear		},
		{"CL4"					, kQAOptional_CL4			},
		{"CL8"					, kQAOptional_CL8			}
	};
	const int optionalFeatureTableSize = 13;
	
	static S3tFeatureTable	fastFeatureTable[] = {
		{"Line"					, kQAFast_Line			},
		{"Gouraud"				, kQAFast_Gouraud		},
		{"Texture"				, kQAFast_Texture		},
		{"TextureHQ"			, kQAFast_TextureHQ		},
		{"Blend"				, kQAFast_Blend			},
		{"Antialiasing"			, kQAFast_Antialiasing	},
		{"ZSorted"				, kQAFast_ZSorted		},
		{"CL4"					, kQAOptional_CL4		},
		{"CL8"					, kQAOptional_CL8		}
	};
	const int fastFeatureTableSize = 9;

	const int leftEngine = 0;
	const int rightEngine = 1;
	int currentEngine;
	
	char* responseString[256];
	long responseLong = 0;
	unsigned long responseUnsignedLong = 0;
	TQAError error = kQANoErr;
	TQAEngine* engine = nil;
	int i;
	
		// this loop is executed twice, once for each engine
	for (currentEngine = leftEngine; currentEngine <= rightEngine; currentEngine++) {
		if (currentEngine == leftEngine) {
				// first make all the gestalt calls on the left engine
			engine = S3TestInfo_GetLEngine(testInfo);
			S3Test_Print("%sLeft engine:\r\r", kS3Divider_Hyphen);
		} else {
				// now make them on the right engine
			engine = S3TestInfo_GetREngine(testInfo);
			S3Test_Print("%sRight engine:\r\r", kS3Divider_Hyphen);
		} 
			
			// get the engine's name
	 	error = QAEngineGestalt(engine,
			 					kQAGestalt_ASCIINameLength, 
			 					&responseLong);
		S3Test_Print("  ASCIINameLength: ");
		if (error == kQANoErr) {
			S3Test_Print("%d\r", responseLong);
		} else {
			S3Test_Print("*** Error: %s ***\r", S3TestUtil_TQAErrorToString(error));
		}
		
		if (responseLong > 255) {
				// the engine's name is too long to fit into our string buffer
			S3Test_Print("  ASCIIName: *** ASCIINameLength > 255 ***\r");
		} else {
		 	error = QAEngineGestalt(engine,
				 					kQAGestalt_ASCIIName, 
				 					&responseString);
			S3Test_Print("  ASCIIName: ");
			if (error == kQANoErr) {
				S3Test_Print("%s\r", responseString);
			} else {
				S3Test_Print("*** Error: %s ***\r", S3TestUtil_TQAErrorToString(error));
			}
		}
		
			// get the vendor ID
	 	error = QAEngineGestalt(engine,
			 					kQAGestalt_VendorID, 
			 					&responseLong);
		S3Test_Print("  VendorID: ");
		if (error == kQANoErr) {
			S3Test_Print("%d\r", responseLong);
		} else {
			S3Test_Print("*** Error: %s ***\r", S3TestUtil_TQAErrorToString(error));
		}
		
			// get the engine ID
	 	error = QAEngineGestalt(engine,
			 					kQAGestalt_EngineID, 
			 					&responseLong);
		S3Test_Print("  EngineID: ");
		if (error == kQANoErr) {
			S3Test_Print("%d\r", responseLong);
		} else {
			S3Test_Print("*** Error: %s ***\r", S3TestUtil_TQAErrorToString(error));
		}
		
			// get the revision number
	 	error = QAEngineGestalt(engine,
			 					kQAGestalt_Revision, 
			 					&responseLong);
		S3Test_Print("  Revision: ");
		if (error == kQANoErr) {
			S3Test_Print("%d\r", responseLong);
		} else {
			S3Test_Print("*** Error: %s ***\r", S3TestUtil_TQAErrorToString(error));
		}
		
			// get all the fast features
	 	error = QAEngineGestalt(engine,
			 					kQAGestalt_FastFeatures, 
			 					&responseUnsignedLong);
		S3Test_Print("  FastFeatures:\r");
		if (error == kQANoErr) {
			for (i = 0; i < fastFeatureTableSize; i++) {
				if (responseUnsignedLong & fastFeatureTable[i].feature) {
						// the engine has this FastFeature bit turned on
					S3Test_Print("    %s\r", fastFeatureTable[i].name);
				}
			} // endfor
		} else {
			S3Test_Print("*** Error: %s ***\r", S3TestUtil_TQAErrorToString(error));
		}
		
			// get all the optional features
	 	error = QAEngineGestalt(engine,
			 					kQAGestalt_OptionalFeatures, 
			 					&responseUnsignedLong);
		S3Test_Print("  OptionalFeatures:\r");
		if (error == kQANoErr) {
			for (i = 0; i < optionalFeatureTableSize; i++) {
				if (responseUnsignedLong & optionalFeatureTable[i].feature) {
						// the engine has this OptionalFeatures bit turned on
					S3Test_Print("    %s\r", optionalFeatureTable[i].name);
				}
			} // endfor
		} else {
			S3Test_Print("*** Error: %s ***\r", S3TestUtil_TQAErrorToString(error));
		}
		
		S3Test_Print("\r%s\r", kS3Divider_Hyphen);
	} // endfor

	return kS3Status_Completed;
}
