// ===========================================================================
//
// 	GSColorTable.c 
//	
//	Copyright (C) 1996 Apple Computer, Inc.  All rights reserved.
//
// ===========================================================================


#include <stdlib.h>

#if (kQAPlatform == kQAMacOS)
	#include <PictUtil.h>
#endif

#include "RAVE.h"

#include "GSColorTable.h"
#include "GSError.h"


// ===========================================================================
// 	Constants
// ===========================================================================

#define		 				kCL32bit_ElementSize 			4
#define		 				kCL8_ElementCount 				256
#define		 				kCL4_ElementCount 				16

#define		 				kARGB32shift_alpha				24
#define		 				kARGB32shift_red				16
#define		 				kARGB32shift_green				8
#define		 				kARGB32shift_blue				0

#define		 				kARGB32mask_alpha				0xFF << kARGB32shift_alpha
#define		 				kARGB32mask_red					0xFF << kARGB32shift_red
#define		 				kARGB32mask_green				0xFF << kARGB32shift_green
#define		 				kARGB32mask_blue				0xFF << kARGB32shift_blue


// ===========================================================================
//	Types
// ===========================================================================

typedef struct TGSColorTable {
	TQAColorTableType		mType;
	Byte*					mDataP;
} TGSColorTable;


// ===========================================================================
//	Public Routines
// ===========================================================================

// ===========================================================================
//	GSColorTable_New
// ===========================================================================
	TGSError
GSColorTable_New(
	TGSColorTable**			inColorTable)
{
	*inColorTable = (TGSColorTable*) malloc(sizeof(TGSColorTable));
	
	if (*inColorTable == nil) {
		return kGSError_NotEnoughMemory;
	}
	
	(**inColorTable).mType = kQAPixel_RGB32;
	(**inColorTable).mDataP = nil;
	
	return kGSError_None;
}


#if (kQAPlatform == kQAMacOS)

// ===========================================================================
//	GSColorTable_NewFromMacColorTable
// ===========================================================================
	TGSError
GSColorTable_NewFromMacColorTable(
	TGSColorTable**			inColorTable,
	TQAColorTableType	 	inColorTableType,
	CTabHandle				inMacColorTableH)
{
	short					elementCount = 0;
	short					i;
	TGSError				gsError;
	
	GSAssert_(inMacColorTableH);
	GSAssert_(inColorTableType >= kQAColorTable_CL8_RGB32 && inColorTableType <= kQAColorTable_CL4_RGB32);
	
	gsError = GSColorTable_New(inColorTable);
	
	if (gsError != kGSError_None) {
		GSColorTable_Delete(*inColorTable);
		return gsError;
	}
	
	(**inColorTable).mType = inColorTableType;
	
	switch ((**inColorTable).mType) {
		case kQAColorTable_CL8_RGB32:
			elementCount = kCL8_ElementCount;
			break;
		
		case kQAColorTable_CL4_RGB32:
			elementCount = kCL4_ElementCount;
			break;
		
		default:
			GSAssert_(!"Found unrecognized TQAColorTableType in GSColorTable_NewFromMacColorTable");
			break;
	} // endswitch
	
	(**inColorTable).mDataP = malloc(elementCount * kCL32bit_ElementSize);
	
	if ((**inColorTable).mDataP == nil) {
		GSColorTable_Delete(*inColorTable);
		return kGSError_NotEnoughMemory;
	}

	for (i = 0; i < elementCount; i++) {
		unsigned long*		elementP = (unsigned long*) (**inColorTable).mDataP;
		unsigned long 		channelValue;
		
			// turn on all the bits in the alpha channel.  we do this
			// for every entry in the TQAColorTable, regardless of how
			// many entries are in inMacColorTableH.  this also sets
			// the color to black.
		elementP[i] = kARGB32mask_alpha;
		
		if (i <= (**inMacColorTableH).ctSize) {
				// we haven't run out of ColorTable entries yet, so reduce
				// each 16-bit channel in the entry to 8 bits and incorporate
				// it into the TQAColorTable entry.
			channelValue = (**inMacColorTableH).ctTable[i].rgb.red >> 8;
			elementP[i] |= ((channelValue << kARGB32shift_red) & kARGB32mask_red);

			channelValue = (**inMacColorTableH).ctTable[i].rgb.green >> 8;
			elementP[i] |= ((channelValue << kARGB32shift_green) & kARGB32mask_green);

			channelValue = (**inMacColorTableH).ctTable[i].rgb.blue >> 8;
			elementP[i] |= ((channelValue << kARGB32shift_blue) & kARGB32mask_blue);
		}
	} // endfor
	
	return kGSError_None;
}

#endif


// ===========================================================================
//	GSColorTable_Delete
// ===========================================================================
	void
GSColorTable_Delete(
	TGSColorTable*			inColorTable)
{
	if (inColorTable == nil) {
		return;
	}
	
	free(inColorTable);
}


// ===========================================================================
//	GSColorTable_ChangeBrightness
// ===========================================================================
	TGSError
GSColorTable_ChangeBrightness(
	TGSColorTable*			inColorTable,
	float					inBrightnessPercentage)
{
	short					elementCount = 0;
	short					i;

	GSAssert_(inColorTable);
	GSAssert_(inColorTable->mDataP != nil);
	
	switch (inColorTable->mType) {
		case kQAColorTable_CL8_RGB32:
			elementCount = kCL8_ElementCount;
			break;
		
		case kQAColorTable_CL4_RGB32:
			elementCount = kCL4_ElementCount;
			break;
		
		default:
			GSAssert_(!"Found unrecognized TQAColorTableType in GSColorTable_NewFromMacColorTable");
			break;
	} // endswitch
	
	for (i = 0; i < elementCount; i++) {
		Byte*				channelP = (Byte*) inColorTable->mDataP;
		unsigned long		channelValue;
		
			// skip past the alpha channel, which is the first one
		channelP++;
		
		channelValue = (unsigned long) (*channelP * inBrightnessPercentage);
		*channelP = (Byte) (channelValue > 0xFF) ? 0xFF : channelValue;
		channelP++;

		channelValue = (unsigned long) (*channelP * inBrightnessPercentage);
		*channelP = (Byte) (channelValue > 0xFF) ? 0xFF : channelValue;
		channelP++;

		channelValue = (unsigned long) (*channelP * inBrightnessPercentage);
		*channelP = (Byte) (channelValue > 0xFF) ? 0xFF : channelValue;
		channelP++;
	} // endfor
	
	return kGSError_None;
}


// ===========================================================================
//	GSColorTable_GetData
// ===========================================================================
	void*
GSColorTable_GetData(
	TGSColorTable*			inColorTable)
{
	GSAssert_(inColorTable);
	
	return inColorTable->mDataP;
}


// ===========================================================================
//	GSColorTable_GetType
// ===========================================================================
	TQAColorTableType
GSColorTable_GetType(
	TGSColorTable*			inColorTable)
{
	GSAssert_(inColorTable);
	
	return inColorTable->mType;
}


